/*
 * Copyright (C) 2002-2011 Edscott Wilson Garcia
 * EMail: edscott@xfce.org
 *
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif


#include "rfm_libs.h"

/* private symbols */

#include "primary-completion.i"
gchar *
rfm_file_completion(widgets_t *widgets_p,  gchar *token){
    if (!token) {
	msg_help_text(widgets_p);
	return NULL;
    }
    if (token) g_strchug (token);
    if (strlen(token) == 0) {
	msg_help_text(widgets_p);
	return NULL;
    }
    gint match_count;
    return bash_file_completion(widgets_p, token, &match_count);
}

gchar *
rfm_bash_completion(widgets_t *widgets_p,  gchar *token){
    //
    //
    // variable, starts with a $ (pending)
    // username, starts with a ~ (pending)
    // hostname, starts with a @ (pending)

    //
    // Exec completion.
    // First, get path...
    if (!token) {
	msg_help_text(widgets_p);
	return NULL;
    }
    if (token) g_strchug (token);
    if (strlen(token) == 0) {
	msg_help_text(widgets_p);
	return NULL;
    }

    // First, get appropriate token.
    //

    gchar *command_token = NULL;
    gchar *file_token = NULL;
    gchar *next_token = NULL;
    const gchar *token_link=" ";

    if (strrchr(token, ';') || strrchr(token, '&') ||
	    strrchr(token, '|') || strrchr(token, '>')){
	// Which one is the last?
	if (strrchr(token, ';') > strrchr(token, '&') &&
	    strrchr(token, ';')	> strrchr(token, '|') &&
	    strrchr(token, ';')	> strrchr(token, '>')){
	    next_token = strrchr(token, ';');
	    token_link=";";
	} 
	if (strrchr(token, '&') > strrchr(token, ';') &&
	    strrchr(token, '&')	> strrchr(token, '|') &&
	    strrchr(token, '&')	> strrchr(token, '>')){
	    next_token = strrchr(token, '&');
	    token_link="&";
	} 
	if (strrchr(token, '|') > strrchr(token, ';') &&
	    strrchr(token, '|')	> strrchr(token, '&') &&
	    strrchr(token, '|')	> strrchr(token, '>')){
	    next_token = strrchr(token, '|');
	    token_link="|";
	} 
	if (strrchr(token, '>') > strrchr(token, ';') &&
	    strrchr(token, '>')	> strrchr(token, '&') &&
	    strrchr(token, '>')	> strrchr(token, '|')){
	    next_token = strrchr(token, '>');
	    token_link=">";
	}     
    }
    if (next_token){
	*next_token=0;
	next_token++;
	g_strchug(next_token);
	if (strlen(next_token)==0) return NULL;
    }
    

    gchar *active_token = next_token? next_token: token;

    // Obtain the file token separation, if any.
    if (strstr(active_token, "\\ ")){
	// Using escaped method
	gchar **split = g_strsplit (active_token, "\\ ", -1);
	if (strchr(split[0], ' ')){
	    // Does the first element have a space?
	    const gchar *substring = strrchr(split[0], ' ')+1;
	    file_token = strstr(active_token, substring);
	    substring = strrchr(split[0], ' ');
	    *strstr(active_token, substring)=0;
	} else {
	    command_token = active_token;
	}
	g_strfreev(split);
    } else if (strrchr(active_token, ' ')) {
	file_token = strrchr(active_token, ' ')+1;
	*strrchr(active_token, ' ') = 0;
    } else {
	command_token = active_token;
	next_token=NULL;
    }

    // Now that we have our token defined, we need to check
    // if the user is using quotes or not.
    gchar *quote=NULL;
    if (command_token) {
	if (command_token[0]=='\'') quote="\'";
	if (command_token[0]=='\"') quote="\"";
	if (quote) {
	    *command_token=0;
	    command_token++;
	}
    } else if (file_token) {
	if (file_token[0]=='\'') quote="\'";
	if (file_token[0]=='\"') quote="\"";
	if (quote) {
	    *file_token = 0;
	    file_token++;
	}
    }
//    if (quote) *active_token = 0;

    gchar *suggest = NULL;
    gint match_count = 0;
    if (command_token) {
	suggest = bash_exec_completion(widgets_p, command_token, &match_count);	
    }
    else if (file_token) {
	suggest = bash_file_completion(widgets_p, file_token, &match_count);
    }

    if (suggest) {
      if (quote) {
	gchar *s;
	if (match_count==1 && suggest[strlen(suggest)-1] != '/') {
	    s = g_strconcat (quote, suggest, quote, " ", NULL);
	} else {
	    s = g_strconcat (quote, suggest, NULL);
	}
	g_free(suggest);
	suggest=s;

      } else {
	if (strchr(suggest, ' ')){
	    gchar **split = g_strsplit (suggest, " ", -1);
	    gchar *s = g_strjoinv ("\\ ", split);
	    g_free(suggest);
	    suggest=s;
	    g_strfreev(split);
	}
	if (match_count==1 && suggest[strlen(suggest)-1] != '/'){
	    gchar *s=g_strconcat (suggest, " ", NULL);
	    g_free(suggest);
	    suggest=s;
	}
      }
    }


    if (suggest && command_token) {
	gchar *full_suggest;
	if (command_token == token) {
	    full_suggest = g_strconcat(suggest, NULL);
	    
	} else {
	    full_suggest = g_strconcat(token, token_link, " ", suggest, NULL);
	}
	g_free(suggest);
	suggest = full_suggest;
    } 
    else if (suggest && file_token){
	// file_token is always accompanied by token in bash completion.
	gchar *full_suggest;
	if (next_token){
	    full_suggest = g_strconcat(token, token_link, " ", 
		    next_token, " ", suggest, NULL);
	} else {
	    full_suggest = g_strconcat(token, token_link, suggest, NULL);
	}
	g_free(suggest);
	suggest = full_suggest;
    } 

    return suggest;
}


gchar *
rfm_history_completion (widgets_t * widgets_p,  char *complete) {
    view_t *view_p = widgets_p->view_p;
    TRACE ("rfm_history_completion\n");
    if (!complete) {
	msg_help_text(widgets_p);
	return NULL;
    }
    if (complete) g_strchug (complete);
    if (strlen(complete) == 0) {
	msg_help_text(widgets_p);
	return NULL;
    }
    GList *matches=NULL;
    GList *p;
    for(p = g_list_first (view_p->sh_command); p && p->data; p = p->next) {
	gchar *data = (char *) p->data;
	if(strncmp (complete, data, strlen (complete)) == 0) {
	    if(g_list_find_custom (matches, data, ya_strcmp) == NULL) {
		TRACE ("COMPLETE: match! %s <= %s\n", complete, data);
		matches = g_list_append (matches, p->data);
	    }
	}
	TRACE ("COMPLETE: ?? %s\n", data);
    }

    gchar *suggest = complete_it(widgets_p, &matches, MATCH_HISTORY);
    g_list_free (matches);
    
    if(!suggest && strncmp (complete, "history", strlen (complete)) == 0) {
	suggest = g_strdup ("history");
    }
    return suggest;
}


