//
//
static gint
maximum_completion_options(void){
    const gchar *env_maximum=getenv("RFM_MAXIMUM_COMPLETION_OPTIONS");
    if (!env_maximum || !strlen(env_maximum)) return 52;
    errno=0;
    glong amount = strtol(env_maximum, (char **) NULL, 10);
    if (errno) return 52;
    return amount;
}


#define BASH_COMPLETION_OPTIONS maximum_completion_options()
static void
msg_too_many_matches(widgets_t *widgets_p){
    if (!widgets_p) return;
    rfm_show_text(widgets_p);
    gchar *message1=g_strdup_printf("%s (> %d)",
	    _("Too many matches"), BASH_COMPLETION_OPTIONS);
    gchar *message2=g_strdup_printf("%s %s", _("Options:"), message1);
    rfm_diagnostics (widgets_p, "xffm/stock_dialog-info", NULL);
    rfm_diagnostics (widgets_p, "xffm_tag/blue", "[",  _("Text Completion"),"] ", NULL);
    rfm_diagnostics (widgets_p, "xffm_tag/red", message2, "\n", NULL);
    g_free(message1);
    g_free(message2);
}

static void
msg_show_match(widgets_t *widgets_p, gint match_type, const gchar *match){
    if (!widgets_p) return;
    if (!match) {
	const gchar *option_type=NULL;
	switch (match_type){
	    case MATCH_COMMAND:
		option_type = _("Command");
		break;
	    case MATCH_FILE:
		option_type = _("File");
		break;
	    case MATCH_HISTORY:
		option_type = _("Command history");
		break;
	}
	rfm_diagnostics (widgets_p, "xffm/stock_dialog-error",NULL);
	match = _("Found no match");
	rfm_diagnostics (widgets_p, "xffm_tag/red", "(", 
		option_type, ")", NULL);
    } 
    rfm_diagnostics (widgets_p, "xffm_tag/blue", " ", match, "\n", NULL);
    
}

static void
msg_help_text(widgets_t *widgets_p){
    if (!widgets_p) return;
    rfm_show_text(widgets_p);

    rfm_diagnostics (widgets_p, "xffm/stock_dialog-info", NULL);
    rfm_diagnostics (widgets_p, "xffm_tag/green",
	    _("Completion mode:"), 
	    " bash ", _("command"), "/", _("file"), " --> ", NULL);
    rfm_diagnostics (widgets_p, "xffm_tag/red", "TAB.",  NULL);



    rfm_diagnostics (widgets_p, "xffm_tag/green", "   ", 
	    _("Completion mode:")," ", _("Command history")," --> ", NULL); 
    rfm_diagnostics (widgets_p, "xffm_tag/red", "CTRL+TAB.\n",  NULL);
}

static void
msg_result_text(widgets_t *widgets_p, gint match_type){
    if (!widgets_p) return;
    rfm_show_text(widgets_p);
    rfm_diagnostics (widgets_p, "xffm/stock_dialog-info", NULL);
    const gchar *option_type=NULL;
    switch (match_type){
	case MATCH_COMMAND:
	    option_type = _("Command");
	    break;
	case MATCH_FILE:
	    option_type = _("File");
	    break;
	case MATCH_HISTORY:
	    option_type = _("Command history");
	    break;
    }

    rfm_diagnostics (widgets_p, "xffm_tag/green", _("Options >>"), NULL);
    rfm_diagnostics (widgets_p, "xffm_tag/red", "(", option_type, ")\n", NULL);
}

static gint
ya_strcmp (
    gconstpointer a,
    gconstpointer b
) {
    return strcmp ((char *) a, (char *) b);
}

static const gchar *
get_workdir(widgets_t *widgets_p){
    if (!widgets_p) return g_get_home_dir();
    view_t *view_p=widgets_p->view_p;
    if (!view_p) return g_get_home_dir();
    record_entry_t *en=view_p->en;
    if (!en || !en->path) return g_get_home_dir();
    if (!g_file_test(en->path, G_FILE_TEST_IS_DIR)) return g_get_home_dir();
    return en->path;
}

static int
length_equal_string(const gchar *a, const gchar *b){
    int length=0;
    int i;
    for (i = 0; i < strlen(a) && i < strlen(b); i++){
	if (strncmp(a,b,i+1)) {
	    length=i;
	    break;
	} else {
	    length=i+1;
	}
    }
     TRACE ("%s --- %s differ at length =%d\n", a,b,length);
   return length;
}

static gchar *
list_matches (widgets_t * widgets_p, GList **matches_p, gint match_type){
    *matches_p = g_list_sort (*matches_p, ya_strcmp);
    GList *a;
    GList *b;
    int length;
    GList *p = g_list_first (*matches_p);
    char *suggest = g_strdup ((gchar *) (p->data));
    int equal_length=strlen(suggest);
    for(a = *matches_p; a && a->data; a = a->next) {
	TRACE("comparing a %s\n", (gchar *)a->data);
	for(b = a->next; b && b->data; b = b->next) {
	    TRACE("comparing b %s\n", (gchar *)b->data);
	    length=length_equal_string(
		    (const gchar *)(a->data), (const gchar *)(b->data));
	    if(length < equal_length) {
		equal_length = length;
	    }
	    TRACE("comparing %s to %s: length=%d\n", (gchar *)a->data, (gchar *)b->data, length);
	}
    }
    suggest[equal_length]=0;
    TRACE("string=%s, equal_length=%d, suggest=%s\n",
	    (gchar *) (p->data), equal_length, suggest);


    int i;
    for(i = 1, p = g_list_first (*matches_p); p && p->data; p = p->next, i++) {
	msg_show_match(widgets_p, match_type, (const gchar *)p->data);
    }
    return suggest;
}

static gchar *
complete_it(widgets_t * widgets_p, GList **matches_p, gint match_type){
        gchar *suggest = NULL;
        if(*matches_p) {
            TRACE ("COMPLETE: matches %d\n", g_list_length (*matches_p));
            if(g_list_length (*matches_p) == 1) {
                suggest = g_strdup ((*matches_p)->data);
            } else {
		rfm_show_text(widgets_p);
		msg_result_text(widgets_p, match_type);
		suggest=list_matches(widgets_p, matches_p, match_type);
            }
        }
	return suggest;
}
static gchar *
bash_file_completion(widgets_t *widgets_p,  
	const char *in_file_token, gint *match_count_p){
    if (!in_file_token) return NULL;
    if (strlen(in_file_token) == 0) return NULL;

    gchar *file_token=NULL;
    if (in_file_token[0] == '~'){
	file_token = g_strconcat(g_get_home_dir(),in_file_token+1, NULL);
    } else {
	file_token = g_strdup(in_file_token);
    }
    GList *matches=NULL;


    gchar *directory;
    gchar *relative_directory=NULL;
    if (g_path_is_absolute(file_token)) {
	directory = g_strdup_printf("%s*", file_token);

    } else {
	directory = g_strdup_printf("%s/%s*", get_workdir(widgets_p), file_token);
	relative_directory = g_path_get_dirname(file_token);
	if (strcmp(relative_directory, ".")==0 && 
		strncmp(file_token, "./", strlen("./")) != 0) {
	    g_free(relative_directory);
	    relative_directory=NULL;
	}
    }
    
    TRACE ("file_token=%s\ndirectory=%s\n", file_token, directory);

    
    
    glob_t stack_glob_v;
    gint flags=0;

    glob(directory, flags, NULL, &stack_glob_v);
    g_free(directory);

    if (stack_glob_v.gl_pathc == 0){
	    msg_show_match(widgets_p, MATCH_FILE, NULL);
    } else if (stack_glob_v.gl_pathc <= BASH_COMPLETION_OPTIONS){
	gint i;
	for (i=0; i<stack_glob_v.gl_pathc; i++){
	    gboolean isdir=g_file_test(stack_glob_v.gl_pathv[i], G_FILE_TEST_IS_DIR);
	    gchar *base;
	    if (g_path_is_absolute(file_token)) {
		base = g_strdup(stack_glob_v.gl_pathv[i]);
	    } else {
		base = g_path_get_basename(stack_glob_v.gl_pathv[i]);
		if (relative_directory) {
		    gchar *b=g_build_filename(relative_directory, base, NULL);
		    g_free(base);
		    base=b;
		}
	    }
	    if (isdir) {
		gchar *d=g_strconcat(base, "/", NULL);
		g_free(base);
		base = d;
	    }

	    //rfm_diagnostics (widgets_p, "xffm_tag/green", "file option = ", base, "\n", NULL);
	    matches = g_list_append (matches, base);
	}
    } else {
	msg_too_many_matches(widgets_p);
    }

	    
    g_free(relative_directory);
    globfree(&stack_glob_v);
    gchar *suggest = complete_it(widgets_p, &matches, MATCH_FILE);
    TRACE("match = \"%s\"\n",suggest? suggest : "no match!");


    if (suggest) {
	/*gchar *absolute_suggest=g_build_filename(get_workdir(widgets_p), suggest, NULL);
	if (g_list_length(matches)==1 && 
	    absolute_suggest[strlen(absolute_suggest)-1] != '/'){
	    gchar *s;
	    TRACE("adding final space...\n");
	    s=g_strconcat(suggest, " ", NULL);
	    g_free(suggest);
	    suggest = s;
	}*/
	*match_count_p = g_list_length(matches);
    } else *match_count_p = 0;

    GList *p=matches;
    for (;p && p->data; p=p->next){
	g_free(p->data);
    }
    g_list_free (matches);

    g_free(file_token);
    return suggest;
}


static gchar *
bash_exec_completion(widgets_t *widgets_p, const char *token, gint *match_count_p){
    GList *matches=NULL;

	
    glob_t stack_glob_v;
    gboolean straight_path = g_path_is_absolute(token) ||
	strncmp(token, "./", strlen("./"))==0 ||
	strncmp(token, "../", strlen("../"))==0;

    if (straight_path) {
	if (chdir(get_workdir(widgets_p)) < 0){ g_warning("chdir %s",get_workdir(widgets_p));}
	
	gchar *directory = g_strdup_printf("%s*", token);
	glob(directory, 0, NULL, &stack_glob_v);
	g_free(directory);

    }     
    else if (getenv("PATH") && strlen(getenv("PATH"))){
	gchar *path_v = g_strdup(getenv("PATH"));
	gchar **path_pp = rfm_split(path_v,':');
	gchar **pp = path_pp;
	    
	for (; pp && *pp; pp++){
	    if (strlen(*pp)==0) continue;

	    gint flags;
	    gchar *directory=g_strdup_printf("%s/%s*", *pp, token);
	    if (pp==path_pp) flags = 0;
	    else flags =  GLOB_APPEND;
	    //gint glob_result = 
	    glob(directory, flags, NULL, &stack_glob_v);
	    g_free(directory);
	    if (stack_glob_v.gl_pathc > BASH_COMPLETION_OPTIONS) break;
	}
	g_free(path_pp);
	g_free(path_v);
    }

    if (stack_glob_v.gl_pathc == 0){
	    msg_show_match(widgets_p, MATCH_COMMAND, NULL);
    } else if (stack_glob_v.gl_pathc <= BASH_COMPLETION_OPTIONS){
	struct stat st;
	gint i;
	for (i=0; i<stack_glob_v.gl_pathc; i++){
	    stat (stack_glob_v.gl_pathv[i], &st);
	    if (S_IXOTH & st.st_mode){
		gchar *base;
		if (straight_path) {
		    base = g_strdup(stack_glob_v.gl_pathv[i]);
		} else {
		    base = g_path_get_basename(stack_glob_v.gl_pathv[i]);
		}
		//rfm_diagnostics (widgets_p, "xffm_tag/red", "exec option = ", base, "\n", NULL);
		matches = g_list_append (matches, base);
	    }
	}
    } else {
	msg_too_many_matches(widgets_p);
    }

    if (chdir(g_get_home_dir()) < 0){ g_warning("chdir %s",g_get_home_dir());}

    globfree(&stack_glob_v);


    gchar *suggest = complete_it(widgets_p, &matches, MATCH_COMMAND);
    if (!suggest) {
        suggest = bash_file_completion(widgets_p, token, match_count_p);
	if (suggest) {

	    gchar *absolute_suggest=
		g_build_filename(get_workdir(widgets_p), suggest, NULL);
	    g_strchomp (absolute_suggest);
	    TRACE("file absolute_suggest=%s (%s)\n", absolute_suggest, suggest);
	    if (access(absolute_suggest, X_OK) != 0){
	    //if (!g_file_test(absolute_suggest, G_FILE_TEST_IS_EXECUTABLE)     
		g_free(suggest);
		suggest=NULL;
		TRACE("access \"%s\": %s\n", absolute_suggest, strerror(errno));
	    } 
	    g_free(absolute_suggest);
	}
    }
    TRACE ("suggest=%s\n", suggest);
    *match_count_p = g_list_length(matches);
  /*  if (g_list_length(matches)==1){

	gchar *s;
	if (g_file_test(suggest, G_FILE_TEST_IS_DIR)){
	    s=g_strconcat(suggest, "/", NULL);
	} else {
	    s=g_strconcat(suggest, " ", NULL);
	}
	g_free(suggest);
	suggest = s;
    }*/
    GList *p=matches;
    for (;p && p->data; p=p->next){
	g_free(p->data);
    }
    g_list_free (matches);

    return suggest;

}


