//
//
static GMutex *winner_mutex = NULL;
// internal hash record structure
static GHashTable *pixbuf_hash = NULL;
static GMutex *pixbuf_hash_mutex = NULL;

typedef struct pixbuf_t {
    time_t mtime; // stat mtime info for thumbnails
    gint   size;  // pixbuf icon size
    off_t  st_size; // stat st_size for thumbnails
    ino_t  st_ino; // stat st_ino for thumbnails
    GdkPixbuf *pixbuf;
    union {
        gchar *mime_id;
        gchar *path;
    };
} pixbuf_t;

static
GdkPixbuf *
get_pixbuf (const gchar * key, gint size, gboolean hash);

static 
const gchar *
resolve_folder_icon_type (gint type){
 	/*if (en->st && rfm_write_ok(en->st)) {
	    return g_strconcat (id, "/composite/emblem_write-ok", NULL);
	}*/
	const gchar *t;
        if(IS_NOACCESS_TYPE (type)) {
		t = "xffm/stock_directory/composite/emblem_no-access";
        } else if(IS_NOWRITE_TYPE (type)) {
        // without write access: xffm/stock_directory
		t = "xffm/stock_directory";
        } else {
	// with write access: xffm/stock_directory
		t = "xffm/stock_directory/composite/emblem_write-ok";
        }
    return t;
}

/* some of the code in this function contributed by Thomas Leonard: */
static void
save_thumbnail (const gchar * thumbnail_path, 
	const gchar * original_path, 
	GdkPixbuf * tgt)
{
    if(!GDK_IS_PIXBUF (tgt)) return;
    
    gint original_width, original_height;
    gchar *swidth, *sheight, *ssize, *smtime;
    struct stat st;
    GError *error = NULL;

    if(!thumbnail_path || !original_path)
        return;
    TRACE ("creating thumbnail for %s (%s)\n", original_path, thumbnail_path);

    original_width = gdk_pixbuf_get_width (tgt);
    original_height = gdk_pixbuf_get_height (tgt);

    if(stat (original_path, &st) != 0)
        return;
    if (!pixbuf_mutex) pixbuf_mutex = g_mutex_new();

    swidth = g_strdup_printf ("%d", original_width);
    sheight = g_strdup_printf ("%d", original_height);
    ssize = g_strdup_printf ("%ld", (long int)st.st_size);
    smtime = g_strdup_printf ("%ld", st.st_mtime);

	    g_mutex_lock(pixbuf_mutex);
    gdk_pixbuf_save (tgt, thumbnail_path, "png", &error,
                     /*"tEXt::Thumb::Image::Width", swidth,
                        "tEXt::Thumb::Image::Height", sheight,
                        "tEXt::Thumb::Size", ssize, 
                        "tEXt::Thumb::MTime", smtime, */
                     "tEXt::Software", "Rodent", NULL);
	    g_mutex_unlock(pixbuf_mutex);
    if(error)
        g_error_free (error);
    g_free (swidth);
    g_free (sheight);
    g_free (ssize);
    g_free (smtime);
}

static
    void
insert_pixbuf_tag (const GdkPixbuf * tag, GdkPixbuf * composite_pixbuf, const gchar * where, double scale_factor, gint overall_alpha) {
    if (!GDK_IS_PIXBUF(composite_pixbuf)) return;
    int width = gdk_pixbuf_get_width (composite_pixbuf);
    int height = gdk_pixbuf_get_height (composite_pixbuf);
    // "SW"
    int dest_x = 0;
    int dest_y = 0;
    int dest_width = width;
    int dest_height = height;
    double scale_x = 1.0 / scale_factor;
    double scale_y = 1.0 / scale_factor;
    // default SW
    double offset_x = 0.0;
    double offset_y = height - ((double)height) / scale_factor;
    if(strcmp (where, "SE") == 0) {
        offset_x = width - ((double)width) / scale_factor;
    } else if(strcmp (where, "NW") == 0) {
        offset_y = 0.0;
    } else if(strcmp (where, "NE") == 0) {
        offset_x = width - ((double)width) / scale_factor;
        offset_y = 0.0;
    } else if(strcmp (where, "C") == 0) {
        offset_x = ((double)width) / 2 - ((double)width) / scale_factor / 2;
        offset_y = ((double)height) / 2 - ((double)height) / scale_factor / 2;
    } else if(strcmp (where, "SC") == 0) {
        offset_x = ((double)width) / 2 - ((double)width) / scale_factor / 2;
    } else if (strcmp (where, "NC") == 0) {
        offset_y = 0.0;
        offset_x = ((double)width) / 2 - ((double)width) / scale_factor / 2;
    }
    if (!pixbuf_mutex) pixbuf_mutex = g_mutex_new();

	    g_mutex_lock(pixbuf_mutex);
    gdk_pixbuf_composite (tag, composite_pixbuf,
                          dest_x, dest_y,
                          dest_width, dest_height, offset_x, offset_y, scale_x, scale_y, GDK_INTERP_BILINEAR, overall_alpha);
	    g_mutex_unlock(pixbuf_mutex);
    return;
}

typedef struct composite_t{
    const gchar *sub_id;
    const gchar *where;
    double scale_factor;
    gint overall_alpha;    
}composite_t;

// This routine is mainly here to cover up when the icon module
// plugin is not available or fails to load.
static GdkPixbuf *
fallback_pixbuf(const gchar * id, int size){
    GdkPixbuf *pixbuf=NULL;
    gchar *fullpath=NULL;
	gint gtksize;
        if (size >= SMALL_ICON_SIZE)
	    gtksize = GTK_ICON_SIZE_DIALOG;
	else 
	    gtksize = GTK_ICON_SIZE_LARGE_TOOLBAR;
	    //gtksize = GTK_ICON_SIZE_BUTTON;
	gchar *gtk_id=NULL;
	if (strstr(id, "xffm/stock_")){
	    gtk_id = g_strdup_printf("gtk-%s", 
		    id+strlen("xffm/stock_"));
	}
	    
	TRACE("fallback icon for id=%s at size %d\n", id, size);
    if (!pixbuf_mutex) pixbuf_mutex = g_mutex_new();


	if (gtk_id){
	    // XXX: these should most probably have gdk mutex enabled...
	    //      but at this point this would block main thread...
	    //      so let's just use the pixbuf mutex...
	    g_mutex_lock(pixbuf_mutex);
	    GtkWidget *image=gtk_image_new_from_file("whatever");
	    pixbuf = 
		gtk_widget_render_icon (image, gtk_id, 
			gtksize,NULL);
	    g_mutex_unlock(pixbuf_mutex);
	    if (pixbuf && size > SMALL_ICON_SIZE) {
		g_mutex_lock(pixbuf_mutex);
		GdkPixbuf *new_pixbuf = 
		    gdk_pixbuf_scale_simple (pixbuf, 
			    size, size, GDK_INTERP_BILINEAR);
		g_mutex_unlock(pixbuf_mutex);
		g_object_unref(pixbuf);
		pixbuf=new_pixbuf;
	    }
	    
	    g_free(gtk_id);
	    if (pixbuf) return pixbuf;
	} else {
	    if (strcmp(id, "xffm/emblem_write-ok")==0) {
		fullpath = g_strdup_printf("%s/%s/%s", 
			PACKAGE_ICON_DIR, "emblems", "emblem-write-ok.svg");
	    }
	    else if (strcmp(id, "xffm/emblem_link")==0) {
		fullpath = g_strdup_printf("%s/%s/%s", 
			PACKAGE_ICON_DIR, "emblems", "emblem-symbolic-link.svg");
	    }
	    else if (strcmp(id, "xffm/emblem_unreadable")==0 ||
		    strcmp(id, "xffm/emblem_no-access")==0 ||
		    strcmp(id, "xffm/emblem_no-read")==0) {
		fullpath = g_strdup_printf("%s/%s/%s", 
			PACKAGE_ICON_DIR, "emblems", "emblem-unreadable.svg");
	    }
	    else if (strcmp(id, "xffm/emblem_show-hidden")==0 ) {
		fullpath = g_strdup_printf("%s/%s/%s", 
			PACKAGE_ICON_DIR, "emblems", "emblem-show-hidden.svg");
	    }
	    else if (strcmp(id, "xffm/emblem_show-previews")==0 ) {
		fullpath = g_strdup_printf("%s/%s/%s", 
			PACKAGE_ICON_DIR, "places", "folder-images.svg");
	    }
	    else if (strcmp(id, "xffm/device_computer")==0 ) {
		fullpath = g_strdup_printf("%s/%s/%s", 
			PACKAGE_ICON_DIR, "devices", "computer.svg");
	    }
	    else if (strcmp(id, "xffm/places_user-bookmarks")==0 ) {
		fullpath = g_strdup_printf("%s/%s/%s", 
			PACKAGE_ICON_DIR, "places", "user-bookmarks.svg");
	    }

	}

	if (fullpath) {
	    pixbuf = 
		get_pixbuf(fullpath, size, TRUE);
	    if (!pixbuf) {
		DBG("incorrect fullpath: %s", fullpath);
	    }
	    g_free(fullpath);
	}
	if (!pixbuf && !strstr(id, "xffm/")) {
	    g_mutex_lock(pixbuf_mutex);
	    GtkWidget *image=gtk_image_new_from_file("whatever");
	    pixbuf = gtk_widget_render_icon (image, "gtk-file", 
			gtksize,NULL);
	    g_mutex_unlock(pixbuf_mutex);
	    if (pixbuf && size > SMALL_ICON_SIZE) {
		g_mutex_lock(pixbuf_mutex);
		GdkPixbuf *new_pixbuf = 
		    gdk_pixbuf_scale_simple (pixbuf, 
			    size, size, GDK_INTERP_BILINEAR);
		g_mutex_unlock(pixbuf_mutex);
		g_object_unref(pixbuf);
		pixbuf=new_pixbuf;
	    }
	}

	return pixbuf;
/*	pixbuf = 
		gtk_widget_render_icon (image, GTK_STOCK_FILE, 
			gtksize,NULL);*/
}

static GdkPixbuf *
make_pixbuf_from_id (const gchar * id, int size, gboolean hash) {
    GdkPixbuf *pixbuf=NULL;

    TRACE ("ICONx:make_pixbuf_from_id: %s\n", id);
    if(!id || !strlen (id))
        return NULL;
    const gchar *c_key="/composite";
    const gchar *cc_key="/composite/";
    const gchar *ccc_key="composite/";
	    const gchar *position="C";
	    gdouble scale_f=1.8;
	    gint overall_alpha=180;
    if (strstr(id, "/composite2/")){
	c_key="/composite2";
	cc_key="/composite2/";
	ccc_key="composite2/";
	position="NE";
	scale_f=2.0;
	overall_alpha=150;
    }
    if (!pixbuf_mutex) pixbuf_mutex = g_mutex_new();

    if(strstr (id, c_key)) {
        gchar *basic_type = g_strdup (id);
        TRACE ("composite type = %s\n", id);

        GdkPixbuf *basic_pixbuf, *composite_pixbuf;
        *(strstr (basic_type, c_key)) = 0;

        basic_pixbuf = get_pixbuf (basic_type, size, hash);
        g_free (basic_type);
        if(GDK_IS_PIXBUF(basic_pixbuf)) {
	    g_mutex_lock(pixbuf_mutex);
            composite_pixbuf = gdk_pixbuf_copy (basic_pixbuf);
	    g_mutex_unlock(pixbuf_mutex);
        } else {
            composite_pixbuf = NULL;
            TRACE ("cannot get basic pixbuf for %s", id);
            return NULL;
        }

        if(strstr (id, cc_key) == NULL) {
            // incomplete composite specification.
            return composite_pixbuf;
        }
	composite_t composite_a[]={
	    {"/stock_dialog-error", "SE", 2.0, 180},
	//    {"/show-previews", "SW", 6.0, 180},
	    {"/book", "NE", 2.0, 220},
	    {"/places_user-bookmarks", "NE", 2.0, 220},
	//    {"/show-hidden", "NE", 3.0, 180},
	    {"/stock_yes", "NW", 3.0, 180},
	    {"/stock_no", "NW", 3.0, 180},
	    {"/stock_connect", "NW", 3.0, 180},
	//    {"/stock_directory", "SE", 1.7, 180},
	    {"/stock_execute", "NE", 1.7, 255},
	    {"/emblem_no-read", "SE", 2.0, 180},
	    {"/emblem_no-access", "NE", 2.0, 180},
	    {"/emblem_write-ok", "SE", 1.7, 180},
	    {NULL, NULL, 0.0, 180}
	};
	

        gchar *composite_id = g_strdup(strstr(id, cc_key) + strlen(cc_key));

	// split
	gint count=0;
	gchar *p;
	gchar *comp_id[10];
	p=comp_id[count]=composite_id;
	count++;
	while (strchr(p,'/')){
	    *strchr(p,'/')=0;
	    p += (strlen(p)+1);
	    if (strncmp(p,ccc_key,strlen(ccc_key)!=0)){
		TRACE(" composite type= %s\n", p);
		comp_id[count]=p;
		count++; 
	    }
	}
	gint j;
	const gchar *group="xffm";
	for (j=0; j<count; j++){
	    gchar *groups[]={"application", "audio", "chemical", "image", "inode",
		"message", "model", "multipart", "text", "video", "x-conference",
		"x-content", "x-epoc", "x-world",NULL};
	    gchar **pp=groups;
	    for(; pp && *pp; pp++){
		if (strcmp(comp_id[j],*pp)==0){
		    group=*pp;
		    continue;
		}
	    }
	    // Don't composite with group names.
	    if (strcmp(group, comp_id[j])==0) continue;
	    gchar *c_id = NULL;

	    gint i;
	    // special position pixbufs (xffm group):
	    for(i = 0; composite_a[i].sub_id; i++) {
		if(strstr (composite_a[i].sub_id, comp_id[j])) {
		    position=composite_a[i].where;
		    c_id = g_strdup_printf ("xffm%s", composite_a[i].sub_id);
		    scale_f=composite_a[i].scale_factor;
		    overall_alpha=composite_a[i].overall_alpha;
		    break;
		}
	    }
	    if (!c_id) {
		// no special position found
		c_id = g_strdup_printf ("%s/%s", group, comp_id[j]);
	    }
	    TRACE("composite=%s\n", c_id);
		    
	    GdkPixbuf *tag = get_pixbuf (c_id, size, hash);
	    if(!tag) {
		g_warning ("cannot get pixbuf for id: %s", c_id);
		return NULL;
	    }
	    insert_pixbuf_tag (tag, composite_pixbuf, position, scale_f, overall_alpha);
	    //g_object_unref(front);
	    g_free(c_id);
	}
	g_free(composite_id);
        return composite_pixbuf;
    }

    GError *error = NULL;
    gchar *file = NULL;
    file = ICON_get_filename_from_id (id);

    if(!file) {
	// fallback for missing icons plugin or borked id...
	TRACE("fallback for %s", id);
	pixbuf = fallback_pixbuf(id, size);	
    }
    TRACE ("ICON_get_filename_from_id(%s) ---> %s\n", id, file);
    /////////////////////////////////////////////////
    //        gdk_pixbuf_new is not thread safe and must
    //        be run with GDK mutex, but not from main thread.
    //        (how about a private mutex, just for gdkpixbuf functions)?
    //        That should work if gdkpixbuf does not overstep its
    //        own boundaries...
    if (file && rfm_g_file_test(file, G_FILE_TEST_EXISTS)){
	g_mutex_lock(pixbuf_mutex);
	pixbuf = gdk_pixbuf_new_from_file_at_size (file, size, size, &error);
	g_mutex_unlock(pixbuf_mutex);
    }

    if (!pixbuf) {
        TRACE ("make_pixbuf_from_id(%s, %d): returns NULL!\n", id, size);
    }
    g_free(file);
    return pixbuf;
}

// This is used by the preview thread manager to determine if 
// population_pp under study is still valid.
static gboolean
invalid_population (preview_manager_t *preview_manager_p, gint population_serial) {
    gboolean result = FALSE;
    // Is the view gone?
    if (preview_manager_p->diagnostics_window == NULL &&
	preview_manager_p->diagnostics != NULL &&
       *(preview_manager_p->diagnostics)==NULL){
	TRACE("invalid_population 2\n");
       	return TRUE;
    }    
    g_mutex_lock (preview_manager_p->view_p->mutexes.population_serial);
    if(preview_manager_p->view_p->population_serial != population_serial) {
	TRACE("invalid population serial %d != %d\n", 
		preview_manager_p->view_p->population_serial, population_serial);
        result = TRUE;
    }
    g_mutex_unlock (preview_manager_p->view_p->mutexes.population_serial);
    return result;
}


static gpointer
thread_preview_manager (gpointer data) {
    preview_manager_t *preview_manager_p = data;
    view_t *view_p = preview_manager_p->view_p;
    gint population_serial = preview_manager_p->population_serial;
    TRACE ("PREVIEWS: thread_preview_manager\n");
    ///// READLOCK ON //////////
    // Get thread's private copy of population_pp.
    rfm_population_read_lock (view_p);
    // Once we have obtained the read lock, is the population_serial
    // still valid? If not, abort.
    if(invalid_population (preview_manager_p, population_serial)) {
        g_free (preview_manager_p);
	TRACE("* thread_preview_manager: invalid_population!\n");
        rfm_population_read_unlock (view_p);
        return NULL;
    }
 
	    g_mutex_lock(view_p->mutexes.janitor);
	    view_p->janitor.children++;
	    g_mutex_unlock(view_p->mutexes.janitor);

    // gtk icon theme barrier:
    // if barrier is requested from another thread, this will block
    // otherwise barrier will be disabled until decremented.
    // Currently the barrier will be requested when the icontheme
    // changes.
    rfm_barrier_stop_or_increment();
  
    rfm_threadwait();
    //
    // What elements of the population items do we really need after the
    // initial read lock?
    // For the manager:
    // 1. population_p->flags, to check if the item is already thumbnailed.
    // NO 2a. population_p->en, for rfm_is_image() test.  
    // 2b. population_p->en->path, for mimetype winner tag.  
    // 2c. population_p->en->type, for continue condition in loop
    //
    // So we can get away with a specialized structure for each population_p:
    //
    // typedef struct fireup_thread_manager_t {
    //     population_t *population_p;
    //     gint flags;
    //     record_entry_t *en;
    //     preview_manager_t preview_manager_v;
    // } fireup_thread_manager_t;
    //
    // A GSList of this stuff instead of a copy of population_pp array.
    // The thread manager requires only read access.
    //
    // How about the worker threads? These require access to:
    // 1. view_p (view may be gone before last worker completes).
    //           The check for valid view_p is done with the *widgets_p->diagnostics
    //           or  *widgets_p->diagnostics_window. This would also be needed
    //           by the manager, although the manager would never lose the
    //           race.
    // 2. population_p: requires a read lock.
    //    a) population_p->pixbuf
    //    b) population_p->pixbufW
    //    c) population_p->pixbufH
    //    d) population_p->flags
    //
    // typedef struct preview_manager_t{
    //	   view_t *view_p;
    //	   gint population_serial;
    //	   GtkWidget **diagnostics;
    //	   GtkWidget **diagnostics_window;
    //} preview_manager_t;  
    //
    // preview_manager_v instead of preview_manager_p because
    // preview_manager_p will be disposed of by the thread manager.
    //
    GSList *fireup_list=NULL;
    gint j=0;
    // create a list of items to be considered, only.
    // Not stuff for directory contents:  // __NOWRITE_TYPE
    // ((__UP_TYPE | __NOACCESS_TYPE | __SLNK |__MOUNTED_TYPE) & 0xffffffff)
    gint not_stuff=((__UP_TYPE | __NOACCESS_TYPE | __SLNK |__MOUNTED_TYPE) & 0xffffffff);
    // 
    gboolean user_requested_abort=FALSE;
    gboolean have_mime_plugin=GPOINTER_TO_INT(rfm_void(PLUGIN_DIR, "mime", "module_active"));
    gboolean have_icons_plugin=GPOINTER_TO_INT(rfm_void(PLUGIN_DIR, "icons", "module_active"));

    // XXX The smart thing here would be to see which icons are on the
    //     visible view, and update them first. A bit complicated to code...
    for (; j < view_p->xfdir_p->pathc && view_p->population_pp[j]; j++){
	record_entry_t *en=view_p->population_pp[j]->en;
	// Check for abort condition
	if (view_p->flags.user_wants_lock || view_p->flags.status==STATUS_EXIT) {
	    user_requested_abort=TRUE;
	    break;
	}
	// update pixbuf with mime and mimemagic,
	// but only if this has not already been done.
	gboolean update_icon = 
	    view_p->population_pp[j]->flags & POPULATION_SHORT_ICON 
	    &&
	    !(view_p->population_pp[j]->flags & POPULATION_THUMBNAILED);
	    /*
	    view_p->population_pp[j]->en && 
	    !IS_UP_TYPE (view_p->population_pp[j]->en->type) &&
	     view_p->population_pp[j]->en->mimetype == NULL &&
	     view_p->population_pp[j]->en->mimemagic == NULL &&
	    !(view_p->population_pp[j]->flags & POPULATION_THUMBNAILED);*/

	if (update_icon){
	    GdkPixbuf *pixbuf = NULL;
	    TRACE("updating icon...\n");
	    //if (have_mime_plugin) 
	    {
	      const gchar *id = rfm_resolve_icon_id (view_p->population_pp[j]->en);
	      //if (id)
	      {
	        gchar *composite_id = 
		  rfm_get_composite_icon_id (view_p->population_pp[j]->en, id);
	        pixbuf =
	       	  get_pixbuf (composite_id, REDUCED_ICONSIZE(view_p), TRUE);
	        g_free(composite_id);
	      }
	      view_p->population_pp[j]->flags &= (POPULATION_SHORT_ICON ^ 0xffffffff);
	      
	    } 
	    if (pixbuf) {
		view_p->population_pp[j]->pixbuf = pixbuf;
	    }
	    GdkRectangle rect;
	    rfm_get_population_icon_rect(view_p, view_p->population_pp[j], &rect);
	    rfm_thread_expose_rect (gtk_widget_get_window(view_p->widgets.paper), &rect);
	}


	// default is to show thumbnails.
	gboolean directory_condition = 
	    have_mime_plugin && have_icons_plugin &&
	    en && !view_p->module &&
	    IS_SDIR(en->type) && 
	    !(en->type & not_stuff) &&
	    en->path &&
	    strcmp(en->path, g_get_home_dir());
        if(view_p->en && 
		(directory_condition || rfm_is_image (en))) {
    	    fireup_thread_manager_t *fireup_thread_manager_p = 
		(fireup_thread_manager_t *)malloc(sizeof(fireup_thread_manager_t));
	    memset(fireup_thread_manager_p, 0, sizeof(fireup_thread_manager_t));
	    fireup_thread_manager_p->population_p = view_p->population_pp[j];
	    fireup_thread_manager_p->flags = view_p->population_pp[j]->flags;
	    fireup_thread_manager_p->path = g_strdup(view_p->population_pp[j]->en->path);
	    fireup_thread_manager_p->type = view_p->population_pp[j]->en->type;
	    memcpy(&(fireup_thread_manager_p->preview_manager_v),
	       preview_manager_p, sizeof(preview_manager_t));
	    fireup_list = g_slist_prepend(fireup_list, fireup_thread_manager_p);
	}
    }

    view_p->flags.thumbnailer_active=TRUE;
    rfm_population_read_unlock (view_p);
    ///// READLOCK OFF //////////
    //if (view_p->icon_size < SMALL_ICON_SIZE) goto abort;
    if (user_requested_abort) goto abort;
    if (fireup_list) fireup_list = g_slist_reverse(fireup_list);

    /* schedule image previews... */
    gint active_threads = 0;
    GThread *thread_id[PREVIEW_THREAD_LIMIT];


    // This is to tell monitor thread to pause and wait for the mutex.
    // Monitor and thread manager will not run at the same time. 
    if(view_p->mutexes.fireup_thread_manager == NULL) {
	g_error("thread_preview_manager(): fireup_thread_manager_mutex == NULL");
    }
    g_mutex_lock (view_p->mutexes.fireup_thread_manager);
    //for(i = 0; thread_population_pp[i]; i++) {
    GSList *list = fireup_list;
    for(; list && list->data; list=list->next) {
	if (view_p->flags.user_wants_lock) {
	    TRACE("-- view_p->flags.user_wants_lock\n");
	    break;
	}
	if (view_p->flags.status == STATUS_EXIT) break;
	fireup_thread_manager_t *fireup_thread_manager_p = list->data;
	if(invalid_population (preview_manager_p, population_serial)){
	    TRACE("-- invalid_population\n");
	    break;
	}	
	if (view_p->flags.thumbnailer_exit) {
	    TRACE("-- view_p->flags.thumbnailer_exit\n");
	    break;
	}
	if(fireup_thread_manager_p->flags & POPULATION_THUMBNAILED) {
            TRACE ("fireup_thread_manager_p->flags & POPULATION_THUMBNAILED\n");
            continue;
        }
	GError *error=NULL;
	gint retries=1;
retry:
	
       thread_id[active_threads] = 
	    g_thread_create (rfm_preview_thread_f, 
		    (gpointer) fireup_thread_manager_p, TRUE, &error);
	if (thread_id[active_threads] == NULL){
	    if (retries++ >= 3){
		g_warning("cannot create thread %d for %s: %s",
		    active_threads, fireup_thread_manager_p->population_p->en->path,
		    (error)?error->message:"shucks! no error message!");
		continue;
	    }
	    TRACE("g_thread_create: %s. Taking a breath...\n",
		    (error)?error->message:"shucks! no error message!");
	    if (error) g_error_free(error);

	    rfm_threadwait();
	    goto retry;
	}
        active_threads++;

        //every PREVIEW_THREAD_LIMIT threads, don't spawn more until joined
	if(active_threads == PREVIEW_THREAD_LIMIT) {
	    do {
		TRACE ("SEMAPHORE: thread_preview_manager: inner-join %d\n", active_threads);
		g_thread_join (thread_id[active_threads - 1]);
		active_threads--;
	    }
	    while(active_threads > 0);
	}
    }

    while(active_threads > 0) {
        TRACE ("SEMAPHORE: thread_preview_manager outer-join %d\n", active_threads);
        g_thread_join (thread_id[active_threads - 1]);
        active_threads--;
    }
abort:
    if (fireup_list) g_slist_free(fireup_list);
    // check if view is still valid?
    // 
    view_p->flags.thumbnailer_active=FALSE;
    view_p->flags.thumbnailer_exit=FALSE;
    // this is to tell monitor thread to continue
    g_mutex_unlock (view_p->mutexes.fireup_thread_manager);

	    g_mutex_lock(view_p->mutexes.janitor);
	    view_p->janitor.children--;
	    g_mutex_unlock(view_p->mutexes.janitor);

    g_free (preview_manager_p);

    // enable barrier again.    
    rfm_barrier_decrement();

    return NULL;
}


static GdkPixbuf *
create_preview (const gchar * file, gint size) {
    GdkPixbuf *src=NULL;
    GError *error = NULL;
    gchar *thumbnail_path = rfm_get_thumbnail_path (file, size);
    if(!file)
        return NULL;
    if (!pixbuf_mutex) pixbuf_mutex = g_mutex_new();

    if(thumbnail_path && rfm_g_file_test (thumbnail_path, G_FILE_TEST_EXISTS)) {
        struct stat st;
        struct stat file_st;
        stat (thumbnail_path, &st);
        stat (file, &file_st);
        if(st.st_mtime > file_st.st_mtime) {
            TRACE ("PREVIEW thread: retrieving thumbnail for %s (%s)\n", file, thumbnail_path);
	    if (rfm_g_file_test(thumbnail_path, G_FILE_TEST_EXISTS)){
	        g_mutex_lock(pixbuf_mutex);
		src = gdk_pixbuf_new_from_file (thumbnail_path, NULL);
	        g_mutex_unlock(pixbuf_mutex);

	    }
            if(src) {
                g_free (thumbnail_path);
                return src;
            }
        }
    }
    TRACE ("PREVIEW thread: no thumbnail for %s (%s)\n", file, thumbnail_path);

    // Try the width at specified size
    if (rfm_g_file_test(file, G_FILE_TEST_EXISTS)){
	TRACE("file=%s at size %d\n", file, size);
	g_mutex_lock(pixbuf_mutex);
	src = gdk_pixbuf_new_from_file_at_size (file, size, -1, &error);
	g_mutex_unlock(pixbuf_mutex);
	if(error) {
	    TRACE ("%s:%s", error->message, file);
	    g_error_free (error);
	    error = NULL;
	    g_free (thumbnail_path);
	    return NULL;
	}
	gint ph = gdk_pixbuf_get_height (src);
	gint pw = gdk_pixbuf_get_width (src);
	if (ph > pw) {
	    // other way around
	    g_object_unref(src);
	    src = gdk_pixbuf_new_from_file_at_size (file, -1, size, &error);
	    ph = gdk_pixbuf_get_height (src);
	    pw = gdk_pixbuf_get_width (src);
	}
    }

    if(src) {
	save_thumbnail (thumbnail_path, file, src);
	TRACE ("save_thumbnail %s\n", thumbnail_path);
    }
    g_free (thumbnail_path);
    return src;
}




static
    void
init_pixbuf_hash () {
    if(!pixbuf_hash) {
        pixbuf_hash = g_hash_table_new (g_str_hash, g_str_equal);
    }
    if(!pixbuf_hash_mutex) {
        pixbuf_hash_mutex = g_mutex_new ();
    }
    return;
}

static void
update_pixbuf_f (gpointer key, gpointer value, gpointer data){
    pixbuf_t *pixbuf_p = value;
    if (g_path_is_absolute(pixbuf_p->path)){
	// don't touch thumbnails.
	return;
    }
    GdkPixbuf *pixbuf=get_pixbuf(pixbuf_p->mime_id, pixbuf_p->size, FALSE);
    if (GDK_IS_PIXBUF(pixbuf)) {
	    TRACE("updating pixbuf: %s (%d)\n", pixbuf_p->mime_id, pixbuf_p->size);
	    gint width=gdk_pixbuf_get_width(pixbuf_p->pixbuf);
	    gint height=gdk_pixbuf_get_height(pixbuf_p->pixbuf);
	    gdk_pixbuf_scale  (pixbuf, pixbuf_p->pixbuf,
			    0, 0,
                          width, height, 
			  0, 0, 
			  1.0, 1.0, 
			  GDK_INTERP_NEAREST);
                                                          
	    g_object_unref(pixbuf);
    }
}

static
pixbuf_t *
find_in_pixbuf_hash (const gchar * key, int size) {
    if (!key) return NULL;
    if (size > MAX_PIXBUF_HASH_SIZE) return NULL;
    init_pixbuf_hash ();
    gchar *hash_key = rfm_get_hash_key (key, size);
    //TRACE("HASH: find in hash: %s --> %s\n",key, hash_key);

    g_mutex_lock (pixbuf_hash_mutex);
    pixbuf_t *pixbuf_p = (pixbuf_hash)?g_hash_table_lookup (pixbuf_hash, hash_key):NULL;
    g_mutex_unlock (pixbuf_hash_mutex);
    // mime_id or NULL return immediately:
    if(!pixbuf_p) return NULL;
    //TRACE("found  %s --> %s in hash!\n",key, hash_key);
    if(!g_path_is_absolute (key))
        return pixbuf_p;
    // check if item is out of date
    if(rfm_g_file_test (key, G_FILE_TEST_EXISTS)) {
        struct stat st;
        stat (key, &st);
        if(pixbuf_p->mtime != st.st_mtime || 
		pixbuf_p->st_size != st.st_size||
		pixbuf_p->st_ino != st.st_ino)
       	{
	    // Obsolete item must be removed from pixbuf hash
	    // and eliminated from thumnail cache.
	    //
	    //Eliminate from thumbnail cache:
	    gchar *thumbnail_path = rfm_get_thumbnail_path (key, size);
	    if (g_file_test(thumbnail_path, G_FILE_TEST_EXISTS)) {
		if (unlink(thumbnail_path) < 0) {
		    g_warning("Cannot unlink thumbnail file: %s (%s)",
			thumbnail_path, strerror(errno));
		}
	    }
	    g_free (thumbnail_path);
	    // Eliminate from pixbuf hash:
            g_mutex_lock (pixbuf_hash_mutex);
            g_hash_table_steal (pixbuf_hash, hash_key);
            g_mutex_unlock (pixbuf_hash_mutex);
            g_free (pixbuf_p->path);
            g_object_unref (pixbuf_p->pixbuf);
            g_free (pixbuf_p);
            g_free (hash_key);
            return NULL;
        }
    }
    g_free (hash_key);
    return pixbuf_p;
}


static
GdkPixbuf *
get_pixbuf (const gchar * key, gint size, gboolean hash) {
    TRACE ("rfm_get_pixbuf(%s, %d)\n", key, size);
    if(!key) {
        TRACE ("rfm_get_pixbuf(): key is NULL!\n");
	return NULL;
    }
    // find in pixbuf hash
    pixbuf_t *pixbuf_p = (hash)?find_in_pixbuf_hash (key, size):NULL;
    if(pixbuf_p) {
        TRACE ("HASH-primary-icons.c: %s found in pixbuf hash\n", key);
        return pixbuf_p->pixbuf;
    }
    pixbuf_p = (pixbuf_t *) malloc (sizeof (pixbuf_t));
    memset (pixbuf_p, 0, sizeof (pixbuf_t));

    gchar *fullpath=NULL;
    if(g_path_is_absolute (key)) {
	if (rfm_g_file_test (key, G_FILE_TEST_EXISTS)){
	    fullpath=g_strdup(key);
	} else { 
	    g_free (pixbuf_p);
	    TRACE ("rfm_get_pixbuf(%s): file does not exist!", key);
	    return NULL;
	}
    } 
    if(fullpath) {
	TRACE ("rfm_get_pixbuf(%s) maps to %s\n", key, fullpath);
        pixbuf_p->pixbuf = create_preview (fullpath, size);
        if(!pixbuf_p->pixbuf) {
            g_free (pixbuf_p);
            TRACE ("create_preview(%s, %d): returns NULL!\n", fullpath, size);
            return NULL;
        }
        struct stat st;
        stat (fullpath, &st);
        pixbuf_p->mtime = st.st_mtime;
        pixbuf_p->st_size = st.st_size;
        pixbuf_p->st_ino = st.st_ino;
	g_free(fullpath);
    } else {
	TRACE ("doing make_pixbuf_from_id\n");
        pixbuf_p->pixbuf = make_pixbuf_from_id (key, size, hash);
        if(!pixbuf_p->pixbuf) {
            g_free (pixbuf_p);
            TRACE ("rfm_get_pixbuf(%s, %d): returns NULL!\n", key, size);
            return NULL;
        }
    }
    pixbuf_p->path = g_strdup (key);
    pixbuf_p->size = size;
    if(hash && pixbuf_hash && size <= MAX_PIXBUF_HASH_SIZE) {
        gchar *hash_key = rfm_get_hash_key (key, size);
        g_mutex_lock (pixbuf_hash_mutex);
	if (!G_IS_OBJECT (pixbuf_p->pixbuf)){
	    g_warning("!G_IS_OBJECT (pixbuf) at rfm_get_pixbuf()");
	}
        g_object_ref (pixbuf_p->pixbuf);
        TRACE ("HASH: inserting into pixbuf hash: %s --> %s\n", key, hash_key);
        g_hash_table_insert (pixbuf_hash, hash_key, pixbuf_p);
        g_mutex_unlock (pixbuf_hash_mutex);
    }
    return pixbuf_p->pixbuf;
}

