//
//
/*
 * Copyright (C) 2002-2011 Edscott Wilson Garcia
 * EMail: edscott@xfce.org
 *
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

// FIXME: this static variable is thread-unsafe. (though at the present
//        status of the code, this is irrelevant).
static int relative_sort_column = TYPE_SORT;

#define MODULE_STAT(x,y) GPOINTER_TO_INT(rfm_rational(PLUGIN_DIR, xfdir_p->view_p->module, x,y, "module_stat"))
#define MODULE_COUNT(x) GPOINTER_TO_INT(rfm_natural(PLUGIN_DIR, xfdir_p->view_p->module, x, "module_count"))
#define MODULE_DATA(x) (rfm_natural(PLUGIN_DIR, xfdir_p->view_p->module, x, "module_data"))
// check if monitor should wait:
/**
 * xfdir_monitor_skip:
 * @xfdir_p: pointer to an xfdir_t structure
 * Returns: TRUE if monitor should skip tests on current loop
 * or FALSE if monitor should go ahead with the update tests.
 * 
 * Checks whether the xfdir monitor should skip monitor tests on the current
 * loop or go ahead with monitor tests. It is an exported symbol so that modules
 * may use the same monitor algorithm as the xfdir monitor thread.
 **/
static
  gboolean
xfdir_monitor_skip (xfdir_t * xfdir_p) {
    static gboolean said = FALSE;
    gboolean skip = FALSE;
    if(xfdir_p == NULL) {
        g_error ("monitor_skip: xfdir_p ==NULL!");
    }
    view_t *view_p = xfdir_p->view_p;
    /*XXX view_p not set up yet */
    /*if (xfdir_p->active) {
       skip = TRUE;
       TRACE("thread monitor: xfdir_p->active\n");
       } */
    /* update in progress */

    if(view_p->mutexes.monitor_loop == NULL) {
	g_error("xfdir_monitor_skip(): monitor_loop_mutex == NULL");
    }
    g_mutex_lock (xfdir_p->view_p->mutexes.monitor_loop);
    if(xfdir_p->view_p->flags.refresh) {
        skip = TRUE;
        TRACE ("thread monitor is skipping: xfdir_p->refresh\n");
    }
    g_mutex_unlock (xfdir_p->view_p->mutexes.monitor_loop);

    if(view_p->mutexes.fireup_thread_manager == NULL) {
	g_error("xfdir_monitor_skip(): fireup_thread_manager_mutex == NULL");
    }
    g_mutex_lock (view_p->mutexes.fireup_thread_manager);
    // This is a traffic light. If the fireup thread manager
    // is running, the monitor will stop here to wait for the mutex.
    // The funtionality of this mutex lies not in what is inside
    // the mutex, but the ability to pause 
    TRACE ("thread monitor: fireup_thread_manager_mutex OK\n");
    g_mutex_unlock (view_p->mutexes.fireup_thread_manager);

    /* popup active? */
    GtkWidget *w = view_p->widgets.popup;
    if(w && gtk_widget_get_visible (w)) {
        skip = TRUE;
        TRACE ("thread monitor: popup visible\n");
    }
    // private (module) popup active?
    w =	g_object_get_data(G_OBJECT(view_p->widgets.paper), 
		"private_popup_widget");
    if(w && gtk_widget_get_visible (w)) {
        skip = TRUE;
        TRACE ("thread monitor: private_popup_widget visible\n");
    }

    /* entry window is mapped? */
    else if(view_p->widgets.rename) {
        skip = TRUE;
        TRACE ("thread monitor: entry window is mapped\n");
    }
    /* tip window is mapped? This could be disabled, but better be safe. */

    else if(view_p->tip_event.tooltip_active) {
        // But is it a valid active state?
	// State should be valid here since
	// condition is taken care of in rodent_mouse.c
	// with the on_leave signal
        TRACE ("thread monitor: widgets.tooltip_active\n");
        skip = TRUE;
    }
    /* button down? */
    else if(view_p->mouse_event.boxX != -1 && view_p->mouse_event.boxY != -1) {
        skip = TRUE;
        TRACE ("thread monitor: mouse_event.boxX != -1\n");
    }
    /* dragging? */
    else if(view_p->mouse_event.doing_drag_p)
        skip = TRUE;
    /* saving cache? */
    if(!skip && said) {
        said = FALSE;
        TRACE ("thread monitor: active\n");
    }
    if(!skip && said) {
        said = FALSE;
        TRACE ("thread monitor: active\n");
    }
    if(skip) {
        if(!said) {
            said = TRUE;
            TRACE ("thread monitor: sleeping\n");
        }
        rfm_threadwait ();
    }
    return skip;
}

/**
 * xfdir_free_data:
 * @xfdir_p:  pointer to an xfdir_t structure
 *
 *  Frees allocated resources of a loaded xfdir structure
 *
 **/
static
    int
xfdir_free_data (xfdir_t * xfdir_p) {

    if(!xfdir_p) {
        g_warning ("xfdir_p==NULL in xfdir_free_data");
        return -1;
    }
    if(xfdir_p->gl != NULL && xfdir_p->pathc) {
        int i;
        for(i = 0; i < xfdir_p->pathc; i++) {
            TRACE ("thread monitor: will free xfdir_p->gl[i].pathv=0x%lx\n", (long unsigned)xfdir_p->gl[i].pathv);
            g_free (xfdir_p->gl[i].pathv);
            TRACE ("thread monitor: will destroy xfdir_p->gl[i].en=0x%lx\n", (long unsigned)xfdir_p->gl[i].en);
            rfm_destroy_entry (xfdir_p->gl[i].en);
        }
        TRACE ("thread monitor: will free xfdir_p->gl=0x%lx\n", (long unsigned)xfdir_p->gl);
        g_free (xfdir_p->gl);
        xfdir_p->gl = NULL;
        xfdir_p->pathc = 0;
    }
    TRACE ("thread monitor: will free xfdir_p->data=0x%lx\n", (long unsigned)xfdir_p->data);
    g_free (xfdir_p->data);
    return 0;
}

static
  gboolean
exit_monitor (xfdir_t * xfdir_p) {
    gboolean result = FALSE;
    g_mutex_lock (xfdir_p->view_p->mutexes.monitor_id);
    if(xfdir_p->view_p->flags.monitor_id != xfdir_p->monitor_id)
        result = TRUE;
    g_mutex_unlock (xfdir_p->view_p->mutexes.monitor_id);
    return result;
}

/**
 * xfdir_monitor_continue:
 * @xfdir_p:  pointer to an xfdir_t structure
 * Returns: TRUE if monitor should continue alive or FALSE if monitor should 
 * finish.
 *
 * This function is a generic check whether the local file monitor thread
 * should continue alive or exit. It is an exported symbol so that modules
 * may use the same monitor algorithm as the xfdir monitor thread.
 **/
static
  gboolean
xfdir_monitor_continue (xfdir_t * xfdir_p) {
    if(!xfdir_p) {
        TRACE ("thread monitor: monitor_continue !xfdir_p\n");
        return FALSE;
    }
    /* signaled exit: */
    if(exit_monitor (xfdir_p)) {
        TRACE ("thread monitor: exit_monitor(xfdir_p) = 1\n");
        return FALSE;
    }
    /* invalid view: */
    if(!xfdir_p->view_p)
        return FALSE;
    /* root window */
    if(!xfdir_p->view_p->en) {
        TRACE ("thread monitor: exit on !xfdir_p->view_p->en\n");
        return FALSE;
    }
    /* directory has vanished */
    if(xfdir_p->view_p->module==NULL) {
	if(xfdir_p->path == NULL || !rfm_g_file_test (xfdir_p->path, G_FILE_TEST_EXISTS)) {
	    // this is the test for local files.
	    TRACE ("monitor: !rfm_g_file_test(view_p->xfdir_p->path,G_FILE_TEST_EXISTS)\n");
	    return FALSE;
	}
    } else {// a module defined test...
    } 
    return TRUE;
}

static off_t
st_sum (struct stat *st) {
    off_t sum;
    if(!st)
        sum = 0;
    else {
        sum = st->st_mtime + st->st_size + st->st_mode + st->st_nlink + st->st_uid + st->st_gid;
    }
    return sum;
}

static off_t
st_sum_dir (struct stat *st) {
    off_t sum;
    int build=atoi(BUILD);
    if(!st){
        sum = 0;
    } else {
        sum = st->st_ino + st->st_mtime + st->st_ctime + st->st_size + st->st_mode + st->st_nlink + st->st_uid + st->st_gid + build;
    }
    return sum;
}

static
    void
set_sort_column (int column) {
    relative_sort_column = column;
}

static
int update_xfdir (xfdir_t * xfdir_p, gboolean fullbones);

static
    void
set_deepest_dir (view_t * view_p) {
    if(view_p->en && view_p->en->path && rfm_g_file_test (view_p->en->path, G_FILE_TEST_IS_DIR)) {
        if(!view_p->deepest_dir || !strstr (view_p->deepest_dir, view_p->en->path)) {
            g_free (view_p->deepest_dir);
            view_p->deepest_dir = g_strdup_printf ("%s/.", view_p->en->path);
        }
    } else {
        //g_warning("cannot set deepest dir");
    }
}

// Two possible races with other processes:
//  1. There are more files present than those counted
//  2. There are less files present than those counted
static gint
read_files_local (xfdir_t * xfdir, gboolean fullread, gboolean fullbones) {
    DIR *directory;
    gint count = 0;
//    gint smallcount = 0;
    struct dirent *d;

    directory = opendir (xfdir->path);
    if(!directory){
        return count;
    }

    TRACE("SHOWS_HIDDEN (xfdir->type)=%d\n", SHOWS_HIDDEN (xfdir->type));

    while((d = readdir (directory)) != NULL) {
	if (xfdir->limited_read && count >= 250) break;
        gchar *fullpath = NULL;
        gchar *basepath = NULL;
        if(strcmp (d->d_name, ".") == 0) continue;
        if(!SHOWS_HIDDEN (xfdir->type) && d->d_name[0] == '.' && strcmp (d->d_name, "..") != 0) continue;
        if(!fullread || xfdir == NULL) {// only counting files...
            count++;
	} else {   
	    if(strcmp (d->d_name, "..") == 0){
		fullpath = g_path_get_dirname (xfdir->path);
		basepath = g_path_get_basename (fullpath);
                if(strcmp (xfdir->path, "/") == 0) {
                    xfdir->gl[count].en = NULL;
                    xfdir->gl[count].pathv = g_strdup (g_get_host_name ());
                } else {
		    // up icon will be stated even in barebones mode
		    xfdir->gl[count].en = rfm_stat_entry (fullpath, 0);
                    xfdir->gl[count].pathv = basepath;
                    xfdir->gl[count].en->path = fullpath;
                    SET_DUMMY_TYPE (xfdir->gl[count].en->type);
                    SET_UP_TYPE (xfdir->gl[count].en->type);
                }
            } else {
		fullpath = g_build_filename (xfdir->path, d->d_name, NULL);
		basepath = g_strdup(d->d_name);
		xfdir->gl[count].en = rfm_stat_entry (fullpath, xfdir->type);
		xfdir->gl[count].en->path = fullpath;
		switch (d->d_type){
		  case DT_BLK:
		    //This is a block device.
		    SET_SBLK(xfdir->gl[count].en->type);
		    break;
		  case DT_CHR:
		    //This is a character device.
		    SET_SCHR(xfdir->gl[count].en->type);
		    break;
		  case DT_DIR:
		    //This is a directory.
		    SET_SDIR(xfdir->gl[count].en->type);
		    break;
		  case DT_FIFO:
		    //This is a named pipe (FIFO).
		    SET_SFIFO(xfdir->gl[count].en->type);
		    break;
		  case DT_LNK:
		    //This is a symbolic link.
		    SET_SLNK(xfdir->gl[count].en->type);
		    break;
		  case DT_REG:
		    //This is a regular file.
		    SET_SREG(xfdir->gl[count].en->type);
		    break;
		  case DT_SOCK:
		    //This is a Unix domain socket.
		    SET_SSOCK(xfdir->gl[count].en->type);
		    break;
		  case DT_UNKNOWN:
		    //The file type is unknown.
		    break;			
		}
                xfdir->gl[count].pathv = basepath;
            }
            count++;
            // ensure that we only read files for which we 
	    // have enough malloc'd memory (race condition 
	    // safeguard, since this is a two step mechanism):
            if(count >= xfdir->pathc) {
                break;
            }
        } 
    }
    closedir (directory);
    // ignore malloc'd space for files which have disappeared between count and read steps:
    if(fullread && xfdir != NULL && count < xfdir->pathc) {
        xfdir->pathc = count;
    }
    return (count);
}

// this function checks if a directory (or module) currently in 
// xfdir structure needs to be updated in the view (in other words,
// is a new xfdir structure necessary?)
static
  gboolean
reload_condition (xfdir_t * xfdir_p, gint * current_count, struct stat *current_st) {
    gboolean instant_return = FALSE;

    // Test 1. Short circuit on instruction from another thread.
    //
    // put a lock on xfdir structure and use the reload variable
    // to shortcircuit test on condition set by other threads
    g_mutex_lock (xfdir_p->view_p->mutexes.monitor_loop);
    instant_return = xfdir_p->reload;
    xfdir_p->reload = FALSE;
    g_mutex_unlock (xfdir_p->view_p->mutexes.monitor_loop);
    if(instant_return){
	TRACE("reload_condition: instant_return\n");
        return TRUE;
    }

    // Test 2. inode count condition.
    // This will count the number of icons in the view.
    gint instant_count=0;
    if(xfdir_p->view_p->module) {
        TRACE ("1. xfdir_p->view_p->module\n"); 
        // module defined count test.
        instant_count = MODULE_COUNT (xfdir_p);
	TRACE ("MODULE_COUNT current count= %d, instant=%d\n", *current_count, instant_count);
    } else {
        // local directory test 
        instant_count = read_files_local (xfdir_p, FALSE, FALSE);
    }
    TRACE ("MONITOR current count= %d, instant=%d\n", *current_count, instant_count);
    if(*current_count != instant_count) {
        TRACE ("reload condition true on COUNT condition %d != %d\n", 
		*current_count, instant_count);
	// We update the current count, otherwise the reload condition
	// will not cease, and it must cease to trigger the actual reload.
	*current_count=instant_count;
        return TRUE;
    }

    // Test 3. Check stat information of current view.
    struct stat instant_st;
    if(xfdir_p->view_p->module) {
        TRACE ("2. xfdir_p->view_p->module\n"); 
	memset(&instant_st, 0, sizeof(struct stat));
        // module defined test.
        TRACE ("2.1. xfdir_p->view_p->module\n"); 
        if(MODULE_STAT (NULL, NULL) && !MODULE_STAT (NULL, &instant_st)) {
	    TRACE ("reload_condition(): module_stat not correctly defined for %s\n", 
		    xfdir_p->view_p->module);
        }
	TRACE("MODULE_STAT current= %d, instant=%d\n",
	    (gint)st_sum_dir (current_st), (gint)st_sum_dir (&instant_st));
        TRACE ("2.2. xfdir_p->view_p->module\n"); 
    } else {
        // local file test
        stat (xfdir_p->path, &instant_st);
    }
    TRACE("current= %d, instant=%d\n",
	    (gint)st_sum_dir (current_st), (gint)st_sum_dir (&instant_st));

    if(current_st && st_sum_dir (current_st) != st_sum_dir (&instant_st)) {
 	// We update the current stat, otherwise the reload condition
	// will not cease, and it must cease to trigger the actual reload.
        TRACE ("MONITOR: reload condition true on STAT condition\n");
        memcpy (current_st, &instant_st, sizeof (struct stat));
        return TRUE;
    }

    // Test 4. Module specific test: comparison of character data 
    // 	       in xfdir structure.
    if(xfdir_p->view_p->module && xfdir_p->data) {
        TRACE ("3. xfdir_p->view_p->module\n"); 
        void *data = MODULE_DATA (xfdir_p);
        int different = strcmp ((char *)(xfdir_p->data), (char *)data);
        if(different) {
            g_free (xfdir_p->data);
            xfdir_p->data = data;
            TRACE ("MONITOR: reload condition true on DATA condition\n");
            return TRUE;
        }
        g_free (data);
    }

    return FALSE;
}

// If a full reload is not necessary, this function just updates
// stat information and thumbnails.
static gboolean
update_stat_info (xfdir_t * xfdir_p) {
    int i;
    view_t *view_p = xfdir_p->view_p;
    gboolean something_changed=FALSE;


    // We skip first element (which is the dummy "up" icon).
    for(i = 1; i < xfdir_p->pathc; i++) {
	// Is the work still necessary or has the thread met an exit condition?
        if(exit_monitor (xfdir_p)){
            return FALSE;
	}
	// This is a safeguard, it really should not happen though.
        if(xfdir_p->gl[i].en == NULL){
            continue;
	}

        struct stat st;
	{
            // local files test
            if (stat (xfdir_p->gl[i].en->path, &st) < 0 
		    &&
		lstat (xfdir_p->gl[i].en->path, &st) < 0) {
		// Failed stat instruction means that the file count test
		// (which passed by now) is incorrect and does not reflect
		// the actual directory condition.
		TRACE ("local files test failed stat on: %s\n", xfdir_p->gl[i].en->path);
		return TRUE;
	    }
	    // if its a directory, check mount status
	    // (greenball updates)
	    /*gboolean isofs=
		(xfdir_p->gl[i].en->mimetype && 
		 strstr(xfdir_p->gl[i].en->mimetype, "application/x-iso9660-image")) ||
		(xfdir_p->gl[i].en->mimemagic && 
		 strstr(xfdir_p->gl[i].en->mimemagic, "application/x-iso9660-image"));*/
	    if (IS_SDIR(xfdir_p->gl[i].en->type) ||
		 IS_SBLK(xfdir_p->gl[i].en->type) 
		// || isofs
	       ){
		if (IS_MOUNTED_TYPE(xfdir_p->gl[i].en->type) && 
			!FSTAB_is_mounted(xfdir_p->gl[i].en->path)){
		    UNSET_MOUNTED_TYPE(xfdir_p->gl[i].en->type);
		    memset(xfdir_p->gl[i].en->st, 0, sizeof(struct stat));
		    TRACE("%s is now unmounted\n",xfdir_p->gl[i].en->path); 
		} else if (!IS_MOUNTED_TYPE(xfdir_p->gl[i].en->type) && 
			FSTAB_is_mounted(xfdir_p->gl[i].en->path)){
		    SET_MOUNTED_TYPE(xfdir_p->gl[i].en->type);
		    memset(xfdir_p->gl[i].en->st, 0, sizeof(struct stat));
		    TRACE("%s is now mounted\n",xfdir_p->gl[i].en->path); 
		}

	    }
        }
	
	// This is the stat comparison test.
        if(st_sum (xfdir_p->gl[i].en->st) != st_sum (&st)) {
	    // Stat information update is necessary.
            TRACE ("POP_SEM:  stat differs for %s (%ld != %ld\n",
                 xfdir_p->gl[i].en->path, 
		 (long)st_sum (xfdir_p->gl[i].en->st), 
		 (long)st_sum (&st));
	    // Step 1. update stat information in xfdir structure.
            memcpy (xfdir_p->gl[i].en->st, &st, sizeof (struct stat));
            TRACE ("POP_SEM:  update_stat_info lock..\n");
	    // Step 2. update stat information in corresponding view population item.
            // First get the population semaphore. A simple read lock is OK since
	    // the view_p->population_pp is not affected.
            rfm_population_read_lock (view_p);
	    /*This is borked. Green ball will not
	     * appear if trylock is used.
	     if(!rfm_population_try_read_lock (view_p)){
		TRACE("cannot get readlock\n");
                continue;
	    }*/

            if(exit_monitor (xfdir_p)) {
		// Now, is the thread doing required work or not?
		// Condition may have changed after the read lock wait.
		TRACE("exit_monitor condition\n");
                rfm_population_read_unlock (view_p);
                return FALSE;
            }

            // Do the update on current view population.
            population_t **tmp;
            for(tmp = view_p->population_pp; *tmp; tmp++) {
                population_t *population_p = *tmp;
		gboolean found =  population_p->en
			       && population_p->en->path
			       && strcmp (population_p->en->path, xfdir_p->gl[i].en->path) == 0;
                if(found) {
		    something_changed=TRUE; 
                    TRACE ("POP_SEM: found population_p for %s\n", xfdir_p->gl[i].en->path);
		    // Update current population item stat information.
                    memcpy (population_p->en->st, &st, sizeof (struct stat));
		    // Update current population item stat information, along
		    // with stat dependend binary flags:
		    //rfm_stat_entry_update(population_p->en, population_p->en->path); 
		    // Are we in the list view?
		    // If so we will need to update the displayed stat information.
		    if (view_p->view_layout.icon_size < TINY_ICON_SIZE/2) {
			// recreate population layout
			rfm_xfdir_make_layout(view_p, population_p);
			// Expose layout (threaded)
			GdkRectangle rect;
			rfm_get_population_label_rect(view_p, population_p, &rect);
			rfm_thread_expose_rect (gtk_widget_get_window(view_p->widgets.window), &rect);
		    }
		    // Does the referred item have a thumbnail?
                    if(population_p->flags & POPULATION_THUMBNAILED) {
                        // It does have a thumbnail. This should be redone.
			population_p->flags &= (POPULATION_THUMBNAILED ^ 0xffffffff);
                        TRACE ("THREAD: thread(rfm_preview_thread_f) at primary-xfdir.i\n");
                        // First we invalidate preview popup so that it will be
			// regenerated the next time the user triggers a tooltip preview.
                        population_p->preview_pixbuf_timestamp = 0;
			// Now assign memory block which is to be used by the thumbnail
			// preview thread.
			fireup_thread_manager_t *fireup_thread_manager_p = 
			    (fireup_thread_manager_t *)malloc(sizeof(fireup_thread_manager_t));
			memset(fireup_thread_manager_p, 0, sizeof(fireup_thread_manager_t));
			fireup_thread_manager_p->population_p = population_p;
			fireup_thread_manager_p->flags = population_p->flags;
			fireup_thread_manager_p->path = g_strdup(population_p->en->path);
			fireup_thread_manager_p->preview_manager_v.view_p = view_p;
			fireup_thread_manager_p->preview_manager_v.population_serial = 
			    view_p->population_serial;
			fireup_thread_manager_p->preview_manager_v.diagnostics = 
			    view_p->widgets.diagnostics;
			fireup_thread_manager_p->preview_manager_v.diagnostics_window = 
			    view_p->widgets.diagnostics_window;

			// Fire up preview thread and forget (this will expose).
			TRACE("stat change thumbnailing: %s\n", population_p->en->path);
                        g_thread_create (rfm_preview_thread_f, (gpointer) fireup_thread_manager_p, FALSE, NULL);
			// Exit the item in population search.
			break; 
                    } else { 
			// update the icon pixmap, for directories so that
			// greenball dissappears if unmounted or appears
			// if mounted. 
			if (IS_SDIR(population_p->en->type) ||
			    IS_SBLK(population_p->en->type)){
			    TRACE("greenball update now\n");
			    if (FSTAB_is_mounted(population_p->en->path)){
				SET_MOUNTED_TYPE(population_p->en->type);
			    } else {
				UNSET_MOUNTED_TYPE(population_p->en->type);
			    }
			    gchar *icon_id = 
				rfm_get_composite_icon_id (population_p->en, NULL); 
			    GdkPixbuf *pixbuf = 
				rfm_get_pixbuf (icon_id, REDUCED_ICONSIZE(view_p));
			    g_free (icon_id);
			    population_p->pixbuf = pixbuf;
			    GdkRectangle rect;
			    rfm_get_population_rect(view_p, population_p, &rect);
			    rfm_thread_expose_rect (gtk_widget_get_window(view_p->widgets.window), &rect);
			}
		    }
                }
            } // This is the end of the item in population search.
            // Release the population read lock.
            rfm_population_read_unlock (view_p);
	    // directory is updated
        } // This is the end of the stat comparison test.
    } // This is the end of the stat comparison loop on all items of xfdir structure
    if (something_changed && 
	    !(relative_sort_column == NAME_SORT ||relative_sort_column == TYPE_SORT)) 
    {
	// we need to resort the icons
	// FIXME: should do this in a soft way
	//      i.e., sort the population_pp and expose
	// returning TRUE is overkill, because it is a non necessary
	// soft reload (and it doesn't even work!)
	//return TRUE;
    }
	
    return FALSE;
}

static gint
pathv_compare_up (const gchar * a, const gchar * b) {
    if(!a && !b)
        return 0;
    if(!a)
        return -1;
    if(!b)
        return 1;
    if(strchr (a, G_DIR_SEPARATOR))
        a = strrchr (a, G_DIR_SEPARATOR) + 1;
    if(strchr (b, G_DIR_SEPARATOR))
        b = strrchr (b, G_DIR_SEPARATOR) + 1;
    return strcmp (a, b);
}

static gint
pathv_compare_down (const gchar * a, const gchar * b) {
    if(!a && !b)
        return 0;
    if(!a)
        return 1;
    if(!b)
        return -1;
    if(strchr (a, G_DIR_SEPARATOR))
        a = strrchr (a, G_DIR_SEPARATOR) + 1;
    if(strchr (b, G_DIR_SEPARATOR))
        b = strrchr (b, G_DIR_SEPARATOR) + 1;
    return strcmp (a, b);
}

static
  gint
entry_compare_up (int caso, record_entry_t * en_a, record_entry_t * en_b) {
    gboolean d1d,
      d2d;                      /* d1, d2 is directory flag */
    if(!en_a && !en_b)
        return 0;
    if(!en_a)
        return 1;
    if(!en_b)
        return -1;
    if(IS_DUMMY_TYPE (en_a->type))
        return 1;
    if(IS_DUMMY_TYPE (en_b->type))
        return -1;

    if(!en_a->path && !en_b->path)
        return 0;
    if(!en_a->path)
        return -1;
    if(!en_b->path)
        return 1;

    d1d = IS_SDIR (en_a->type);
    d2d = IS_SDIR (en_b->type);

    if(d1d && !d2d)
        return 1;
    if(!d1d && d2d)
        return -1;

    /* subsorting... */
    if(caso == TYPE_SORT && IS_FILE (en_a->type) && IS_FILE (en_b->type)) {
        char *a,
         *b;
        a = strrchr (en_a->path, '.');
        b = strrchr (en_b->path, '.');
        if(a || b) {
            if(!a)
                return -1;
            if(!b)
                return 1;
            if(strcmp (a, b))
                return strcmp (a, b);
        }
    }

    switch (caso) {
    case TYPE_SORT:
    case NAME_SORT:
        {
            gchar *a = en_a->path,
                *b = en_b->path;
            if(strchr (a, G_DIR_SEPARATOR))
                a = strrchr (a, G_DIR_SEPARATOR) + 1;
            if(strchr (b, G_DIR_SEPARATOR))
                b = strrchr (b, G_DIR_SEPARATOR) + 1;
            return strcmp (a, b);
        }
    default:
        if(!en_a->st && !en_b->st)
            return 0;
        if(!en_a->st)
            return -1;
        if(!en_b->st)
            return 1;
        switch (caso) {
        case SIZE_SORT:
            return en_a->st->st_size - en_b->st->st_size;
        case DATE_SORT:
            return en_a->st->st_mtime - en_b->st->st_mtime;
        case OWNER_SORT:
            return en_a->st->st_uid - en_b->st->st_uid;
        case GROUP_SORT:
            return en_a->st->st_gid - en_b->st->st_gid;
        case MODE_SORT:
            return en_a->st->st_mode - en_b->st->st_mode;
        }
    }
    return 0;
}

static
  gint
entry_compare_down (int caso, record_entry_t * en_a, record_entry_t * en_b) {
    gboolean d1d,
      d2d;                      /* d1, d2 is directory flag */
    if(!en_a && !en_b)
        return 0;
    if(!en_a)
        return -1;
    if(!en_b)
        return 1;
    if(IS_DUMMY_TYPE (en_a->type))
        return -1;
    if(IS_DUMMY_TYPE (en_b->type))
        return 1;

    if(!en_a->path && !en_b->path)
        return 0;
    if(!en_a->path)
        return 1;
    if(!en_b->path)
        return -1;

    d1d = IS_SDIR (en_a->type);
    d2d = IS_SDIR (en_b->type);

    if(d1d && !d2d)
        return -1;
    if(!d1d && d2d)
        return 1;

    /* subsorting... */
    if(caso == TYPE_SORT && IS_FILE (en_a->type)
       && IS_FILE (en_b->type)) {
        char *a,
         *b;
        a = strrchr (en_a->path, '.');
        b = strrchr (en_b->path, '.');
        if(a || b) {
            if(!a)
                return -1;
            if(!b)
                return 1;
            if(strcmp (a, b))
                return strcmp (a, b);
        }
    }

    switch (caso) {
    case TYPE_SORT:
    case NAME_SORT:
        {
            gchar *a = en_a->path,
                *b = en_b->path;
            if(strchr (a, G_DIR_SEPARATOR))
                a = strrchr (a, G_DIR_SEPARATOR) + 1;
            if(strchr (b, G_DIR_SEPARATOR))
                b = strrchr (b, G_DIR_SEPARATOR) + 1;
            return strcmp (a, b);
        }
    default:
        if(!en_a->st && !en_b->st)
            return 0;
        if(!en_a->st)
            return 1;
        if(!en_b->st)
            return -1;
        switch (caso) {
        case SIZE_SORT:
            return en_a->st->st_size - en_b->st->st_size;
        case DATE_SORT:
            return en_a->st->st_mtime - en_b->st->st_mtime;
        case OWNER_SORT:
            return en_a->st->st_uid - en_b->st->st_uid;
        case GROUP_SORT:
            return en_a->st->st_gid - en_b->st->st_gid;
        case MODE_SORT:
            return en_a->st->st_mode - en_b->st->st_mode;
        }
    }
    return 0;
}

static
    int
xfdir_compare_up (const void *a, const void *b) {
    int result;
    const dir_t *d1 = (const dir_t *)a;
    const dir_t *d2 = (const dir_t *)b;
    if(!d1 && !d2)
        return 0;
    if(!d1)
        return -1;
    if(!d2)
        return 1;

    /* we sort here by displayed name, not by the path's basename,
     * which is what entry compare will do... */

    if(relative_sort_column == NAME_SORT) {
        if(!d1->en || IS_DUMMY_TYPE (d1->en->type))
            return 1;
        if(!d2->en || IS_DUMMY_TYPE (d2->en->type))
            return -1;
        if(d1->en && d2->en) {  /* directories on top */
            gboolean d1d = IS_SDIR (d1->en->type);
            gboolean d2d = IS_SDIR (d2->en->type);
            if(d1d && !d2d)
                return -1;
            if(!d1d && d2d)
                return 1;
        }

        result = pathv_compare_up (d1->pathv, d2->pathv);
        if(result)
            return result;
        /* break name ties with date: */
        return (entry_compare_up (DATE_SORT, d1->en, d2->en));
    }
    result = entry_compare_up (relative_sort_column, d1->en, d2->en);
    if(result)
        return result;
    /* break other ties with name */
    return pathv_compare_up (d1->pathv, d2->pathv);

    /*return (entry_compare(relative_sort_column,d1->en,d2->en)); */
}

static
    int
xfdir_compare_down (const void *a, const void *b) {
    int result;
    const dir_t *d1 = (const dir_t *)a;
    const dir_t *d2 = (const dir_t *)b;
    if(!d1 && !d2)
        return 0;
    if(!d1)
        return -1;
    if(!d2)
        return 1;

    /* we sort here by displayed name, not by the path's basename,
     * which is what entry compare will do... */

    if(relative_sort_column == NAME_SORT) {
        if(!d1->en || IS_DUMMY_TYPE (d1->en->type))
            return -1;
        if(!d2->en || IS_DUMMY_TYPE (d2->en->type))
            return 1;
        if(d1->en && d2->en) {  /* directories on top */
            gboolean d1d = IS_SDIR (d1->en->type);
            gboolean d2d = IS_SDIR (d2->en->type);
            if(d1d && !d2d)
                return -1;
            if(!d1d && d2d)
                return 1;
        }

        result = pathv_compare_down (d1->pathv, d2->pathv);
        if(result)
            return result;
        /* break name ties with date: */
        return (entry_compare_down (DATE_SORT, d1->en, d2->en));
    }
    result = entry_compare_down (relative_sort_column, d1->en, d2->en);
    if(result)
        return result;
    /* break other ties with name */
    return pathv_compare_down (d1->pathv, d2->pathv);

    /*return (entry_compare(relative_sort_column,d1->en,d2->en)); */
}

static
    int
xfdir_compare_default (const void *a, const void *b) {
    int result;
    const dir_t *d1 = (const dir_t *)a;
    const dir_t *d2 = (const dir_t *)b;
    if(!d1 && !d2)
        return 0;
    if(!d1)
        return -1;
    if(!d2)
        return 1;

    {
        /* equivalent to no sort (treeview uses 3 states:
         * no sort is ascending by name and
         * subsorted by filetype).*/
        d2 = (const dir_t *)a;
        d1 = (const dir_t *)b;
    }
    /* we sort here by displayed name, not by the path's basename,
     * which is what entry compare will do... */

    if(relative_sort_column == NAME_SORT) {
        if(!d1->en || IS_DUMMY_TYPE (d1->en->type))
            return 1;
        if(!d2->en || IS_DUMMY_TYPE (d2->en->type))
            return -1;
        if(d1->en && d2->en) {  /* directories on top */
            gboolean d1d = IS_SDIR (d1->en->type);
            gboolean d2d = IS_SDIR (d2->en->type);
            if(d1d && !d2d)
                return 1;
            if(!d1d && d2d)
                return -1;
        }

        result = pathv_compare_up (d1->pathv, d2->pathv);
        if(result)
            return result;
        /* break name ties with date: */
        return (entry_compare_up (DATE_SORT, d1->en, d2->en));
    }
    result = entry_compare_up (relative_sort_column, d1->en, d2->en);
    if(result)
        return result;
    /* break other ties with name */
    return pathv_compare_up (d1->pathv, d2->pathv);

    /*return (entry_compare(relative_sort_column,d1->en,d2->en)); */
}

/**
 * xfdir_sort:
 * @xfdir_p: pointer to an xfdir_t structure
 *
 * Sorts an xfdir_t structure according to the settings in xfdir_p->ascending 
 * and xfdir_p->sort_column. Ascending is TRUE or FALSE and sort_column may be: 
 * 
 * TYPE_SORT, NAME_SORT, SIZE_SORT,
 * DATE_SORT, OWNER_SORT, GROUP_SORT, MODE_SORT.
 * <para>
 * qsort library routine is used as a sorting method.
 * </para>
 * 
 **/
static void
xfdir_sort (xfdir_t * xfdir_p) {
    TRACE ("reload_iconview: do quicksort here: sortcolumn is %d, ascending=%d\n", xfdir_p->sort_column, xfdir_p->ascending);
    TRACE ("xfdir_p=0x%lx,pathc is %ld\n", (unsigned long)xfdir_p, (long)xfdir_p->pathc);

    set_sort_column (xfdir_p->sort_column);

    switch (xfdir_p->ascending) {
    case 0:
        qsort ((void *)xfdir_p->gl, xfdir_p->pathc, sizeof (dir_t),
#ifdef __COMPAR_FN_T
               (__compar_fn_t)
#endif
               xfdir_compare_down);
        break;
    case 1:
        qsort ((void *)xfdir_p->gl, xfdir_p->pathc, sizeof (dir_t),
#ifdef __COMPAR_FN_T
               (__compar_fn_t)
#endif
               xfdir_compare_up);
        break;
    default:
        qsort ((void *)xfdir_p->gl, xfdir_p->pathc, sizeof (dir_t),
#ifdef __COMPAR_FN_T
               (__compar_fn_t)
#endif
               xfdir_compare_default);
        break;
    }
}


// function to do a background reload and set things up for the update
static xfdir_t *
immediate_reload_action (xfdir_t *xfdir_p){
    // allocate memory for reloaded xfdir structure
    xfdir_t *new_xfdir_p = 
	new_xfdir_p = (xfdir_t *) malloc (sizeof (xfdir_t));

    TRACE ("monitor new_xfdir_p=0x%lx\n", (long unsigned)new_xfdir_p);

    // initialize new xfdir structure with current xfdir structure information
    memcpy (new_xfdir_p, xfdir_p, sizeof (xfdir_t));

    // not necesary: new_xfdir_p->view_p = xfdir_p->view_p;
    //
    // new xfdir structure specific initializations:
    new_xfdir_p->pathc = 0;
    new_xfdir_p->gl = NULL;
    new_xfdir_p->xfdir_p = NULL;

    // reload directory (or module) into new xfdir structure
    // Since this is background thread, fullbones is best.
    update_xfdir (new_xfdir_p, TRUE);

    // sort directory (or module) using current user settings
    xfdir_sort (new_xfdir_p);

    // safeguard check: if updated xfdir structure is no longer
    // required, discard updated directory 
    if(exit_monitor (xfdir_p)) {
	TRACE ("monitor now exiting and dumping new_xfdir_p!\n");
	// rodent no longer needs this.
	xfdir_free_data (new_xfdir_p);
	g_free (new_xfdir_p);
	TRACE ("monitor: id= %d is now defunct (updated xfdir discarded)\n",
		xfdir_p->monitor_id);
	return xfdir_p;
    }
    
    // this mutex is for working on the current xfdir structure
    // without confusing other threads which may require access
    // to the xfdir structure. multiple view_p in a single rodent instance
    // (XXX actually, I wonder if it is really necessary...)
    g_mutex_lock (xfdir_p->view_p->mutexes.monitor_loop);
    // clean up xfdir playground
    if(xfdir_p->xfdir_p) {
	TRACE ("monitor freeing xfdir_p->xfdir_p=0x%lx\n", (long unsigned)xfdir_p->xfdir_p);
	xfdir_free_data (xfdir_p->xfdir_p);
	g_free (xfdir_p->xfdir_p);
	xfdir_p->xfdir_p = NULL;
    }

    // save a pointer to current xfdir structure
    new_xfdir_p->xfdir_p = xfdir_p;
    // mutex end.
    g_mutex_unlock (xfdir_p->view_p->mutexes.monitor_loop);

    return new_xfdir_p;
}


// thread_monitor_f is in charge of xfdir_p cleanup!
// AND view_p cleanup!
static gpointer
thread_monitor_f (gpointer data) {
    gboolean disabled = FALSE;

    // memory structure should be received from parent thread
    xfdir_t *xfdir_p = (xfdir_t *) data;
    if(xfdir_p == NULL || !xfdir_p->path) {
        return NULL;
    }
    g_mutex_lock(xfdir_p->view_p->mutexes.janitor);
	xfdir_p->view_p->janitor.monitors++;
    g_mutex_unlock(xfdir_p->view_p->mutexes.janitor);
	
    rfm_barrier_stop_or_increment();

    // This is the main monitor loop. It is the event loop
    // which controls the monitor thread (outer loop)
     do { // outer loop (1)      
	TRACE("outer loop\n");
        gboolean refresh_pending = FALSE;
        gboolean refresh_action = FALSE;

	// count items present in directory (or module)
        gint current_count=0;
        if(xfdir_p->view_p->module) {
            current_count = MODULE_COUNT (xfdir_p);
	    TRACE("module current_count=%d\n",current_count);
        } else {
	    // simple file count may prove buggy
	     current_count = read_files_local (xfdir_p, FALSE, FALSE);
	    // maybe we do an inode count, which is more complete
	    // (sum the inodes of all files)
	    //current_count = inode_count();
        }

	// check whether stat condition has changed on directory (or module)
        struct stat current_st;
        if(xfdir_p->view_p->module) {
	    memset(&current_st, 0, sizeof(struct stat));
            if(MODULE_STAT (NULL, NULL) && !MODULE_STAT (NULL, &current_st)) {
                // if the module stat function is broken, enter suspend mode
		// until event loop is broken on a reload condition
            g_warning ("thread_monitor_f(): module_stat not correctly defined for %s\n", 
		    xfdir_p->view_p->module);
                disabled = TRUE;
            }
        } else {
            stat (xfdir_p->path, &current_st);
        }

	// Setup to enter inner loop is complete. For this we have a count of 
	// items which should be present (inode count) and a stat structure 
	// with current information.
        TRACE ("monitor %d setup for %s\n", xfdir_p->monitor_id, xfdir_p->path);
        do { // inner loop (2)
	    // decrement monitor count to allow barrier
	    //  (value has been incremented on entering function
	    //   and will be incremented on repeating this loop) 
	    rfm_barrier_decrement();
	    // if barrier is requested from another thread, this will block
	    // otherwise barrier will be disabled.
	    rfm_barrier_stop_or_increment();

	    // check if the inner event loop must be broken
	    // this happens if monitor thread will exit or a reload condition
	    // has been detected
	    if(!xfdir_monitor_continue (xfdir_p)) {
		TRACE("!xfdir_monitor_continue\n");
                rfm_threadwait ();
                break;
            }
            // check if the event loop must be placed in suspend mode
	    if(disabled || xfdir_monitor_skip (xfdir_p)) {
		TRACE("disabled || xfdir_monitor_skip (xfdir_p)\n");
                rfm_threadwait ();
                continue;
            }
	    if(xfdir_p->view_p->module) {
		// This asks the module if an unconditional refresh is pending
		TRACE("checking for module monitor_refresh_action\n");

		if (rfm_natural(PLUGIN_DIR,xfdir_p->view_p->module, 
			    &(xfdir_p->view_p->widgets),
			    "monitor_refresh_action")){
		    TRACE("monitor_refresh_action is TRUE\n"); 
		    // since conditions may have changed during the call,
		    // xfdir_monitor_skip should follow
		    if (refresh_pending) refresh_action=TRUE;
		    else refresh_pending = TRUE;
		} 
	    }
            // check if the event loop must be placed in suspend mode
	    if(disabled || xfdir_monitor_skip (xfdir_p)) {
		TRACE("disabled || xfdir_monitor_skip (xfdir_p)\n");
		refresh_pending = refresh_action = FALSE;
                rfm_threadwait ();
                continue;
            }

            // check for an immediate reload condition which requires
	    // immediate reload action (gboolean refresh_action)
            if(refresh_action) {
                TRACE ("monitor now setting up refresh!\n");
		// obtain new xfdir structure
		xfdir_p = immediate_reload_action(xfdir_p);
		// is the monitor still active?
		if(exit_monitor (xfdir_p)) {
		   rfm_threadwait ();
		   break;
		}
		// assign new xfdir structure to view
		xfdir_p->view_p->xfdir_p = xfdir_p;
	        // tell main thread that refresh information is ready
		xfdir_p->view_p->flags.refresh = 1;
		// put current monitor thread in suspend state until 
		// reload has completed.
		refresh_action = FALSE;
		// take a breath
		rfm_threadwait ();
                break;
            }
            // if a refresh is not pending, check for a full reload condition: 
            else if(!refresh_pending && 
		    reload_condition (xfdir_p, &current_count, &current_st)) {
		// reload condition is true, so we queue a refresh
		// refresh is queued so that refresh will not actually
		// occur until reload condition stops occurring. 
		// (this is to avoid a cascade of unnecessary reloads
		// on a rapidly changing directory)
                TRACE ("MONITOR: reload_condition is TRUE\n");
                refresh_pending = TRUE;
		// here we used to have a status line condition
		// message, but that will now be annoying to lpterm users (me).
            }
            // check if reload pending and changes have ceased 
            else if(refresh_pending && 
		    !reload_condition (xfdir_p, &current_count, &current_st)) {
                TRACE ("MONITOR: reload_condition has ceased. Now scheduling refresh\n");
		// Since reload condition has ceased, we can schedule an 
		// immediate reload action on the next loop iteration
                refresh_action = TRUE;
            } else {
		// reload condition has not occurred, but stat information
		// or thumbnail image may have changed for an existing item
		// without affecting directory stat information. In this case
		// we update stat information for individual xfdir structure items. 
                TRACE ("MONITOR: now checking update_stat_info...\n");
                if (update_stat_info (xfdir_p)){
		    TRACE ("refresh_pending on update_stat_info()...\n");
		    // A TRUE return value indicates that the directory
		    // count does not correspond to the actual files 
		    // which currently exist and the file count is wrong.
		    // In this case we queue a refresh.
		    refresh_pending = TRUE;
		} 
            }
	    // take a breath
            rfm_threadwait ();
        } while (2);   // end inner loop 2
        TRACE ("monitor %d loop broken for %s\n", xfdir_p->monitor_id, xfdir_p->path);
	// inner loop has exited. This means one of two things:
	// 	1. monitor thread has met an exit condition and is no longer necessary
	// 	2. a reload condition has occurred and monitor thread must pause
	// 	   until reload has been accomplished.
	//
        // Here we consider case 1, monitor thread had met an exit condition:
        if(exit_monitor (xfdir_p)) {
            TRACE ("MONITOR: all is cool to release xfdir memory\n");
            TRACE ("monitor: id= %d ending now\n", xfdir_p->monitor_id);
            TRACE ("monitor: will free xfdir_p=0x%lx\n", (long unsigned)xfdir_p);
            TRACE ("monitor: id= %d is now defunct\n", xfdir_p->monitor_id);

	    // monitor thread ends here: 
	    g_mutex_lock(xfdir_p->view_p->mutexes.janitor);
		xfdir_p->view_p->janitor.monitors--;
	    g_mutex_unlock(xfdir_p->view_p->mutexes.janitor);
	    // current xfdir structure (no longer needed) cleanup
            if(xfdir_p->xfdir_p) {
                xfdir_free_data (xfdir_p->xfdir_p);
                g_free (xfdir_p->xfdir_p);
            }
            xfdir_free_data (xfdir_p);
            g_free (xfdir_p);
            xfdir_p = NULL;
	    rfm_barrier_decrement();

            return NULL;
        }
        // main thread must free view_p->xfdir_p upon condition where
        // xfdir_p->xfdir_p replaces it on upload...
        TRACE ("monitor: id= %d continues with xfdir_p->xfdir_p\n", xfdir_p->monitor_id);

    } while(1);
    // monitor event loop ends here (end outer loop)
}

/**
 * xfdir_start_monitor:
 * @view_p: A Rodent view_t pointer
 * @en: An icon entry structure.
 *
 * Starts a monitor thread for the loaded @xfdir_p
 * structure. This is done automatically for local files
 * but modules should define a FIXME function to 
 * test for updates and a FIXME function to reload.
 * 
 **/
static
    void
xfdir_start_monitor (view_t * view_p, xfdir_t * xfdir_p) {

    // (mutex lock might be overkill here)
    g_mutex_lock (view_p->mutexes.monitor_id);
    xfdir_p->monitor_id = ++view_p->flags.monitor_id;
    view_p->flags.monitor_enabled=TRUE;
    g_mutex_unlock (view_p->mutexes.monitor_id);

    // start the background monitor thread
    xfdir_p->thread = g_thread_create (thread_monitor_f, (gpointer) xfdir_p, FALSE, NULL);
    if (xfdir_p->thread == NULL) {
        g_warning("cannot create monitor thread at primary_xfdir.i");
        view_p->flags.monitor_enabled=FALSE;
    }

}

static long long 
get_tama(xfdir_t *xfdir_p){
    long long tama=0;
    if (!xfdir_p) return tama;
    int j;
    for(j = 0; j < xfdir_p->pathc; j++)
            if(xfdir_p->gl[j].en && xfdir_p->gl[j].en->st) {
                tama += xfdir_p->gl[j].en->st->st_size;
    }
    return tama;
}

// update xfdir structure for directory (or module)
static int
update_xfdir (xfdir_t * xfdir_p, gboolean fullbones) {
    if(xfdir_p->view_p->module) {
        TRACE ("obtaining view_p->xfdir_p from module %s\n", xfdir_p->view_p->module);
        if(MODULE_STAT (NULL, NULL) && !MODULE_STAT (NULL, &xfdir_p->st)) {
            g_warning ("update_xfdir(): module_stat not correctly defined for %s\n", 
		    xfdir_p->view_p->module);
        }
        xfdir_p->pathc = MODULE_COUNT (xfdir_p);
        xfdir_p->data = MODULE_DATA (xfdir_p);
        xfdir_p->view_p->submodule_name = 
	    rfm_void(PLUGIN_DIR, xfdir_p->view_p->module_name, "submodule_name");
	TRACE("module=%s\n", xfdir_p->view_p->module);
	gint retval = 
	    GPOINTER_TO_INT(rfm_natural(PLUGIN_DIR, 
			xfdir_p->view_p->module, xfdir_p, 
			"module_xfdir_get"));
	return retval;
    } else {
        xfdir_p->view_p->submodule_name = NULL;
    }
    //int checksum;

    xfdir_p->tama = 0;

    stat (xfdir_p->path, &xfdir_p->st);

    /* count step */
    xfdir_p->pathc = read_files_local (xfdir_p, FALSE, FALSE);
    xfdir_p->tama = 0;
    if(xfdir_p->pathc <= 0) {
        //g_warning ("read_files_local returned %zd", xfdir_p->pathc);
        TRACE ("read_files_local returned %zd", xfdir_p->pathc);
        return -1;
    }
    TRACE ("monitor file count=%zd xfdir_p=0x%lx xfdir_p->xfdir_p=0x%lx\n", xfdir_p->pathc, (long unsigned)xfdir_p,
         (long unsigned)xfdir_p->xfdir_p);
    /* read part */
    if(xfdir_p->pathc) {
        xfdir_p->gl = (dir_t *) malloc (xfdir_p->pathc * sizeof (dir_t));
        if(!xfdir_p->gl) {
            g_warning ("unable to malloc struct dir_t");
            return -1;
        } else {
            memset (xfdir_p->gl, 0, xfdir_p->pathc * sizeof (dir_t));
        }

    //    checksum = st_sum_dir (&(xfdir_p->st)) + SHOWS_HIDDEN (xfdir_p->type);

	// if count is less than one age, full bones is not a significant
	// disadvantage... This is done automatically by read_files_local()
	// for record elements with index less than one age.
	// if (xfdir_p->pathc<=102){ fullbones=TRUE;}
        /* read step (read <= count step) */
        if(read_files_local (xfdir_p, TRUE, fullbones) < 0) {
            g_warning ("read_files_local returned < 0");
            return -1;
        }
        xfdir_p->tama = get_tama(xfdir_p);
        
        // a write lock is placed on cache
  /*      save_xfdir_cache (xfdir_p, checksum);
        TRACE ("monitor: cache saved...\n");*/

    }

    return 1;
}

