#ifndef RFM_PRIMARY_H
# define RFM_PRIMARY_H

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/wait.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <limits.h>
#include <pwd.h>
#include <grp.h>
#include <math.h>
#include <locale.h>

#ifdef HAVE_STDARG_H
# include <stdarg.h>
#elif HAVE_VARARGS_H
# include <varargs.h>
#endif

#include <ctype.h>
#include <glob.h>
#include <memory.h>
#include <signal.h>
#include <dirent.h>
#include <regex.h>

#include <X11/Xlib.h>
#ifdef HAVE_LIBSM
# include <X11/SM/SMlib.h>
#endif
# include <X11/Xmd.h>
# include <X11/SM/SMlib.h>
#include <X11/Xproto.h>
#include <X11/Xatom.h>

#include <cairo-xlib.h>
#include <gio/gio.h>
#include <gtk/gtk.h>
#include <gdk/gdkx.h>
#include <gdk/gdkkeysyms.h>
#include <gtk/gtksignal.h>
#include <gmodule.h>
#include <time.h>
#include <sys/timeb.h>
#include <sys/mman.h>

#ifdef HAVE_PTHREAD_TRYJOIN
# ifndef _GNU_SOURCE
#  define _GNU_SOURCE
# endif
# include <pthread.h>
#endif

#ifdef HAVE_SYS_PARAM_H
#include <sys/param.h>
#endif
#ifdef HAVE_SYS_UCRED_H
#include <sys/ucred.h>
#endif
#ifdef HAVE_SYS_MOUNT_H
#include <sys/mount.h>
#endif
#ifdef HAVE_FSTAB_H
#include <fstab.h>
#endif
#ifdef HAVE_MNTENT_H
# include <mntent.h>
#endif
#ifdef HAVE_PATHS_H
# include <paths.h>
#endif

#ifdef HAVE_SYS_FS_TYPES_H
# include <sys/fs_types.h>
#endif

#ifdef HAVE_SYS_MNTTAB_H
# include <sys/mnttab.h>
#endif

#ifdef HAVE_SYS_VFS_H
# include <sys/vfs.h>
#endif

#ifdef HAVE_SYS_VFSTAB_H
# include <sys/vfstab.h>
#endif

#include <libxml/parser.h>
#include <libxml/tree.h>
#include <magic.h>


#include "primary-environment.i"

gchar *
rfm_get_hash_key (const gchar * key, gint size);

// This will decrement active thread count. When active thread count reaches
// zero, barrier will be granted.
void
rfm_barrier_decrement(void);

// If barrier is requested, this function will block, otherwise active
// thread count will increment. While active thread count is
// greater than zero, barrier will not be granted.
void
rfm_barrier_stop_or_increment(void);

// This will request a barrier. Function will block until barrier
// is obtained. Barrier will not be obtained until the active thread
// count is equal to zero.
void
rfm_request_barrier(void);

// This function should only be called after a barrier has been obtained
// and will release the barrier, allowing all blocked threads to continue.
void
rfm_release_barrier(void);

gint
rfm_get_default_size(void);

gboolean
rfm_g_file_test(const gchar *path, gint test);

gchar *
rfm_get_paste_buffer (void );

void
rfm_clear_paste_buffer(void);

void
rfm_store_paste_buffer(gchar *buffer, gint len);

GtkWidget *
rfm_dialog_button (const char *icon_id, const char *text);

void 
rfm_get_population_rect(view_t * view_p, const population_t * population_p, GdkRectangle *rect);

void 
rfm_get_population_icon_rect(view_t * view_p, const population_t * population_p, GdkRectangle *rect);

void 
rfm_get_population_label_rect(view_t * view_p, const population_t * population_p, GdkRectangle *rect);

void
rfm_thread_expose_rect (GdkWindow * window, GdkRectangle * rect);

void
rfm_expose_item (view_t * view_p, const population_t * population_p);

void
rfm_expose_icon (view_t * view_p, const population_t * population_p);

void
rfm_expose_label (view_t * view_p, const population_t * population_p);


void
  smc_connect (gint argc, gchar ** argv, CARD8 style_hint);

//

gboolean
  rfm_write_ok (struct stat *st);
gboolean
  rfm_read_ok (struct stat *st);

  // deprecated:
GtkWidget *
  rfm_get_widget_by_name (GtkWidget * parent, const gchar * name);

  // deprecated:
void
  rfm_set_widget_by_name (GtkWidget * parent, const gchar * name, 
	  GtkWidget * widget);

void
  rfm_lock_fd_read (int fd);

void
  rfm_lock_fd_write (int fd);

void
  rfm_unlock_fd (int fd);

int
  rfm_mkdir (const gchar * dir);

void
  rfm_save_to_go_history (char *p);

void
  rfm_threadwait (void);

void
  rfm_init_env (void);

// rfm_setenv() is an internal command. If you want to set environment
// for child processes, please use ../mods/mcs_set_var:
// void *
// mcs_set_var (const gchar * setting_name, const gchar * setting_value)
void
  rfm_setenv (const char *name, char *value, gboolean verbose);

gchar *
  rfm_esc_string (const gchar * string);

const gchar *
  rfm_chop_excess (gchar * b);

gchar *
  rfm_time_to_string (time_t when);

gchar *
  rfm_mode_string (mode_t mode);

gchar *
rfm_user_string (struct stat *st);

gchar *
rfm_group_string (struct stat *st);

gchar *
rfm_date_string (time_t the_time);

gchar *
rfm_date_access_string (struct stat *st);

gchar *
rfm_date_status_string (struct stat *st);


gchar *
  rfm_utf_string (const gchar * t);

gchar *
  rfm_sizetag (off_t tama, gint count);

gchar *
  rfm_host_name (Window xid);

gboolean 
  rfm_population_write_lock (view_t * view_p);

void
  rfm_population_write_unlock (view_t * view_p);

gboolean 
  rfm_population_read_lock (view_t * view_p);

gboolean 
  rfm_population_try_read_lock (view_t * view_p);

void
  rfm_population_read_unlock (view_t * view_p);

int
  rfm_valid_pasteboard (view_t * view_p);

int
  rfm_in_pasteboard (view_t * view_p, record_entry_t * en);

gboolean 
  rfm_update_pasteboard (view_t * view_p);

void
  rfm_soft_reload (view_t * view_p);

    void
rfm_save_view_preferences (view_t * view_p, record_entry_t *target_en);

void
rfm_get_view_preferences (view_t * view_p, record_entry_t *target_en);

gboolean
rfm_write_ok_path(const gchar *path);

gboolean
rfm_read_ok_path(const gchar *target_path);
#endif
