//
//
/*
 * 
 * Copyright 2003-2011 under GNU/GPL which means what it means.
 * Edscott Wilson Garcia 
 *
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#define CPY_CANCEL      0
#define CPY_YES         1
#define CPY_YES_TO_ALL  2
#define CPY_NO          3
#define CPY_NO_TO_ALL   4

static gchar *
mktgpath (const gchar * target_dir, char *source) {
    gchar *name = g_path_get_basename (source);
    gchar *target = g_build_filename (target_dir, name, NULL);
    g_free (name);
    return target;
}

static int
warn_target_exists (widgets_t * widgets_p, const gchar * target, const gchar * src, gboolean force_override) {
    TRACE("force_override=%d\n", force_override);
    int result = CPY_YES;
    if(!force_override) {
        result = rodent_make_overwrite_dialog (widgets_p, target, src);
    }
    return result;
}

static int
ok_input (widgets_t * widgets_p, char *target, const gchar *path, gboolean force_override) {
    if(!g_file_test (target, G_FILE_TEST_EXISTS)) {
	TRACE("%s does not exist\n", target);
        return (CPY_YES);
    } 
    if(getenv ("VERSION_CONTROL")
       && strlen (getenv ("VERSION_CONTROL"))
       && strcmp (getenv ("VERSION_CONTROL"), "none") == 0) {
        return warn_target_exists (widgets_p, target, path, force_override);
    } 
    return CPY_YES;
}

static void
verify_list (widgets_t * widgets_p, GList * in_list, GSList ** out_list, const gchar * target_path) {
    gboolean force_override = FALSE;
    *out_list = NULL;
    if(g_list_length (in_list) > 1 && !g_file_test (target_path, G_FILE_TEST_IS_DIR)) {
        rfm_diagnostics (widgets_p, "xffm/stock_dialog-warning", NULL);
        rfm_diagnostics (widgets_p, "xffm_tag/stderr", target_path, ": ", strerror (ENOTDIR), "\n", NULL);
        //g_warning("target must be a directory: %s",target_path);
        return;
    }
    GList *list;
    for(list = in_list; list && list->data; list = list->next) {
	// Duplicate path for verified list
	gchar *path=g_strdup((gchar *)list->data);
        //record_entry_t *s_en = rfm_stat_entry (url, 0);

        gchar *target = (g_file_test(target_path, G_FILE_TEST_IS_DIR))?
	    mktgpath (target_path, path):
	    g_strdup(target_path);

        TRACE ("verify_list: target=%s\n", target_path);

        switch (ok_input (widgets_p, target, path, force_override)) {
        case CPY_NO:
            rfm_diagnostics (widgets_p, "xffm_tag/green", _("Cancel"), " :", path, "\n", NULL);
	    g_free(path);
            break;
        case CPY_CANCEL:       
            rfm_diagnostics (widgets_p, "xffm/stock_dialog-warning", strerror (ECANCELED), "\n", NULL);
	    g_free(path);
	    GSList *tmp=*out_list;
	    for (; tmp && tmp->data; tmp=tmp->next){
		g_free(tmp->data);
	    }
            g_slist_free (*out_list);
            *out_list = NULL;
            g_free (target);
	    target=NULL;
            return;
        case CPY_YES_TO_ALL:
	    force_override = TRUE;
        case CPY_YES:
        default:
            *out_list = g_slist_prepend (*out_list, path);
            break;
        }
        g_free (target);
    }
    return;
}


static gpointer
private_rodent_cp (int mode, widgets_t *widgets_p, GSList **verified_list, const gchar * target_path){
    view_t *view_p=widgets_p->view_p;
    gchar *sources[MAX_COMMAND_ARGS];
    int i = 0;
// Choose the GNU command to execute:
    switch (mode) {
    case (TR_MOVE):
    case (TR_RENAME):
        sources[i++] = "mv";
        break;
    case (TR_LINK):
        sources[i++] = "ln";
        break;
    default:
        sources[i++] = "cp";
        break;
    }
// Set the common options: force, verbose.
    sources[i++] = "-f"; 
    if(getenv ("RFM_CP_VERBOSE") && strlen (getenv ("RFM_CP_VERBOSE"))) {
	sources[i++] = "-v";
    }
// Set the options for each command. These options are listed as environment
// variables in the file primary-environment.i.
    switch (mode) {
    case (TR_COPY):
        if(getenv ("RFM_CP_A") && strlen (getenv ("RFM_CP_A"))) {
            sources[i++] = "--archive";
        }
        if(getenv ("RFM_CP_RR") && strlen (getenv ("RFM_CP_RR"))) {
            sources[i++] = "--recursive";
        }
#ifdef GNU_CP
        if(getenv ("RFM_CP_B") && strlen (getenv ("RFM_CP_B"))) {
            sources[i++] = "-b";
        }
        if(getenv ("RFM_CP_U") && strlen (getenv ("RFM_CP_U"))) {
            sources[i++] = "--update";
        }
        if(getenv ("RFM_CP_D") && strlen (getenv ("RFM_CP_D"))) {
            sources[i++] = "-d";
        }
        if(getenv ("RFM_CP_S") && strlen (getenv ("RFM_CP_S"))) {
            sources[i++] = "--symbolic-link";
        }
#endif
        if(getenv ("RFM_CP_L") && strlen (getenv ("RFM_CP_L"))) {
            sources[i++] = "--link";
        }
        if(getenv ("RFM_CP_LL") && strlen (getenv ("RFM_CP_LL"))) {
            sources[i++] = "--dereference";
        }
        if(getenv ("RFM_CP_N") && strlen (getenv ("RFM_CP_N"))) {
            sources[i++] = "--no-clobber";
        }
        if(getenv ("RFM_CP_PP") && strlen (getenv ("RFM_CP_PP"))) {
            sources[i++] = "--no-dereference";
        }
        if(getenv ("RFM_CP_P") && strlen (getenv ("RFM_CP_P"))) {
            sources[i++] = "-p";
        }
        if(getenv ("RFM_CP_X") && strlen (getenv ("RFM_CP_X"))) {
            sources[i++] = "--one-file-system";
        }
        if(getenv ("RFM_CP_H") && strlen (getenv ("RFM_CP_HH"))) {
            sources[i++] = "-H";
        }
        break;
    case (TR_MOVE):
#ifdef GNU_CP
        if(getenv ("RFM_MV_B") && strlen (getenv ("RFM_MV_B"))) {
            sources[i++] = "-b";
        }
        if(getenv ("RFM_MV_U") && strlen (getenv ("RFM_MV_U"))) {
            sources[i++] = "--update";
        }
        if(getenv ("RFM_MV_N") && strlen (getenv ("RFM_MV_N"))) {
            sources[i++] = "--no-clobber";
        }
#endif
        break;
    case (TR_RENAME):
#ifdef GNU_CP
        if(getenv ("RFM_MV_B") && strlen (getenv ("RFM_MV_B"))) {
            sources[i++] = "-b";
        }
        sources[i++] = "--no-target-directory"; // or -T
#endif
        break;
    case (TR_LINK):
        if(getenv ("RFM_LN_S") && strlen (getenv ("RFM_LN_S"))) {
            sources[i++] = "--symbolic";
        }
#ifdef GNU_CP
        if(getenv ("RFM_LN_B") && strlen (getenv ("RFM_LN_B"))) {
            sources[i++] = "-b";
        }
        if(getenv ("RFM_LN_L") && strlen (getenv ("RFM_LN_L"))) {
            sources[i++] = "--logical";
        }
        if(getenv ("RFM_LN_N") && strlen (getenv ("RFM_LN_N"))) {
            sources[i++] = "--no-dereference";
        }
        if(getenv ("RFM_LN_P") && strlen (getenv ("RFM_LN_P"))) {
            sources[i++] = "--physical";
        }
#endif
        break;
    }
    // Options are now set. What follows is listing the parameters to
    // the respective GNU command.

    // If we are in link mode, we make all targets relative to 
    // avoid absolute paths for symlinks.
    GSList *list;
    GSList *ln_list=NULL;
    gchar *tgtdir = NULL;
    gboolean ok = TRUE;
    gboolean readOK = TRUE;
    if (mode==TR_LINK) {
	const gchar *ln_path=(*verified_list)->data;
	
	tgtdir = g_strdup (target_path);
        gchar *base_dir = g_path_get_dirname (ln_path);
        int up = 0;
        while(base_dir && tgtdir && !strstr (base_dir, tgtdir)
              && strchr (tgtdir, '/')) {
            up++;
            *(strrchr (tgtdir, '/')) = 0;
            TRACE ("CPY: tgtdir=%s\n", tgtdir);
        }
        g_free (base_dir);

	for(list = *verified_list; list && list->data; list = list->next) {
	    const gchar *orig_path=list->data;
	    const gchar *source_substring = orig_path + strlen (tgtdir) + 1;
	    int newsource_length = up * strlen ("../") + strlen (source_substring) + 1;
	    gchar *newsource = (gchar *)malloc (newsource_length);
	    memset (newsource, 0, newsource_length);
	    int k;
	    for(k = 0; k < up; k++) {
		strcat (newsource, "../");
	    }
	    strcat (newsource, source_substring);
	    // relative links may fail if the path we are building
	    // from contains a symlink itself...
	    gchar *target;
	    if(g_file_test (target_path, G_FILE_TEST_IS_DIR)) {
		target = g_strdup (target_path);
	    } else {
		target = g_path_get_dirname (target_path);
	    }
	    gchar *relative_path=g_build_filename(target, newsource, NULL);
	    g_free(target);
	    if (!g_file_test(relative_path, G_FILE_TEST_EXISTS)){
		// in this case, fall back to absolute path
		TRACE("%s does not exist! Fallback= %s\n",
			newsource, orig_path);
		g_free(newsource);
		newsource=g_strdup(orig_path);
	    } else {
		TRACE("%s is ok\n", newsource);
	    }
	    g_free(relative_path);
	    sources[i++] = newsource;
	    ln_list=g_slist_prepend(ln_list, newsource);
	    if(i == MAX_COMMAND_ARGS - 3) {
		rfm_diagnostics (widgets_p, "xffm/stock_dialog-warning", NULL);
                gchar *max=g_strdup_printf("%d",MAX_COMMAND_ARGS);
                rfm_diagnostics (widgets_p, "xffm_tag/stderr", sources[0], ": ", strerror(E2BIG)," (> ",max,")","\n", NULL);
                g_free(max);
		ok = FALSE;
		break;
	    }
	}
	g_free (tgtdir);

	gboolean exists=g_file_test (target_path, G_FILE_TEST_IS_DIR);

        if(exists) {
            tgtdir = g_strdup (target_path);
        } else {
            tgtdir = g_path_get_dirname (target_path);
        }
        sources[i++] = tgtdir;
        if(!g_file_test(tgtdir, G_FILE_TEST_IS_DIR)) {
            g_warning ("Cannot chdir to %s\n", tgtdir);
            g_free (tgtdir);
	    tgtdir=NULL;
	    ok = FALSE;
        } else {
            g_free (widgets_p->workdir);
            widgets_p->workdir = g_strdup(tgtdir);
        }
    } else {
	for(list = *verified_list; list && list->data; list = list->next) {
	    sources[i++] = (gchar *)list->data;
	    if (!rfm_read_ok_path((gchar *)list->data)){
		TRACE("readOK=false %s\n", (gchar *) list->data);
		readOK=FALSE;
	    }
	    if(i == MAX_COMMAND_ARGS - 3) {
		rfm_diagnostics (widgets_p, "xffm/stock_dialog-warning", NULL);
                gchar *max=g_strdup_printf("%d",MAX_COMMAND_ARGS);
                rfm_diagnostics (widgets_p, "xffm_tag/stderr", sources[0], ": ", strerror(E2BIG)," (> ",max,")","\n", NULL);
                g_free(max);
		ok = FALSE;
		break;
	    }
	}
        TRACE ("CPY: mode=%d ?= %d, list length=%d\n", mode, TR_LINK, g_slist_length (*verified_list));
        sources[i++] = (gchar *) target_path;
    }

    sources[i++] = NULL;

    // note: with run_argv we do not need to escape paths.
    if(ok) {
	if (view_p->flags.type != DESKVIEW_TYPE) {
	    rfm_show_text(widgets_p);
	}
	// Not for dummies. On cp/mv/ln we will check for permissions on target
	// directory, and if permission do not allow the operation,
	// then we will try to do it with sudo.
	gchar *src_dir=NULL;
	if (mode==TR_MOVE || mode==TR_RENAME) {
	    src_dir=g_path_get_dirname(sources[i-3]);
	}
	
	gboolean need_sudo = !rfm_write_ok_path(target_path) ||
	    (src_dir && !rfm_write_ok_path(src_dir)) ||
	     !readOK;
	if (!need_sudo) {
	    rfm_thread_run_argv (widgets_p, sources, FALSE);
	} else {

	    gchar *failed=NULL;
	    gchar *tgt=NULL;
	    const gchar *operation;
	    if (g_file_test (target_path, G_FILE_TEST_IS_DIR)){
		tgt = g_strdup (target_path);
	    } else {
		tgt = g_path_get_dirname (target_path);
	    }
	    switch (mode){
		case TR_COPY:
		    failed=g_strdup_printf("%s.", _("Failed to copy file"));
		    operation="cp";
		    break;
		case TR_MOVE:
		case TR_RENAME:
		    if (mode==TR_RENAME){
			failed=g_strdup(_("Error: Rename failed."));
		    } else {
			failed = g_strdup_printf("%s: %s", _( "Move files"), 
				strerror(EACCES));
		    }

		    if (src_dir && !rfm_write_ok_path(src_dir)) {
			g_free(tgt);
			tgt=g_strdup(src_dir);
		    }
		    operation="mv";
		    break;
		case TR_LINK:
		    failed=g_strdup_printf(_("Failed to link %s to %s"), 
			    _("File"), _("Destination"));
		    operation="ln";
		    break;
		default:
		    g_warning("non valid condition in private_rodent_cp()");
	    }
	    if (rfm_confirm_sudo(widgets_p, tgt, failed, operation)){
		rfm_try_sudo (widgets_p, sources, FALSE);
	    }
	    g_free(failed);
	    g_free(tgt);
	}  
	g_free(src_dir);
    } else {
        TRACE ("PASTE CP: ok == FALSE\n");
    }
    // We should do partial cleanup here...
    // Clean up stuff generated in link mode:
    g_free(tgtdir);
    for(list = ln_list; list && list->data; list = list->next) {
        g_free(list->data);
    }
    g_slist_free (ln_list);
    // 1. Clean up input selection_list
    for(list = *verified_list; list && list->data; list = list->next) {
        g_free(list->data);
    }
    g_slist_free (*verified_list);
    return NULL;
}


