//
//
/* */
/*
 * Copyright (C) 2002-2011 Edscott Wilson Garcia
 * EMail: edscott@xfce.org
 *
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
static gint
rodent_query_rm (widgets_t * widgets_p, const gchar * nfile, const gchar *ofile, gboolean overwrite) {

    gboolean result=rodent_make_overwrite_dialog (widgets_p, nfile, ofile);
    if(result) {
        if(!overwrite) {
            if(unlink (nfile) < 0) {
                rfm_diagnostics (widgets_p, "xffm_tag/red", _("Error while deleting."), ": unlink(", nfile, ")--> ", strerror (errno), "\n", NULL);
            }
        }
        return TRUE;
    }
    return FALSE;
}

static int
entry_rename (widgets_t * widgets_p, const gchar * nfile, const gchar * ofile) {
    TRACE ("at rename\n");
    if(!widgets_p || !nfile || !ofile || !strlen (nfile) || !strlen (ofile))
        return FALSE;

    GList *in_list = NULL;
    in_list = g_list_append (in_list, (void *)ofile);
    // If the target exists, and is a directory, or a symlink, 
    // we should first remove it so that the mv command
    // does not put our source inside the directory.
    // For this we use the option -T, --no-target-directory...
 /*   if (g_file_test(nfile, G_FILE_TEST_IS_DIR) || 
	    g_file_test(nfile, G_FILE_TEST_IS_DIR)){
	gchar *backup=g_strconcat(nfile, "~", NULL);
        if (rename(nfile, backup) < 0) {
	    gchar *argv[]={"mv", "-f", 
	    rfm_trysudo(widgets_p, gchar ** argv, gboolean interm);

);
	}
	
    }*/
    rodent_cp (TR_RENAME, widgets_p, in_list, nfile);
    g_list_free (in_list);

    return TRUE;
}

static int
entry_duplicate (widgets_t * widgets_p, const gchar * nfile, const gchar * ofile) {
    if(!widgets_p || !nfile || !ofile || !strlen (nfile) || !strlen (ofile))
        return FALSE;

    GList *in_list = NULL;
    in_list = g_list_append (in_list, (void *)ofile);
    rodent_cp (TR_COPY, widgets_p, in_list, nfile);
    g_list_free (in_list);

    return TRUE;
}

static int
entry_symlink (widgets_t * widgets_p, const gchar * nfile, const gchar * ofile) {

    struct stat st;

    if(!widgets_p || !nfile || !ofile || !strlen (nfile) || !strlen (ofile))
        return FALSE;
#if 0
    // this does not work since target is not a directory
    GList *in_list = NULL;
    in_list = g_list_append (in_list, (void *)ofile);
    rodent_cp (TR_LINK, widgets_p, in_list, nfile);
    g_list_free (in_list);
#else
    if(lstat (nfile, &st) >= 0) {
        if(!rodent_query_rm (widgets_p, nfile, ofile, FALSE))
            return FALSE;
    }

    g_free(widgets_p->workdir);
    // Relative links, always...
    widgets_p->workdir=g_path_get_dirname(ofile);
    gchar *b_ofile=g_path_get_basename(ofile);
    gchar *b_nfile=g_path_get_basename(nfile);
    gchar *command = g_strdup_printf ("ln -s \"%s\" \"%s\"", b_ofile, b_nfile);
    rfm_thread_run2argv (widgets_p, command, FALSE);
    g_free (command);
    g_free(b_ofile);
    g_free(b_nfile);
#endif

    return TRUE;
}

static void
entry_activate (GtkEntry * entry, view_t * view_p, int caso) {
    gchar *actual_tag;
    gchar *tag;
    gchar *b;
    int result = FALSE;
    gchar *path;
    widgets_t *widgets_p=&(view_p->widgets);
    
    if(!rfm_population_try_read_lock (view_p)) {
	rfm_diagnostics(widgets_p, "xffm/stock_dialog-warning", "Unable to obtain readlock... aborting entry activate. Please retry.",NULL);
        return;
    }

    gtk_widget_hide (GTK_WIDGET (view_p->widgets.rename));

    path = g_object_get_data (G_OBJECT (view_p->widgets.rename), "path");

    TRACE ("entry_activate\n");

    if(!path) {
	rfm_population_read_unlock (view_p);
        return;
    }
    actual_tag = gtk_editable_get_chars ((GtkEditable *) entry, 0, -1);
    g_strstrip(actual_tag);
    tag = g_locale_from_utf8 (actual_tag, -1, NULL, NULL, NULL);
    g_free (actual_tag);
    actual_tag = tag;
    b = g_path_get_basename (path);

    gchar *d = g_path_get_dirname (path);
    gchar *p = g_build_filename (d, actual_tag, NULL);
    g_free (d);
    switch (caso) {
    case 0:
        result = entry_rename (&(view_p->widgets), p, path);
        if(result) {
            TRACE ("renaming %s to %s\n", path, p);
        }
        break;
    case 1:
        result = entry_duplicate (&(view_p->widgets), p, path);
        if(result) {
            TRACE ("duplication %s to %s\n", path, p);
        }
        break;
    case 2:
        result = entry_symlink (&(view_p->widgets), p, path);
        if(result) {
            TRACE ("symlinking %s to %s\n", path, p);
        }
        break;
    }
    g_free (p);

    g_free (b);
    g_free (actual_tag);
    rodent_done_with_rename ((gpointer) view_p);
    rfm_population_read_unlock (view_p);    
}

static void
entry_activate_rename (GtkEntry * entry, gpointer data) {
    entry_activate (entry, (view_t *) data, 0);
}

static void
entry_activate_duplicate (GtkEntry * entry, gpointer data) {
    entry_activate (entry, (view_t *) data, 1);
}

static void
entry_activate_symlink (GtkEntry * entry, gpointer data) {
    entry_activate (entry, (view_t *) data, 2);
}

static void
destroy_dialog (GtkWidget * widget, gpointer data) {
    rodent_done_with_rename (data);
}

static gint
on_key_press (GtkWidget * entry, GdkEventKey * event, gpointer data) {
    if(event->keyval == GDK_Escape) {
        rodent_done_with_rename (data);
        //rfm_status (&(view_p->widgets), NULL, _("Omitting"), NULL);
        return TRUE;
    }
    return FALSE;
}

static gboolean
grab_focus (GtkWidget * widget, GdkEventCrossing * event, gpointer data) {
    XSetInputFocus (gdk_x11_display_get_xdisplay(gdk_display_get_default()), 
	    GDK_WINDOW_XID (gtk_widget_get_parent_window (widget)), RevertToParent, CurrentTime);
    XUngrabPointer (gdk_x11_display_get_xdisplay(gdk_display_get_default()), 
	    CurrentTime);
    return TRUE;
}
