
/*
 * Copyright (C) 2002-2011 Edscott Wilson Garcia
 * EMail: edscott@xfce.org
 *
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include "rodent_libs.h"
#include "rodent_expose.i"

void
rodent_expose_all (view_t * view_p) {
    TRACE (">> rodent_expose_all\n");
    GdkRectangle rect;
    rect.x = rect.y = 0;
    rect.width = view_p->widgets.window->allocation.width;
    rect.height = view_p->widgets.window->allocation.height;
    gdk_window_invalidate_rect (view_p->widgets.window->window, &rect, TRUE);
    TRACE ("exposeall for grid w=%d h=%d\n", rect.width, rect.height);
}

#if 0
// this break things after going to /usr/bin and doing rubber band selection...
void
rodent_expose_all_items(view_t *view_p){
    GdkRectangle rect;
    // get rectangle
    rect.x = 0;
    rect.width = view_p->widgets.window->allocation.width;

    if (GTK_IS_SCROLLED_WINDOW (view_p->widgets.scrolled_window)) {
	double page = gtk_adjustment_get_page_size (
	    gtk_scrolled_window_get_vadjustment (view_p->widgets.scrolled_window));
	double value = gtk_adjustment_get_value  (
	    gtk_scrolled_window_get_vadjustment (view_p->widgets.scrolled_window));
	rect.y = value;
	rect.height = page;
    } else {
	rect.y = 0;
	rect.height = view_p->widgets.window->allocation.height;
    }
    TRACE("exposing %d,%d %d,%d\n", rect.x, rect.y, rect.width, rect.height);
    // find items in rectangle
    GSList *render_list = rodent_find_items_in_rectangle(view_p, &rect, FALSE);
    // expose items
    rodent_expose_slist(view_p, render_list, &rect);
    // cleanup
    g_slist_free(render_list);

}
#endif

static
population_t *find_grid_element(view_t * view_p, gdouble x, gdouble y) {
    if(!view_p->population_pp)  return NULL;
    
    // rectangle x,y coordinates
    gint row = y / CELLHEIGHT(view_p);
    gint column = x / CELLWIDTH(view_p);
    if(column >= view_p->view_layout.grid_columns) return NULL;
    if(row >= view_p->view_layout.grid_rows) return NULL;
    if(x < 0 || y < 0)  return NULL;

    gint element;
    if (view_p->flags.type == ICONVIEW_TYPE) {
	element= row * view_p->view_layout.grid_columns + column;
    } else { // transposed
	element= column * view_p->view_layout.grid_rows + row;
    }
    if(element >= view_p->view_layout.max_elements) return NULL;

    population_t *population_p = *(view_p->population_pp + element);

    if (population_p && 
		(population_p->row != row || population_p->column != column)) {
	// This is a sanity hack test. Should never happen in a perfect world.
	g_warning ("THIS IS WRONG: row, column: current (%d,%d) correct (%d,%d) element=%d columns=%d\n",
	    population_p->row, population_p->column,
	    row, column, element,
	    view_p->view_layout.grid_columns);
	population_p->row = row;
	population_p->column = column;
    }

    return population_p;
}


const 
population_t *
rodent_find_in_population (view_t * view_p, gdouble x, gdouble y) {

    // Margin considerations:
    x -= view_p->view_layout.margin_left;
    y -= view_p->view_layout.margin_top;

    population_t *population_p = find_grid_element(view_p, x, y);
    if(!population_p) return NULL;



    gint X = (gint)x % CELLWIDTH(view_p);
    gint x_spacing = (view_p->view_layout.icon_size >= SMALL_ICON_SIZE)?
	(CELLWIDTH(view_p) - population_p->pixbufW) / 2:
	0;
    if(x_spacing < 0) x_spacing = 0;
    if(X < x_spacing || X > x_spacing + population_p->pixbufW) return NULL;

    gint Y = (gint)y % CELLHEIGHT(view_p);
    gint y_spacing = (ICON_SIZE(view_p) - population_p->pixbufH)/2;

    if(Y < y_spacing || Y > y_spacing + population_p->pixbufH) return NULL;
    return population_p;
}

const
population_t *
rodent_find_in_labels (view_t * view_p, gdouble x, gdouble y) {
    // Margin considerations:
    x -= view_p->view_layout.margin_left;
    y -= view_p->view_layout.margin_top;
	
    // Find grid element.
    population_t *population_p = find_grid_element(view_p, x, y);
    if(!population_p) return NULL;
;

    GdkRectangle label_rect;
    rfm_get_population_label_rect(view_p, population_p, &label_rect);
    if(x >= label_rect.x && x < label_rect.x + label_rect.width 
	    &&
       y >= label_rect.y && y < label_rect.y + label_rect.height)
    {
	return population_p;
    }
    return NULL;
}

    

// Find elements (or icons, specified by the respective gboolean)
// within an expose rectangle. The icons restriction is to do rubberband
// selecting considering only the icons to select.

GSList *rodent_find_items_in_rectangle(view_t * view_p, GdkRectangle *rect, gboolean icons) {

    if(!rect->width || !rect->height) return NULL;
    if (!rfm_population_try_read_lock(view_p)) return NULL;
    if(!view_p->population_pp){
	rfm_population_read_unlock (view_p);
        return NULL;
    }

    GSList *list=NULL;


    gint low_X = rect->x;
    gint low_Y = rect->y;
    // Margin considerations:
    low_X -= view_p->view_layout.margin_left;
    low_Y -= view_p->view_layout.margin_top;
    gint high_X = low_X + rect->width;// - 1;
    gint high_Y = low_Y + rect->height;// - 1;

    if(low_X < 0){ low_X = 0; }
    if(low_Y < 0){ low_Y = 0; }

    gint low_column  = low_X / CELLWIDTH(view_p);
    gint low_row     = low_Y / CELLHEIGHT(view_p);
    gint high_column = high_X / CELLWIDTH(view_p);
    gint high_row    = high_Y / CELLHEIGHT(view_p);

    if ( low_X % CELLWIDTH(view_p)) low_column--;
    if ( low_Y % CELLHEIGHT(view_p)) low_row--;
    if (low_column < 0) low_column = 0;
    if (low_row < 0) low_row = 0;

    if (high_column > view_p->view_layout.grid_columns - 1) {
	TRACE("epa! high_column (%d) > view_p->view_layout.grid_columns - 1 (%d)\n",
		high_column, view_p->view_layout.grid_columns - 1);
	high_column = view_p->view_layout.grid_columns - 1;
    }
    if (high_row > view_p->view_layout.grid_rows - 1) {
	TRACE("epa! high_row (%d) > view_p->view_layout.grid_rows - 1 (%d)\n",
		high_row, view_p->view_layout.grid_rows - 1);
	high_row = view_p->view_layout.grid_rows - 1;
    }

    TRACE("rows = %d-%d, columns = %d-%d cellwidth=%d  cellheight=%d\n", 
	    low_row, high_row, low_column, high_column,
	    CELLWIDTH(view_p), CELLHEIGHT(view_p));

    // Just check if the element corresponding to the given
    // row and column is within the selected rows and columns.
    gint column;
    for (column=low_column; column<=high_column; column++){
	gint row;
	for (row=low_row; row<=high_row; row++) {
	    gint element;
	    if (view_p->flags.type == ICONVIEW_TYPE) {
		element = row * view_p->view_layout.grid_columns + column;
	    } else { // transposed
		element = column * view_p->view_layout.grid_rows + row;
	    }
	    if (element >= view_p->view_layout.max_elements) {
		TRACE("epa! element (%d) >= view_p->view_layout.max_elements (%d)\n",
			element, view_p->view_layout.max_elements);
		continue;
	    }
	    population_t *population_p = view_p->population_pp[element];
	    TRACE("element: %d/%d (%s)\n",
		    element, view_p->view_layout.max_elements-1,
		    (population_p->en)?population_p->en->path:NULL);
	    // Ok, we item is within the rows and columns, but are we only interested
	    // in the icons? 
	    gboolean found = FALSE;
	    if (!icons){
		found = TRUE;
	    } else {
		// This is just to consider the icons in the selection
		// process (for rubberband selection, mainly).
		GdkRectangle icon_rect;
		rfm_get_population_icon_rect(view_p, population_p, &icon_rect);
		gint low_x = icon_rect.x;
		gint high_x = icon_rect.x + icon_rect.width;
		gint low_y = icon_rect.y;
		gint high_y = icon_rect.y + icon_rect.height;
		gboolean X_in=FALSE;
		gboolean Y_in=FALSE;
		X_in = (low_x >= low_X && low_x < high_X) ||
		       (high_x >= low_X && high_x < high_X);
		Y_in = (low_y >= low_Y && low_y < high_Y) ||
		       (high_y >= low_Y && high_y < high_Y);
		if(X_in && Y_in){
		    found = TRUE;
		}
	    }
	    if (found) {
		TRACE("**found: %s\n", population_p->en->path);
		list=g_slist_prepend(list, population_p);	
	    }
	}
    }
    rfm_population_read_unlock(view_p);
    return list;
}


#if 0
void
rodent_render(view_t *view_p, GSList *render_list){
    GdkRectangle area;
    area.x = area.y =9999999;
    area.width = area.height =0;

    GSList *list=render_list;
    for (; list && list->data; list=list->next){
	GdkRectangle rect;
	const population_t * population_p=list->data;
	rfm_get_population_rect(view_p, population_p, &rect);
	if (rect.x < area.x) area.x = rect.x;
	if (rect.y < area.y) area.y = rect.y;
	if (rect.x + rect.width > area.width) area.width = rect.x + rect.width;
	if (rect.y + rect.height > area.height) area.height = rect.y + rect.height;
    }
    area.width = area.width - area.x;
    area.height = area.height - area.y;
    rodent_expose_slist(view_p, render_list, &area);
    return;
}
#endif


void
rodent_expose_item(view_t * view_p, const population_t * population_p) {
    rfm_expose_item (view_p, population_p);
}

void
rodent_expose (GtkWidget * widget, GdkEventExpose * event, gpointer data) {
    TRACE ("TRACE  on_expose x,y=%d,%d w,h=%d,%d\n", event->area.x, event->area.y, event->area.width, event->area.height);
    view_t *view_p = (view_t *) data;
    if(!CELLWIDTH(view_p) || !CELLHEIGHT(view_p)) return;
    if(!view_p || !view_p->widgets.paper)  return;


    // Put a read lock on population.
    // This presents a race condition problem:
    //   if we lose the race, no expose gets done (this happens)
    //   if we do a lock-wait instead, deadlock occurs...
    //
    if (!rfm_population_try_read_lock(view_p)) {
	// We lost the race! 
	// Since the expose could not proceed, keep generating
	// expose signal until it works. Expose signal is generated
	// in its own thread to avoid any deadlock potential.
	TRACE("We lost the race! Threading an expose signal now");
	// Clean the area anyways, for that we don't need a readlock.
	cairo_t *gdk_context =  
	    gdk_cairo_create(gtk_widget_get_window(view_p->widgets.paper));
	clean_rectangle (view_p, 
		event->area.x, event->area.y, 
		event->area.width, event->area.height,
		gdk_context, 0, 0);
	cairo_destroy(gdk_context);
	//gdk_flush();

	rfm_thread_expose_rect (gtk_widget_get_window(view_p->widgets.paper), &(event->area));
	return;
    }

    // check for background pixbuf (deskview)
    // this check will create background image when changed
    background_test(view_p);

    TRACE("Got write lock. on_expose area= %d, %d width=%d, height=%d\n", 
	    event->area.x, event->area.y, event->area.width, event->area.height);   
    // Find items which need to be redrawn.
    // take into consideration margin displacement

    GSList *render_list=rodent_find_items_in_rectangle(view_p, &(event->area), FALSE);

    GdkRectangle rect;
    rect.x = event->area.x; 
    rect.y = event->area.y; 
    rect.width = event->area.width;
    rect.height = event->area.height;
    rect.x=get_minimum_x(view_p, render_list, event->area.x);
    rect.y=get_minimum_y(view_p, render_list, event->area.y);
    rect.width = event->area.width + (event->area.x - rect.x);
    rect.height = event->area.height + (event->area.y - rect.y);

    TRACE("*** rect= %d,%d width=%d, height=%d\n", rect.x, rect.y, rect.width, rect.height);
    expose_slist_t *expose_slist_p = 
	mk_expose_slist_p(view_p, render_list, &rect);
    // not possible: g_thread_create(expose_slist_f, expose_slist_p, FALSE, NULL);
    expose_slist_f(expose_slist_p);
    //expose_slist(view_p, render_list, &rect);
}

void
rodent_insert_all_population (view_t * view_p) {
    int j;
    for(j = 0; j < view_p->xfdir_p->pathc; j++) {
        TRACE ("TRACE: inserting population %s\n", view_p->xfdir_p->gl[j].pathv);
        insert_population_item (view_p, NULL, j);
        view_p->flags.count++;
    }
}

void 
rodent_erase_deskview_geometries (void) {
    gchar *f = g_build_filename (g_get_user_config_dir (), DESKVIEW_GEOMETRY_DBH_FILE, NULL);
    if(g_file_test (f, G_FILE_TEST_EXISTS)) {
	unlink(f);
    }
    g_free (f);
    return;
}

void
rodent_save_deskview_geometry (const gchar * fullpath, int x, int y) {
    DBHashTable *preferences;
    GString *gs;
    gchar *f;
    deskview_geometry_t deskview_geometry;

    if(!fullpath || !strlen (fullpath))
        return;

    f = g_build_filename (g_get_user_config_dir (), DESKVIEW_GEOMETRY_DBH_FILE, NULL);
    if(!g_file_test (f, G_FILE_TEST_EXISTS)) {
        TRACE ("Creating preferences file: %s", f);
        preferences = dbh_create (f, 11);
    } else {
        preferences = dbh_open (f);
        if(!preferences) {
            unlink (f);
            preferences = dbh_create (f, 11);
        }
    }
    if(!preferences) {
        g_warning ("cannot open %s", f);
        g_free (f);
        return;
    }
    g_free (f);

    gs = g_string_new (fullpath);
    sprintf ((char *)DBH_KEY (preferences), "%10u", g_string_hash (gs));
    g_string_free (gs, TRUE);

    deskview_geometry.x = x;
    deskview_geometry.y = y;

    memcpy (DBH_DATA (preferences), &deskview_geometry, sizeof (deskview_geometry_t));
    dbh_set_recordsize (preferences, sizeof (deskview_geometry_t));

    if(!dbh_update (preferences)) {
        g_warning ("!dbh_update(deskview_geometry preferences)");
    }
    dbh_close (preferences);
    TRACE ("saved deskview_geometry preferences with key=%s\n", fullpath);
    return;
}

void
rodent_recalc_population_geometry (view_t * view_p) {
   if (!view_p || view_p->population_pp==NULL || *view_p->population_pp==NULL || view_p->view_layout.max_elements==0){
       TRACE("rodent_recalc_population_geometry: skipping ...\n");
       return;
   }
   TRACE("rodent_recalc_population_geometry: grid_columns=%d grid_rows=%d max_elements=%d\n", 
	    view_p->view_layout.grid_columns, view_p->view_layout.grid_rows, view_p->view_layout.max_elements);

   gint row;
   gint column;
   for(column = 0; column < view_p->view_layout.grid_columns; column++) {
       for(row = 0; row < view_p->view_layout.grid_rows; row++) {
	   gint element;
	   if (view_p->flags.type == ICONVIEW_TYPE){
	       element = row * view_p->view_layout.grid_columns + column;
	   } else { //transposed...
		element = column * view_p->view_layout.grid_rows + row;
	   }
	   if (element >= view_p->view_layout.max_elements) break;
	   if (view_p->population_pp[element] == NULL) break;
	   // if element is repeated as first column of next row, skip it 
	   if(element == (row + 1) * view_p->view_layout.grid_columns){
	       TRACE("skipping %s\n",view_p->population_pp[element]->en->path);
	       continue;
	   }                
	   view_p->population_pp[element]->column = column;
	   view_p->population_pp[element]->row = row;

       }
   }
   return;
}

