
/*
 * Copyright (C) 2002-2011 Edscott Wilson Garcia
 * EMail: edscott@xfce.org
 *
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif


#include "rodent_libs.h"
#include "rodent_gridview.i"


void 
rodent_new_gridview(widgets_t *widgets_p, const gchar *in_path){
    gchar *esc_path = NULL;
    const gchar *exec="rodent-forked";
    TRACE("in_path: %s\n", in_path);

    if (in_path!=NULL) {
	if (strncmp(in_path,"exec:",strlen("exec:"))==0){
	    exec=in_path+strlen("exec:");
	} 
    }
    gchar *argv[]={(gchar *)exec, (gchar *)in_path, NULL};
    TRACE("child_constructor: %s \"%s\"\n", argv[0], (argv[1])?argv[1]:"NULL");
    rfm_thread_run_argv (widgets_p, argv, FALSE);
    g_free(esc_path);
}

view_t *
rodent_get_current_view ( GtkWidget * window) {
    if (!window){
	g_warning("rodent_get_current_view() window==NULL");
    }
    GtkWidget *notebook=g_object_get_data(G_OBJECT(window), "notebook");   
    view_t *view_p;
    rfm_global_p->window_view_list_readers++;
    if (notebook==NULL && rfm_global_p->window_view_list) { // single view: desktop
	view_p = (view_t *)(rfm_global_p->window_view_list->data);
    } else {
	// we get the current page ...
	gint npage = gtk_notebook_get_current_page (GTK_NOTEBOOK (notebook));
	if (npage < 0) return NULL;
	GtkWidget *child =gtk_notebook_get_nth_page(GTK_NOTEBOOK (notebook), npage);
	view_p = g_object_get_data (G_OBJECT (child), "view_p");
    }
    rfm_global_p->window_view_list_readers--;
    return view_p;
}

int 
rodent_text_color(population_t *population_p){
    int text_color = TEXT_COLOR;
    //view_t *view_p=population_p->view_p;
    if(population_p && population_p->en && population_p->en->path && !g_utf8_validate (population_p->en->path, -1, NULL)){
        text_color = INVALID_UTF8_TEXT_COLOR;
    }
    if(population_p->flags & LABEL_SATURATED){
//        text_color = SATURATE_LABEL_COLOR;
    }
    if(population_p->flags & POPULATION_SELECTED){
        text_color = SELECT_TEXT_COLOR;
    }
    else if(population_p->flags & POPULATION_SATURATED){
        text_color = SATURATE_TEXT_COLOR;
    }
    return text_color;
}

void
rodent_init_grid (view_t * view_p, int pathc) {
    if (pathc == 0){
	TRACE("init_grid: skipping...\n");
	return;
    }

    view_p->view_layout.max_elements = pathc;

    //Do we have a user selected geometry?
    //(this is too nerdy. Let the user make her choices)
    //This does one thing for direct read and a different
    //one for cache read. Too buggy.
    /*if (!IS_USER_PREFERENCES(view_p->preferences)) {
	//We do not have a user selected geometry.
	//In this case we may select tiny icons on path count
	if (pathc >= TINY_ICONS_PATHC){
	    view_p->view_layout.icon_size = TINY_ICON_SIZE;
	}
    }*/

    // this is CELLWIDTH(view_p) !!!
    if (view_p->view_layout.icon_size >= SMALL_ICON_SIZE) {
	view_p->view_layout.column_width = ICON_SIZE(view_p) * 2; 
    } else if (view_p->view_layout.icon_size >= TINY_ICON_SIZE) {
	view_p->view_layout.column_width = SMALL_ICON_SIZE * 4; 
    } else {
	view_p->view_layout.column_width = view_p->widgets.window->allocation.width; 
    }

    if(!view_p->widgets.window) {
        g_warning ("init_grid: view_p->widgets.window==NULL\n");
        view_p->view_layout.grid_columns = MINIMUM_ICON_COLUMNS;
        return;
    }

    // minus margin
    gint width = view_p->widgets.window->allocation.width - (CELLWIDTH(view_p) / 2);

    TRACE ("init_grid: view_p->view_layout.max_elements=%d CELLWIDTH(view_p)=%d\n", pathc, CELLWIDTH(view_p));
    if (view_p->flags.type == ICONVIEW_TYPE) {
	view_p->view_layout.grid_columns = ((double)width / CELLWIDTH(view_p)) + 0.5;
	view_p->view_layout.grid_rows = (view_p->view_layout.max_elements) / view_p->view_layout.grid_columns;
	if(view_p->view_layout.max_elements % view_p->view_layout.grid_columns)
	    view_p->view_layout.grid_rows++;
    } else { // transposed...
	gint height = view_p->widgets.window->allocation.height - view_p->view_layout.margin_bottom - view_p->view_layout.margin_top;
	width = view_p->widgets.window->allocation.width - view_p->view_layout.margin_right - view_p->view_layout.margin_left;

	view_p->view_layout.grid_rows = ((double)height / CELLHEIGHT(view_p)) + 0.5;
	view_p->view_layout.grid_columns = ((double)width / CELLWIDTH(view_p)) + 0.5;
	if (view_p->view_layout.grid_rows * view_p->view_layout.grid_columns < view_p->view_layout.max_elements) {
	    view_p->view_layout.max_elements = view_p->view_layout.grid_rows * view_p->view_layout.grid_columns;
	}
    }



    TRACE ("init_grid (%d):width= %d, rows=%d columns=%d\n", 
	    pathc, view_p->widgets.window->allocation.width,
	    view_p->view_layout.grid_rows, view_p->view_layout.grid_columns);
    view_p->view_layout.grid_area = view_p->view_layout.grid_rows * view_p->view_layout.grid_columns;
    view_p->view_layout.paperX = CELLWIDTH(view_p);
    view_p->view_layout.paperY = view_p->view_layout.grid_rows * CELLHEIGHT(view_p);
    return;
}

/* restore window size   */
void
rodent_restore_view_size (view_t * view_p) {
    /* watchout, this is for paper resize */
    TRACE (">> geometry: rodent_restore_view_size\n");
    view_geometry_t *view_geometry_p;
    view_geometry_p = rodent_get_view_geometry_p (view_p);
    if(!view_geometry_p) {
        TRACE ("rodent_restore_view_size: no view_geometry_p!\n");
        return;
    }
    TRACE ("rodent_restore_view_size: apply_window_size(%d,%d)\n", view_geometry_p->w, view_geometry_p->h);

    apply_window_size (view_p, view_geometry_p->w, view_geometry_p->h);

    //gdk_flush ();
    TRACE ("PAPER rodent_restore_view_size: paper size=(%d,%d), allocation=(%d,%d)\n", view_p->view_layout.paperX, view_p->view_layout.paperY,
         view_p->widgets.paper->allocation.width, view_p->widgets.paper->allocation.height);


}

/**
 * rodent_on_size:
 *
 * figures out the number of columns and rows in the paper
 * from resize event data *
 * **/

gboolean
rodent_on_size (view_t * view_p, int width, int height) {
    TRACE (">> rodent_on_size\n");
    int w,
      h;
    GtkAdjustment *adjustment;
    TRACE ("rodent_on_size:: w+h=area... %d x %d = %d\n", width, height, width * height);
    if(!view_p || !view_p->widgets.paper) {
        g_warning ("!view_p || !view_p->widgets.paper");
        return TRUE;
    }
    if(!CELLWIDTH(view_p)) {
        TRACE ("!CELLWIDTH(view_p)\n");
        return TRUE;
    } else {
        TRACE ("CELLWIDTH(view_p)=%d\n", CELLWIDTH(view_p));
    }

    w = view_p->widgets.paper->allocation.width;
    h = view_p->widgets.paper->allocation.height;
    TRACE ("paper... w*h=area...%d x %d = %d\n", w, h, w * h);
    TRACE ("window... w*h=area...%d x %d = %d\n",
           view_p->widgets.window->allocation.width,
           view_p->widgets.window->allocation.height,
           view_p->widgets.window->allocation.width * view_p->widgets.window->allocation.height);

    TRACE ("rodent_on_size:rows and columns: %d,%d\n", view_p->view_layout.grid_rows, view_p->view_layout.grid_columns);
    if(GTK_IS_SCROLLED_WINDOW (view_p->widgets.scrolled_window)) {
        adjustment = gtk_scrolled_window_get_vadjustment (view_p->widgets.scrolled_window);
        TRACE ("adjustment upper=%lf, adjustment pagesize=%lf\nf", adjustment->upper, adjustment->page_size);
        adjustment->upper = view_p->view_layout.grid_rows * CELLHEIGHT(view_p);
    }
    return TRUE;
}

// this arbitrarily set scroll position
void
rodent_set_scroll (view_t * view_p, double scroll_position) {
    TRACE ("SCROLL: scroll_position = %lf\n", scroll_position);
    if(!GTK_IS_SCROLLED_WINDOW (view_p->widgets.scrolled_window))
        return;

    double upper = gtk_adjustment_get_upper (
	    gtk_scrolled_window_get_vadjustment (
		view_p->widgets.scrolled_window));
    double page = gtk_adjustment_get_page_size (
	    gtk_scrolled_window_get_vadjustment (
		view_p->widgets.scrolled_window));
    if(scroll_position < upper - page) {
        gtk_adjustment_set_value (
		gtk_scrolled_window_get_vadjustment (
		    view_p->widgets.scrolled_window), scroll_position);

    } else {
        TRACE ("SCROLL: geometryZ3: setting scroll position to 0 \n");
        gtk_adjustment_set_value (
		gtk_scrolled_window_get_vadjustment (
		    view_p->widgets.scrolled_window), upper - page);
    }
    gtk_adjustment_changed (
	    gtk_scrolled_window_get_vadjustment (
		view_p->widgets.scrolled_window));
}

// this sets scroll position from saved geometry
void
rodent_set_scroll_position (view_t * view_p) {
    if(!GTK_IS_SCROLLED_WINDOW (view_p->widgets.scrolled_window))
        return;
    int window_area = view_p->widgets.window->allocation.height * view_p->widgets.window->allocation.width;
    int icon_area = view_p->widgets.window->allocation.width * view_p->view_layout.grid_rows * CELLHEIGHT(view_p);

    TRACE ("SCROLL: icon_area=%d, window_area=%d\n", icon_area, window_area);
    TRACE ("SCROLL: geometry: setting scroll position grid_rows= %d grid_columns=%d\n", view_p->view_layout.grid_rows, view_p->view_layout.grid_columns);
    view_geometry_t *view_geometry_p = rodent_get_view_geometry_p (view_p);
    if(icon_area < window_area) {
        TRACE ("SCROLL: geometryZ1: setting scroll position to 0 \n");
        gtk_adjustment_set_value (gtk_scrolled_window_get_vadjustment (view_p->widgets.scrolled_window), 0.0);
        gtk_adjustment_changed (gtk_scrolled_window_get_vadjustment (view_p->widgets.scrolled_window));
        return;
    }
	// inverse:
	// floor((upper+page) * scrollX +0.5)
    gdouble upper=gtk_adjustment_get_upper (
	    gtk_scrolled_window_get_vadjustment (
		view_p->widgets.scrolled_window));
    gdouble page=gtk_adjustment_get_page_size (
	    gtk_scrolled_window_get_vadjustment (
		view_p->widgets.scrolled_window));
    
    double scroll_position = floor((upper+page) * view_geometry_p->scrollX +0.5);
    TRACE ("SCROLL: geometryZ2: setting scroll position to %lf geometry height=%d\n", scroll_position, view_geometry_p->h);
    rodent_set_scroll (view_p, scroll_position);

}

static void
apply_new_size (view_t *view_p){
    if (view_p->module){
	// Set the iconsize in the module for the view_p.
	gint size=(view_p->view_layout.icon_size)?view_p->view_layout.icon_size:-1;
	rfm_rational(PLUGIN_DIR,view_p->module, view_p, 
		GINT_TO_POINTER(size), "module_icon_size");
    }
    record_entry_t *en = rfm_copy_entry (view_p->en);
    rfm_save_view_preferences (view_p, view_p->en);  
    if (!rodent_refresh (&(view_p->widgets), en)){ 
	rfm_destroy_entry(en); 
    }
}

static GMutex *sweep_mutex=NULL;
static void
reset_saved_iconsize (DBHashTable * old) {
    DBHashTable *new = old->sweep_data;
    gint record_size = DBH_RECORD_SIZE(old);

    dbh_set_recordsize (new, record_size);
    memcpy(DBH_KEY(new), DBH_KEY(old), DBH_KEYLENGTH(old));
    memcpy(new->data, old->data, record_size);
    view_preferences_t *view_preferences_p = new->data;
    view_preferences_p->icon_size = rfm_get_default_size();
    TRACE("sweep now, record size=%d\n", record_size);
    dbh_update(new);
    return;
}

void
rodent_default_iconsize_all (GtkButton * button, gpointer data){
    // The easy way would be just to unlink the preferences file,
    // but that would take down sort orders and show hidden
    // attributes as well. So instead we just set each saved preference
    // item to user default size.
    widgets_t *widgets_p =  data;
    view_t *view_p = widgets_p->view_p;
    gint default_size = rfm_get_default_size();

    if (!sweep_mutex) sweep_mutex = g_mutex_new();
    gchar *f;
    if (view_p->flags.type==ICONVIEW_TYPE) {
	f = g_build_filename (g_get_user_config_dir (), GRIDVIEW_PREFERENCES_DBH_FILE, NULL);
    } else {
	f = g_build_filename (g_get_user_config_dir (), DESKVIEW_PREFERENCES_DBH_FILE, NULL);
    }
    DBHashTable *old = dbh_open_ro (f);
    if(old == NULL) {
	g_free(f);
	return;
    }

    gchar *ff=g_strconcat(f,"-new",NULL);
    g_mutex_lock(sweep_mutex);
    DBHashTable *new = dbh_create (ff, DBH_KEYLENGTH (old));

    if(new == NULL) {
	dbh_close(old);
	g_warning("cannot create file %s", ff);
	g_free(f);
	g_free(ff);
	g_mutex_unlock(sweep_mutex);
	return;
    }
    old->sweep_data = new;
    TRACE("sweep now\n");
    dbh_foreach_sweep (old, reset_saved_iconsize);

    dbh_close(old);
    dbh_close(new);

    /*if (unlink(f) < 0) {
	g_warning("unlink(%s): %s", f, strerror(errno));
    }*/
    TRACE("rename(%s, %s)\n", f, ff);
    if (rename (ff,f) < 0) {
	g_warning("rename(%s, %s): %s", f, ff, strerror(errno));
    }

	
    g_free(f);
    g_free(ff);
	
    g_mutex_unlock(sweep_mutex);

    if (default_size != view_p->view_layout.icon_size) {
	view_p->view_layout.icon_size = default_size;
	apply_new_size (view_p);
    }
}

void
rodent_default_iconsize (GtkButton * button, gpointer data){
    widgets_t *widgets_p =  data;
    view_t *view_p = widgets_p->view_p;
    gint default_size = rfm_get_default_size();

    if (default_size != view_p->view_layout.icon_size) {
	view_p->view_layout.icon_size = default_size;
	apply_new_size (view_p);
    }
}

void
rodent_increase_iconsize (GtkButton * button, gpointer data){
    widgets_t *widgets_p =  data;
    view_t *view_p = widgets_p->view_p;
    gint original_size = view_p->view_layout.icon_size;
    view_p->view_layout.icon_size += 24;
    if (view_p->view_layout.icon_size > BIG_ICON_SIZE) {
	view_p->view_layout.icon_size = BIG_ICON_SIZE;
    }
    if (original_size != view_p->view_layout.icon_size) {
	apply_new_size (view_p);
    }
}

void
rodent_decrease_iconsize (GtkButton * button, gpointer data){
    widgets_t *widgets_p =  data;
    view_t *view_p = widgets_p->view_p;
    gint original_size = view_p->view_layout.icon_size;
    view_p->view_layout.icon_size -= 24;
    if (view_p->view_layout.icon_size < 0) {
	view_p->view_layout.icon_size = 0;
    }
    if (original_size != view_p->view_layout.icon_size) {
	apply_new_size (view_p);
    }
}

void
rodent_details_iconsize (GtkButton * button, gpointer data){
    widgets_t *widgets_p =  data;
    view_t *view_p = widgets_p->view_p;
    view_p->view_layout.icon_size = 0;
    apply_new_size (view_p);
}

void
rodent_compact_iconsize (GtkButton * button, gpointer data){
    widgets_t *widgets_p =  data;
    view_t *view_p = widgets_p->view_p;
    view_p->view_layout.icon_size = 24;
    apply_new_size (view_p);
}

void
rodent_normal_iconsize (GtkButton * button, gpointer data){
    widgets_t *widgets_p =  data;
    view_t *view_p = widgets_p->view_p;
    view_p->view_layout.icon_size = 48;
    apply_new_size (view_p);
}

void
rodent_big_iconsize (GtkButton * button, gpointer data){
    widgets_t *widgets_p =  data;
    view_t *view_p = widgets_p->view_p;
    view_p->view_layout.icon_size = 72;
    apply_new_size (view_p);
}

void
rodent_huge_iconsize (GtkButton * button, gpointer data){
    widgets_t *widgets_p =  data;
    view_t *view_p = widgets_p->view_p;
    view_p->view_layout.icon_size = 96;
    apply_new_size (view_p);
}


void  
rodent_size_scale (GtkRange *range,  gpointer  data){
    widgets_t *widgets_p =  data;
    view_t *view_p = widgets_p->view_p;
    gdouble value = gtk_range_get_value (range);
    if (value < TINY_ICON_SIZE / 2){
	// ls -l format
	value = 0;
    } else if (value < TINY_ICON_SIZE +(SMALL_ICON_SIZE - TINY_ICON_SIZE)/ 2){
	// compact format
	value = TINY_ICON_SIZE;
    } else if (value < SMALL_ICON_SIZE +(BIG_ICON_SIZE - SMALL_ICON_SIZE)/ 4){
	value = SMALL_ICON_SIZE;
    } else if (value < SMALL_ICON_SIZE +3*(BIG_ICON_SIZE - SMALL_ICON_SIZE) /4){
	value = SMALL_ICON_SIZE +(BIG_ICON_SIZE - SMALL_ICON_SIZE)/ 2;
    } else {
	value = BIG_ICON_SIZE;
    }
    gint i_value=value;
    if (i_value==view_p->view_layout.icon_size) return;
    
    view_p->view_layout.icon_size = value;
    gtk_range_set_value (range, value);
    apply_new_size (view_p);
}

view_geometry_t *
rodent_get_view_geometry_p (view_t * view_p) {
    TRACE (">> rodent_get_view_geometry_p\n");
    DBHashTable *geometry;
    const gchar *key;
    GString *gs;
    static view_geometry_t view_geometry;
    view_geometry_t path_view_geometry;
    gchar *f;

    if(!view_p)
        return NULL;
    /* defaults */
    view_geometry.w = DEFAULT_WIDTH;
    view_geometry.h = DEFAULT_HEIGHT;
    view_geometry.x = view_geometry.y = -1;
    view_geometry.scrollX = 0;

    if(view_p->en && view_p->en->path && strlen (view_p->en->path))
        key = view_p->en->path;
    else
        key = "RODENT_ROOT";

    f = g_build_filename (g_get_user_config_dir (), GRIDVIEW_GEOMETRY_DBH_FILE, NULL);
    TRACE ("looking for geometry file %s with key=%s\n", f, key);
    if(!g_file_test (f, G_FILE_TEST_EXISTS)) {
        TRACE ("no geometry file: %s\n", f);
        g_free (f);
        return &view_geometry;
    }

    geometry = dbh_open (f);
    g_free (f);
    if (!geometry) return NULL;

    /* load root record geometry */
    gs = g_string_new ("RODENT_ROOT");
    sprintf ((char *)DBH_KEY (geometry), "%10u", g_string_hash (gs));
    g_string_free (gs, TRUE);
    if(!dbh_load (geometry)) {
        TRACE ("no geometry record for %s (%s)", "RODENT_ROOT", (char *)DBH_KEY (geometry));
        dbh_close (geometry);
        return &view_geometry;
    }
    memcpy (&view_geometry, DBH_DATA (geometry), sizeof (view_geometry_t));

    gs = g_string_new (key);
    sprintf ((char *)DBH_KEY (geometry), "%10u", g_string_hash (gs));
    g_string_free (gs, TRUE);

    if(!dbh_load (geometry)) {
        TRACE ("no geometry record for %s (%s)\n", key, (char *)DBH_KEY (geometry));
    } else {
        memcpy (&path_view_geometry, DBH_DATA (geometry), sizeof (view_geometry_t));
        view_geometry.scrollX = path_view_geometry.scrollX;
#ifdef PER_PATH_GEOMETRY
        /* this is disabled by default */
        view_geometry.x = path_view_geometry.x;
        view_geometry.y = path_view_geometry.y;
        view_geometry.w = path_view_geometry.w;
        view_geometry.h = path_view_geometry.h;
#endif
    }
    //view_geometry.scrollX *= view_p->view_layout.grid_rows;
    TRACE ("(graphics_on_size) got geometry with key=%s  (scroll %lf,x=%d y=%d, w=%d h=%d)\n", key, view_geometry.scrollX,
         view_geometry.x, view_geometry.y, view_geometry.w, view_geometry.h);
    dbh_close (geometry);
    if(view_geometry.w <= 0 || view_geometry.h <= 0) {
        view_geometry.w = DEFAULT_WIDTH;
        view_geometry.h = DEFAULT_HEIGHT;
    }
    return &view_geometry;
}

void
rodent_save_view_geometry_p (view_t * view_p, const char *inkey) {
    TRACE (">> rodent_save_view_geometry_p\n");
    view_geometry_t view_geometry;
    DBHashTable *geometry;
    GString *gs;
    gchar *f;
    const gchar *key;
    if(!view_p)
        return;

    if(!view_p->en || !view_p->en->path)
        key = "RODENT_ROOT";
    else if(inkey)
        key = inkey;
    else
        key = view_p->en->path;

    f = g_build_filename (g_get_user_config_dir (), GRIDVIEW_GEOMETRY_DBH_FILE, NULL);
    if(!g_file_test (f, G_FILE_TEST_EXISTS)) {
        TRACE ("Creating geometry file: %s", f);
        geometry = dbh_create (f, 11);
    }

    geometry = dbh_open (f);

    if(!geometry) {
        g_warning ("cannot open geometry file %s", f);
        g_free (f);
        return;
    }
    g_free (f);

    gs = g_string_new (key);
    sprintf ((char *)DBH_KEY (geometry), "%10u", g_string_hash (gs));
    g_string_free (gs, TRUE);

    view_geometry.w = view_p->widgets.window->allocation.width;
    view_geometry.h = view_p->widgets.window->allocation.height;

    if(strcmp (key, "RODENT_ROOT") == 0)
        view_geometry.scrollX = 0;
    else {
	gdouble value=gtk_adjustment_get_value (
		gtk_scrolled_window_get_vadjustment (
		    view_p->widgets.scrolled_window));
	gdouble upper=gtk_adjustment_get_upper (
		gtk_scrolled_window_get_vadjustment (
		    view_p->widgets.scrolled_window));
	gdouble page=gtk_adjustment_get_page_size (
		gtk_scrolled_window_get_vadjustment (
		    view_p->widgets.scrolled_window));
	// size=upper+page;
	// place=value/size;
        view_geometry.scrollX = value/(upper+page);
	// inverse:
	// floor((upper+page) * scrollX +0.5)
    }

    TRACE ("geometry graphics: saving ajustment at %lf (%s)\n", view_geometry.scrollX, key);
    gtk_window_get_position ((GtkWindow *) view_p->widgets.window, &(view_geometry.x), &(view_geometry.y));

    TRACE ("(graphics_on_size)  saving geometry %d,%d size %d,%d\n",
         view_geometry.x, view_geometry.y, view_geometry.w, view_geometry.h);

    /* here we save the per path geometry */
    memcpy (DBH_DATA (geometry), &view_geometry, sizeof (view_geometry_t));
    dbh_set_recordsize (geometry, sizeof (view_geometry_t));
    if(!dbh_update (geometry))
        g_warning ("!dbh_update(geometry)");
    TRACE ("1.saved geometry with key=%s (%lf,%d,%d)\n", key, view_geometry.scrollX, view_geometry.w, view_geometry.h);

    dbh_close (geometry);
    return;
}

void
rodent_save_local_view_geometry_p (view_t * view_p) {
    rodent_save_view_geometry_p (view_p, NULL);
}

void
rodent_save_root_view_geometry_p (view_t * view_p) {
    rodent_save_view_geometry_p (view_p, "RODENT_ROOT");
}
