
/*
 * Copyright (C) 2002-2011 Edscott Wilson Garcia
 * EMail: edscott@xfce.org
 *
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include "rodent_libs.h"

#define CONTROL_MODE (event->state & GDK_CONTROL_MASK)
#define NO_CLICK 0
#define LABEL_CLICK 1
#define ICON_CLICK 2

#include "rodent_mouse.h"


static void hide_tip (view_t * view_p);

#include "rodent_mouse.i"


void
rodent_saturate_item (view_t *view_p, const population_t * population_p) {
    unsaturate_item (view_p);
    saturate_item (view_p, population_p);
}

void
rodent_unsaturate_item (view_t *view_p) {
    unsaturate_item (view_p);
}
/*******************************************************************/
/*           GTK   signal callbacks */
/******************************************************************/

gboolean
rodent_signal_on_button_press (GtkWidget * widget, GdkEventButton * event, gpointer data) {

    TRACE (">>  rodent_signal_on_button_press: %d\n", event->button);
    view_t *view_p = (view_t *) data;
    if (view_p->flags.protected){
	return TRUE;
    }

    /* cancel any pending tip : */
    view_p->tip_event.tip_timer = 0;
    if(view_p->widgets.rename) {
        rodent_done_with_rename ((gpointer) view_p);
        return TRUE;
    }
    if (view_p->mouse_event.rubberbanding) return TRUE;
     rfm_population_read_lock (view_p);
    // totally useless:
    // if (rfm_population_try_read_lock (view_p)){ return TRUE; }

    gboolean label_click = FALSE;


    if(event->button == 1) {
        view_p->mouse_event.mouseX = event->x;
        view_p->mouse_event.mouseY = event->y;
    } else {
        view_p->mouse_event.mouseX = -1;
        view_p->mouse_event.mouseY = -1;
    }
    TRACE ("on_button_press\n");
    /* ignore size events until button released... */
    view_p->mouse_event.eventtime = event->time;
    if(event->button == 2) {
        g_object_set_data (G_OBJECT (view_p->widgets.paper), "normal_tip", GINT_TO_POINTER (1));
    }
    // get population_p into stack memory (population_v) or return


    record_entry_t en_v;
    population_t population_v;

    const population_t *population_p = rodent_find_in_population (view_p, event->x, event->y);
    if(!population_p) {
        population_p = rodent_find_in_labels (view_p, event->x, event->y);
	TRACE("label click is true\n");
        label_click = TRUE;
    }
    if(population_p) {
        memcpy (&population_v, population_p, sizeof (population_t));
        if(population_p->en) {
            memcpy (&en_v, population_p->en, sizeof (record_entry_t));
            population_v.en = &en_v;
        } else {
            population_v.en = NULL;
        }
    }

    TRACE ("OPEN: rodent_find_in_population=0x%lx\n", (unsigned long)population_p);
    if(event->button == 3) {
        TRACE ("uh-huh. a popup now.\n");
        if(!population_p) {
            rodent_unselect_all_pixbuf (view_p);
            rodent_unsaturate_item (view_p);
            view_p->mouse_event.selected_p = NULL;
        }
        view_p->mouse_event.doing_drag_p = NULL;
        view_p->mouse_event.dragstate = FALSE;
        view_p->mouse_event.rubberbanding = FALSE;
        view_p->mouse_event.old_X = view_p->mouse_event.old_Y = view_p->mouse_event.boxX = view_p->mouse_event.boxY = -1;
        hide_tip (view_p);
        // button_popup locks the monitor loop
        // population_p may or may not be valid, but population_v is 
        // no good here...
        rfm_population_read_unlock (view_p);
        button_popup (event, view_p, population_p);
        return TRUE;
    }
    // this will start the box selection:
    if(!population_p && event->button == 1) {
        view_p->mouse_event.selected_p = NULL;
        TRACE ("nothing found at %f %f\n", event->x, event->y);
        if(!(event->state & GDK_CONTROL_MASK)) {
            rodent_unselect_all_pixbuf (view_p);
            rodent_unsaturate_item (view_p);
        }
        view_p->mouse_event.rubberbanding = TRUE;
        view_p->mouse_event.boxX = event->x;
        view_p->mouse_event.boxY = event->y;
        rfm_population_read_unlock (view_p);
        return TRUE;
    }

    if(event->button == 2) {
        if((event->type == GDK_2BUTTON_PRESS)) {
            view_p->mouse_event.doing_drag_p = NULL;
            view_p->mouse_event.dragstate = FALSE;
	    view_p->mouse_event.rubberbanding = FALSE;
            view_p->mouse_event.old_X = view_p->mouse_event.old_Y = view_p->mouse_event.boxX = view_p->mouse_event.boxY = -1;

            rodent_unselect_all_pixbuf (view_p);
            rodent_unsaturate_item (view_p);
            rodent_select_pixbuf (view_p, population_p);
	    rodent_expose_item (view_p, population_p);
            rfm_population_read_unlock (view_p);
            TRACE ("unlock for doubleclick\n");
            rodent_double_click (view_p, &population_v, TRUE);
        } else {
            rfm_population_read_unlock (view_p);
        }
        return TRUE;
    }

    view_p->mouse_event.dragstate = FALSE;

    if(event->state & GDK_SHIFT_MASK && !label_click) {
	if (population_p && view_p->population_pp){
	    population_t **pp, **to=NULL, **from=NULL;
	    for (pp=view_p->population_pp; pp && *pp; pp++){
		population_t *p=*pp;
		gboolean from_c=
		    !from && 
		    (find_in_selection_list(view_p, p->en)
		     || p==population_p);
		if (from_c) from=pp;
		gboolean to_c=
		    from &&  
		    (find_in_selection_list(view_p, p->en)
		     || p==population_p);
		if (to_c && pp!=from) to=pp;
	    }
	    if (*from != population_p && *to  != population_p) to=NULL;
	    if (!to) to=from;
	    rodent_unselect_all_pixbuf (view_p);
	    for (pp=from; pp && *pp; pp++){
		rodent_select_pixbuf (view_p, *pp);
		rodent_expose_item (view_p, *pp);
		if (pp==to) break;
	    }
	}
        rfm_population_read_unlock (view_p);
        return TRUE;
    }
    if(!CONTROL_MODE) {
        rodent_unselect_all_pixbuf (view_p);
        rodent_select_pixbuf (view_p, population_p);
	rodent_expose_item (view_p, population_p);
    }

    /* set up drag status */
    view_p->mouse_event.doing_drag_p = NULL;
    view_p->mouse_event.dragstate = FALSE;
    view_p->mouse_event.rubberbanding = FALSE;
    view_p->mouse_event.old_X = view_p->mouse_event.old_Y = view_p->mouse_event.boxX = view_p->mouse_event.boxY = -1;
    int x_spacing;
    x_spacing = (CELLWIDTH(view_p) - population_v.pixbufW) / 2;
    if(x_spacing < 0) {
        x_spacing = 0;
    }
//  view_p->mouse_event.doing_drag_p=population_p;
    TRACE ("doing_drag_p=0x%lx \n", (unsigned long)view_p->mouse_event.doing_drag_p);
    if(event->button == 1) {
	setup_drag_state (view_p, event);
    }

    if(SINGLE_CLICK_MODE) {
        if(label_click) {
            rodent_unselect_all_pixbuf (view_p);
            rodent_unsaturate_item (view_p);
            view_p->mouse_event.single_click_mode_state = LABEL_CLICK;
        } else {
            view_p->mouse_event.single_click_mode_state = ICON_CLICK;
        }
        rfm_population_read_unlock (view_p);
        /* double click status gets processed on button release... */
	TRACE("SINGLE_CLICK_MODE return\n");
        return TRUE;
    }

    /* now we process doubleclick, if any */
    view_p->mouse_event.single_click_mode_state = NO_CLICK;

    TRACE("now we process doubleclick, if any\n");
    if(event->type == GDK_2BUTTON_PRESS) {
        if(!label_click) {
	    TRACE("label click is false\n");
	    if (event->state & GDK_SHIFT_MASK) {
		rfm_population_read_unlock (view_p);
                return TRUE;
	    }
            rodent_unselect_all_pixbuf (view_p);
            rodent_unsaturate_item (view_p);
            rodent_select_pixbuf (view_p, population_p);
	    rodent_expose_item (view_p, population_p);
            rfm_population_read_unlock (view_p);
            rodent_double_click (view_p, &population_v, FALSE);
            return TRUE;
        } else {
	    TRACE("label click...\n");
            gint caso = 0;
            if(event->state & GDK_CONTROL_MASK)
                caso = 1;
            if(event->state & GDK_MOD1_MASK)
                caso = 2;
            if(event->state & GDK_MOD5_MASK)
                caso = 2;
            if(event->state & GDK_SHIFT_MASK && event->state & GDK_CONTROL_MASK)
                caso = 2;
            rodent_unselect_all_pixbuf (view_p);
            rodent_unsaturate_item (view_p);
            rodent_select_pixbuf (view_p, population_p);
	    rodent_expose_item (view_p, population_p);

            if(population_v.en && !IS_ROOT_TYPE (population_v.en->type)
               && !IS_DUMMY_TYPE (population_v.en->type)
               && population_v.en->path) {
                //view_p->mouse_event.selected_p = population_p;
                if(g_path_is_absolute(population_v.en->path) && 
		    (IS_BROKEN_LNK(population_v.en->type) ||
			g_file_test (population_v.en->path, G_FILE_TEST_EXISTS))) {
                    // here we respecify labelX and labelY to avoid
                    // race with reload after a finished rename 
                    //population_p->labelX = CELLWIDTH(view_p) * population_p->column;
                    //population_p->labelY =
                    //    CELLHEIGHT(view_p) * population_p->row + ICON_SIZE(view_p);
                    rodent_mk_text_entry (view_p, &population_v, caso);
                } else if(view_p->module_name) {
                    // if it's not a valid path, and loaded by a module
                    // then the module should know what to do here...
                    g_warning ("not a valid path loaded by module... \nmodule should know what to do here!\n");
                }
            }
        }
    }
    TRACE ("rodent_signal_on_button_press() done!\n");
    rfm_population_read_unlock (view_p);

    return TRUE;
}

gboolean
rodent_signal_on_button_release (GtkWidget * widget, GdkEventButton * event, gpointer data) {
    TRACE ("RECT>> rodent_signal_on_button_release event x=%lf y=%lf\n", event->x, event->y);

    view_t *view_p = (view_t *) data;
    if (view_p->flags.protected){
	return TRUE;
    }
    if (view_p->mouse_event.rubberbanding && event->button != 1) return TRUE;
    const population_t *population_p;
    /*view_p->mouse_event.mouseX=event->x;
       view_p->mouse_event.mouseY=event->y; */
    TRACE ("button release...\n");
    if(event->x < 0)
        event->x = 0;
    if(event->y < 0)
        event->y = 0;
    TRACE ("RECT>> rodent_signal_on_button_release event x=%lf y=%lf\n", event->x, event->y);

    view_p->mouse_event.mouseX = -1;
    view_p->mouse_event.mouseY = -1;
    if(event->button == 2) {
        g_object_set_data (G_OBJECT (view_p->widgets.paper), "normal_tip", NULL);
    }
    if (view_p->mouse_event.doing_drag_p) return TRUE;
    // get population_p into stack memory (population_v) or return
    if(!rfm_population_try_read_lock (view_p)) {
        return TRUE;
    }
    record_entry_t en_v;
    population_t population_v;

    TRACE ("population_sem: rodent_signal_on_button_release() obtained...\n");
    population_p = rodent_find_in_population (view_p, (gdouble) event->x, (gdouble) event->y);
    if(!population_p) {
        population_p = rodent_find_in_labels (view_p, event->x, event->y);
    }
    if(population_p) {
        memcpy (&population_v, population_p, sizeof (population_t));
        if(population_p->en) {
            memcpy (&en_v, population_p->en, sizeof (record_entry_t));
            population_v.en = &en_v;
        } else {
            population_v.en = NULL;
        }
    }

    if(population_p && (event->state & GDK_SHIFT_MASK) && ! (event->state & GDK_CONTROL_MASK)) {
            rfm_population_read_unlock (view_p);
            return TRUE;
    }
    if(population_p && CONTROL_MODE) {
        if(!(population_p->flags  & POPULATION_SELECTED)) {
            rodent_select_pixbuf (view_p, population_p);
	    rodent_expose_item (view_p, population_p);
        } else {
            unselect_pixbuf (view_p, population_p);
	    rodent_expose_item(view_p, population_p);
        }
    }


    if(population_p && SINGLE_CLICK_MODE && (event->button == 1)
       && view_p->mouse_event.single_click_mode_state) {
        if(CONTROL_MODE && view_p->mouse_event.single_click_mode_state != LABEL_CLICK) {
            view_p->mouse_event.single_click_mode_state = NO_CLICK;
            view_p->mouse_event.doing_drag_p = NULL;
	    view_p->mouse_event.rubberbanding = FALSE;
            view_p->mouse_event.old_X = view_p->mouse_event.old_Y = view_p->mouse_event.boxX = view_p->mouse_event.boxY = -1;
            rfm_update_status_line (view_p);
            rfm_population_read_unlock (view_p);
            return TRUE;
        }

        if(view_p->mouse_event.single_click_mode_state == ICON_CLICK) {
	    TRACE("ICON_CLICK\n");
            rodent_unselect_all_pixbuf (view_p);
            rodent_select_pixbuf (view_p, population_p);
            //rodent_unsaturate_item (view_p);
	    rodent_expose_item (view_p, population_p);
            rfm_population_read_unlock (view_p);
	    // this is threaded, so leave saturated...
            rodent_double_click (view_p, &population_v, FALSE);
            return TRUE;
        } else { // LABEL_CLICK
	    if (view_p->module && 
		    rfm_natural(PLUGIN_DIR,view_p->module, population_v.en, "label_click")){
		TRACE("do nothing here\n");
	    } else if(population_v.en && !IS_ROOT_TYPE (population_v.en->type)
               && !IS_DUMMY_TYPE (population_v.en->type)
               && population_v.en->path) {
		TRACE("do something here\n");
                int caso = 0;
                if(event->state & GDK_CONTROL_MASK)
                    caso = 1;
                if(event->state & GDK_MOD1_MASK)
                    caso = 2;
                if(event->state & GDK_MOD5_MASK)
                    caso = 2;
                if(event->state & GDK_SHIFT_MASK && event->state & GDK_CONTROL_MASK)
                    caso = 2;
                //view_p->mouse_event.selected_p = population_p;
                if(g_path_is_absolute(population_v.en->path) &&
		    (IS_BROKEN_LNK(population_v.en->type) ||
			g_file_test (population_v.en->path, G_FILE_TEST_EXISTS))) {
                    /* here we respecify labelX and labelY to avoid
                     * race with reload after a finished rename */
                    //population_p->labelX = CELLWIDTH(view_p) * population_p->column;
                    //population_p->labelY =
                    //    CELLHEIGHT(view_p) * population_p->row + ICON_SIZE(view_p);
                    rodent_mk_text_entry (view_p, &population_v, caso);
                } /*else if(view_p->module_name) {
                    TRACE ("FIXME: ask module\n");
                }*/
            }
        }
        view_p->mouse_event.single_click_mode_state = NO_CLICK;
    }

    view_p->mouse_event.doing_drag_p = NULL;
    rubber_band (view_p, event->x, event->y, FALSE);


    view_p->mouse_event.rubberbanding = FALSE;
    view_p->mouse_event.old_X = view_p->mouse_event.old_Y = view_p->mouse_event.boxX = view_p->mouse_event.boxY = -1;
    rfm_update_status_line (view_p);
    rfm_population_read_unlock (view_p);

    TRACE ("rodent_signal_on_button_release() done!\n");
    return TRUE;
}

gboolean
rodent_signal_on_motion (GtkWidget * widget, GdkEventMotion * event, gpointer data) {

    TRACE (">> rodent_signal_on_motion\n");
    view_t *view_p = (view_t *) data;
    if (view_p->flags.protected){
	return TRUE;
    }
    if(event->x < 0) event->x = 0;
    if(event->y < 0)  event->y = 0;
    view_p->mouse_event.current_mouseX = event->x;
    view_p->mouse_event.current_mouseY = event->y;
    //
    // if we are window selecting, don't do individual saturation 
    // (at least not yet)
    // no read lock is necessary...
    //if(view_p->mouse_event.boxX != -1 && view_p->mouse_event.boxY != -1) {
    if(view_p->mouse_event.rubberbanding) {
	Display *display=gdk_x11_display_get_xdisplay(gdk_display_get_default());
	XGrabServer(display);
        rubber_band (view_p, event->x, event->y, TRUE);
        gint X = event->x - view_p->mouse_event.boxX;
        gint Y = event->y - view_p->mouse_event.boxY;
        if(X * X + Y * Y > 2)
            view_p->mouse_event.single_click_mode_state = NO_CLICK;
        /* reselect */
        /* here you could select_population_area() on the fly.
         * the only difficulty is unselecting on the fly
         * as the user changes her mind */
        /*select_population_area(view_p); */
	// nothing else to do here...
	XUngrabServer(display);
        return TRUE;
    }  

    // get semaphore or return:
    if(!rfm_population_try_read_lock (view_p)) {
        return TRUE;
    }
    TRACE ("population_sem: rodent_signal_on_motion() obtained...\n");


    if(view_p->mouse_event.current_mouseX >= 0 &&
       (abs (view_p->mouse_event.current_mouseX - view_p->mouse_event.mouseX) >
        ICON_SIZE(view_p) / 4 || abs (view_p->mouse_event.current_mouseY - view_p->mouse_event.mouseY) > ICON_SIZE(view_p) / 4)) 
    {
        view_p->mouse_event.doing_drag_p = rodent_find_in_population (view_p, view_p->mouse_event.mouseX, view_p->mouse_event.mouseY);
	if (view_p->mouse_event.doing_drag_p && view_p->mouse_event.doing_drag_p->en) {
	    TRACE ("turning drag on... %s current_mouseX=%d mouseX=%lf current_mouseY=%d mouseY=%lf\n", 
		    view_p->mouse_event.doing_drag_p->en->path,
		    view_p->mouse_event.current_mouseX, view_p->mouse_event.mouseX,
		    view_p->mouse_event.current_mouseY, view_p->mouse_event.mouseY
		    );
	}
    }

    if(view_p->mouse_event.doing_drag_p) {
	gdouble h = (event->x - view_p->mouse_event.mouseX) * (event->x - view_p->mouse_event.mouseX) + (event->y - view_p->mouse_event.mouseY) * (event->y - view_p->mouse_event.mouseY);
	TRACE ("h=%f\n", h);
	if(h >= 4)
	    enter_drag_state (view_p);
    }

    population_t *population_p = (population_t *)
	rodent_find_in_population (view_p, event->x, event->y);
    
    TRACE ("on motion\n");

    if(population_p) {
	rodent_saturate_item (view_p, population_p);    
	/*TRACE("x=%f,y=%f (%f,%f)\n",
           event->x,event->y,view_p->mouse_event.mouseX,view_p->mouse_event.mouseY); */


	if(getenv ("RFM_ENABLE_TIPS")==NULL || strlen (getenv ("RFM_ENABLE_TIPS"))==0) {
	    // tips disabled condition 
	    TRACE ("tips disabled by environment variable\n");
	} 
	else {
	    gtk_widget_set_tooltip_text (view_p->widgets.window, "Rodent");
	    if (!(population_p->flags  & POPULATION_TIP_BUSY)) {
#if 10
	    // don't do any of this if preview is busy
	    population_p->flags |= POPULATION_TIP_BUSY;
	    activate_tip_t *activate_tip_p=
		(activate_tip_t *)malloc(sizeof(activate_tip_t));
	    activate_tip_p->view_p=view_p;
	    activate_tip_p->population_serial=view_p->population_serial;
	    activate_tip_p->population_p=population_p;
	    g_thread_create(activate_tip_thread, (gpointer) activate_tip_p, FALSE, NULL);
	    }
	    gtk_widget_trigger_tooltip_query(view_p->widgets.window);
#endif
	}
	rfm_population_read_unlock (view_p);
	return TRUE;
    } 
    unsaturate_item (view_p);    
    
    TRACE ("population_sem: rodent_signal_on_motion() released!\n");
    population_p = 
	    (population_t *)rodent_find_in_labels (view_p, event->x, event->y);
    if (!population_p) {
	view_p->tip_event.tooltip_active = FALSE;
	// allow tip thread to fireup again
	// XXX this loop might be taxing to main thread...
	gint i;
	for (i=0; view_p->population_pp && view_p->population_pp[i]; i++) {
	    if (view_p->population_pp[i] && POPULATION_TIP_BUSY) {
		view_p->population_pp[i]->flags &= (POPULATION_TIP_BUSY ^ 0xffffffff);
	    }
	}

    }
    rodent_label_event (view_p, population_p);
    rfm_population_read_unlock (view_p);
    return TRUE;
}

gboolean
rodent_signal_on_enter (GtkWidget * widget, GdkEventCrossing * event, gpointer data) {
    TRACE (">> rodent_signal_on_enter\n");
    //view_t *view_p = (view_t *)data;   
    return TRUE;
}

gboolean
rodent_signal_on_leave_paper (GtkWidget * widget, GdkEventCrossing * event, gpointer data) {
    TRACE ("rodent_signal_on_leave_paper\n");
    view_t *view_p = (view_t *) data;
    // this does not always work since there are ways to get 
    // around this signal. like exiting the window by cruising
    // on top of a tooltip preview beyond the parent window's 
    // border...
    view_p->tip_event.tooltip_active=FALSE;
    view_p->mouse_event.current_mouseX = -1;
    view_p->mouse_event.current_mouseY = -1;   
    unsaturate_item (view_p);
    rodent_label_event(view_p, NULL);

    TRACE ("population_sem: rodent_signal_on_leave() released!\n");
    return TRUE;
}

gboolean
rodent_signal_on_leave (GtkWidget * widget, GdkEventCrossing * event, gpointer data) {
    TRACE ("rodent_signal_on_leave\n");
    view_t *view_p = (view_t *) data;

    // Is the view still valid (valid_view?)
    rfm_global_p->window_view_list_readers++;
    if (!g_slist_find(rfm_global_p->window_view_list, view_p)){
	rfm_global_p->window_view_list_readers--;
	return FALSE;
    } else {
	rfm_global_p->window_view_list_readers--;
    }
    
    widgets_t *widgets_p=&(view_p->widgets);

	    
    rodent_signal_on_leave_paper (widget, event, data);
  
    if (widgets_p->rename) {
       // get pointer position
       GdkRectangle pointer;
       gdk_window_get_pointer(
          gtk_widget_get_window(widgets_p->window), 
	  &pointer.x, &pointer.y, NULL);
        TRACE("pointer: x=%d, y=%d\n", pointer.x, pointer.y);
      // get window position
       gint windowX, windowY;
       gdk_window_get_position (
	       gtk_widget_get_window(widgets_p->window), 
    	        &windowX, &windowY);
        pointer.x += windowX;
        pointer.y += windowY;
        TRACE("window: x=%d, y=%d\n", windowX, windowY);
        GdkRectangle entry_extent;
        //gdk_window_get_geometry((GTK_WIDGET(widgets_p->window))->window, NULL, NULL, &entry_extent.width, &entry_extent.height, NULL);
        gdk_window_get_position (
		gtk_widget_get_window(widgets_p->rename), 
    	        &entry_extent.x, &entry_extent.y);
        gdk_window_get_geometry(
		gtk_widget_get_window(widgets_p->rename), 
                NULL, NULL, &entry_extent.width, &entry_extent.height, NULL);
        entry_extent.width += entry_extent.x;
        entry_extent.height += entry_extent.y;
        TRACE("rename: x,y= (%d, %d), w,h=( %d, %d) \n", 
                entry_extent.x, entry_extent.y, 
		entry_extent.x+entry_extent.width, 
                entry_extent.y+entry_extent.height);
        gboolean x_condition=pointer.x >= entry_extent.x && pointer.x < entry_extent.width;
        gboolean y_condition=pointer.y >= entry_extent.y && pointer.y < entry_extent.height;
	TRACE ("pointerY=%d < ? entry_extent.height=%d\n",
		pointer.y, entry_extent.height);

        if (x_condition && y_condition) {
                TRACE("OK\n");
        } else {
                rodent_done_with_rename(view_p);
		TRACE("zap\n");
        }
    }

    TRACE ("population_sem: rodent_signal_on_leave() released!\n");
    return FALSE;
}

/**  drag events **********************************************/
// This signal is received by the receiving end of the drag/drop action
void
rodent_signal_drag_data (GtkWidget * widget,
                  GdkDragContext * context,
                  gint x, gint y, 
		  GtkSelectionData * selection_data, 
		  guint info, 
		  guint time, 
		  gpointer data) {
    TRACE ("DND>> rodent_signal_drag_data\n");

    const population_t *population_p;
    view_t *view_p = (view_t *) data;
    if (view_p->flags.protected){
	g_warning("rodent_signal_drag_data(): this should never happen!");
	return;
    }
    record_entry_t *target_en;
    TRACE ("drag_data: drag_data...\n");
    /*view_p = (view_t *)g_object_get_data(G_OBJECT(widget),"view_p"); */
    if(!view_p) {
        g_warning ("rodent_signal_drag_data() view_p==NULL\n");
        gtk_drag_finish (context, FALSE, FALSE, time);
        return;
    }
    target_en = view_p->en;

    if(!target_en || !target_en->path) {
        TRACE ("drag_data: !target_en || !target_en->path\n");
        TRACE ("--DND>>rodent_signal_drag_data !target_en || !target_en->path\n");
        gtk_drag_finish (context, FALSE, FALSE, time);
        return;
    }
    rfm_cursor_wait (view_p->widgets.window);

    // drop will proceed...
    rfm_population_read_lock (view_p);
    //gint population_serial=rodent_get_population_serial(view_p);
    record_entry_t en_v;
    population_t population_v;

    TRACE ("population_sem: rodent_signal_drag_data() obtained...\n");
    population_p = rodent_find_in_population (view_p, x, y);
    if(!population_p) {
        population_p = rodent_find_in_labels (view_p, x, y);
    }
    if(population_p) {
        memcpy (&population_v, population_p, sizeof (population_t));
        if(population_p->en) {
            memcpy (&en_v, population_p->en, sizeof (record_entry_t));
            population_v.en = &en_v;
        } else {
            population_v.en = NULL;
        }
    }

    if(population_p && population_v.en && population_v.en->path) {
        if(g_file_test (population_v.en->path, G_FILE_TEST_IS_DIR)){
            target_en = population_v.en;
	} else if (population_v.en->mimetype &&
		strcmp(population_v.en->mimetype, "application/x-desktop")==0){
            target_en = population_v.en;
	}
    }

    TRACE ("DND>>rodent_signal_drag_data: target entry is %s, source is %s\n", (target_en)?target_en->path:NULL, view_p->mouse_event.drag_src);
    // Here we check if drag is within the same view, which 
    // does not make sense for the iconview. For the deskview,
    // we will tabulate a user defined position (but only
    // if a single item is dragged).
    if(!target_en || (drag_view_p==view_p && view_p->mouse_event.drag_src && strcmp (target_en->path, view_p->mouse_event.drag_src) == 0)) {
	TRACE("ignoring drop command! drag_src=%s target=%s\n", view_p->mouse_event.drag_src ,(target_en)?target_en->path:NULL);
	TRACE("drag status=%d (%d, %d, %d)\n",
		context->actions, GDK_ACTION_MOVE, GDK_ACTION_COPY, GDK_ACTION_LINK);
        rfm_cursor_reset (view_p->widgets.window);
        view_p->mouse_event.drag_src = NULL;
        drag_view_p = NULL;
        gtk_drag_finish (context, FALSE, FALSE, time);
        rfm_population_read_unlock (view_p);
        return;
    }

    TRACE ("drag_data...gui_drag_data\n");
    if(gui_drag_data (&(view_p->widgets), target_en, context, x, y, selection_data, info, time)) {
        TRACE ("drag_data...reload 0x%lx->0x%lx\n", (unsigned long)view_p, (unsigned long)view_p->en);
    }

    TRACE ("drag_data...all done\n");
    rfm_cursor_reset (view_p->widgets.window);
    rfm_population_read_unlock (view_p);
    TRACE ("population_sem: rodent_signal_drag_data() released!\n");
    return;
}
//#define TRACE TRACE

void
rodent_signal_drag_leave (GtkWidget * widget, GdkDragContext * drag_context, guint time, gpointer data) {
    TRACE ("DND>> rodent_signal_drag_leave\n");

}

void
rodent_signal_drag_delete (GtkWidget * widget, GdkDragContext * context, gpointer data) {
    TRACE ("DND>> rodent_signal_drag_delete\n");
}

void
rodent_signal_drag_end (GtkWidget * widget, GdkDragContext * context, gpointer data) {
    view_t *view_p = (view_t *) data;
    TRACE ("DND>> rodent_signal_drag_end: %s\n", view_p->mouse_event.drag_src);
    gui_drag_end (widget, context, data);
    drag_view_p = NULL;
    view_p->mouse_event.drag_src = NULL;
    rfm_cursor_wait (view_p->widgets.window);
    rfm_cursor_reset (view_p->widgets.window);
    TRACE ("drag_end... alldone\n");
}

gboolean
rodent_signal_drag_motion (GtkWidget * widget, GdkDragContext * dc, gint x, gint y, guint t, gpointer data) {
    TRACE ("DND>> rodent_signal_drag_motion\n");

    gboolean target_ok = FALSE;
    view_t *view_p = (view_t *) data;
    if (view_p->flags.protected){
	return TRUE;
    }
    rfm_population_read_lock (view_p);
    //    if(!rfm_population_try_read_lock (view_p)) { return TRUE; }
        

    TRACE ("population_sem: rodent_signal_drag_motion() obtained...\n");

    const population_t *population_p = 
	rodent_find_in_population (view_p, x, y);
    record_entry_t en_v;
    population_t population_v;
    if(population_p) {
        memcpy (&population_v, population_p, sizeof (population_t));
        if(population_p->en) {
            memcpy (&en_v, population_p->en, sizeof (record_entry_t));
            population_v.en = &en_v;
        } else {
            population_v.en = NULL;
        }
    }

    if(getenv ("RFM_DRAG_DOES_MOVE") && strlen (getenv ("RFM_DRAG_DOES_MOVE")))
        view_p->mouse_event.drag_action = GDK_ACTION_MOVE;
    else
        view_p->mouse_event.drag_action = GDK_ACTION_COPY;

    if(dc->actions == GDK_ACTION_MOVE)
        gdk_drag_status (dc, GDK_ACTION_MOVE, t);
    else if(dc->actions == GDK_ACTION_COPY)
        gdk_drag_status (dc, GDK_ACTION_COPY, t);
    else if(dc->actions == GDK_ACTION_LINK)
        gdk_drag_status (dc, GDK_ACTION_LINK, t);
    else if(dc->actions & view_p->mouse_event.drag_action)
        gdk_drag_status (dc, view_p->mouse_event.drag_action, t);
    else
        gdk_drag_status (dc, 0, t);
    TRACE ("on_drag_motion...x=%d, y= %d, population_p=0x%lx\n", x, y, (unsigned long)population_p);
    hide_tip (view_p);

    if(population_p) {
        /* if not valid drop target, return */
        if(view_p->module) {
            if(rfm_natural (PLUGIN_DIR, view_p->module_name, population_v.en, "valid_drop_site"))
                target_ok = TRUE;
        } else {                /* local */
	    if (population_v.en && 
		population_v.en->path) {
		target_ok = 
		    g_file_test(population_v.en->path, G_FILE_TEST_IS_DIR); 
		if (population_v.en->mimetype && 
			strcmp(population_v.en->mimetype,
			    "application/x-desktop")==0) {
		    target_ok = TRUE;
		}
	    }
	}
    }
    
    if(view_p->mouse_event.saturated_p != population_p) {
	unsaturate_item (view_p);
    }
    if (target_ok) {
	rodent_saturate_item (view_p, population_p);
    }

    if(view_p->mouse_event.doing_drag_p) {
        TRACE ("widget ok\n");
    }
    rfm_population_read_unlock (view_p);
    TRACE ("population_sem: rodent_signal_drag_motion() released!\n");
    return (TRUE);
}

// This signal is received by the sending end of the drag/drop event
void
rodent_signal_drag_data_get (GtkWidget * widget,
                      GdkDragContext * context, GtkSelectionData * selection_data, guint info, guint time, gpointer data) {
    TRACE ("DND>> rodent_signal_drag_data_get\n");
    view_t *view_p = (view_t *) data;
    hide_tip (view_p);
    gui_drag_data_get (&(view_p->widgets), view_p->selection_list, context, selection_data, info, time);
    TRACE ("drag_data_get: all done\n");
}

void
rodent_signal_drag_begin (GtkWidget * widget, GdkDragContext * drag_context, gpointer data) {
    view_t *view_p = (view_t *) data;
    drag_view_p = view_p;
    TRACE ("DND>> rodent_signal_drag_begin: %s\n", view_p->mouse_event.drag_src);
    TRACE ("-------------------------on_drag_begin...\n");
    view_p->mouse_event.drag_event.context = drag_context;
    hide_tip (view_p);
}

/* end of drag signals ********************************/

void
rodent_create_target_list (view_t * view_p) {
    if(view_p->mouse_event.target_list)
        return;
    view_p->mouse_event.target_list = gtk_target_list_new (target_table, NUM_TARGETS);
/*    this does not seem to be necessary
 *    gtk_drag_source_set ((GtkWidget *) view_p->widgets.paper,
                         GDK_BUTTON1_MASK | GDK_BUTTON2_MASK, target_table,
                         NUM_TARGETS, GDK_ACTION_MOVE | GDK_ACTION_COPY | GDK_ACTION_LINK);*/
    gtk_drag_dest_set ((GtkWidget *) view_p->widgets.paper,
                       GTK_DEST_DEFAULT_DROP, target_table, NUM_TARGETS, GDK_ACTION_MOVE | GDK_ACTION_COPY | GDK_ACTION_LINK);
    return;
}


/**
 * rodent_select_pixbuf:
 *
 * Selects the icon which belongs to the parameter @population_p
 * in the iconview window @view_p
 *
 * **/

void
rodent_select_pixbuf (view_t * view_p, const population_t * population_p) {
    TRACE (">> rodent_select_pixbuf\n");
    if(!population_p) {
        TRACE ("!population_p\n");
        return;
    }
    if(!population_p->en)
        return;

    if(IS_DUMMY_TYPE (population_p->en->type))
        return;
    if(view_p->module_name) {
        /* ask the module whether the element is selectable 
         * (by default they will not be)*/
        if(!rfm_natural (PLUGIN_DIR, view_p->module_name, population_p->en, "is_selectable")) {
            return;
        }
    }
    if(IS_ROOT_TYPE (population_p->en->type) && population_p->en->module) {
        /*if (!population_p->en->module) return; */
        if(!rfm_natural (PLUGIN_DIR, population_p->en->module, population_p->en, "get_dnd_path"))
            return;
    }

    if (!(population_p->flags  & POPULATION_SELECTED)) {
	((population_t *)population_p)->flags  |= POPULATION_SELECTED;
//	rodent_expose_item (view_p, population_p);
    }
    
    if (!find_in_selection_list(view_p, population_p->en)) {
	record_entry_t *en=rfm_copy_entry(population_p->en);
        view_p->selection_list = g_slist_append (view_p->selection_list, en);
    }

    view_p->mouse_event.selected_p = population_p;
}

void
rodent_unselect_all_pixbuf (view_t * view_p) {
    unselect_all_pixbuf_with_expose(view_p, TRUE);
}

void
rodent_label_event (view_t * view_p, const population_t * population_p) {
    TRACE (">> rodent_label_event\n");
    if(!view_p)
        return;
    if(!population_p && view_p->mouse_event.label_p) {
        TRACE ("unsaturating label\n");
	((population_t *)view_p->mouse_event.label_p)->flags &= (LABEL_SATURATED ^ 0xffffffff);
	hide_tip (view_p);
        rodent_expose_item (view_p, view_p->mouse_event.label_p);
        //rfm_expose_label (view_p, view_p->mouse_event.label_p);
        view_p->mouse_event.label_p = NULL;
        return;
    }
    if(!population_p){
        return;
    }
    if(population_p->flags  & POPULATION_SELECTED){
	((population_t *)population_p)->flags &= (LABEL_SATURATED ^ 0xffffffff);
        return;
    }

    if(population_p && !view_p->mouse_event.label_p && population_p->en && !IS_ROOT_TYPE (population_p->en->type)
       && !IS_DUMMY_TYPE (population_p->en->type)) {
	gboolean file_exists = rfm_g_file_test (population_p->en->path, 
		G_FILE_TEST_EXISTS) || IS_BROKEN_LNK(population_p->en->type);
	gchar * module_tip = rfm_natural (PLUGIN_DIR, view_p->module_name, population_p->en, 
		    "entry_label_tip");
        /*if(view_p->module_name) {
            TRACE ("FIXME: ask module\n");
        } else */
	if(file_exists || module_tip) {
            TRACE ("saturating label\n");
            view_p->mouse_event.label_p = population_p;
            gtk_widget_set_tooltip_text (view_p->widgets.window, "Rodent");
	    normal_tip(population_p, FALSE);
	    ((population_t *)population_p)->flags |= LABEL_SATURATED;
            
	    rodent_expose_item (view_p, view_p->mouse_event.label_p);
	    //rfm_expose_label (view_p, view_p->mouse_event.label_p);
            TRACE("tip=%s\n",population_p->en->path);
        }
	g_free(module_tip);
    }
}

///////////////////////
void
rodent_double_click (view_t * view_p, const population_t * population_p, gboolean control) {
    TRACE("rodent_double_click\n");
    view_t *view_target = view_p;
    // semaphore need not be set, for this should use a stack location
    // for a copy of selected population_p with copy of entry record.
    widgets_t *widgets_p = &(view_p->widgets);
    if(!view_p || !population_p)
        return;
    if(view_p->flags.status) {
        switch (view_p->flags.status) {
        case STATUS_EXIT:
            /* ignore doubleclick if closing in progress */
            return;
        case STATUS_IDLE:
        default:
            break;
        }
    }
    hide_tip (view_p);

    //rfm_cursor_wait(view_p->widgets.window);

    if(population_p->en) {
        TRACE ("submodule_name=%s\n", population_p->en->module);
        if(IS_NOACCESS_TYPE (population_p->en->type)) {
            rfm_show_text(widgets_p);
            rfm_diagnostics (widgets_p, "xffm/stock_dialog-error", NULL);
            rfm_diagnostics (widgets_p, "xffm_tag/stderr", population_p->en->path, ": ", strerror (EACCES), "\n", NULL);

            return;
        }
    }

    /* does the module have something else in mind for the
       double click on the particular entry ? */
    if(view_p->module && population_p->en) {
        TRACE ("click: looking for alternate module double click here\n");
        if(rfm_rational (PLUGIN_DIR, view_p->module_name, &(view_p->widgets), population_p->en,  "double_click")) {
            rfm_cursor_reset (view_p->widgets.window);
            return;
        }
        TRACE ("click: no alternate found, going for default...\n");
    }

    if(population_p->en && !population_p->en->module && !rfm_g_file_test (population_p->en->path, G_FILE_TEST_IS_DIR)) {
        record_entry_t *en = population_p->en;
        TRACE ("OPEN: !population_p->en->module\n");
	if (population_p->en->mimetype && strcmp(population_p->en->mimetype, "application/x-desktop")==0){
	    if (rfm_rational(PLUGIN_DIR, "dotdesktop", widgets_p, population_p->en,  "double_click")) {
		return;
	    }
	}

        if(rfm_g_file_test (en->path, G_FILE_TEST_EXISTS)) {
            TRACE ("OPEN: G_FILE_TEST_EXISTS %s\n", en->path);
            record_entry_t *en = population_p->en;
            rfm_cursor_reset (view_p->widgets.window);
            if(!en->mimetype)
                en->mimetype = MIME_magic (en->path);
            gchar *prg = NULL;
            if(en->mimetype)
                prg = MIME_command (en->mimetype);
            if(IS_EXE_TYPE (en->type)) {     /* run it (if not registered) */
		TRACE("running %s\n", en->path);
                if(prg){
		    TRACE("using rodent_open_with\n");
                    rodent_open_with (&(view_p->widgets), en);
		} else {
		    TRACE("using rfm_thread_run2argv\n");
		    g_free(widgets_p->workdir);
		    widgets_p->workdir=g_strdup(view_p->en->path);

                    /*  assume in_term to be safe */
                    rfm_thread_run2argv (widgets_p, en->path, TRUE);
                    //rfm_thread_run2argv (widgets_p, en->path, FALSE);
                }
            } else {            /* open with */


               TRACE ("OPEN: calling rodent_open_with()\n");
                rodent_open_with (&(view_p->widgets), en);
            }
            g_free (prg);
            return;
        }
	gchar *text=g_strdup_printf("\n%s: %s", en->path, strerror(ENOENT));
	rfm_confirm(widgets_p, GTK_MESSAGE_ERROR,
		text, NULL, _("Accept"));
	g_free(text);
        //g_warning ("nothing programed here with double click\n");
        rfm_cursor_reset (view_p->widgets.window);
        return;
    } else {
        record_entry_t *en = rfm_copy_entry (population_p->en);
        const gchar *submodule_name = NULL;
        if(population_p->en)
            submodule_name = population_p->en->module;

        TRACE ("RELOADING HERE... module_name=%s control=%d constructor=0x%x popup_constructor=0x%x\n", 
		view_target->module_name,
		control,
		GPOINTER_TO_INT(view_p->constructor),
		GPOINTER_TO_INT(view_p->child_constructor));

        if(en && en->path && rfm_g_file_test (en->path, G_FILE_TEST_IS_DIR)) {
            rfm_save_to_go_history (en->path);
        }

        //if(control && (view_p->constructor || view_p->popup_constructor || view_p->child_constructor)) {
        if(view_p->flags.type == DESKVIEW_TYPE && (view_p->constructor||view_p->child_constructor))
	{
            TRACE ("create_iconview \n");
            if(view_p->constructor) {
                TRACE ("create iconview with %s\n", (en) ? en->path : "NULL entry");
                view_target = (*view_p->constructor) (en);
                view_target->flags.preferences = view_p->flags.preferences;
                view_target->flags.sortcolumn = view_p->flags.sortcolumn;
                view_target->module_name = submodule_name;
		if (!rodent_refresh (&(view_target->widgets), en)){ 
		    rfm_destroy_entry(en);
		}

            } else if (view_p->child_constructor) {
		TRACE ("child_constructor\n");
		if (en && en->module) {
		    gchar *path=
			g_strdup_printf("exec:%s", (gchar *)rfm_void(PLUGIN_DIR, en->module, "exec_name"));
		    (*view_p->child_constructor) (widgets_p, path);
		    g_free(path);
		} else {
		    (*view_p->child_constructor) (widgets_p, (en)?en->path:NULL);
		}
            } 
	    rfm_cursor_reset (view_target->widgets.window);
            return;
        }

        if(en && view_p->child_constructor) {
            TRACE ("DOUBLE_CLICK: en && view_p->child_constructor\n");
            (*view_p->child_constructor) (widgets_p, population_p->en->path);
            rfm_cursor_reset (view_target->widgets.window);
            return;
        }

        rodent_push_view_go_history (view_p);

        record_entry_t *out_en = rfm_copy_entry (en);
        if(out_en) {
            TRACE ("RELOADING: default goto action: %s !\n", out_en->path);
        }
        if(rodent_refresh (&(view_target->widgets), out_en)) {
            if(out_en && out_en->path && rfm_g_file_test (out_en->path, G_FILE_TEST_IS_DIR))
                rfm_save_to_go_history (out_en->path);
            rfm_cursor_reset (view_target->widgets.window);
        } else {
	    rfm_destroy_entry(out_en);
	}
        return;
    }
    TRACE ("population_sem: released!\n");
    return;
}


