//
//
/* This file is included by rodent_mouse.c */
/*
 * Copyright (C) 2002-2011 Edscott Wilson Garcia
 * EMail: edscott@xfce.org
 *
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*************************************************************************/
/******************   dnd functions *************************************/
/************************************************************************/

#define MAXURILEN 4096          /* Longest URI to allow */

#define DRAG_TYPE_UNDEFINED	0
#define DRAG_TYPE_LOCAL		0x01
#define DRAG_TYPE_NET		0x02
#define DRAG_TYPE_INCONSISTENT	0x04

static GtkTargetEntry target_table[] = {
    {"text/uri-list", 0, TARGET_URI_LIST},
    {"text/x-moz-url", 0, TARGET_MOZ_URL},
    {"text/plain", 0, TARGET_PLAIN},
    {"UTF8_STRING", 0, TARGET_UTF8},
    {"STRING", 0, TARGET_STRING}
};

#define NUM_TARGETS (sizeof(target_table)/sizeof(GtkTargetEntry))
static view_t *drag_view_p = NULL;
static GMutex *tooltip_text_mutex = NULL;
enum {
    DND_NONE,
    DND_MOVE,
    DND_COPY,
    DND_LINK
};

static void
unselect_pixbuf (view_t * view_p, const population_t * population_p);
static gpointer
tip_preview_thread_f (gpointer data);

typedef struct activate_tip_t {
    view_t *view_p;
    const population_t *population_p;
    gint population_serial;
} activate_tip_t;
    

static char *dnd_data = NULL;
/*************   core *****************/
#if 0
void
on_drag_data_delete (GtkWidget * widget, GdkDragContext * context, gpointer data) {
    TRACE ("TRACE: on_drag_data_delete!\n\n");

    return;
}
#endif

static void
unsaturate_item (view_t *view_p) {
    if (!view_p->mouse_event.saturated_p) {
	return;
    }
    //view_p->tip_event.tooltip_active = FALSE;
    TRACE("unsaturate_item: view_p->tip_event.tooltip_active = FALSE\n");
    // unsaturate item
    ((population_t *)view_p->mouse_event.saturated_p)->flags &= (POPULATION_SATURATED ^ 0xffffffff);

    TRACE("0x%x set unsaturated flag now-->%d\n",
	    GPOINTER_TO_INT(view_p->mouse_event.saturated_p), 
	    view_p->mouse_event.saturated_p->flags & POPULATION_SATURATED );
    rodent_expose_item(view_p, view_p->mouse_event.saturated_p);
    view_p->mouse_event.saturated_p = NULL;
}

static void
saturate_item (view_t *view_p, const population_t * population_p) {
    rodent_label_event(view_p, NULL);
    if(population_p->pixbuf) {
	view_p->mouse_event.dnd_pixbuf=population_p->pixbuf;
    }
    if(view_p->mouse_event.saturated_p != population_p) {
	if (view_p->mouse_event.saturated_p){
	    TRACE("0x%x  (2) set unsaturated flag now\n",
		    GPOINTER_TO_INT(view_p->mouse_event.saturated_p));
	    ((population_t *)view_p->mouse_event.saturated_p)->flags &=
		(POPULATION_SATURATED ^ 0xffffffff);
	}
	// saturate item
	TRACE("0x%x set saturated flag now\n",GPOINTER_TO_INT(population_p));
	((population_t *)population_p)->flags |= POPULATION_SATURATED;
	view_p->mouse_event.saturated_p = population_p;
	rodent_expose_item(view_p, view_p->mouse_event.saturated_p);
	//rfm_expose_icon(view_p, view_p->mouse_event.saturated_p);
    }
}

static void
gui_drag_end (GtkWidget * widget, GdkDragContext * context, gpointer data) {
    TRACE ("*DND drag end\n");
    /* check if this is in treeviewlib, then delete:
     * if (scrolltimer)g_source_remove(scrolltimer);
     scrolltimer=0,scrollx=-1, scrolly=-1;*/
    if(!widget)
        return;

    if(dnd_data) {
        g_free (dnd_data);
        dnd_data = NULL;
    }
    TRACE ("DND drag end done\n");
    return;
}

/*
 * DND sender: prepare data for the remote receiver.
 * event: drag_data_get
 */
static void
gui_drag_data_get (widgets_t * widgets_p,
                   GSList * drag_entry_list, 
		   GdkDragContext * context, 
		   GtkSelectionData * selection_data, 
		   guint info, 
		   guint time) {
    char *files;
    GSList *tmp;
    record_entry_t *en;
    int selection_len;
    int drag_type;
    gchar *format = NULL;
    gchar *me,
     *she;

    if(!drag_entry_list || !g_slist_length (drag_entry_list)
       || !drag_entry_list->data) {
	// This should never happen.
	g_warning("gui_drag_data_get(): no selection list");
        return;
    }
    en = (record_entry_t *) drag_entry_list->data;

    me = g_strdup (g_get_host_name ());
    she = rfm_host_name (GDK_WINDOW_XID (context->dest_window));

    view_t *view_p = widgets_p->view_p;
    if(view_p->en) {
        TRACE ("*DND on_drag_data_get (%s), me=%s --> she=%s\n", view_p->en->path, me, she);
    }

    if(en->module) {
        const gchar *fmt = rfm_natural (PLUGIN_DIR, en->module, en, "get_dnd_format");
        if(fmt)
            format = g_strdup (fmt);
        TRACE ("DND module format=%s\n", (format ? format : "null"));
    } else
        TRACE ("DND not module format\n");
    if(!format) {
        drag_type = DRAG_TYPE_LOCAL;
        if(strcmp (me, she)) {
            struct passwd *pw = getpwuid (getuid ());
            if(pw) {
                format = g_strdup_printf ("file://%s@%s", pw->pw_name, me);
            } else {
                format = g_strdup_printf ("file://%s", me);
            }
        } else
            format = g_strdup ("file:");
    }
    g_free (me);
    g_free (she);
    TRACE ("DND on_drag_data_get, format=%s\n", (format ? format : "null"));

    /* prepare data for the receiver */
    switch (info) {
    case TARGET_RAW:
        g_warning ("TARGET_RAW");
    case TARGET_UTF8:
        g_warning ("TARGET_UTF8");
    case TARGET_URI_LIST:
        TRACE ("DND TARGET_URI_LIST\n");
    default:
        selection_len = 0;
        if(dnd_data) {
            g_free (dnd_data);
            dnd_data = NULL;
        }
        /* count length of bytes to be allocated */
        for(tmp = drag_entry_list; tmp; tmp = tmp->next) {
            const gchar *dndpath;
            en = (record_entry_t *) tmp->data;
            if(!en || !en->path || !strlen (en->path))
                continue;
            if(en->module && rfm_natural (PLUGIN_DIR, en->module, en, "get_dnd_path")) {
                dndpath = rfm_natural (PLUGIN_DIR, en->module, en, "get_dnd_path");
            } else {
                dndpath = en->path;
            }
            /* 2 is added for the \r\n */
            selection_len += (strlen (dndpath) + strlen (format) + 2);
        }
        /* 1 is added for terminating null character */
        dnd_data = files = g_malloc (selection_len + 1);
        memset (files, 0, selection_len + 1);
        for(tmp = drag_entry_list; tmp; tmp = tmp->next) {
            const gchar *dndpath;
            en = (record_entry_t *) tmp->data;
            if(!en || !en->path || !strlen (en->path))
                continue;
            if(en->module && rfm_natural (PLUGIN_DIR, en->module, en, "get_dnd_path")) {
                dndpath = rfm_natural (PLUGIN_DIR, en->module, en, "get_dnd_path");
            } else {
                dndpath = en->path;
            }
            sprintf (files, "%s%s\r\n", format, dndpath);
            files += (strlen (format) + strlen (dndpath) + 2);
        }
        break;
    }
    TRACE ("DND drag data is:%s\n", dnd_data);

    gtk_selection_data_set (selection_data, selection_data->target, 8, (const guchar *)dnd_data, selection_len);
    /*rfm_details->comm |= COMM_IGNORE_RELEASE;
    TRACE ("DND drag get done, format=%s\n", format);
    TRACE ("DND ignore second releas=%d... \n", rfm_details->comm & COMM_IGNORE_RELEASE);*/
    g_free (format);
}

static gboolean
gui_drag_data (widgets_t * widgets_p,
               record_entry_t * target_en,
               GdkDragContext * context,
	       gint x, gint y, 
	       GtkSelectionData * data,
	       guint info,
	       guint time) {
    int the_mode;
    int nitems,
      action;
    gchar *url;
    int mode = 0;
    GList *list = NULL;
    gboolean result = FALSE;
    gchar *he,
     *me;

    if(!target_en || data->length < 0 || data->format != 8) {
	// We should never get this warning. If we do, something is 
	// terribly wrong.
        g_warning ("gui_drag_data(): !target_en || data->length < 0 || data->format != 8");
        goto drag_over;         /* of course */
    }

    me = g_strdup (g_get_host_name ());
    he = rfm_host_name (GDK_WINDOW_XID (context->source_window));

    view_t *view_p = widgets_p->view_p;
    if(view_p->en) {
        TRACE ("*DND on_drag_data_get (%s), me=%s --> she=%s\n", view_p->en->path, me, he);
    }

    /* remote instance may have full format specification,
     * or borked specification. Here we must consider
     * both cases */

    if(context->action <= GDK_ACTION_DEFAULT) {
        if(getenv ("RFM_DRAG_DOES_MOVE")
           && strlen (getenv ("RFM_DRAG_DOES_MOVE"))) {
            action = GDK_ACTION_MOVE;
        } else {
            action = GDK_ACTION_COPY;
        }
    } else
        action = context->action;

    TRACE ("DND info=%d (%d,%d)\n", info, TARGET_STRING, TARGET_URI_LIST);
    if(!(info == TARGET_STRING) && !(info == TARGET_URI_LIST)
       && !(info == TARGET_MOZ_URL)) {
	// Here we have something unknown in the drag.
        goto drag_over;         /* of course */
    }

    TRACE ("DND: action=%d\n", action);
    if(action == GDK_ACTION_MOVE)
        the_mode = mode = TR_MOVE;
    else if(action == GDK_ACTION_COPY)
        the_mode = mode = TR_COPY;
    else if(action == GDK_ACTION_LINK)
        the_mode = mode = TR_LINK;
    else {
	g_warning("Drag drop mode is not GDK_ACTION_MOVE | GDK_ACTION_COPY | GDK_ACTION_LINK");
        goto drag_over;         /* of course */
    }

    TRACE ("DND drag data=%s", (const char *)data->data);

    nitems = rfm_uri_parse_list ((const char *)data->data, &list);

    TRACE ("DND nitems=%d, drag data=%s", nitems, (const char *)data->data);
    if(!nitems) {
	g_warning("number of items in drag is zero!");
        goto drag_over;         /* of course */
    }

    /***/


    /* if target is a plugin, let the plugin take care of business. */
    if(target_en->module) {
        rfm_uri_remove_file_prefix_from_list (list, he, me);
        TRACE ("en->module=%s\n", target_en->module);
        if(rfm_natural (PLUGIN_DIR, target_en->module, target_en, "valid_drop_site")) {
            TRACE ("module: valid_drop_site for %s\n", target_en->module);
            rfm_natural (PLUGIN_DIR, target_en->module, target_en, "set_drop_entry");
            if(rfm_complex (PLUGIN_DIR, target_en->module, widgets_p, target_en->path, list,  "process_drop")) {
                TRACE ("module: process_drop ok\n");
                result = TRUE;
            }
            rfm_void (PLUGIN_DIR, target_en->module, "clear_drop_entry");
            list = rfm_uri_free_list (list);
            goto drag_over;
        }
    }


    /* now determine whether cp or scp should be used 
     * (we ignore the format for this determination, 
     * because other applications may bork this)*/
    rfm_uri_remove_file_prefix_from_list (list, he, me);

    /* if target is dotdesktop type, take individual action */
    if (target_en->mimetype &&
		strcmp(target_en->mimetype, "application/x-desktop")==0){
	if (rfm_complex(PLUGIN_DIR, "dotdesktop", widgets_p, target_en->path, list, "process_drop")) {
	    TRACE("Target is dotdesktop: %s\n", target_en->path);
	}
	list = rfm_uri_free_list (list);
	result = TRUE;
	goto drag_over;
    }

#if 0
    // This figures out if scp should be attempted
    // XXX this is out of date and I really never used it
    //     so why keep it?
    if(strcmp (me, he) != 0) {
        int l;
        char **srcs;

        TRACE ("dnd: doing scp\n");

        l = g_list_length (list);
        srcs = (char **)malloc ((l + 1) * sizeof (char *));
        if(!srcs) {
            list = rfm_uri_free_list (list);
            TRACE ("dnd: no srcs\n");
            goto drag_over;
        }
        srcs[l] = NULL;
        for(l = 0, tmp = list; tmp; tmp = tmp->next, l++) {
            srcs[l] = tmp->data;
        }
        rodent_scp (widgets_p, (const gchar **)srcs, target_en->path);
        g_free (srcs);
        list = rfm_uri_free_list (list);
        result = TRUE;
        goto drag_over;
    }
#endif

    /* local file cp/ln/mv */
    url = list->data;
    {
        /* nonsense check */
        struct stat st;
        lstat (url, &st);
	// Here we check if the file source and destination is actually 
	// the same thing, this time by stat information instead of
	// path string.
	// This is a more robust test. We must test *both* st_ino and
	// st_dev, because stuff on different devices may (and do) have
	// equal st_ino.
        if(target_en->st && 
		st.st_ino == target_en->st->st_ino &&
		st.st_dev != target_en->st->st_dev)
	{
            list = rfm_uri_free_list (list);
            rfm_diagnostics(&(view_p->widgets),"xffm/stock_dialog-warning",NULL);
            rfm_diagnostics (widgets_p, "xffm_tag/stderr", " ", strerror (EEXIST), ": ", target_en->path, "\n", NULL);
            goto drag_over;
        }
#if 0
	// This is another check. Most probably superceded by inode test above
        if(!S_ISDIR (st.st_mode) && strchr (url, '/')) {
            char *p;
            p = g_strdup (url);
            *(strrchr (p, '/')) = 0;
            if(target_en->path && strcmp (p, target_en->path) == 0) {
                list = rfm_uri_free_list (list);
                g_free (p);
                p = NULL;
                rfm_show_text(widgets_p);
                rfm_diagnostics (widgets_p, "xffm/stock_dialog-error", NULL);
                rfm_diagnostics (widgets_p, "xffm_tag/stderr", strerror (EEXIST), ": ", target_en->path, "\n", NULL);

                /*fprintf(stderr,"dbg:nonsense 2\n"); */
                goto drag_over;
            }
            g_free (p);
            p = NULL;
        }
#endif
    }

    TRACE ("DND: local\n");
    rodent_cp (mode, widgets_p, list, target_en->path);
    list = rfm_uri_free_list (list);
    result = TRUE;
  drag_over:
    g_free (me);
    g_free (he);
    gtk_drag_finish (context, TRUE, (the_mode & TR_MOVE) ? TRUE : FALSE, time);
    TRACE ("DND: on drag data: drag_over\n");
    return result;
}

/*************************************************************************/
/******************   end of dnd functions *******************************/
/*************************************************************************/

/*************************************************************************/
/******************* tip functions ***************************************/
/*************************************************************************/

static void hide_tip (view_t * view_p);

//static
//void show_tip (population_t * population_p);
#ifdef FAST_PROCESSOR
# undef FAST_PROCESSOR
#endif

static gchar *
path_info (record_entry_t * en) {
    gchar *s1 = NULL,
        *s2 = NULL;
    gchar *info = NULL;
    if(!en || !en->path)
        return NULL;
    if(IS_ROOT_TYPE (en->type) && !g_file_test (en->path, G_FILE_TEST_EXISTS))
        return NULL;

#ifdef FAST_PROCESSOR
    if (!en->mimetype) en->mimetype=MIME_type (en->path, en->st);
    if (!en->mimemagic) en->mimemagic= MIME_magic(en->path);
    if (!en->filetype) en->filetype= MIME_file(en->path);
    /* printout full file information: */
#endif

    gchar *b = g_path_get_basename (en->path);
    rfm_chop_excess (b);
    gchar *vpath = rfm_utf_string (b);
    g_free (b);
    if(IS_SLNK (en->type)) {
        char lpath[_POSIX_PATH_MAX + 1];
        memset (lpath, 0, _POSIX_PATH_MAX + 1);
        if(readlink (en->path, lpath, _POSIX_PATH_MAX) > 0) {
            rfm_chop_excess (lpath);
            gchar *q = rfm_utf_string (lpath);
            gchar *linkto=g_strdup_printf (_("Link to %s"), q);
            s1 = g_strdup_printf ("%s -> %s\n\n", vpath, linkto);
            g_free(linkto);
            //s1 = g_strdup_printf (_("Symbolic link: %s -> %s\n\n"), vpath, q);
            g_free (q);
        }
    } else {
        gchar *p = g_strdup_printf ("%s %s\n\n",_("Name:"), vpath);
        if(en->module) {
	    gchar *plugin_info=rfm_void(PLUGIN_DIR, en->module, "plugin_info");
	    if (plugin_info) {
		s1 = g_strconcat (p, plugin_info, NULL);
		g_free (p);
		g_free(plugin_info);
	    } else {
		gchar *q = g_strdup_printf ("%s: %s", _("Installed Plugins"), en->module);
		s1 = g_strconcat (p, q, NULL);
		g_free (p);
		g_free (q);
	    }
        } else {
            s1 = p;
        }
    }
    g_free (vpath);

    if(rfm_g_file_test (en->path, G_FILE_TEST_EXISTS) && en->st) {
        gchar *grupo=rfm_group_string(en->st);
        gchar *owner=rfm_user_string(en->st);
        gchar *tag = rfm_sizetag ((off_t) en->st->st_size, -1);

        //    gchar *ss= rfm_time_to_string(en->st->st_mtime);   

        gchar *dirname = g_path_get_dirname (en->path);
	gchar *mode_string=rfm_mode_string (en->st->st_mode);
#ifdef FAST_PROCESSOR
        gchar **apps = MIME_apps (en->mimemagic);
        gchar *app_txt = g_strdup ("");;
        gchar *openwith=g_strdup_printf(_("Open with %s"),": ");
        if(apps) {
            gchar **p,
             *pp;
            for(p = apps; p && *p; p++) {
                pp = g_strconcat (app_txt, "\n\t", *p, NULL);
                g_free (app_txt);
                app_txt = pp;
            }

        } else {
            app_txt = openwith;
        }
        g_strfreev (apps);
        s2 = g_strdup_printf (
                "%s/%s: %s/%s\n%s: %s\n%s: %s\n\n%s: %s\n%s: %s\n%s: %s%s\n\n%s: %s\n\n%s: %s",
                _("Owner"),_("Group"), owner, grupo,
                _("Permissions"), mode_string,
                _("Folder"), dirname, 
                _("File Type"), en->filetype,
                _("MIME Type"), (en->mimetype) ? en->mimetype : _("unknown"),
                _("MIME Type"),  " (file)", en->mimemagic,
                openwith,  app_txt,
                _("Size"),  tag);

        g_free (app_txt);
        g_free (openwith);

#else
        s2 = g_strdup_printf (
                "%s/%s: %s/%s\n%s: %s\n%s: %s\n\n%s: %s",
                 _("Owner"),_("Group"), owner, grupo,
                _("Permissions"), mode_string,
                _("Folder"), dirname, 
                _("Size"),  tag);
#endif
        //    g_free(q);
        g_free (owner);
        g_free (grupo);
        g_free (tag);
        g_free (dirname);
        g_free (mode_string);
    }
    if(rfm_g_file_test (en->path, G_FILE_TEST_EXISTS) && en->st) {
        gchar buf[1024];

	gchar *date_string=rfm_date_string(en->st->st_ctime);

        sprintf (buf, "Status change= %s", date_string);
	g_free(date_string);

        gchar *s3 = g_strconcat (s2, "\n", buf, NULL);
        g_free (s2);
        s2 = s3;

	date_string=rfm_date_string(en->st->st_mtime);
        sprintf (buf, "Modification time= %s", date_string);
	g_free(date_string);


        s3 = g_strconcat (s2, "\n", buf, NULL);
        g_free (s2);
        s2 = s3;

	date_string=rfm_date_string(en->st->st_atime);
        sprintf (buf, "Access time= %s", date_string);
	g_free(date_string);

        s3 = g_strconcat (s2, "\n", buf, NULL);
        g_free (s2);
        s2 = s3;

        s3 = g_strdup_printf ("%s\n\nNumber of hard links: %ld", s2, (long)en->st->st_nlink);
        g_free (s2);
        s2 = s3;
    }

    if(!s1)
        s1 = g_strdup ("");
    if(!s2)
        s2 = g_strdup ("");
    info = g_strconcat (s1, s2, NULL);
    g_free (s1);
    g_free (s2);
#ifdef FAST_PROCESSOR
    g_free (mimetype);
    g_free (mimemagic);
    g_free (filetype);
#endif
    return info;
}

static gchar *
get_tip_text (view_t * view_p, const population_t * population_p, const gchar * prepend_txt) {
    TRACE ("TIP: SHOW_TIP population_p->preview_pixbuf = 0x%lx\n", (long unsigned)view_p->tip_event.tooltip_pixbuf);
    const gchar *warning = NULL;
    gchar *info = path_info (population_p->en);
    gchar *g;
    if(population_p->en && IS_SDIR (population_p->en->type)) {
        warning = _("Folder");
//        warning = _("Directory:\n");
        gchar *b = g_path_get_basename (population_p->en->path);
        gint files = rfm_xfdir_count_files (population_p->en->path);
        gint hidden = rfm_xfdir_count_hidden_files (population_p->en->path);
        gchar *info = path_info (population_p->en);
        gchar *hidden_string = NULL;
        gchar *files_string = NULL;

        gchar *plural_string = g_strdup_printf (ngettext (" (containing %'d item)", " (containing %'d items)", files),files);
        files_string = g_strdup_printf ("%s%s.",_("Folder"), plural_string);
        g_free(plural_string);
    
        plural_string = g_strdup_printf(ngettext ("%'u item","%'u items",hidden), hidden);
        hidden_string = g_strdup_printf ("%s: %s.",_("Hidden"), plural_string);
        g_free(plural_string);

        if(files) {
            g = g_strdup_printf ("%s %s\n%s", files_string, hidden_string, info);
        } else if (rfm_g_file_test(population_p->en->path, G_FILE_TEST_EXISTS)){
            g = g_strdup_printf ("%s: %s\n%s", _("Folder"), _("The location is empty."), info);
        } else {
            g = g_strdup_printf ("%s: \n%s", _("Folder"), info);
	}


        g_free (hidden_string);
        g_free (files_string);
        g_free (b);
    } else {
        if (warning) g = g_strdup_printf ("%s:\n%s", warning, info);
        else  g = g_strdup_printf ("%s", info);
    }
    // modules can prepend text to the tooltip
    if(prepend_txt) {
        gchar *gg = g_strconcat (prepend_txt, "\n", g, NULL);
        g_free (g);
        g = gg;
    }
    g_free (info);
    return g;
}

static void
normal_tip(const population_t * population_p, gboolean full){
    TRACE ("normal_tip: %s\n", (population_p->en)?population_p->en->path:"null");
    view_t *view_p = population_p->view_p;
    
    if(population_p == NULL || !population_p->en) {
	TRACE ("population_p == NULL\n");
        return;
    }
    if(getenv ("RFM_ENABLE_TIPS")==NULL || strlen (getenv ("RFM_ENABLE_TIPS"))==0) {
        return;
    }

    gchar *module_txt = NULL;
    if(view_p->module_name) {
        module_txt = rfm_natural (PLUGIN_DIR, view_p->module_name, population_p->en, "entry_tip");
    }

    gint isize;
    if (!tooltip_text_mutex){
	tooltip_text_mutex=g_mutex_new();
    }
    g_mutex_lock(tooltip_text_mutex);
    gchar *oldtext=view_p->tip_event.tooltip_text;
    if (full){
        view_p->tip_event.tooltip_text = get_tip_text (view_p, population_p, module_txt);
        isize = BIG_ICON_SIZE;
    } else {
	gchar *basename=NULL;
	if(view_p->module_name) {
	    basename = rfm_natural (PLUGIN_DIR, view_p->module_name, population_p->en, 
		    "entry_label_tip");    
	}
	if (!basename) {
           basename=g_path_get_basename(population_p->en->path);
	}
        view_p->tip_event.tooltip_text = rfm_utf_string(basename);
        g_free(basename);
        isize = SIZE_BUTTON;
    }
	
    g_free (oldtext);
    g_free (module_txt);

    // Clear additional ref to tooltip pixbuf.
    if (view_p->tip_event.tooltip_pixbuf) {
	g_object_unref(view_p->tip_event.tooltip_pixbuf);
    }

    if(view_p->view_layout.icon_size == BIG_ICON_SIZE) {
        view_p->tip_event.tooltip_pixbuf = population_p->pixbuf;
    } else {
	const gchar *id = rfm_resolve_icon_id (population_p->en);
	TRACE("ICON:%s --> %s\n",population_p->en->path, id);
	gchar *composite_id = 
	    rfm_get_composite_icon_id (population_p->en, id);
	TRACE("ICON:%s --> %s\n",population_p->en->path, composite_id);
	view_p->tip_event.tooltip_pixbuf =
	    rfm_get_pixbuf (composite_id, isize);
	g_free(composite_id);
	// Fallback
	if (view_p->tip_event.tooltip_pixbuf == NULL){
	    view_p->tip_event.tooltip_pixbuf =
		rfm_get_pixbuf (id, isize);
	}
    }
    // double fallback
    if (view_p->tip_event.tooltip_pixbuf == NULL) {
	view_p->tip_event.tooltip_pixbuf =
	    rfm_get_pixbuf ("xffm/emote_cool", isize);
    }

    // additional ref to pixbuf
    if (!G_IS_OBJECT (view_p->tip_event.tooltip_pixbuf)){
	TRACE("!G_IS_OBJECT (pixbuf) at normal_tip()\n");
    } else {
	g_object_ref(view_p->tip_event.tooltip_pixbuf);
    }
    
    record_entry_t *olden=view_p->tip_event.tooltip_entry;
    view_p->tip_event.tooltip_entry = rfm_copy_entry (population_p->en);
    rfm_destroy_entry(olden);
    TRACE ("TIP: SHOW_TIP %s -> %s\n",
	    (population_p->en) ? population_p->en->path : "Null entry", 
	    view_p->tip_event.tooltip_text);
    view_p->tip_event.tooltip_active = TRUE;
    g_mutex_unlock(tooltip_text_mutex);
    
}

static gboolean 
preliminary_tip(activate_tip_t *activate_tip_p){
    TRACE("preliminary_tip\n");
    view_t *view_p=activate_tip_p->view_p;
    int population_serial=activate_tip_p->population_serial;
    rfm_population_read_lock (view_p);
    if (view_p->population_serial != population_serial){
	rfm_population_read_unlock (view_p);
	return FALSE;
    }
    gboolean retval=TRUE;
    if (activate_tip_p->population_p->en->mimemagic==NULL){
	activate_tip_p->population_p->en->mimemagic=
	    MIME_magic(activate_tip_p->population_p->en->path);
    }
    population_t *population_p = (population_t *)malloc(sizeof(population_t));
    memcpy(population_p, activate_tip_p->population_p, sizeof(population_t));
    population_p->en = rfm_copy_entry(activate_tip_p->population_p->en);

    // get additional references for pixbuf, preview_pixbuf and initial pixbuf...
    if (population_p->preview_pixbuf && G_IS_OBJECT (population_p->preview_pixbuf)){
	g_object_ref(population_p->preview_pixbuf);
    }
    if (population_p->initial_preview_pixbuf && G_IS_OBJECT (population_p->initial_preview_pixbuf)){
	g_object_ref(population_p->initial_preview_pixbuf);
    }
    if (population_p->pixbuf && G_IS_OBJECT (population_p->pixbuf)){
	g_object_ref(population_p->pixbuf);
    }

    rfm_population_read_unlock (view_p);

    const gchar *type=(population_p->en->mimetype)?
	population_p->en->mimetype:population_p->en->mimemagic;
    
    // is the work already done?
    gboolean preview_ok = (population_p->preview_pixbuf
		       && population_p->preview_pixbuf_timestamp <= 
		          population_p->en->st->st_mtime);
    if (!tooltip_text_mutex){
	tooltip_text_mutex=g_mutex_new();
    }

    if(preview_ok) {
	g_mutex_lock(tooltip_text_mutex);
	if (view_p->tip_event.tooltip_pixbuf) {
	    g_object_unref(view_p->tip_event.tooltip_pixbuf);
	}
	view_p->tip_event.tooltip_pixbuf = population_p->preview_pixbuf;
	if (!G_IS_OBJECT (view_p->tip_event.tooltip_pixbuf)){
	    g_warning("!G_IS_OBJECT (pixbuf) at [3] preliminary_tip()");
	}
	g_object_ref(view_p->tip_event.tooltip_pixbuf);
	// 
	record_entry_t *olden=view_p->tip_event.tooltip_entry;
	view_p->tip_event.tooltip_entry = rfm_copy_entry (population_p->en);
	rfm_destroy_entry(olden);
	if (view_p->tip_event.tooltip_text) {
	    g_free(view_p->tip_event.tooltip_text);
	    view_p->tip_event.tooltip_text=NULL;
	}
        view_p->tip_event.tooltip_active = TRUE;
	g_mutex_unlock(tooltip_text_mutex);
	// FALSE return value indicates to calling thread that
	// the pixbuf preview does not have to be recreated.
	retval=FALSE;
	goto cleanup;
    } 
    else {
	TRACE("###### preview not ok %ld<%ld not met\n",
		(long)population_p->preview_pixbuf_timestamp,
		(long)population_p->en->st->st_mtime);
    }

    // initial image:
    // This creates a reffed pixbuf that must be unreffed later on.
    if (population_p->initial_preview_pixbuf == NULL) {
	gint width=0;
	gint height=0;
	if (population_p->pixbuf && GDK_IS_PIXBUF (population_p->pixbuf)) {
	    width=gdk_pixbuf_get_width(population_p->pixbuf);
	    height=gdk_pixbuf_get_height(population_p->pixbuf);
	}
	if (width != height){
	    // this is a definite thumbnail
	    if (width > height) {
		height = height * PREVIEW_IMAGE_SIZE / width;
		width = PREVIEW_IMAGE_SIZE;
	    } else {
		width = width * PREVIEW_IMAGE_SIZE / height;
		height = PREVIEW_IMAGE_SIZE;
	    }
	    // this refs:
	    population_p->initial_preview_pixbuf = 
		    gdk_pixbuf_scale_simple (population_p->pixbuf,
			width,
			height,
			GDK_INTERP_NEAREST);
	    // population_p->initial_preview_pixbuf: ref==1
	    retval=TRUE;
	} else {
	    // scale up svg icon (ignore square thumbnails)
	    // the ref is not held in the pixbuf hash since 
	    // the size is larger than 
	    // the maximum pixbuf hash size. 
	    // This res:
	    population_p->initial_preview_pixbuf=rfm_get_pixbuf(type, PREVIEW_IMAGE_SIZE);
	    // ref==1 but is not in pixbuf hash, because of size
	    // PREVIEW_IMAGE_SIZE  > MAX_PIXBUF_HASH_SIZE 
	    // (400 > 128 last time I looked)
            if (G_IS_OBJECT (population_p->initial_preview_pixbuf) &&
		    PREVIEW_IMAGE_SIZE  <= MAX_PIXBUF_HASH_SIZE){
		// this is for the case where the initial reference of the pixbuf
		// belongs to the pixbuf hash. In this case we need an additional
		// reference. 
		g_object_ref (population_p->initial_preview_pixbuf);
	    }
	    if (population_p->initial_preview_pixbuf == NULL) {
		g_warning("population_p->initial_preview_pixbuf == NULL!");
		retval=FALSE;
		view_p->tip_event.tooltip_active = FALSE;
		goto cleanup;
	    }
	    rfm_population_read_lock (view_p);
	    if (view_p->population_serial != population_serial){
		retval=FALSE;
		TRACE("tooltip_active FALSE: view_p->population_serial != population_serial\n");
		view_p->tip_event.tooltip_active = FALSE;
		rfm_population_read_unlock (view_p);
		goto cleanup;
	    }
	    ((population_t *)activate_tip_p->population_p)->initial_preview_pixbuf=
		population_p->initial_preview_pixbuf;
	    if (!G_IS_OBJECT (activate_tip_p->population_p->initial_preview_pixbuf)){
		g_warning("!G_IS_OBJECT (pixbuf) at [5] preliminary_tip()");
	    }
	    g_object_ref(activate_tip_p->population_p->initial_preview_pixbuf);
	    rfm_population_read_unlock (view_p);
	    retval=TRUE;
	}
    } // end of population_p->initial_preview_pixbuf == NULL


    g_mutex_lock(tooltip_text_mutex);
    if (view_p->mouse_event.saturated_p && view_p->mouse_event.saturated_p->en &&
	    view_p->mouse_event.saturated_p->en->path &&
	    strcmp(view_p->mouse_event.saturated_p->en->path,population_p->en->path)==0) { 
	// race check
      if (view_p->tip_event.tooltip_text) {
        g_free(view_p->tip_event.tooltip_text);
        view_p->tip_event.tooltip_text=NULL;
      }
      if(view_p->tip_event.tooltip_pixbuf){
       g_object_unref(view_p->tip_event.tooltip_pixbuf);
      }
      // population_p->initial_preview_pixbuf has an additional ref
      view_p->tip_event.tooltip_pixbuf= population_p->initial_preview_pixbuf;
      if (!G_IS_OBJECT (view_p->tip_event.tooltip_pixbuf)){
	    g_warning("!G_IS_OBJECT (pixbuf) at [6] preliminary_tip()");
      }
      g_object_ref(view_p->tip_event.tooltip_pixbuf);
      // the entry will resolve any race condition when the real
      // preview is done.
      record_entry_t *olden=view_p->tip_event.tooltip_entry;
      view_p->tip_event.tooltip_entry = rfm_copy_entry (population_p->en);
      rfm_destroy_entry(olden);
    }
    g_mutex_unlock(tooltip_text_mutex);
		
    view_p->tip_event.tooltip_active = TRUE;

cleanup:
    // private population cleanup
    if(population_p->pixbuf) {
        g_object_unref(population_p->pixbuf);
    }
    if(population_p->preview_pixbuf) {
        g_object_unref(population_p->preview_pixbuf);
    }
    if(population_p->initial_preview_pixbuf) {
        g_object_unref(population_p->initial_preview_pixbuf);
    }
    rfm_destroy_entry(population_p->en);
    g_free(population_p);
		

    return retval;
}


static gboolean 
is_normal_condition(population_t *population_p){
    gboolean normal_tips = FALSE;
    view_t *view_p=population_p->view_p;
    if (population_p->en==NULL || population_p->en->st==NULL || population_p->en->st->st_size == 0) {
	TRACE ("normal_tips; population_p->en==NULL || population_p->en->st==NULL\n");
	normal_tips = TRUE;
    }
    if (population_p->en && population_p->en->mimetype && 
	    strcmp(population_p->en->mimetype, "application/x-desktop")==0){
	normal_tips = TRUE;
    }

    
    
    if (! (view_p->flags.preferences & __SHOW_IMAGES)) {
	TRACE("normal_tips; ! (view_p->flags.preferences & __SHOW_IMAGES)\n");
	normal_tips = TRUE;
    }
    if(g_object_get_data (G_OBJECT (view_p->widgets.paper), "normal_tip")) {
	TRACE("normal_tips; g_object_get_data (G_OBJECT (view_p->widgets.paper), normal_tip)\n");
	normal_tips = TRUE;
    }
    gboolean normal_condition=TRUE;
    if (normal_tips) {
	normal_condition=TRUE; // user option
    } else if (population_p->en->module) {
	normal_condition=TRUE; // for now, preview are disabled in modules
    } else if (population_p->preview_pixbuf) {
	normal_condition=FALSE; // already done
	TRACE("already done: %s\n", population_p->en->path);
    } else if ((population_p->flags & POPULATION_NORMAL_TIP)) {
	TRACE("normal_tips; POPULATION_NORMAL_TIP\n");
	normal_condition=TRUE; // unable to do
    } else if (!population_p->en ) {
	TRACE("normal_tips; population_p->normal_tip !population_p->en\n");
	normal_condition=TRUE; // unable to do
    } else if (!population_p->en->st || 
	    population_p->en->st->st_size == 0) {
	TRACE("normal_tips; population_p->normal_tip !population_p->en->st\n");
	normal_condition=TRUE; // unable to do
    }
    else if (rfm_is_image (population_p->en)) {
	normal_condition=FALSE; // condition applies for images
	TRACE("rfm_is_image: %s\n", population_p->en->path);
    }
    else if((population_p->en->mimetype && strstr(population_p->en->mimetype, "opendocument"))
	    ||
	    (population_p->en->mimemagic && strstr(population_p->en->mimemagic, "opendocument"))){
	normal_condition=FALSE; // condition applies zipped opendocument files
    }
    else if(rfm_natural(PLUGIN_DIR, "mime", population_p->en, "want_imagemagick_preview")) {
	normal_condition=FALSE; // condition applies for PDF and TXT
	TRACE("want_imagemagick_preview: %s\n", population_p->en->path);
    }
    return normal_condition;
}

static void *
activate_tip_thread(void *data){
    activate_tip_t *activate_tip_p=data;
    TRACE("activate_tip_thread: 0x%x\n", GPOINTER_TO_INT(g_thread_self()));

    view_t *view_p = activate_tip_p->view_p;
    gint population_serial=activate_tip_p->population_serial;

    // thread memory setup (read lock protected)
    rfm_population_read_lock (view_p);
    if (view_p->population_serial != population_serial){
	rfm_population_read_unlock (view_p);
	// This is dumb. Why unmark it from busy state if 
	// the population serial is no longer valid?
	//((population_t *)activate_tip_p->population_p)->flags &=
	//	(POPULATION_TIP_BUSY ^ 0xffffffff);
	g_free(activate_tip_p);
	return NULL;
    }
    // Now that we have a valid population_p and a read lock on
    // the population_pp, we can proceed to make a valid copy
    population_t *population_p = (population_t *)malloc(sizeof(population_t));
    memcpy(population_p, activate_tip_p->population_p, sizeof(population_t));
    // Record entry need to be a new one so we can unlock the population_pp.
    // All other references in the population_p are not to be used.
    population_p->en = rfm_copy_entry(activate_tip_p->population_p->en);
    rfm_population_read_unlock (view_p);

    if (is_normal_condition(population_p)) {
	TRACE ("****  Normal tip\n");
        normal_tip(population_p, TRUE);  
	rfm_destroy_entry(population_p->en);
	g_free(population_p);
	GDK_THREADS_ENTER();
	    gtk_widget_trigger_tooltip_query(view_p->widgets.window);
	GDK_THREADS_LEAVE();
        g_free(activate_tip_p);
	return NULL;
    } 
    TRACE ("Not normal tip\n");
    rfm_destroy_entry(population_p->en);
    g_free(population_p);

   // preliminary_tip will set up the preliminary tip preview
   // if it returns TRUE, then preliminary preview is now in place. 
   // We can proceed with the final preview.
    if (preliminary_tip(activate_tip_p)) {
	TRACE("THREAD: g_thread_create(tip_preview_thread_f) at rodent_mouse.c\n");
	tip_preview_thread_f((gpointer) activate_tip_p);
    }
    GDK_THREADS_ENTER();
	gtk_widget_trigger_tooltip_query(view_p->widgets.window);
    GDK_THREADS_LEAVE();
    // end of not normal tips condition
    g_free(activate_tip_p);

    return NULL;
}


static void
hide_tip (view_t * view_p) {
    if(!view_p->tip_event.tooltip_active) {
	TRACE("view_p->tip_event.tooltip_active==FALSE\n");
	return;
    } 
    // free tip text and icon
    gtk_widget_set_tooltip_text (view_p->widgets.window, NULL);
    if (!tooltip_text_mutex){
	tooltip_text_mutex=g_mutex_new();
    }
    g_mutex_lock(tooltip_text_mutex);
    if(view_p->tip_event.tooltip_pixbuf) {
        view_p->tip_event.tooltip_pixbuf = NULL;
    }
    if(view_p->tip_event.tooltip_entry) {
        rfm_destroy_entry (view_p->tip_event.tooltip_entry);
        view_p->tip_event.tooltip_entry = NULL;
    }
    if(view_p->tip_event.tooltip_text) {
        g_free (view_p->tip_event.tooltip_text);
        view_p->tip_event.tooltip_text = NULL;
    }
    g_mutex_unlock(tooltip_text_mutex);

    TRACE ("SHOW_TIP: hide_tip view_p\n");
    view_p->tip_event.tooltip_active = FALSE;
    gtk_widget_trigger_tooltip_query(view_p->widgets.window);
}


gboolean
rodent_tip_function (
        GtkWidget * window, 
        gint x, 
        gint y, 
        gboolean keyboard_mode, 
        GtkTooltip * tooltip,
        gpointer user_data
) {
    view_t *view_p = rodent_get_current_view (window);
    TRACE ("RODENT_TIP_FUNCTION\n");

    record_entry_t *en = NULL;
    if(view_p->tip_event.tooltip_entry) {
        en = view_p->tip_event.tooltip_entry;
    } else {
        TRACE ("RODENT_TIP_FUNCTION; view_p->tip_event.tooltip_entry is NULL\n");
        gtk_widget_set_tooltip_text (view_p->widgets.window, NULL);
	view_p->tip_event.tooltip_active = FALSE;
        return FALSE;
    }

    if(IS_UP_TYPE (en->type)) {
        TRACE ("RODENT_TIP_FUNCTION : IS_UP_TYPE\n");  
        gtk_widget_set_tooltip_text (view_p->widgets.window, NULL);
	view_p->tip_event.tooltip_active = FALSE;
        return FALSE;           // omit up icons
    }

    if(view_p->tip_event.tooltip_pixbuf && view_p->tip_event.tooltip_active) {
        TRACE ("RODENT_TIP_FUNCTION (pixbuf) %s\n", view_p->tip_event.tooltip_text);
        // this conditional is just to avoid a useless warning on race condition
        // with threads (not really avoid, just reduce possibility).
	// But it doesn't always avoid crash, so let's get a read lock
	// instead...
	if (!rfm_population_try_read_lock (view_p))
	    return TRUE;
        if (GDK_IS_PIXBUF (view_p->tip_event.tooltip_pixbuf)) {
            gtk_tooltip_set_icon (tooltip, view_p->tip_event.tooltip_pixbuf);
        }
        gtk_tooltip_set_text (tooltip, view_p->tip_event.tooltip_text);
	rfm_population_read_unlock (view_p);
	
        return TRUE;
    } else {
        TRACE ("RODENT_TIP_FUNCTION tooltip_pixbuf==0x%x view_p->tip_event.tooltip_active=%d\n", 
		GPOINTER_TO_INT(view_p->tip_event.tooltip_pixbuf), view_p->tip_event.tooltip_active);
        gtk_widget_set_tooltip_text (view_p->widgets.window, NULL);
	view_p->tip_event.tooltip_active = FALSE;
        return FALSE;
    }
}

static gpointer
tip_preview_thread_f (gpointer data) {
    TRACE ("SHOW_TIP: tip_preview_thread_f NOW...\n");
    activate_tip_t *activate_tip_p=data;
    view_t *view_p = activate_tip_p->view_p;
    int population_serial=activate_tip_p->population_serial;
    rfm_population_read_lock (view_p);
    if (view_p->population_serial != population_serial){
	rfm_population_read_unlock (view_p);
	return NULL;
    }
    if (activate_tip_p->population_p->flags & POPULATION_PREVIEW_BUSY) {
	rfm_population_read_unlock (view_p);
	return NULL;
    }
    ((population_t *)activate_tip_p->population_p)->flags |= POPULATION_PREVIEW_BUSY;
    population_t *population_p = (population_t *)malloc(sizeof(population_t));
    memcpy(population_p, activate_tip_p->population_p, sizeof(population_t));
    population_p->en = rfm_copy_entry(activate_tip_p->population_p->en);
    rfm_population_read_unlock (view_p);

    // Do the real preview.
    
    if (!tooltip_text_mutex){
	tooltip_text_mutex=g_mutex_new();
    }

       
    GdkPixbuf *result=rfm_natural(PLUGIN_DIR, "mime", population_p, "mime_preview");
    if (GDK_IS_PIXBUF(result)) {
	// update tooltip pixbuf, subject to
        // race condition check
        gboolean norace = (view_p->tip_event.tooltip_entry 
            && strcmp(view_p->tip_event.tooltip_entry->path,population_p->en->path)==0);
        if (norace){
            TRACE("SHOW_TIP: population_p->preview_pixbuf OK!\n");
    g_mutex_lock(tooltip_text_mutex);
	    if (view_p->mouse_event.saturated_p && view_p->mouse_event.saturated_p->en 
		    && view_p->mouse_event.saturated_p->en->path &&
		strcmp(view_p->mouse_event.saturated_p->en->path,population_p->en->path)==0) { 
		if (view_p->tip_event.tooltip_pixbuf) {
		    g_object_unref(view_p->tip_event.tooltip_pixbuf);
		}
		view_p->tip_event.tooltip_pixbuf = result;
		if (!G_IS_OBJECT (view_p->tip_event.tooltip_pixbuf)){
		    g_warning("!G_IS_OBJECT (pixbuf) at [2] tip_preview_thread_f()");
		}
		g_object_ref(view_p->tip_event.tooltip_pixbuf);
	    }
    g_mutex_unlock(tooltip_text_mutex);
            //view_p->tip_event.tooltip_entry = rfm_copy_entry (population_p->en);
            view_p->tip_event.tooltip_active = TRUE;
        } 
    }
	   
   
    // update public population_p here
    rfm_population_read_lock (view_p);
    if (view_p->population_serial != population_serial){
	rfm_population_read_unlock (view_p);
	goto cleanup;
    }
    if (!result) {
            TRACE("result==FALSE!\n");

	// busy flag will remains to avoid retries.
	// above: activate_tip_p->population_p->flags |= POPULATION_PREVIEW_BUSY;
	// normal tip is set to short circuit
	((population_t *)activate_tip_p->population_p)->flags |= POPULATION_NORMAL_TIP;
    } else {
	if (view_p->population_serial == population_serial){
	    if (activate_tip_p->population_p->preview_pixbuf) {
		g_object_unref(activate_tip_p->population_p->preview_pixbuf);
	    }
	    ((population_t *)activate_tip_p->population_p)->preview_pixbuf=result;
	    if (!G_IS_OBJECT (activate_tip_p->population_p->preview_pixbuf)){
		g_warning("!G_IS_OBJECT (pixbuf) at [3] tip_preview_thread_f()");
	    }
	    g_object_ref(activate_tip_p->population_p->preview_pixbuf);
	}
	((population_t *)activate_tip_p->population_p)->flags &= (POPULATION_PREVIEW_BUSY ^ 0xffffffff);
	((population_t *)activate_tip_p->population_p)->preview_pixbuf_timestamp = 
	    activate_tip_p->population_p->en->st->st_mtime;	
    }
    rfm_population_read_unlock (view_p);

    
    //population_p->flags &= (POPULATION_PREVIEW_BUSY ^ 0xffffffff);

    // cleanup:
    TRACE("SHOW_TIP: tip_preview_thread_f DONE...\n");
cleanup:
    if (result) {
	 g_object_unref(result);
    }
    rfm_destroy_entry(population_p->en);
 
    g_free(population_p);

    return NULL;
}


/*************************************************************************/
/******************* end of tip functions ********************************/
/*************************************************************************/


static void setup_drag_state (view_t * view_p, GdkEventButton * event);

static void rubber_band (view_t * view_p, int x, int y, gboolean draw);

static void 
enter_drag_state (view_t * view_p) {
    TRACE ("enter dragstate \n");

    if(view_p->mouse_event.dragstate) {
        TRACE ("dragstate true\n");
        return;
    }
    TRACE ("now entering dragstate: event=0x%lx\n", (unsigned long)(&(view_p->mouse_event.drag_event)));
    //TRACE("now entering dragstate: G_IS_OBJECT (event)=%d\n",G_IS_OBJECT (&(view_p->mouse_event.drag_event)));

    view_p->mouse_event.drag_event.context = gtk_drag_begin (view_p->widgets.paper,
                                                 view_p->mouse_event.target_list,
                                                 GDK_ACTION_MOVE |
                                                 GDK_ACTION_COPY |
                                                 GDK_ACTION_LINK, 
						 1, //drag button
						 (GdkEvent *) (&(view_p->mouse_event.drag_event)));

    TRACE ("drag begun...\n");
    if(!view_p->mouse_event.drag_event.context)
        return;
    gdk_drag_status (view_p->mouse_event.drag_event.context, view_p->mouse_event.drag_action, view_p->mouse_event.drag_event.time);
    TRACE ("drag status...\n");


    gchar *plural_text=g_strdup_printf (
        ngettext ("%'u item", "%'u items",g_slist_length(view_p->selection_list)),
	g_slist_length(view_p->selection_list));
    gchar *g = g_strdup_printf ("%s: %s", _("Selection"), plural_text);
    g_free(plural_text);

    rfm_status (&(view_p->widgets), "xffm/stock_dialog-info", g, NULL);
    g_free (g);
    
    if(g_slist_length(view_p->selection_list) > 1) {
        TRACE ("selection_count > 1\n");
        gtk_drag_set_icon_stock (view_p->mouse_event.drag_event.context, GTK_STOCK_DND_MULTIPLE, 0, 0);
    } else if(view_p->mouse_event.dnd_pixbuf) {
        TRACE ("setting view_p->mouse_event.dnd_pixbuf\n");
        gtk_drag_set_icon_pixbuf (view_p->mouse_event.drag_event.context, view_p->mouse_event.dnd_pixbuf, 0, 0);
    } else {
        TRACE ("view_p->mouse_event.dnd_pixbuf is null, using GTK_STOCK_DND\n");
        gtk_drag_set_icon_stock (view_p->mouse_event.drag_event.context, GTK_STOCK_DND, 0, 0);
    }
    view_p->mouse_event.dragstate = TRUE;
    TRACE ("enter dragstate done\n");

}

static void
setup_drag_state (view_t * view_p, GdkEventButton * event) {
    TRACE ("+DND>>setup_drag_state\n");
    if(view_p->en && view_p->en->path)
        view_p->mouse_event.drag_src = view_p->en->path;
    else
        view_p->mouse_event.drag_src = NULL;
    TRACE ("+DND>>setup_drag_state: view_p->mouse_event.drag_src=%s\n", view_p->mouse_event.drag_src);
    view_p->mouse_event.drag_event.type = GDK_DRAG_ENTER;
    view_p->mouse_event.drag_event.x_root = event->x;
    view_p->mouse_event.drag_event.y_root = event->y;
    view_p->mouse_event.drag_event.time = event->time + 2;
    view_p->mouse_event.drag_event.window = event->window;
    view_p->mouse_event.drag_event.send_event = event->send_event;
    TRACE ("event time=0x%x\n", event->time);
}

static void
unselect_all_pixbuf_with_expose(view_t * view_p, gboolean expose) {
    TRACE (">> unselect_all_pixbuf_with_expose\n");
    population_t **tmp;
    for(tmp = view_p->population_pp; tmp && *tmp; tmp++) {
        population_t *population_p = *tmp;
        if(!population_p)
            continue;
        if(population_p == view_p->mouse_event.doing_drag_p)
            continue;

        if (population_p->flags  & POPULATION_SELECTED) {
	    unselect_pixbuf (view_p, population_p);
	    if (expose) rodent_expose_item(view_p, population_p);
        }
    }
    if(view_p->selection_list){
	GSList *tmp=view_p->selection_list;
	for (;tmp && tmp->data; tmp=tmp->next){
	    record_entry_t *en=tmp->data;
	    rfm_destroy_entry(en);
	}
        g_slist_free (view_p->selection_list);
    }
    view_p->selection_list = NULL;

}


static gboolean
scroll_business(view_t *view_p, gint y, GdkRectangle *area){
    if (!GTK_IS_SCROLLED_WINDOW (view_p->widgets.scrolled_window)) return FALSE;
    double upper = gtk_adjustment_get_upper (
	    gtk_scrolled_window_get_vadjustment (view_p->widgets.scrolled_window));
    double page = gtk_adjustment_get_page_size (
	    gtk_scrolled_window_get_vadjustment (view_p->widgets.scrolled_window));
    double value = gtk_adjustment_get_value  (
	    gtk_scrolled_window_get_vadjustment (view_p->widgets.scrolled_window));
    gboolean set_scroll=FALSE;
    gdouble new_value = 0.0;
    if (y > value + page && value + page < upper) {
        TRACE("scrolldown: y= %d value =%lf, upper=%lf page=%lf \n",
		 y, value, upper, page);
	new_value = (y - page < upper - page)? 
	    y - page : upper - page;
	TRACE("scrolldown to %lf\n", new_value );
	set_scroll=TRUE;
    } else if (y < value) {
	new_value = y;
	TRACE("scrollup to %lf\n", new_value);
        set_scroll=TRUE;
    } else {
         TRACE("scroll noop: y= %d value =%lf, upper=%lf page=%lf \n",
		 y, value, upper, page);
   }
    if (set_scroll){
        gtk_adjustment_set_value (
		gtk_scrolled_window_get_vadjustment (
		    view_p->widgets.scrolled_window), 
		new_value);
    }
    return set_scroll;
}



static void
rubber_band (view_t * view_p, int x, int y, gboolean draw) {
    TRACE (">> rubber_band\n");
    if(view_p->mouse_event.boxX == -1 && view_p->mouse_event.boxY == -1)
        return;
    if(view_p->mouse_event.old_X == -1 && view_p->mouse_event.old_Y == -1) {
        view_p->mouse_event.old_X = view_p->mouse_event.boxX;
        view_p->mouse_event.old_Y = view_p->mouse_event.boxY;
	return;
    } 

    // old rectangle
    gint lowX = (view_p->mouse_event.old_X > view_p->mouse_event.boxX) ? view_p->mouse_event.boxX : view_p->mouse_event.old_X;
    gint lowY = (view_p->mouse_event.old_Y > view_p->mouse_event.boxY) ? view_p->mouse_event.boxY : view_p->mouse_event.old_Y;
    gint highX = (view_p->mouse_event.old_X < view_p->mouse_event.boxX) ? view_p->mouse_event.boxX : view_p->mouse_event.old_X;
    gint highY = (view_p->mouse_event.old_Y < view_p->mouse_event.boxY) ? view_p->mouse_event.boxY : view_p->mouse_event.old_Y;

    // new rectangle
    view_p->mouse_event.old_X = x;
    view_p->mouse_event.old_Y = y;
    gint new_lowX = (view_p->mouse_event.old_X > view_p->mouse_event.boxX) ? view_p->mouse_event.boxX : view_p->mouse_event.old_X;
    gint new_lowY = (view_p->mouse_event.old_Y > view_p->mouse_event.boxY) ? view_p->mouse_event.boxY : view_p->mouse_event.old_Y;
    gint new_highX = (view_p->mouse_event.old_X < view_p->mouse_event.boxX) ? view_p->mouse_event.boxX : view_p->mouse_event.old_X;
    gint new_highY = (view_p->mouse_event.old_Y < view_p->mouse_event.boxY) ? view_p->mouse_event.boxY : view_p->mouse_event.old_Y;
#if 10

/*
GXxor Combines the source and existing bits in such a way that, if the
operation is repeated, the drawable is returned to its condition
just before the two operations. 

It is important that these two operations
occur without intervening manipulation of the selected bits (for windows,
the server should be grabbed but for a very short time).

When the server is grabbed, the client that grabbed it has sole control 
over the server and the screen. All other clients are put on hold; 
the server saves up events queued for them and does not change the 
screen on their behalf until the server grab is released.
The server is grabbed and released with XGrabServer() and XUngrabServer(), 
but this should be done only when really necessary. Otherwise, the second
XOR operation will not leave the drawable unchanged. GXxor has these
properties on both monochrome and color screens.

When you create a GC of your own, its default values are slightly different
from the values of the default GC. Its foreground and background values are 0 and 1,
respectively, so they are not necessarily black and white or contrasting.
That's why you must always set foreground and background when you create a GC.
Also, the default font is implementation dependent, and it may not be loaded. 
Therefore, you must always load the font before attempting to draw with it.
*/

    // Nothing is as good as getting back to basics...
    static GC *xorGC=NULL;
    //Window window=gdk_x11_get_default_root_xwindow();
    Display *display=gdk_x11_display_get_xdisplay(gdk_display_get_default());
    //Visual *visual = gdk_x11_visual_get_xvisual(gdk_visual_get_system());
    Drawable drawable = gdk_x11_drawable_get_xid(gtk_widget_get_window(view_p->widgets.paper));
    if (!xorGC){
	// create xor gc
	//Screen *screen = gdk_x11_screen_get_xscreen (gdk_screen_get_default());
        XGCValues xGCv;
	/*xGCv.foreground= BlackPixel(display, 0);
	xGCv.background = WhitePixel(display, 0);
	xGCv.function = GXxor;
	*xorGC = XCreateGC (display, drawable, GCForeground | GCBackground | GCFunction, & xGCv);*/
	xorGC = (GC *)malloc(sizeof(GC));
	*xorGC = XCreateGC (display, drawable, 0, & xGCv);
	XSetForeground(display, *xorGC, BlackPixel(display, 0) ^  WhitePixel(display, 0));
	XSetBackground(display, *xorGC, WhitePixel(display, 0));
	XSetFunction(display, *xorGC, GXxor);
    }
#endif
    /* clean old rectangle */
    GdkEventExpose old_event;
    old_event.area.x=lowX;
    old_event.area.y=lowY;
    old_event.area.width=highX - lowX;
    old_event.area.height=highY - lowY;

#if 10
    if(lowX - highX || lowY - highY){
	XDrawRectangle(display, drawable, *xorGC, 
		lowX,  
		lowY, 
		highX - lowX, 
		highY - lowY);
    }
#endif

    GSList *tmp;
    // find all items in old rectangle (restricted to icons)
    GSList *old_list=rodent_find_items_in_rectangle(view_p, &(old_event.area), FALSE);
    gint old_items=g_slist_length(old_list);
    if(!draw) {
	// here we are just erasing the last rubberband
	tmp=old_list;
	for (;tmp && tmp->data; tmp=tmp->next){
	    const population_t *population_p=tmp->data;
	    rodent_expose_item(view_p, population_p);
	}
	g_slist_free(old_list);
        return;
    }
   
 
    GdkEventExpose new_event;
    new_event.area.x=new_lowX;
    new_event.area.y=new_lowY;
    new_event.area.width=new_highX - new_lowX;
    new_event.area.height=new_highY - new_lowY;
	
    // find all items in new rectangle (restricted to icons)
    GSList *new_list=rodent_find_items_in_rectangle(view_p, &(new_event.area), TRUE);
    gint new_items=g_slist_length(new_list);

    // unselect all unselected items
    for (tmp=old_list; tmp && tmp->data; tmp=tmp->next){
	if (!g_slist_find(new_list, tmp->data)){
	    unselect_pixbuf (view_p, (population_t *)tmp->data);
	}
    }

    // this will do any scrolling
    scroll_business(view_p, y, &(new_event.area));

    // select all selected items
    for (tmp=new_list; tmp && tmp->data; tmp=tmp->next){
	rodent_select_pixbuf (view_p, (population_t *)tmp->data);
    }
    g_slist_free(new_list);
    new_list=NULL;
    new_list=rodent_find_items_in_rectangle(view_p, &(new_event.area), FALSE);

    // expose afected items
    if (new_items != old_items) {
	GSList *tmp=old_list;
	for (;tmp && tmp->data; tmp=tmp->next){
	    if(!g_slist_find(new_list, tmp->data)){
		new_list =g_slist_prepend(new_list, tmp->data);
	    }
	}
	TRACE("expose necessary: items=%d (old=%d)\n",new_items,old_items);
	TRACE("expose lists items=%d\n", g_slist_length(new_list));

	tmp=new_list;
	for (;tmp && tmp->data; tmp=tmp->next){
	    const population_t *population_p=tmp->data;

	    // here we need a realtime expose, but will the grab
	    // affect? is a read lock set?
	    rodent_expose_item(view_p, population_p);
	}
    }
    g_slist_free(old_list);
    g_slist_free(new_list);
 
    /* draw new XOR rectangle */
    if(new_lowX - new_highX || new_lowY - new_highY){
	XDrawRectangle(display, drawable, *xorGC, 
		new_lowX, 
		new_lowY, 
		new_highX - new_lowX, 
		new_highY - new_lowY);
    }


    gchar *string=NULL;
    if (new_items) {
	string = g_strdup_printf (ngettext ("%'d item selected", "%'d items selected", new_items), new_items);
    } else {
	string = g_strdup_printf("%s: %s", _("Selection"), _("None"));
    }
    rfm_status(&(view_p->widgets), "xffm/stock_dialog-info", string, NULL);
    g_free(string);



}

/* button press */
static void
button_popup (GdkEventButton * event, view_t * view_p, const population_t * population_p) {
    if(population_p) {
        if(!(population_p->flags  & POPULATION_SELECTED)) {
            rodent_unselect_all_pixbuf (view_p);
        }
        rodent_select_pixbuf (view_p, population_p);
	rodent_expose_item (view_p, population_p);
    }
    rodent_do_popup (view_p, population_p, event);
}

static record_entry_t *
find_in_selection_list(view_t *view_p, record_entry_t *population_en){
    if (!view_p->selection_list) return FALSE;
    record_entry_t *found=NULL;
    GSList *tmp=view_p->selection_list;
    for (; tmp && tmp->data; tmp=tmp->next){
	record_entry_t *en=tmp->data;
	if (!en->path || !population_en->path) continue;
	if (strcmp(en->path, population_en->path)==0){
	    found=tmp->data;
	    break;
	}
    }
    return found;
}
/*
static void
remove_ghost_selections(view_t * view_p){
    GSList *ghosts=NULL;
    GList *tmp=view_p->selection_list;
    for (; tmp && tmp->data; tmp=tmp->next){
	record_entry_t *en=tmp->data;
	if (en->path && !rfm_g_file_test(en->path, G_FILE_TEST_EXISTS)){
	    ghosts = g_slist_prepend(ghosts, tmp->data);
	}
    }
    GSList *stmp = ghosts;
    for (; stmp && stmp->data; stmp=stmp->next){
	record_entry_t *en=stmp->data;
	view_p->selection_list = g_list_remove(view_p->selection_list, stmp->data);
	rfm_entry_destroy(en);
    }
    g_slist_free(ghosts);
    if (g_list_length(view_p->selection_list)==0){
	g_list_free(view_p->selection_list);
	view_p->selection_list = NULL;
    }  
}
*/

static void
unselect_pixbuf (view_t * view_p, const population_t * population_p) {
    TRACE (">> unselect_pixbuf\n");
    if(!population_p) {
        g_warning ("!population_p");
        return;
    }

    if (population_p->flags  & POPULATION_SELECTED) {
	((population_t *)population_p)->flags  &= (POPULATION_SELECTED ^ 0xffffffff);
//	rodent_expose_item (view_p, population_p);
    }
   
    record_entry_t *item = find_in_selection_list(view_p, population_p->en);
    if(item) {
        view_p->selection_list = g_slist_remove (view_p->selection_list, item);
	rfm_destroy_entry(item);
        if(!g_slist_length (view_p->selection_list)) {
            g_slist_free (view_p->selection_list);
            view_p->selection_list = NULL;
        }
    }

    
}


