//
//
/* this file is included by rodent_population.c */


static void
mk_logical_rect (PangoLayout * layout, PangoRectangle * logical_rect) {
    TRACE (">> mk_logical_rect\n");
    if(layout) {
        pango_layout_get_pixel_extents (layout, NULL, logical_rect);
        /* ++ because zero is actually unity, not nonexistance */
        logical_rect->width++;
        logical_rect->height++;
    }
}

//FIXME: we could just use pango internal functions to do the cutting,
//       that would definitely be better for non latin utf8 character sets...
static void
mk_layouts (view_t * view_p, gchar * tag, PangoLayout ** layout, PangoLayout ** layout2) {
    TRACE (">> rfm_mk_layouts\n");
    int i;
    gchar *optimal_chars = " .-_";
    static gchar *optimal_set[4] = { NULL, NULL, NULL, NULL };
    PangoRectangle logical_rect;
    gchar *tag_copy = NULL,
        *the_tag = NULL;

    if(!layout || !layout2)
        return;
    /* cleanup first */
    if(*layout)
        g_object_unref (*layout);
    if(*layout2)
        g_object_unref (*layout2);
    *layout = *layout2 = NULL;

    if(!tag || !strlen (tag))
        return;

    if(!optimal_set[0]) {
        for(i = 0; i < strlen (optimal_chars); i++) {
            gchar c[2] = { 0, 0 };
            c[0] = optimal_chars[i];
            optimal_set[i] = rfm_utf_string (c);
        }
    }

    /* new method... */
    /* get width of string
     * string MUST be a validated utf string here. */
    the_tag = g_strdup (tag);
    *layout = gtk_widget_create_pango_layout (view_p->widgets.paper, the_tag);
    pango_layout_get_pixel_extents (*layout, NULL, &(logical_rect));
    gint max_width=CELLWIDTH(view_p) - 2 * TEXTSPACING;
    if (view_p->view_layout.icon_size == TINY_ICON_SIZE) max_width -= TINY_ICON_SIZE;

    if(view_p->view_layout.icon_size < TINY_ICON_SIZE ||
        logical_rect.width + 1 <= max_width) 
    {
        /* if it fits, or < TINY_ICON_SIZE, were all done */
        TRACE ("no cut necessary for %s\n", tag);
        g_free (the_tag);
        return;
    }
    /* shucks, it doesn't fit. Let's work with a copy of the tag */
    TRACE ("shucks, it doesn't fit: %d > %d\n", logical_rect.width + 1, max_width);
    tag_copy = g_strdup (the_tag);
    while(strlen (tag_copy) > 1 && logical_rect.width + 1 > max_width) {
        gchar *rear;
        int length = strlen (tag_copy);
        gchar *p = tag_copy + length - 1;
        rear = g_utf8_find_prev_char (tag_copy, p);
        length = strlen (rear);
        memset (rear, 0, length);
        g_object_unref (*layout);
        *layout = gtk_widget_create_pango_layout (view_p->widgets.paper, tag_copy);
        pango_layout_get_pixel_extents (*layout, NULL, &(logical_rect));
    }
    if (view_p->view_layout.icon_size >= TINY_ICON_SIZE && view_p->view_layout.icon_size < SMALL_ICON_SIZE){
	// this is compact view, so we don't care about layout2
        g_object_unref (*layout);
	gchar *compact_tag=g_strdup_printf("%s...", tag_copy);
        *layout = gtk_widget_create_pango_layout (view_p->widgets.paper, compact_tag);
        pango_layout_get_pixel_extents (*layout, NULL, &(logical_rect));
	g_free (compact_tag);
	g_free (the_tag);
	g_free (tag_copy);
        return;
    }
    TRACE ("\"%s\" cut to \"%s\"\n", the_tag, tag_copy);
    /* Now the first part fits, but the split may not be optimum
     * we must now look for one in " .-_" to do the split, but the
     * character search must be done in utf */
    /* is the cut optimal? */
    {
        int cut_location = strlen (tag_copy);
        gchar *optimal_ch = NULL;
        gchar *p = the_tag + (cut_location - 1);
        gchar *optimal_cut_p = NULL;
        p = g_utf8_find_prev_char (the_tag, p); /* previous validated location */

        p = g_utf8_find_next_char (p, NULL);    /* validated location */

        for(i = 0; i < strlen (optimal_chars); i++) {
            if(strncmp (p, optimal_set[i], strlen (optimal_set[i])) == 0) {
                optimal_ch = optimal_set[i];
                optimal_cut_p = p;
                TRACE ("optimal cut found for %s\n", tag);
                break;
            }
        }
        if(optimal_cut_p)
            goto go_on;

        for(i = 0; i < strlen (optimal_chars); i++) {
            gchar *g = g_utf8_strrchr (tag_copy, -1, g_utf8_get_char (optimal_set[i]));

            if(g) {
                int diff = strlen (tag_copy) - strlen (g);
                g = the_tag + diff + 1;
                *layout2 = gtk_widget_create_pango_layout (view_p->widgets.paper, g);
                pango_layout_get_pixel_extents (*layout2, NULL, &(logical_rect));
                if(logical_rect.width + 1 <= max_width) {
                    TRACE ("might improve cut for %s\n", tag);
                    g_object_unref (*layout);
                    memset (tag_copy + diff + 1, 0, strlen (tag_copy + diff + 1));
                    *layout = gtk_widget_create_pango_layout (view_p->widgets.paper, tag_copy);

                    goto done;
                } else {
                    TRACE ("NOT improveable: %s (could look for another cut char)\n", tag);
                    g_object_unref (*layout2);
                }
                break;
            }
        }
    }

    /* So we don't have an optimal split. Let's construct the rest. */
  go_on:
    if(strlen (tag_copy) && strlen (the_tag) > strlen (tag_copy)) {
        gchar *p = the_tag + (strlen (tag_copy));
        /* XXX should do this for all spaces, not just the first  */
        if(strncmp (p, optimal_set[0], strlen (optimal_set[0])) == 0)
            p = g_utf8_find_next_char (p, NULL);

        *layout2 = gtk_widget_create_pango_layout (view_p->widgets.paper, p);
        /* final question, does it fit? */
        pango_layout_get_pixel_extents (*layout2, NULL, &(logical_rect));
        while(logical_rect.width + 1 > max_width) {
            gchar *g;
            TRACE ("second line does not fit (%s)\n", p);
            g_object_unref (*layout2);
            p = g_utf8_find_next_char (p, NULL);
            g = g_strconcat ("~", p, NULL);
            *layout2 = gtk_widget_create_pango_layout (view_p->widgets.paper, g);
            pango_layout_get_pixel_extents (*layout2, NULL, &(logical_rect));
            g_free (g);
        }

    }
    /* finally we free our copies of the tag */
  done:
    g_free (tag_copy);
    g_free (the_tag);
    return;
}


void
destroy_population_item (population_t * population_p) {
    if(!population_p)
        return;
    /* objects that need to be unreffed: 
     * (tagged as referrence in types.h) */

    if(population_p->preview_pixbuf)
        g_object_unref (population_p->preview_pixbuf);
    if(population_p->initial_preview_pixbuf)
        g_object_unref (population_p->initial_preview_pixbuf);

    if(population_p->layout)
        g_object_unref (population_p->layout);
    if(population_p->layout2)
        g_object_unref (population_p->layout2);
    rfm_destroy_entry (population_p->en);
    g_free (population_p);
}

// a bit complicated, and a bit dificult to debug...
// but it avoids remaking thumnail pixmaps
static void
transfer_population (view_t * view_p) {
    int i;
    population_t **new_population_pp,
    **old_population_pp;
    new_population_pp = rodent_mk_grid_elements (view_p, view_p->xfdir_p->pathc);
    old_population_pp = view_p->population_pp;

    //TRACE(" gl is 0x%lx for fstab\n", (unsigned long)((long)view_p->xfdir_p->gl));
    //TRACE("monitor transfer_population xfdir_p 0x%lx != xfdir_p->xfdir_p 0x%lx\n", (unsigned long)((long)view_p->xfdir_p), (unsigned long)((long)view_p->xfdir_p->xfdir_p));

    TRACE ("MONITOR TRANSFER, pathc is %ld\n", (long)view_p->xfdir_p->pathc);
    // find stuff in old_population which is in new_population
    // ignore up icon (element 0)
    new_population_pp[0] = old_population_pp[0];

    for(i = 1; i < view_p->xfdir_p->pathc; i++) {
        const gchar *newpath;
        if(view_p->xfdir_p->gl[i].en){
            newpath = view_p->xfdir_p->gl[i].en->path;
	} else {
            g_warning("transfer_population(): this is wrong");
	    continue;
	}
        population_t **tmp;
        gboolean hit = FALSE;
        population_t *population_p;
	TRACE ("looking for %s in old population\n", newpath);
        for(tmp = old_population_pp+1; *tmp; tmp++) {
            population_p = *tmp;
            //if (!population_p) break;
            const gchar *oldpath;
            if(population_p->en) {
                oldpath = population_p->en->path;
                if(view_p->xfdir_p->invalidate_icons) {
		    TRACE("invalidate icons is set!\n");
                    // if the invalidate_icons flag is set
                    // then all icons are updated,
                    // that means, nothing is found
                    // in the old population.
                    // This is practically a hard reload.
                    continue;
                }
            } else {
		g_warning("transfer_population(): this is wrong too");
		continue;
	    }
	    // 	    
	    // NOTE: Watch out, modules!
	    // Path must be unique!
            if(strcmp (newpath, oldpath) == 0)
	    {
		hit = TRUE; 
		// if stat information is different, icon may need to be
		// updated, so the easy way is to test this here too.
		gboolean valid_old=FALSE;
		if (population_p->en && population_p->en->st) 
		    valid_old=TRUE;
		gboolean valid_new=FALSE;
		if (view_p->xfdir_p->gl[i].en && view_p->xfdir_p->gl[i].en->st) 
		    valid_new=TRUE;
		if (valid_old && valid_new &&
		       population_p->en->st->st_mtime != 
		       view_p->xfdir_p->gl[i].en->st->st_mtime)  hit=FALSE;  
                break;
            }
        }
        if(hit) {
            // found in old_population
            // and the stat information is the same
            // image preview need not be updated. 
            new_population_pp[i] = population_p;
            /*  update the stat record */
            if(view_p->xfdir_p->gl[i].en) {
                if(!new_population_pp[i]->en->st) {
                    g_warning ("!new_population_pp[i]->en->st");
                }
                if(!view_p->xfdir_p->gl[i].en->st) {
                    g_warning ("!view_p->xfdir_p->gl[i].en->st");
                }
                memcpy (new_population_pp[i]->en->st, view_p->xfdir_p->gl[i].en->st, sizeof (struct stat));

            }
	    // FIXME: if iconsize is details, then you must update layout2 
	    //        as well, either by module or normal file.

            TRACE ("FOUND\n");
            TRACE ("newmonitor: hit for %s\n", newpath);
            // mv population element

        } else {
            //create new population item
            TRACE ("newmonitor: no-hit for %s\n", newpath);
            new_population_pp[i] = 
		rodent_create_population_t (view_p, view_p->xfdir_p->gl[i].en, i, NULL, view_p->xfdir_p->gl[i].pathv);
	    // create mimetype icon
	    const gchar *id=rfm_resolve_icon_id(view_p->xfdir_p->gl[i].en);
	    gchar *composite_id = 
		rfm_get_composite_icon_id (view_p->xfdir_p->gl[i].en, id);
	    new_population_pp[i]->pixbuf = 
		rfm_get_pixbuf (composite_id, REDUCED_ICONSIZE(view_p));
	    g_free(composite_id);

            //  new x,y position
            //  image preview thread

        }
    }
    // reset the state variable
    view_p->xfdir_p->invalidate_icons = FALSE;
/* That done, update the population selection list. */

    GSList *tmp = view_p->selection_list;
    for(; tmp && tmp->data; tmp = tmp->next) {
        record_entry_t *en = (record_entry_t *) tmp->data;
	if (!en->path) continue;
        gint i;
        gboolean found = FALSE;
        for(i = 0; new_population_pp[i]; i++) {
	    if (!new_population_pp[i]->en || !new_population_pp[i]->en->path) {
		continue;
	    }
            if(strcmp(en->path, new_population_pp[i]->en->path)==0) {
                found = TRUE;
                break;
            }
        }
	if(!found) {
            TRACE ("newmonitor: removing  %s from selection list\n", en->path);
            view_p->selection_list = g_slist_remove (view_p->selection_list, en);
        }
    }

    /* assign new population pp */
    view_p->population_pp = new_population_pp;
    // old population is now obsolete

/*
 * Next, cleanup.
 * Destroy the population elements of
 * old_population_pp which are not in the new_population_pp
 *
 * 
 * */

    int j;
    for(j = 0; old_population_pp[j]; j++) {
        population_t *found = NULL;	
        if(!old_population_pp[j]){
            continue;
	}
        int i;
        for(i = 0; new_population_pp[i]; i++) {
            if(old_population_pp[j] == new_population_pp[i]) {
                found = new_population_pp[i];
                break;
            }
        }
        if(!found) {
            TRACE ("newmonitor: removing  %s population_p\n",
                 (old_population_pp[j]->en == NULL) ? "root entry" : old_population_pp[j]->en->path);
            destroy_population_item (old_population_pp[j]);
            //old_population_pp[j] = NULL;
        } else if (view_p->view_layout.icon_size < TINY_ICON_SIZE){
	    // If found, then you might have to update layout2
	    // when iconsize is set to details.
	    if (found->layout2) g_object_unref(found->layout2);
	    if (view_p->en && view_p->en->module && 
		rfm_rational(PLUGIN_DIR,view_p->en->module, NULL, NULL, "make_list_layout2")){
		rfm_rational(PLUGIN_DIR,view_p->en->module, view_p, found, "make_list_layout2");
	    } else {
		rfm_xfdir_make_layout(view_p, found);
	    }
	}
    }


    /* destroy old population pp */

    g_free (old_population_pp);
    
    return;

}

static void
increase_population_serial (view_t * view_p) {
    if(view_p->mutexes.population_serial == NULL) {
        view_p->mutexes.population_serial = g_mutex_new ();
    }
    g_mutex_lock (view_p->mutexes.population_serial);
    view_p->population_serial++;
    g_mutex_unlock (view_p->mutexes.population_serial);

}

static
gpointer
janitor(gpointer data){
    GtkWidget *window=data;
    while (1){
	rfm_threadwait();
	if (rfm_global_p->window_view_list_readers > 0) continue;
	GSList *view_list = rfm_global_p->window_view_list;
	if (!view_list) {
	    TRACE("Janitor exiting now.\n");
	    // No more pages, destroy window.
GDK_THREADS_ENTER();
	    gtk_widget_hide (window);
	    gdk_flush();
	    gtk_widget_destroy (window);
GDK_THREADS_LEAVE();
	    // Break the gtk loop. 
	    // duh, cannot be done from a thread...
	    ///gtk_main_quit ();
	    // Since threads may be busy with stdout/stderr from
	    // spawned processes, we will do a harakiri here so
	    // that these threads don't go wild.
	    //rfm_threadwait();
	    if (strstr(rfm_global_p->argv[0], "rodent-desk")) {
		TRACE("killing off desktop...\n");
		gchar *arg[]={"killall", "rodent-desk", NULL}; 
		execvp(arg[0], arg);
		g_warning ("Rodent-desk: could not execvp to terminate!");
	    } else {
		// harakiri
		TRACE("harakiri\n");
		kill(rfm_global_p->pid, SIGTERM);
	    }
	    return NULL;
	}
	for (; view_list && view_list->data; view_list=view_list->next){
	    view_t *view_p=view_list->data;
	    g_mutex_lock(view_p->mutexes.janitor);
	    gint monitors=view_p->janitor.monitors;
	    gint children=view_p->janitor.children;
	    g_mutex_unlock(view_p->mutexes.janitor);
	    widgets_t *widgets_p = &(view_p->widgets);

	    if (view_p->flags.status == STATUS_EXIT && monitors==0 && children == 0) {
		TRACE("HEY!!! janitor() will now destroy view 0x%x\n",
			GPOINTER_TO_INT(view_p));
		// nulify any further io to diagnostics
		if (widgets_p->diagnostics_window && 
		    *(widgets_p->diagnostics_window) &&
		    widgets_p->diagnostics )
		{
GDK_THREADS_ENTER();
		    gtk_widget_hide(*(widgets_p->diagnostics));
GDK_THREADS_LEAVE();
		}
		if (view_p->widgets.diagnostics_window) {
		    *(view_p->widgets.diagnostics_window)=NULL;
		}
		if (view_p->widgets.diagnostics) {
		    *(view_p->widgets.diagnostics)=NULL;
		}

		view_p->widgets.status=NULL;
		view_p->widgets.paper=NULL;

		// Cleanup go history.
		GList *tmp;
		for(tmp = view_p->go_list; tmp; tmp = tmp->next) {
		    rfm_destroy_entry ((record_entry_t *) (tmp->data));
		}
		g_list_free (view_p->go_list);

		// Cleanup view population (this will block until write lock obtained)
		rfm_population_write_lock (view_p);
		rodent_destroy_population (view_p);
		view_p->population_pp = NULL;
		rfm_population_write_unlock (view_p);

		// Cleanup view
		rfm_destroy_entry (view_p->en);
		g_free (view_p->deepest_dir);
		while (rfm_global_p->window_view_list_readers > 0){
		    rfm_threadwait();
		}
GDK_THREADS_ENTER();
		rfm_global_p->window_view_list = 
		    g_slist_remove (rfm_global_p->window_view_list, view_p);
GDK_THREADS_LEAVE();
		g_free (view_p);
		view_list = rfm_global_p->window_view_list;
		if (!view_list) break;
	    }
	}
    }
}

//*** progress bar stuff ***/
#define SET_PROGRESS_FULL(w,x,y) set_progress_generic(w,x,y,1)

static gchar *
plain_sizetag (off_t tama, off_t tamal) {
    gchar *tag = "KB";
    gchar *tagl = "KB";
    gchar *buf;
    double x = tama;
    double y = tamal;

    if(tama >= 1000 * 1000) {
        x /= (1000 * 1000);
        tag = "GB";
    } else if(tama >= 1000) {
        x /= 1000;
        tag = "MB";
    } else if(tama >= 1000) {
        x /= 1000;
        tag = "MB";
    } 

    if(tamal >= 1000 * 1000) {
        y /= (1000 * 1000);
        tagl = "GB";
    } else if(tamal >= 1000) {
        y /= 1000;
        tagl = "MB";
    } else if(tamal >= 1000) {
        y /= 1000;
        tagl = "MB";
    }

    buf = g_strdup_printf ("%.3lf %s / %.3lf %s", x, tag, y, tagl);

    return buf;
}

static void
set_progress_generic (widgets_t * widgets_p, int count, int total, int caso) {
    static int smallcount = 1;
    GtkWidget *progress;
    gfloat fraction;

    if(!widgets_p)
        return;
    progress = widgets_p->progress;
    if(progress == NULL)
        return;
    if(caso == -1) {
        if(progress)
            gtk_widget_hide (progress);
        if(widgets_p->status)
            gtk_widget_show (widgets_p->status);
        return;
    }

    if(!progress) {
        TRACE ("progress==NULL");
        return;
    }
    if(!gtk_widget_get_visible (progress)) {
        if(widgets_p->status) {
            GtkTextBuffer *buffer = gtk_text_view_get_buffer ((GtkTextView *) widgets_p->status);
            gtk_text_buffer_set_text (buffer, " ", -1);
        }
        gtk_widget_show (progress);
    }
    /* this does not solve the indented status, neither does look good:
       if (widgets_p->status)gtk_widget_hide(widgets_p->status);
     */
    if(count == -1) {
        /*TRACE("TRACE:updating pulse\n"); */
        if(total < 0 || (smallcount++ & (1 << 7))) {
            gtk_progress_bar_pulse ((GtkProgressBar *) progress);
            //gdk_flush();
            smallcount = 1;
        }
    } else {
        gchar texto[_POSIX_PATH_MAX];
	gchar *size_tag;
        if(!total)
            fraction = 0.0;
        else
            fraction = (float)count / (float)total;
        if(fraction < 0 || fraction > 1)
            return;
        gtk_progress_bar_set_fraction ((GtkProgressBar *) progress, fraction);
        if(count >= total)
            texto[0] = 0;
        else
            switch (caso) {
            case 1:
                sprintf (texto, "%d / %d", count, total);
                break;
            case 0:
            default:
		size_tag = plain_sizetag (count, total);
                sprintf (texto, "%s", size_tag);
		g_free(size_tag);
                break;
            }
        gtk_progress_bar_set_text ((GtkProgressBar *) progress, texto);
    }
}


static
    void
set_application_icon (widgets_t * widgets_p, const gchar * icon_id) {
    // only from main thread
    if (rfm_global_p->self != g_thread_self()) {
	g_warning("attempt to call set_application_icon from child thread");
	return;
    }
    GdkPixbuf *icon_pixbuf = rfm_get_pixbuf (icon_id, SIZE_ICON);
    if(icon_pixbuf) {
        gtk_window_set_icon (GTK_WINDOW (widgets_p->window), icon_pixbuf);
    }
}

static
void
set_icon_name (GtkWidget * widget, const gchar * path) {
    gchar *iconname;
    TRACE ("set_icon_name\n");
    if(!path) {
        iconname = rfm_utf_string (g_get_host_name());
    } else if(g_file_test (path, G_FILE_TEST_EXISTS)) {
	//gchar *display_host = rfm_display_host_name(widget); 


        gchar *basename = g_path_get_basename (path);
        gchar *pathname = g_strdup (path);
        gchar *b = rfm_utf_string (basename);   // non chopped
        rfm_chop_excess (pathname);
        gchar *q = rfm_utf_string (pathname);   // non chopped

        g_free (basename);
        g_free (pathname);
	//iconname = g_strconcat (display_host, ":  ", b, " (", q, ")", NULL);
	iconname = g_strconcat (b, " (", q, ")", NULL);
        g_free (q);
        g_free (b);
    } else {
        iconname = rfm_utf_string (path);
        rfm_chop_excess (iconname);
    }
    gtk_window_set_title (GTK_WINDOW (gtk_widget_get_toplevel (widget)), iconname);
    //gdk_flush ();
    g_free (iconname);
}

/////////   icon roots stuff //////
//
//

static population_t *
create_local_root (view_t * view_p, int i, const gchar * path) {
    record_entry_t *en;
    population_t *population_p;
    const gchar *icon_id = NULL;
    const gchar *label = NULL;
    static gchar *root_icon=NULL;

    if(!path) {
        path = "/";
        label = "/";
	if (root_icon == NULL){
	    root_icon = 
		g_strdup_printf("%s/icons/Rodent/scalable/places/folder-root.svg", PACKAGE_DATA_DIR);
	}
        icon_id = root_icon;
    }
    if((en = rfm_stat_entry ((char *)path, 0)) == NULL) {
        g_warning (" rfm_stat_entry((char *)path, 0) != NULL");
        return NULL;
    }
    if(strcmp (path, g_get_home_dir ()) == 0) {
        icon_id = "xffm/stock_home";
        label = _("Home");
    }
    SET_LOCAL_TYPE (en->type);
    SET_ROOT_TYPE (en->type);
    en->tag = g_strdup (label);
    population_p = rodent_create_population_t (view_p, en, i, icon_id, label);
    if (population_p->en) population_p->en->module=NULL;

    return population_p;
}

static gint 
test_root_element (view_t * view_p, const gchar * module_name) {

    if(!module_name) {
        TRACE ("ROOTS: !module_name\n");
        return FALSE;
    }

    if(!rfm_void (PLUGIN_DIR, module_name, "is_root_module")) {
        TRACE (" ROOTS: %s is not root module\n", module_name);
        return FALSE;
    }
    const gchar *icon_id = 
	rfm_void (PLUGIN_DIR, module_name, "module_icon_id");
    gchar *label = rfm_natural (PLUGIN_DIR, module_name, NULL, "module_label");
    if(!icon_id || !label) {
        g_warning ("(!icon_id || !label)");
	g_free(label);
        return FALSE;
    }
    g_free(label);
    return TRUE;
}


static population_t *
create_icon_root_element (view_t * view_p, int i, const gchar * module_name) {
    if (!test_root_element (view_p, module_name)){
	return NULL;
    }
    TRACE ("ROOTS: module create_icon_root_element: %s\n", module_name);
    const gchar *icon_id =
	rfm_void (PLUGIN_DIR, module_name, "module_icon_id");
    gchar *module_label = rfm_natural (PLUGIN_DIR, module_name, NULL, "module_label");
    const gchar *submodule_name = 
	rfm_void (PLUGIN_DIR, module_name, "module_name");
    TRACE (" ROOTS: module_name --> %s \n", module_name);
    TRACE (" ROOTS: icon_id --> %s \n", icon_id);
    TRACE (" ROOTS: label --> %s \n", module_label);
    TRACE (" ROOTS: submodule_name --> %s \n", submodule_name);
    if (!icon_id || !module_label) {
	g_error("create_icon_root_element(): icon_id != NULL  || module_label != NULL not met");
    }
    record_entry_t *en = rfm_mk_entry (0);
    {
        gchar *module_path = (gchar *)rfm_natural (PLUGIN_DIR, module_name, NULL,
                                                           "module_path");
        if(!module_path){
            module_path = g_strdup (module_label);
	}
        en->path = g_strdup (module_path);
        en->tag = g_strdup (module_label);
	g_free(module_path);
    }
    // what the hell was this doing here: view_p->module_name = NULL;
    population_t *population_p = 
	rodent_create_population_t (view_p, en, i, icon_id, module_label);
    if(population_p->en)
        population_p->en->module = submodule_name;
    SET_ROOT_TYPE (en->type);
    g_free(module_label);
    return population_p;
}

    
typedef struct thread_wait_t {
    GThread *thread;
    gboolean wait_done;
    gboolean exit;
    void *retval;
} thread_wait_t;


static void *
thread_wait(gpointer data){
    thread_wait_t *thread_wait_p=data;
    thread_wait_p->retval = g_thread_join(thread_wait_p->thread);
    thread_wait_p->wait_done=TRUE;
    // If exit is set, then nobody is waiting for wait_done condition
    if (thread_wait_p->exit) g_free(thread_wait_p);
    else thread_wait_p->wait_done=TRUE;
    return NULL;
}

static
void *
thread_tryjoin (gpointer (*function) (gpointer), gpointer data, const gchar *text) {
    view_t *view_p=data;
    thread_wait_t *thread_wait_p=(thread_wait_t *)malloc(sizeof(thread_wait_t));
    memset(thread_wait_p, 0, sizeof(thread_wait_t));

    thread_wait_p->thread=g_thread_create(function, data, TRUE, NULL);
    g_thread_create(thread_wait, (void *)thread_wait_p, FALSE, NULL);

    time_t initial_time=time(NULL);
    gint progress_count = view_p->eyecandy.progress_count;
    //gint loop_count=0;
    while (!thread_wait_p->wait_done){
        gboolean progress = view_p->flags.type != DESKVIEW_TYPE && view_p->widgets.progress
            && gtk_widget_get_visible (view_p->widgets.progress);
        if(progress && progress_count != view_p->eyecandy.progress_count) {
            progress_count = view_p->eyecandy.progress_count;
            SET_PROGRESS_FULL (&(view_p->widgets), progress_count, view_p->eyecandy.progress_total);
        }
 
	// XXX: This must be removed for Rodent Delta.
	while(gtk_events_pending ()){
            gtk_main_iteration ();
	}
	gdk_flush();
	// Use wall clock instead of loop count
	// that is more uniform.
	//if (text && (++loop_count) % 50 == 0) {
	// If it takes more than 5 seconds, consider that the disk
	// is slow to respond and show the dialog to give the user
	// the option to abort.
	// sleep(5);
	if (text && time(NULL) - initial_time >= 5) {
	    if (!rfm_confirm(&(view_p->widgets),GTK_MESSAGE_OTHER,
		    text, _("Go back"), NULL)){
		 thread_wait_p->exit=1;
		 // in thread_wait(): g_free(thread_wait_p);
		 return NULL;
	    }
	}
	rfm_threadwait();
    }
    void *retval=thread_wait_p->retval;
    g_free(thread_wait_p);
    return retval;
}

static
gpointer
load_xfdir_structure (gpointer data) {
    view_t *view_p = data;
    /* Generalized xfdir get (includes modules) */
    xfdir_t *xfdir_p = rfm_xfdir_get (view_p, view_p->en, FALSE);
    if(xfdir_p==NULL) {
        g_warning ("load_xfdir_structure: xfdir_p is NULL!");
    }

    //view_p->xfdir_p = xfdir_p;

    return xfdir_p;
}

static
gpointer
load_xfdir_icons (gpointer data) {
    view_t *view_p = data;
    unsigned int smallcount = 0;
    int j;
    /* add entry for each population_t */
    TRACE ("Loading... load_xfdir_icons()\n");
    TRACE ("pathc=%ld\n", (long)view_p->xfdir_p->pathc);

    /* previous value of view_p->population_pp is freed in init_iconview() call */

    view_p->population_pp = 
	rodent_mk_grid_elements (view_p, view_p->view_layout.grid_rows * view_p->view_layout.grid_columns);

    // we could mutex lock this, 
    // but that should not be necesary since there is only one writer (?)
    view_p->eyecandy.progress_count = 0;
    int maxY = 0;
    int lastY = 0;
    if(view_p->xfdir_p == NULL)
        return NULL;
    if(view_p->population_pp == NULL)
        return NULL;
	TRACE("load_xfdir_structure: loop\n");
    for(j = 0; j < view_p->xfdir_p->pathc; j++) {
        if(j >= view_p->view_layout.max_elements) {
            TRACE ("sanity check: j < view_p->view_layout.max_elements not met\n");
            continue;
        }
        /* this is OK, it may be an icon to the rodent root level
         * if (!view_p->xfdir_p->gl[j].en) {
         g_warning("sanity check: view_p->xfdir_p->gl[j].en != NULL not met");
         continue;
         }*/
        if(!view_p->xfdir_p->gl[j].pathv) {
            g_warning ("sanity check: view_p->gl[j]->pathv != NULL not met");
            continue;
        }
        if(view_p->xfdir_p->gl[j].en) {
            view_p->xfdir_p->gl[j].en->module = view_p->submodule_name;
        }
        view_p->population_pp[j] = rodent_create_population_t (view_p, view_p->xfdir_p->gl[j].en, j, NULL, view_p->xfdir_p->gl[j].pathv);
        TRACE ("%d. created %s (st: 0x%x -> 0x%x)\n",
		j + 1, view_p->xfdir_p->gl[j].pathv, 
		(view_p->xfdir_p->gl[j].en)?
		GPOINTER_TO_INT(view_p->xfdir_p->gl[j].en->st):0,
		(view_p->population_pp[j]->en)?
		GPOINTER_TO_INT(view_p->population_pp[j]->en->st):0);
        view_p->flags.count++;

        TRACE ("load_xfdir_icons: new view_p->population_pp[j](%d,%d)=0x%lx\n",
             view_p->population_pp[j]->row, view_p->population_pp[j]->column, (unsigned long)view_p->population_pp[j]);

        if(view_p->flags.type != DESKVIEW_TYPE) {
            if(view_p->population_pp[j]->rect.y - CELLHEIGHT(view_p) > maxY) {
                maxY = view_p->population_pp[j]->rect.y - CELLHEIGHT(view_p);
            }
            if((smallcount++) % 52 == 0) {
                view_p->eyecandy.progress_count = j;
		// Hmm. this is no good since the expose routine with
		// cairo needs a read lock on the view, and here we are
		// working with a write lock...
                /* GdkRectangle rect;
                rect.x = 0;
                rect.y = lastY;
                rect.width = view_p->widgets.window->allocation.width;
                rect.height = maxY;
                rfm_thread_expose_rect (gtk_widget_get_window(view_p->widgets.paper), &rect);
                lastY = maxY; */
            }
        }
        // if destroy event called from above, then:
        if(view_p->flags.status == STATUS_EXIT)
            return NULL;

    }
    //if(view_p->flags.type == DESKVIEW_TYPE)
      //  rodent_recalc_population_geometry (view_p);
    TRACE ("view_p->xfdir_p->pathc=%ld\n", (long)view_p->xfdir_p->pathc);

    view_p->eyecandy.progress_count = view_p->xfdir_p->pathc;
    TRACE ("memory in population structures=%ld\n", (long)sizeof (population_t) * view_p->xfdir_p->pathc);

    // Since no expose gets done until write lock released, lets do
    // a full expose here (lastY == 0)

    // for iconview, this will expose remaining area,
    // for deskview it will expose all.
    GdkRectangle rect;
    rect.x = 0;
    rect.y = lastY;
    rect.width = view_p->widgets.window->allocation.width;
    rect.height = view_p->widgets.window->allocation.height;
    rfm_thread_expose_rect (gtk_widget_get_window(view_p->widgets.paper), &rect);

    return NULL;
}


static
GdkPixbuf *get_population_pixbuf(view_t * view_p,
	population_t *population_p,
	const gchar * icon_id,
	gint element){
    if (!population_p){
	g_warning("get_population_pixbuf() population == NULL!\n");
	return rfm_get_pixbuf ("xffm/emote_cool", REDUCED_ICONSIZE(view_p));
    }

    if (!population_p->en) {
	return rfm_get_pixbuf ("xffm/device_computer", REDUCED_ICONSIZE(view_p));
    }

    GdkPixbuf *pixbuf = NULL;
    if(icon_id) {
        TRACE ("get_population_pixbuf: rfm_get_pixbuf(%s)\n", icon_id);
        pixbuf = rfm_get_pixbuf (icon_id, REDUCED_ICONSIZE(view_p));
	if (pixbuf) {
	    return pixbuf;
	}
	TRACE("cannot get pixbuf for icon=%s\n", icon_id);
    } 
    
    TRACE("+ now resolving icon for path %s\n", population_p->en->path);

#if 10
    // Default is to always show thumbnails.
    // short circuit if image is already hashed
    //
    // this is broken because preview creation thread is started
    // for all fscking files...
    //
    if(population_p->en->path && 
	    rfm_g_file_test(population_p->en->path, G_FILE_TEST_EXISTS) ) {
	// short circuit if image is already hashed
	pixbuf = rfm_get_from_pixbuf_hash (population_p->en->path, 
		REDUCED_ICONSIZE(view_p));
	if(pixbuf && rfm_is_image(population_p->en)) {
	    population_p->flags |= POPULATION_THUMBNAILED;
	    TRACE("Got %s from pixbuf hash!\n", population_p->en->path);
	    return pixbuf;
	}
    }
#endif
    // image is not hashed or not an image
    // here we ge the default icon, to be updated to
    // a preview by the rfm_preview_thread_f if option is set
    // (this is the default).
    TRACE ("HASH-population.c: %s not found in pixbuf hash\n", 
	    population_p->en->path);
    if(IS_UP_TYPE (population_p->en->type)) {
	const gchar *id = rfm_resolve_icon_id (population_p->en);
	    pixbuf = rfm_get_pixbuf (id, REDUCED_ICONSIZE(view_p));
	return pixbuf;
    }

    const gchar *id=NULL;
    gboolean fullmime=FALSE;
     
    // if count is less than one age, full mime is not a significant
    // disadvantage... modules also default to full behaviour
    if (view_p->module || element <=104 || view_p->flags.type == DESKVIEW_TYPE){
	fullmime=TRUE;
    }
    // plain icon id which will be updated by 
    // fireup preview manager 
    // rfm_get_plain_icon_id takes care of all 
    // types except regular and links, because these are
    // further extended by mimetype determination.
    if (fullmime) {
	//stat(population_p->en->path,population_p->en->st);
	id=rfm_resolve_icon_id (population_p->en);
	gchar *composite_id = 
	    rfm_get_composite_icon_id (population_p->en, id);
	TRACE("%s icon for %s (%s %s) 0x%x\n", 
		id, population_p->en->path,
		population_p->en->mimetype, population_p->en->mimemagic,
		GPOINTER_TO_INT(population_p->en->st));

	pixbuf =
	    rfm_get_pixbuf (composite_id, REDUCED_ICONSIZE(view_p));
	g_free(composite_id);
	// Fallback to plain icon if composite (which includes plain
	//  icon names) fails.
	if (!pixbuf) {
	    pixbuf =
		rfm_get_pixbuf (id, REDUCED_ICONSIZE(view_p));
	}

    } else {
	//if (population_p->en) TRACE("%d plain icon for %s\n", element, population_p->en->path);
	id = MIME_hashtype(population_p->en->path);
	if (!id) {
	    id=rfm_get_plain_icon_id (population_p->en);
	}
	pixbuf = rfm_get_pixbuf (id, REDUCED_ICONSIZE(view_p));
	population_p->flags |= POPULATION_SHORT_ICON;
    }
    return pixbuf;
}


// This function sets the iconsize scale in gridview windows to the 
// size saved for the particular directory in the view.
static void
set_icon_size_scale(view_t *view_p){
    if (view_p->widgets.size_scale == NULL){
	return;
    }
    gtk_range_set_value (GTK_RANGE(view_p->widgets.size_scale), view_p->view_layout.icon_size);

    gchar *size_string=g_strdup_printf("%d", view_p->view_layout.icon_size);
    gchar *format;
    switch (view_p->view_layout.icon_size) {
	case BIG_ICON_SIZE:
	    format =g_strdup_printf(_("Huge (%sx%s)"),
			size_string, size_string);
	    break;
	case MEDIUM_ICON_SIZE:
	    format =g_strdup_printf(_("Large (%sx%s)"),
			size_string, size_string);
	    break;
	case SMALL_ICON_SIZE:
	    format =g_strdup_printf(_("Medium (%sx%s)"),
			size_string, size_string);
	    break;
	case TINY_ICON_SIZE:
	    format =g_strdup_printf(_("Small (%sx%s)"),
			size_string, size_string);
	    break;
	case LIST_ICON_SIZE:
	    format =g_strdup_printf("%s (%dx%d)",
		    _("List View"), TINY_ICON_SIZE, TINY_ICON_SIZE);
	    break;
	default:
	    g_warning("non group icon size is dumb");
	    format=g_strdup_printf("%s %sx%s", _("Icon size:"), 
			size_string, size_string);
    }

    
    gchar *tip_text;
    if (view_p->view_layout.icon_size==SMALL_ICON_SIZE) {
	tip_text = g_strdup_printf(_("Icon size: %s pixels (standard size)"),
		format);
    } else {
	tip_text = g_strdup_printf(_("Icon size: %s pixels"),
		format);
    }
    
    gtk_widget_set_tooltip_text (view_p->widgets.size_scale,tip_text);
    DBG("set_icon_size_scale: %s\n", format);
    g_free(format);
    g_free(tip_text);
    g_free(size_string);
    return;
}

static void 
save_environment_desktop_dir(view_t *view_p, record_entry_t *target_en){
    // save the default desktop dir to the environment...
    if(view_p->flags.type != DESKVIEW_TYPE) {
	return;
    }
    g_free (view_p->desktop_dir);
    if (!target_en) {
	view_p->desktop_dir = g_strdup("");
    } else if (target_en->module) {
	if (rfm_void (PLUGIN_DIR, target_en->module, "module_name")) {
	    view_p->desktop_dir = g_strdup_printf ("module:%s",
		(gchar *)rfm_void (PLUGIN_DIR, target_en->module, "module_name"));
	} else {
	    view_p->desktop_dir = g_strdup (target_en->module);
	}
    } else {
	view_p->desktop_dir = g_strdup (target_en->path);
    }
    rfm_setenv ( "RFM_DESKTOP_DIR", view_p->desktop_dir, TRUE);
    // save configuration too
    rfm_rational (PLUGIN_DIR, "settings", (void *) "RFM_DESKTOP_DIR", (void *) view_p->desktop_dir, "mcs_set_var");
    TRACE(" setting RFM_DESKTOP_DIR to %s\n",view_p->desktop_dir);
    return;
}

// rodent_full_reload_view() is a hard reload:
// This implies that a user click or keyboard event triggered the
// reload. This function will be executed by the main thread (GTK
// controller thread), so this function is protected against recursive
// calls from the gtk-loop.


