//
//
/*****    dialog callbacks   *****/
// specify mode as shred or delete
//shred
//rm


static void
do_the_remove (subthread_t *subthread_p, GSList ** rm_list, gboolean unlink_mode) {
    widgets_t *widgets_p=subthread_p->widgets_p;
    gchar *sources[MAX_COMMAND_ARGS];
    gchar *shred_iterations = NULL;
    int i = 0;
    if(unlink_mode == MODE_SHRED) {
        sources[i++] = "shred";
        if(getenv ("RFM_SHRED_ITERATIONS") && strlen (getenv ("RFM_SHRED_ITERATIONS"))) {
            gboolean ok = TRUE;
            char *p;
            for(p = getenv ("RFM_SHRED_ITERATIONS"); *p; p++) {
                if(!isdigit (*p)) {
                    ok = FALSE;
                    break;
                }
            }
            if(ok) {
                shred_iterations = g_strdup_printf ("--iterations=%s", getenv ("RFM_SHRED_ITERATIONS"));
            }
            sources[i++] = shred_iterations;
        }
        if(getenv ("RFM_SHRED_ZERO") && strlen (getenv ("RFM_SHRED_ZERO"))) {
            sources[i++] = "--zero";
        }
        if(getenv ("RFM_SHRED_REMOVE") && strlen (getenv ("RFM_SHRED_REMOVE"))) {
            sources[i++] = "--remove";
        }
/*      sources[i++] = " -s, --size=N"; */
    } else {
        sources[i++] = "rm";
        sources[i++] = "-r";
        if(getenv ("RFM_RM_ONE_FILE_SYSTEM") && strlen (getenv ("RFM_RM_ONE_FILE_SYSTEM"))) {
            sources[i++] = "--one-file-system";
        }
    }

    // common options:
    sources[i++] = "-f";
    if(unlink_mode == MODE_RM &&
	    getenv ("RFM_RM_VERBOSE") && 
	    strlen (getenv ("RFM_RM_VERBOSE"))) {
        sources[i++] = "-v";
    }
    if(unlink_mode == MODE_SHRED && 
	    getenv ("RFM_SHRED_VERBOSE") && 
	    strlen (getenv ("RFM_SHRED_VERBOSE"))) {
        sources[i++] = "-v";
    }
    //shred:
    gboolean ok = FALSE;
    gboolean excess=FALSE;
    while(*rm_list && (*rm_list)->data) {
        gchar *path = (gchar *) (*rm_list)->data;
        if(path && (g_file_test (path, G_FILE_TEST_EXISTS) || g_file_test (path, G_FILE_TEST_IS_SYMLINK))) {
	    if (g_file_test (path, G_FILE_TEST_IS_SYMLINK)
		    && unlink_mode == MODE_SHRED) {
		// file is a symlink
		gchar *text=g_strconcat(_("Symbolic Link"),":\n", path, "\n", NULL);
		if (!rfm_confirm(widgets_p,  GTK_MESSAGE_WARNING,
		text,
		_("Don't follow symlinks"), _("Follow symlinks"))){
		    // skip file.
		    *rm_list = g_slist_remove (*rm_list, (*rm_list)->data);
		    g_free(text);
		    continue;
		}
		g_free(text);
	    }
            sources[i++] = path;
            *rm_list = g_slist_remove (*rm_list, (*rm_list)->data);
            if (i == MAX_COMMAND_ARGS - 1  && *rm_list){
                excess=TRUE;
                // we are at the last argv, for NULL
                // remaining items will not be processed.
                while (*rm_list) {*rm_list = g_slist_remove (*rm_list, (*rm_list)->data);}
                break;
            }
            ok = TRUE;
        } else {
            *rm_list = g_slist_remove (*rm_list, (*rm_list)->data);
        }
    };
    sources[i++] = NULL;

    // note: with run_argv we do not need to escape paths.
    if(ok) {
	// is view still valid?
	gboolean valid_widgets = 
	    (subthread_p->widgets.diagnostics && 
	     *(subthread_p->widgets.diagnostics)) ||
	    subthread_p->widgets.diagnostics_window;
	if (valid_widgets){
	    gchar *tgt= g_path_get_dirname (sources[i-2]);
	    if (rfm_write_ok_path(tgt)){
		rfm_show_text(widgets_p);
		rfm_thread_run_argv ((void *)widgets_p, sources, FALSE );
	    } else {
		gchar *failed;
	        if (unlink_mode == MODE_RM){
		    failed = g_strdup( _("Delete failed"));
		} else {
		    gchar *warning0=g_strdup_printf (_("Unexpected error: %s"), _("Shred"));
		    gchar *warning1=g_strdup_printf ("%s: %s",
			    _("Shred"),
			    _("Are you sure you want to continue?"));
		    failed = g_strconcat(warning0, "\n", warning1, "\n", NULL);
		    g_free(warning0);
		    g_free(warning1);
		}
		if (rfm_confirm_sudo(widgets_p, 
			    tgt, failed, 
			    (unlink_mode == MODE_RM)?"rm":"shred")){
		    rfm_show_text(widgets_p);
		    rfm_try_sudo (widgets_p, sources, FALSE);
		}
		g_free(failed);
	    }
	}
    } else {
        TRACE ("rodent_remove: ok == FALSE\n");
    }
    g_free (shred_iterations);
    if (excess){
        rfm_diagnostics(widgets_p, "xffm/stock_dialog-warning",NULL);
        gchar *max=g_strdup_printf("%d",MAX_COMMAND_ARGS);
        rfm_diagnostics (widgets_p, "xffm_tag/stderr", sources[0], ": ", strerror(E2BIG)," (> ",max,")","\n", NULL);
        g_free(max);
    }


}

static void
apply_action(GtkWidget * button, gpointer data){
    GtkWidget *dialog=g_object_get_data(G_OBJECT(button), "dialog");
    subthread_t *subthread_p=g_object_get_data(G_OBJECT(dialog), "subthread_p");
    gtk_widget_hide(dialog);
    if (!subthread_p){
	gtk_widget_destroy(dialog);
	return;
    }


    GtkWidget *togglebutton=g_object_get_data(G_OBJECT(dialog), "togglebutton");
    gboolean apply_to_all=gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(togglebutton));
    gint result=GPOINTER_TO_INT(data);

    if (result == RM_YES && apply_to_all) result=RM_YES_ALL;
    else if (result == SHRED_YES && apply_to_all) result=SHRED_YES_ALL;
    else if (result == RM_NO && apply_to_all) result=RM_CANCEL;

    // Check if widgets_p is still valid. if not cancel all.
    // This is only applicable to iconview which may have multiple
    // views
    if (subthread_p->widgets.diagnostics_window==NULL){
	if ((subthread_p->widgets.diagnostics 
		&& *(subthread_p->widgets.diagnostics)==NULL)){
	    result=RM_CANCEL;
	}
    }
    

    gint mode = MODE_RM;
    GSList *tmp;
    gchar *path;
    switch (result) {
	case SHRED_YES:
	    mode = MODE_SHRED;
	case RM_YES:
	    TRACE ("REMOVE: RM_YES  \n");
	    GSList *single_list = 
		g_slist_append (NULL, subthread_p->selection_list->data);
	    do_the_remove (subthread_p, &single_list, mode);
	    g_slist_free (single_list);
	    path = (gchar *) subthread_p->selection_list->data;
	    subthread_p->selection_list =
		g_slist_remove (subthread_p->selection_list, 
			subthread_p->selection_list->data);
	    TRACE("RM_YES free %s\n",path);
	    g_free(path);
	    break;
	case RM_NO:
	    TRACE ("REMOVE: RM_NO  \n");
	    path = (gchar *) subthread_p->selection_list->data;
	    subthread_p->selection_list = 
		g_slist_remove (subthread_p->selection_list, 
			subthread_p->selection_list->data);
	    TRACE("RM_NO free %s\n",path);
	    g_free(path);
	    break;
	case SHRED_YES_ALL:
	    mode = MODE_SHRED;
	case RM_YES_ALL:
	    TRACE ("REMOVE: RM_YES_ALL  \n");
	    do_the_remove (subthread_p, &(subthread_p->selection_list), mode);
	    break;
	case RM_CANCEL:
	default:
	    tmp=subthread_p->selection_list;
	    for (; tmp && tmp->data; tmp=tmp->next){
		TRACE("RM_CANCEL free %s\n",(gchar *)tmp->data);
		g_free(tmp->data);
	    }
	    g_slist_free(subthread_p->selection_list);
	    subthread_p->selection_list=NULL;
	    break;
    }
    gtk_widget_destroy(dialog);

    // Regardless if we have or do not have any items in
    // the selection list to process, we will call the rodent_remove
    // thread. This thread will determine if it is time to cleanup
    // or to pop up the next dialog.
    g_thread_create(rodent_remove, subthread_p, FALSE, NULL);
}


static gboolean
on_destroy_event (GtkWidget * dialog, GdkEvent * event, gpointer data) {
    GtkWidget *button=g_object_get_data(G_OBJECT(dialog), "cancelbutton");
    apply_action(button, GINT_TO_POINTER(RM_CANCEL));
    return TRUE;
}



static void
make_dialog (subthread_t *subthread_p) 
{
    TRACE("make_dialog\n");
    gchar *path = (gchar *) subthread_p->selection_list->data;
    gchar *message;
    gchar *text;
    {
	text=_("Remove");
	gchar *b = g_path_get_basename (path);
	gchar *q = rfm_utf_string (rfm_chop_excess (b));
	struct stat st;
	stat(path, &st);
	gchar *s1 = rfm_time_to_string (st.st_mtime);
	gchar *s2 = rfm_sizetag ((off_t) st.st_size, -1);
	message = g_strconcat (q, "\n", "(", s1, " ", s2, ")", NULL);
	g_free (b);
	g_free (q);
	g_free (s1);
	g_free (s2);
    }

GDK_THREADS_ENTER();
    GtkWidget *dialog = rodent_create_remove (subthread_p->widgets_p, text, message);
    if(subthread_p->widgets_p->view_p->flags.type == DESKVIEW_TYPE) {
        gtk_window_set_keep_above (GTK_WINDOW(dialog), TRUE);
        gtk_window_stick (GTK_WINDOW(dialog));
    } else {
	// we do not want dialog to be gtk style transient,
	// that is so much for dummies...
	//gtk_window_set_transient_for(GTK_WINDOW(dialog), 
	//	GTK_WINDOW(subthread_p->widgets_p->window));
    }
    
    g_object_set_data(G_OBJECT(dialog), "subthread_p", subthread_p);
    g_free(message);
    /* dialog specifics */
    GtkWidget *togglebutton=g_object_get_data(G_OBJECT(dialog), "togglebutton");

    if(g_slist_length (subthread_p->selection_list) < 2) {
	gtk_widget_hide(togglebutton);
    }
GDK_THREADS_LEAVE();

    TRACE ("REMOVE: dialog done\n");
}

/**************/

