//
//
/*
 * Copyright (C) 2003-2011 Edscott Wilson Garcia
 * EMail: edscott@xfce.org
 *
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

static char *ftypes[] = {
    N_("Any"),
    N_("Regular"),
    N_("Directory"),
    N_("Symbolic Link"),
    N_("Socket"),
    N_("Block device"),
    N_("Character device"),
    N_("FIFO"),
    NULL
};

static char *ft[] = {
    "any",
    "reg",
    "dir",
    "sym",
    "sock",
    "blk",
    "chr",
    "fifo",
    NULL
};


static gint result_limit_counter;
// These externals are in find-module_gui.c
extern gint result_limit;
extern gboolean default_recursive;
extern gboolean default_recursiveH;
extern gboolean default_xdev;
extern gboolean default_case_sensitive;
extern gboolean default_ext_regexp;
extern gboolean default_look_in_binaries;
extern gint default_type_index;
extern gboolean default_anywhere;
extern gboolean default_match_words;
extern gboolean default_match_lines;
extern gboolean default_match_no_match;
extern gboolean have_grep;
extern gboolean have_gnu_grep;

static gchar *
#if GTK_MAJOR_VERSION==2 && GTK_MINOR_VERSION<24
get_combo_entry (void* combo_info, GtkComboBoxEntry *combobox);
#else
get_combo_entry (void* combo_info, GtkComboBox *combobox);
#endif

static void
grep_options (GtkComboBox *widget, gpointer data) {
    find_struct_t *find_struct_p = data;
    gchar *s = get_combo_entry (find_struct_p->findgrep_combo_info, find_struct_p->combogrep);
    gboolean state=FALSE;
    if (s && strlen(s)) {
	// sensitive
	state=TRUE;
    }   
    g_free(s);
    gtk_widget_set_sensitive (
	    GTK_WIDGET(g_object_get_data(
		    G_OBJECT(find_struct_p->dialog), "case_sensitive")), state);
    if (have_gnu_grep) 
    {
	gtk_widget_set_sensitive ( 
	    GTK_WIDGET(g_object_get_data(
		    G_OBJECT(find_struct_p->dialog), "ext_regexp")), state);
	gtk_widget_set_sensitive ( 
	    GTK_WIDGET(g_object_get_data(
		    G_OBJECT(find_struct_p->dialog), "look_in_binaries")), state);
    }
    gtk_widget_set_sensitive ( 
	    GTK_WIDGET(g_object_get_data(
		    G_OBJECT(find_struct_p->dialog), "label40")), state);
    gtk_widget_set_sensitive (
	    GTK_WIDGET(g_object_get_data(
		    G_OBJECT(find_struct_p->dialog), "anywhere")), state);
    gtk_widget_set_sensitive (
	    GTK_WIDGET(g_object_get_data(
		    G_OBJECT(find_struct_p->dialog), "match_words")), state);
    gtk_widget_set_sensitive ( 
	    GTK_WIDGET(g_object_get_data(
		    G_OBJECT(find_struct_p->dialog), "match_lines")), state);
    gtk_widget_set_sensitive ( 
	    GTK_WIDGET(g_object_get_data(
		    G_OBJECT(find_struct_p->dialog), "match_no_match")), state);

}

// thread function
static
void
stdout_f (void *user_data, void *stream, int childFD) {
    widgets_t *widgets_p = user_data;
    char *line;
    line = (char *)stream;
    TRACE ("FORK stdout: %s\n", line);

    if(line[0] == '\n') return;

    if (result_limit > 0 && result_limit==result_limit_counter) {
	gchar *g=g_strdup_printf("%s. %s %d", _("Results"), _("Upper limit:"), result_limit);
        GDK_THREADS_ENTER ();
        rfm_diagnostics (widgets_p, "xffm/stock_dialog-warning", NULL);
	rfm_diagnostics (widgets_p, "xffm_tag/green", g, "\n", NULL);
        GDK_THREADS_LEAVE ();
	g_free(g);
    }

    if(strncmp (line, "fgr search complete!", strlen ("fgr search complete!")) == 0) {
  	const gchar *tag;
	if (result_limit > 0 && result_limit < result_limit_counter) {
	    tag="xffm_tag/stderr";
	} else {
	    tag="xffm_tag/green";
	}
	gchar *g=g_strdup_printf(" (%s %d)", _("Maximum:"), result_limit_counter);
        GDK_THREADS_ENTER ();
        rfm_diagnostics (widgets_p, "xffm/stock_find", NULL);
        rfm_diagnostics (widgets_p, "xffm_tag/green", _("Search Complete"), NULL);
        rfm_diagnostics (widgets_p, tag, g, NULL);
        rfm_diagnostics (widgets_p, NULL, "\n", NULL);
        GDK_THREADS_LEAVE ();
	g_free(g);
    } else if(strncmp (line, "Tubo-id exit:", strlen ("Tubo-id exit:")) == 0) {
            if(strchr (line, '\n')) *strchr (line, '\n') = 0;
GDK_THREADS_ENTER ();
            rfm_diagnostics (widgets_p, "xffm/stock_stop", NULL);
            rfm_diagnostics (widgets_p, "xffm_tag/command_id", strchr (line, ':') + 1, ".", NULL);
            rfm_diagnostics (widgets_p, NULL, "\n", NULL);
GDK_THREADS_LEAVE ();
    } else {
	result_limit_counter++; 
	if (result_limit ==0 ||
	    (result_limit > 0 && result_limit >= result_limit_counter) ) {
        GDK_THREADS_ENTER ();
	    rfm_diagnostics (widgets_p, NULL, line, NULL);
        GDK_THREADS_LEAVE ();
	}
    }
    return;
}



static void
path_filechooser (GtkButton * button, gpointer user_data) {
    filechooser_t *filechooser_p = (filechooser_t *) user_data;
    const gchar *text;
    if(filechooser_p->filechooser_action == GTK_FILE_CHOOSER_ACTION_OPEN) {
        text = _("Select Files...");
    } else if(filechooser_p->filechooser_action == GTK_FILE_CHOOSER_ACTION_SELECT_FOLDER) {
        text = _("Select folder to search in");
    } else {
        text = "FIXME";
    }
    TRACE ("FIND: parent=0x%lu action=%d, text=%s\n", (long unsigned)filechooser_p->parent, filechooser_p->filechooser_action, text);
    GtkWidget *dialog = gtk_file_chooser_dialog_new (text,
                                                     GTK_WINDOW (filechooser_p->parent),
                                                     filechooser_p->filechooser_action,
                                                     GTK_STOCK_CANCEL,
                                                     GTK_RESPONSE_CANCEL,
                                                     GTK_STOCK_OPEN,
                                                     GTK_RESPONSE_ACCEPT,
                                                     NULL);
    gtk_file_chooser_set_action ((GtkFileChooser *) dialog, filechooser_p->filechooser_action);
    if(filechooser_p->folder) {
        gtk_file_chooser_set_current_folder ((GtkFileChooser *) dialog, filechooser_p->folder);
    }

    if(gtk_dialog_run (GTK_DIALOG (dialog)) == GTK_RESPONSE_ACCEPT) {
        char *filename;
        void *combo_info = filechooser_p->combo_info;
        filename = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (dialog));
	if (rfm_void(PLUGIN_DIR, "combobox", "module_active")){
	    COMBOBOX_set_entry (combo_info, filename);
	} else {
	    gtk_entry_set_text (filechooser_p->entry, filename);
	}
        TRACE ("Got %s\n", filename);
        g_free (filename);
    }
    gtk_widget_destroy (dialog);

}

static void
save_ff_text (char *p) {
    gchar *fname;
    if(!p || !strlen (p))
        return;
    fname = g_build_filename (g_get_user_data_dir (), FIND_DBH_FILE, NULL);
    COMBOBOX_save_to_history (fname, p);
    g_free (fname);
}

static void
save_fpath_text (const char *p) {
    gchar *fname;
    TRACE ("save_fpath_text-> %s\n", p);
    if(!p || !strlen (p))
        return;
    {
        fname = g_build_filename (g_get_user_data_dir (), FIND_PATHS_DBH_FILE, NULL);
    }
    TRACE ("saving to history-> %s\n", p);
    COMBOBOX_save_to_history (fname, p);
    g_free (fname);
}

static void
save_fgrep_text (char *p) {
    gchar *fname;
    if(!p || !strlen (p))
        return;
    fname = g_build_filename (g_get_user_data_dir (), FIND_GREP_DBH_FILE, NULL);
    COMBOBOX_save_to_history (fname, p);
    g_free (fname);
}

static void
destroy_find (GtkWidget * dialog, gpointer data) {
    find_struct_t *find_struct_p = g_object_get_data(G_OBJECT(dialog), "find_struct_p");
    if (!find_struct_p) return;
    gtk_widget_hide (dialog);
    COMBOBOX_destroy_combo(find_struct_p->findpath_combo_info);
    COMBOBOX_destroy_combo(find_struct_p->find_combo_info);
    COMBOBOX_destroy_combo(find_struct_p->findgrep_combo_info);
    g_free(find_struct_p);
    find_struct_p = NULL;
    g_object_set_data(G_OBJECT(dialog),"find_struct_p", NULL);
    widgets_t *widgets_p=g_object_get_data(G_OBJECT(dialog), "widgets_p");
    if (widgets_p) {
        g_free(widgets_p->diagnostics);
	g_free(widgets_p);
    }
    g_object_set_data(G_OBJECT(dialog), "widgets_p", NULL);
    gboolean standalone = GPOINTER_TO_INT(g_object_get_data(G_OBJECT(dialog), "standalone"));
    gtk_widget_destroy (dialog);
    if (standalone) gtk_main_quit();

}

static void
on_find_close (GtkButton * button, gpointer data) {
    GtkWidget *dialog=g_object_get_data(G_OBJECT(button), "dialog");
    destroy_find (dialog, data);
}


static gchar *
#if GTK_MAJOR_VERSION==2 && GTK_MINOR_VERSION<24
get_combo_entry (void* combo_info, GtkComboBoxEntry *combobox) {
#else
get_combo_entry (void* combo_info, GtkComboBox *combobox) {
#endif
    const gchar *choice;
    gchar *s = NULL;

    if (rfm_void(PLUGIN_DIR, "combobox", "module_active")){
	choice = (gchar *) COMBOBOX_get_entry (combo_info);
    } else {
	choice = gtk_entry_get_text (GTK_ENTRY(GTK_BIN(combobox)->child));
    }
    /*TRACE("choice== %s\n", choice);
       TRACE("choice== %s\n", choice); */

    if(choice) {
        s = g_locale_from_utf8 (choice, -1, NULL, NULL, NULL);
    }
    if(s)
        s = g_strchug (s);
    if(s)
        s = g_strchomp (s);
    if(!s)
        s = g_strdup ("");

    return s;
}

static void
on_find_clicked_action (GtkButton * button, gpointer data, gboolean destroy) {
    short int i,
      j;
    gchar *s;
    gchar *argument[MAX_COMMAND_ARGS];
    gchar *find_filter = NULL,
        *filter = NULL,
        *path = NULL,
        *token = NULL;
    find_struct_t *find_struct_p = data;

    GtkWidget *dialog=g_object_get_data(G_OBJECT(button), "dialog");
    widgets_t *widgets_p;
    // if using a notebook (i.e., ICONVIEW_TYPE), output to the page
    if (destroy && find_struct_p->notebook && *find_struct_p->notebook) {
	    gint current_page = gtk_notebook_get_current_page (
		    GTK_NOTEBOOK(*(find_struct_p->notebook)));	
	    GtkWidget *child =
		gtk_notebook_get_nth_page(
		    GTK_NOTEBOOK(*(find_struct_p->notebook)), current_page);
	    widgets_p=g_object_get_data(G_OBJECT(child), "widgets_p");
    } else {
	// output to dialog diagnostics (no window or DESKVIEW_TYPE or nonclose)
	widgets_p=g_object_get_data(G_OBJECT(dialog), "widgets_p");
    }


    if(!widgets_p) {
        g_warning ("on_find_clicked: !widgets_p");
    }

    if(GTK_IS_WIDGET(widgets_p->window) && !gtk_widget_get_visible (widgets_p->window)) {
        gtk_widget_show (widgets_p->window);
    }
    rfm_show_text (widgets_p);
    if (widgets_p->diagnostics_window && *(widgets_p->diagnostics_window) 
	    && GTK_IS_WIDGET(widgets_p->diagnostics_window)  
	    ) {
	gtk_window_deiconify(GTK_WINDOW(*(widgets_p->diagnostics_window)));
    }

    /* get options */

/* get the parameters set by the user... *****/

/* limit */
    result_limit = gtk_spin_button_get_value_as_int (
	    GTK_SPIN_BUTTON (g_object_get_data(G_OBJECT(dialog), "spinbutton1")));
    result_limit_counter=0;

/* the rest */

    s = get_combo_entry (find_struct_p->findpath_combo_info, find_struct_p->combopath);
    if(!s || strlen (s) == 0 || !rfm_g_file_test (s, G_FILE_TEST_EXISTS | G_FILE_TEST_IS_DIR)) {
        rfm_diagnostics (widgets_p, "xffm/stock_dialog-error", strerror (ENOENT), ": ", s, NULL);
        rfm_diagnostics (widgets_p, NULL, "\n", NULL);
	g_free(s);
        return;
    }

    /* tilde expansion */
    else if(s[strlen (s) - 1] == '~'){
	g_free(s);
        s = g_strdup("~/");
    }
    /* environment variables */
    else if(s[0] == '$') {
        const gchar *p = getenv (s + 1);
        if(p){
	    g_free(s);
            s = g_strdup(p);
	}else {
	    g_free(s);
            s = g_strdup("/");
	}
    }

    save_fpath_text (s);
	    
    g_free (widgets_p->workdir);
    widgets_p->workdir=g_strdup (s);
    path = g_strdup (".");
    g_free(s);

    s = get_combo_entry (find_struct_p->find_combo_info, find_struct_p->combo);
    if(s && strlen (s)) {
        filter = g_strdup (s);
        save_ff_text (filter);
    }
    g_free(s);

    s = get_combo_entry (find_struct_p->findgrep_combo_info, find_struct_p->combogrep);

    if(s && strlen (s)) {
        token = g_strdup (s);
        save_fgrep_text (token);
    } else {
        /* if no grep pattern given, its a plain find and
         * we should show the diagnostics */
    }
    g_free(s);

    /* select list */
#if GTK_MAJOR_VERSION==2 && GTK_MINOR_VERSION<24
    s = gtk_combo_box_get_active_text ((GtkComboBox *)
	    g_object_get_data(G_OBJECT(dialog), "file_type_om"));
#else
    s = gtk_combo_box_text_get_active_text (GTK_COMBO_BOX_TEXT(
	    g_object_get_data(G_OBJECT(dialog), "file_type_om")));
#endif
    default_type_index=gtk_combo_box_get_active((GtkComboBox *)
	    g_object_get_data(G_OBJECT(dialog), "file_type_om"));

    i = 0;
    {
        argument[i++] = "fgr";

        /*argument[i++] = "-v"; (verbose output from glob for debugging) */
        /*argument[i++] = "-P"; */

        if(gtk_toggle_button_get_active ((GtkToggleButton *) g_object_get_data(G_OBJECT(dialog), "recursive"))) {
	    default_recursive=TRUE;
            if(gtk_toggle_button_get_active ((GtkToggleButton *) g_object_get_data(G_OBJECT(dialog), "recursiveH"))) {
                argument[i++] = "-r";
		default_recursiveH=TRUE;
            } else {
                argument[i++] = "-R";
		default_recursiveH=FALSE;
            }
        } else {
	    default_recursive=FALSE;
	}
        if(!gtk_toggle_button_get_active ((GtkToggleButton *) g_object_get_data(G_OBJECT(dialog), "case_sensitive"))){
            argument[i++] = "-i";
	    default_case_sensitive=FALSE;
	} else {
	    default_case_sensitive=TRUE;
	}
        if(gtk_toggle_button_get_active ((GtkToggleButton *) g_object_get_data(G_OBJECT(dialog), "xdev"))){
            argument[i++] = "-a";
	    default_xdev=TRUE;
	} else {
	    default_xdev=FALSE;
	}
        if(token) {
            if(gtk_toggle_button_get_active ((GtkToggleButton *) g_object_get_data(G_OBJECT(dialog), "ext_regexp"))){
                argument[i++] = "-E";
		default_ext_regexp=TRUE;
	    } else {
                argument[i++] = "-e";
		default_ext_regexp=FALSE;
	    }
            argument[i++] = token;

            /* options for grep: ***** */
            if(gtk_toggle_button_get_active ((GtkToggleButton *) g_object_get_data(G_OBJECT(dialog), "look_in_binaries"))){
		default_look_in_binaries=TRUE;
	    } else {
                argument[i++] = "-I";
		default_look_in_binaries=FALSE;
	    }
	    default_anywhere=default_match_words=default_match_lines=default_match_no_match=FALSE;
            if(gtk_toggle_button_get_active ((GtkToggleButton *) g_object_get_data(G_OBJECT(dialog), "anywhere"))) {
		default_anywhere=TRUE;
	    }
            else if(gtk_toggle_button_get_active ((GtkToggleButton *) g_object_get_data(G_OBJECT(dialog), "match_words"))){
                argument[i++] = "-w";
		default_match_words=TRUE;
	    }
            else if(gtk_toggle_button_get_active ((GtkToggleButton *) g_object_get_data(G_OBJECT(dialog), "match_lines"))){
                argument[i++] = "-x";
		default_match_lines=TRUE;
	    }
            else if(gtk_toggle_button_get_active ((GtkToggleButton *) g_object_get_data(G_OBJECT(dialog), ""))){
                argument[i++] = "-L";
		default_match_no_match=TRUE;
	    }
        }

        for(j = 0; ftypes[j] != NULL; j++) {
            if(s && strcmp (s, _(ftypes[j])) == 0) {
                argument[i++] = "-t";
                argument[i++] = ft[j];
                break;
            }
        }

        /* apply wildcard filter if not specified (with dotfiles option) */
        argument[i++] = "-f";
        if(filter && strlen (filter)) {
            argument[i++] = filter;
        } else {
            argument[i++] = "*";
            argument[i++] = "-D";
        }
        argument[i++] = "-v";

        /* last argument is the path (which could be NULL) */
        argument[i++] = path;
    }

    argument[i] = (char *)0;
    rfm_diagnostics (widgets_p, "xffm/apps_search", NULL);
    rfm_diagnostics (widgets_p, "xffm_tag/green", _("Searching..."), "\n", NULL);

//    rfm_thread_run_argv (widgets_p, argument, FALSE);
    rfm_thread_run_argv_with_stdout (widgets_p, argument, FALSE, stdout_f);

    TRACE ("FGR: --> ");
    rfm_show_text (widgets_p);
    for(j = 0; j < i; j++) {
        TRACE (" %s", argument[j]);
    }
    TRACE ("\n");
    if (destroy) {
	gtk_widget_destroy (dialog);
	// exit (1);
    }
    g_free(s);
    g_free (token);
    g_free (filter);
    g_free (find_filter);
    g_free (path);
    rfm_status (widgets_p, "xffm/stock_dialog-info", _("Search results"), NULL);

}

static void
on_find_clicked (GtkButton * button, gpointer data) {
    on_find_clicked_action (button, data, TRUE);
}

static void
on_apply_clicked (GtkButton * button, gpointer data) {
    on_find_clicked_action (button, data, FALSE);
}

static void
on_find_clicked_wrapper (GtkEntry * entry, gpointer data) {
    on_find_clicked_action ((GtkButton *) entry, data, TRUE);
}

