/*
 * Copyright (C) 2003-2011 Edscott Wilson Garcia
 * EMail: edscott@xfce.org
 *
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include "rodent_libs.h"

#include "find-module_gui.h"
#include "find-module_callbacks.h"

// default values:
gint result_limit=256;
gboolean default_recursive=TRUE;
gboolean default_recursiveH=FALSE;
gboolean default_xdev=TRUE;
gboolean default_case_sensitive=FALSE;
gboolean default_ext_regexp=FALSE;
gboolean default_look_in_binaries=FALSE;
gint default_type_index=0;
gboolean default_anywhere=TRUE;
gboolean default_match_words=FALSE;
gboolean default_match_lines=FALSE;
gboolean default_match_no_match=FALSE;
    
gboolean have_grep = FALSE;
gboolean have_gnu_grep = FALSE;

static const gchar *
filter_text_help=
	        N_("Basic rules:\n" "\n"
                          "*  Will match any character zero or more times.\n"
                          "?  Will match any character exactly one time\n"
                          "[] Match any character within the [] \n"
                          "^  Match at beginning of string\n" 
			  "$  Match at end of string \n");
static const gchar *
grep_text_help=
		N_("Reserved characters for extended regexp are\n"
                          ". ^ $ [ ] ? * + { } | \\ ( ) : \n"
                          "In  basic regular expressions the metacharacters\n"
                          "?, +, {, |, (, and ) lose their special meaning.\n"
                          "\n"
                          "The  period  .   matches  any  single  character.\n"
                          "The caret ^ matches at the start of line.\n"
                          "The dollar $ matches at the end of line.\n" "\n"
                          "Characters within [ ] matches any single \n"
                          "       character in the list.\n"
                          "Characters within [^ ] matches any single\n"
                          "       character *not* in the list.\n"
                          "Characters inside [ - ] matches a range of\n"
                          "       characters (ie [0-9] or [a-z]).\n" "\n"
                          "A regular expression may be followed by one\n"
                          "       of several repetition operators:\n"
                          "?      The preceding item is optional and matched\n"
                          "       at most once.\n"
                          "*      The preceding item will be matched zero\n"
                          "       or more times.\n"
                          "+      The preceding item will be matched one or\n"
                          "       more times.\n"
                          "{n}    The preceding item is matched exactly n times.\n"
                          "{n,}   The preceding item is matched n or more times.\n"
                          "{n,m}  The preceding item is matched at least n times,\n"
                          "       but not more than m times.\n" "\n"
                          "To match any reserved character, precede it with \\. \n"
                          "\n"
                          "Two regular expressions may be joined by the logical or\n"
                          "       operator |.\n"
                          "Two regular expressions may be concatenated.\n" "\n"
                          "More information is available by typing \"man grep\"\n"
			  );

static void
on_help_filter (GtkToggleButton * button, gpointer data) {
    GtkWidget *dialog=data;
    widgets_t *widgets_p=g_object_get_data(G_OBJECT(dialog), "widgets_p");
    if (gtk_toggle_button_get_active (button)) {
	rfm_show_text (widgets_p); 
	rfm_diagnostics(widgets_p,"xffm_tag/blue", _(filter_text_help), NULL);
    } else {
	rfm_clear_text(widgets_p);
    }
}

static void
on_help_grep (GtkToggleButton * button, gpointer data) {
    GtkWidget *dialog=data;
    widgets_t *widgets_p=g_object_get_data(G_OBJECT(dialog), "widgets_p");
    if (gtk_toggle_button_get_active (button)) {
	rfm_show_text (widgets_p); 
	rfm_diagnostics(widgets_p,"xffm_tag/blue", _(grep_text_help), NULL);
    } else {
	rfm_clear_text(widgets_p);
    }
}

GtkWidget *
fill_string_option_menu (GtkComboBox *om, GSList * strings) {
    if (!om || ! strings){
	g_warning("create_string_option_menu: GtkComboBox==NULL || GList==NULL\n");
	return NULL;
    }
    GSList *p;
    for (p=strings; p && p->data; p=p->next) {
#if GTK_MAJOR_VERSION==2 && GTK_MINOR_VERSION<24
    // this is deprecated...
	gtk_combo_box_append_text (om, (const gchar *)p->data);
#else
        gtk_combo_box_text_append_text(GTK_COMBO_BOX_TEXT(om), (const gchar *)p->data);
#endif
      /* Note: The model will keep a copy of the string internally, 
       * so the list may be freed */
    }
    gtk_combo_box_set_active(om, default_type_index);
    return (GtkWidget *)om;
}

static void
command_help (GtkWidget * button, gpointer data) {
    GtkWidget *dialog=g_object_get_data(G_OBJECT(button), "dialog");
    widgets_t *widgets_p=g_object_get_data(G_OBJECT(dialog), "widgets_p");
    gchar *argv[]={data, "--help", NULL};
    rfm_show_text (widgets_p); 
    rfm_thread_run_argv_with_stderr (widgets_p, argv, FALSE, rfm_dump_output);
}

static void
rtfm (GtkWidget * button, gpointer data) {
    GtkWidget *dialog=g_object_get_data(G_OBJECT(button), "dialog");
    widgets_t *widgets_p=g_object_get_data(G_OBJECT(dialog), "widgets_p");
    gchar *argv[]={"man", data, NULL};
    rfm_show_text (widgets_p); 
    rfm_thread_run_argv_with_stderr (widgets_p, argv, FALSE, rfm_dump_output);
}

GtkWidget *
create_find_dialog (widgets_t * in_widgets_p) {
    GtkWidget *vbox6;
    GtkWidget *label42;
    GtkWidget *hbox25;
    GtkWidget *hbox14;
    GtkWidget *vbox7a;
    GtkWidget *hbox15;
    GtkWidget *label26;
    GtkWidget *path_combo;
    GtkWidget *hbox16;
    GtkWidget *label27;
    GtkWidget *filter_combo;
    GtkWidget *togglebutton2;
    GtkWidget *image2;
    GtkWidget *hbox17;
    GtkWidget *vbox12;
    GtkWidget *recursive;
    GtkWidget *recursiveH;
    GtkWidget *xdev;
    GtkWidget *hbox18;
    GtkObject *spinbutton1_adj;
    GtkWidget *spinbutton1;
    GtkWidget *label30;
    GtkWidget *label29;
    GtkWidget *label45;
    GtkWidget *hbox26;
    GtkWidget *vbox8;
    GtkWidget *hbox19;
    GtkWidget *label33;
    GtkWidget *grep_combo;
    GtkWidget *hbox20;
    GtkWidget *vbox13;
    GtkWidget *case_sensitive;
    GtkWidget *ext_regexp;
    GtkWidget *togglebutton3;
    GtkWidget *image1;
    GtkWidget *hbox27;
    GtkWidget *vbox10;
    GtkWidget *hbox21;
    GtkWidget *label37;
    GtkWidget *file_type_om;
    GtkWidget *hbox28;
    GtkWidget *vbox11;
    GtkWidget *look_in_binaries;
    GtkWidget *hbox24;
    GtkWidget *label40;
    GtkWidget *anywhere;
    GSList *anywhere_group = NULL;
    GtkWidget *match_words;
    GtkWidget *match_lines;
    GtkWidget *match_no_match;
    GtkWidget *label41;
    GtkWidget *hbuttonbox2;
    GtkWidget *apply_button;
    GtkWidget *find_button;
    GtkWidget *close_button;
    GtkAccelGroup *accel_group;
    GdkPixbuf *pb;
    gchar *grep;

    grep = g_find_program_in_path ("grep");
    if(grep) {
        FILE *pipe;
        const gchar *cmd = "grep --version";
        have_grep = TRUE;
        pipe = popen (cmd, "r");
        if(pipe) {
            gchar line[256];
            memset (line, 0, 256);
            if(fgets (line, 255, pipe) == NULL)
                g_warning ("fgets: %s", strerror (errno));
            pclose (pipe);
            if(strstr (line, "GNU"))
                have_gnu_grep = TRUE;
        }
    }
    g_free (grep);

    accel_group = gtk_accel_group_new ();

    GtkWidget *dialog = gtk_window_new (GTK_WINDOW_TOPLEVEL);

    widgets_t *widgets_p;
    widgets_p =(widgets_t *)malloc(sizeof(widgets_t));
    memset(widgets_p, 0, sizeof(widgets_t));
    g_object_set_data(G_OBJECT(dialog), "widgets_p", widgets_p);
    if (in_widgets_p) widgets_p->window = in_widgets_p->window;


    gtk_window_set_title (GTK_WINDOW (dialog), _("Find"));
    gtk_window_set_position (GTK_WINDOW (dialog), GTK_WIN_POS_MOUSE);

        
    //gtk_window_set_keep_above (GTK_WINDOW(dialog), TRUE);


    
    gtk_window_set_default_size (GTK_WINDOW (dialog), 425, -1);
    gtk_window_set_icon ((GtkWindow *) dialog, rfm_get_pixbuf("xffm/stock_find", SIZE_ICON));
	    

    GtkWidget *vbox = gtk_vbox_new (FALSE, 0);
    gtk_widget_show (vbox);
    gtk_container_add (GTK_CONTAINER (dialog), vbox);

    widgets_p->vpane = gtk_vpaned_new ();
    gtk_widget_show (widgets_p->vpane);
    gtk_box_pack_start (GTK_BOX (vbox), widgets_p->vpane, TRUE, TRUE, 0);
    gtk_paned_set_position (GTK_PANED (widgets_p->vpane), 1000);
    vbox6 = gtk_vbox_new (FALSE, 6);
    gtk_paned_pack1 (GTK_PANED (widgets_p->vpane), 
	    GTK_WIDGET (vbox6), FALSE, TRUE);
    gtk_widget_show(vbox6);

    widgets_p->diagnostics =
		    (GtkWidget **)malloc(sizeof(GtkWidget *));
    *(widgets_p->diagnostics) = gtk_text_view_new ();
    GtkWidget *scrolledwindow = gtk_scrolled_window_new (NULL, NULL);
    gtk_widget_show (scrolledwindow);
    gtk_widget_show (*(widgets_p->diagnostics));
    gtk_paned_pack2 (GTK_PANED (widgets_p->vpane), scrolledwindow, TRUE, TRUE);
    gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolledwindow), 
	    GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);

    gtk_container_add (GTK_CONTAINER (scrolledwindow), *(widgets_p->diagnostics));
    gtk_container_set_border_width (GTK_CONTAINER (*(widgets_p->diagnostics)), 2);
    gtk_widget_set_can_focus(*(widgets_p->diagnostics), FALSE);
    gtk_text_view_set_wrap_mode (GTK_TEXT_VIEW (*(widgets_p->diagnostics)), GTK_WRAP_WORD);
    gtk_text_view_set_cursor_visible (GTK_TEXT_VIEW (*(widgets_p->diagnostics)), FALSE);


    
    gtk_container_set_border_width (GTK_CONTAINER (vbox6), 5);


    GtkWidget *hbox= gtk_hbox_new (FALSE, 0);
    gtk_box_pack_start (GTK_BOX (vbox6), hbox, FALSE, FALSE, 0);
    gtk_widget_show(hbox);

    gchar *t=g_strdup_printf("<b>%s <i>(fgr)</i></b>  ", _("Find"));
    label42 = gtk_label_new (t);
    g_free(t);
    gtk_widget_show (label42);
    gtk_box_pack_start (GTK_BOX (hbox), label42, FALSE, FALSE, 0);
    gtk_label_set_use_markup (GTK_LABEL (label42), TRUE);
    gtk_misc_set_alignment (GTK_MISC (label42), 0, 0.5);

 
    GtkWidget *button = rfm_dialog_button ("xffm/stock_help", NULL);
    gtk_widget_set_tooltip_text(button, "man fgr");
    g_signal_connect (G_OBJECT (button), "clicked", G_CALLBACK (rtfm), "fgr");
    g_object_set_data(G_OBJECT(button), "dialog", dialog);
    gtk_box_pack_start (GTK_BOX (hbox), button, FALSE, FALSE, 0);
    gtk_widget_show(button);

    button = rfm_dialog_button ("xffm/stock_dialog-question", NULL);
    gtk_widget_set_tooltip_text(button, "fgr --help");
    g_signal_connect (G_OBJECT (button), "clicked", G_CALLBACK (command_help), "fgr");
    g_object_set_data(G_OBJECT(button), "dialog", dialog);
    gtk_box_pack_start (GTK_BOX (hbox), button, FALSE, FALSE, 0);
    gtk_widget_show(button);

   

    hbox25 = gtk_hbox_new (FALSE, 0);
    gtk_widget_show (hbox25);
    gtk_box_pack_start (GTK_BOX (vbox6), hbox25, FALSE, FALSE, 0);
    gtk_container_set_border_width (GTK_CONTAINER (hbox25), 2);

    hbox14 = gtk_hbox_new (FALSE, 0);
    gtk_widget_show (hbox14);
    gtk_box_pack_start (GTK_BOX (hbox25), hbox14, TRUE, TRUE, 0);
    gtk_container_set_border_width (GTK_CONTAINER (hbox14), 5);

    vbox7a = gtk_vbox_new (FALSE, 5);
    gtk_widget_show (vbox7a);
    gtk_box_pack_start (GTK_BOX (hbox14), vbox7a, TRUE, TRUE, 0);

    hbox15 = gtk_hbox_new (FALSE, 0);
    gtk_widget_show (hbox15);
    gtk_box_pack_start (GTK_BOX (vbox7a), hbox15, FALSE, FALSE, 0);

    t=g_strdup_printf("%s:", _("Path"));
    label26 = gtk_label_new (t);
    g_free(t);
    gtk_widget_show (label26);
    gtk_box_pack_start (GTK_BOX (hbox15), label26, FALSE, FALSE, 0);

#if GTK_MAJOR_VERSION==2 && GTK_MINOR_VERSION<24
    // this is deprecated...
    path_combo = gtk_combo_box_entry_new ();
#else
    path_combo = gtk_combo_box_new_with_entry ();
#endif
    g_object_set_data (G_OBJECT (path_combo), "GladeParentKey", path_combo);
    gtk_widget_show (path_combo);
    gtk_box_pack_start (GTK_BOX (hbox15), path_combo, TRUE, TRUE, 0);
    g_object_set_data(G_OBJECT(dialog), "path_combo", path_combo);


    button = gtk_button_new ();
    GtkWidget *image = gtk_image_new_from_stock (GTK_STOCK_DIRECTORY, GTK_ICON_SIZE_BUTTON);
    gtk_button_set_image ((GtkButton *) button, image);
    vbox = gtk_vbox_new (FALSE, 6);
    gtk_box_pack_start (GTK_BOX (hbox15), vbox, FALSE, FALSE, 0);
    gtk_box_pack_start (GTK_BOX (vbox), button, FALSE, FALSE, 0);
    gtk_widget_show (vbox);
    gtk_widget_show (button);
    g_object_set_data(G_OBJECT(dialog), "fileselector", button);

    hbox16 = gtk_hbox_new (FALSE, 0);
    gtk_widget_show (hbox16);
    gtk_box_pack_start (GTK_BOX (vbox7a), hbox16, TRUE, FALSE, 5);

    gchar *text=g_strdup_printf("%s ", _("Filter:"));
    label27 = gtk_label_new (text);
    g_free(text);
    gtk_widget_show (label27);
    gtk_box_pack_start (GTK_BOX (hbox16), label27, FALSE, FALSE, 0);

#if GTK_MAJOR_VERSION==2 && GTK_MINOR_VERSION<24
    // this is deprecated...
    filter_combo = gtk_combo_box_entry_new ();
#else
    filter_combo = gtk_combo_box_new_with_entry ();
#endif
    g_object_set_data (G_OBJECT (filter_combo), "GladeParentKey", filter_combo);
    gtk_widget_show (filter_combo);
    gtk_box_pack_start (GTK_BOX (hbox16), filter_combo, TRUE, TRUE, 0);
    g_object_set_data(G_OBJECT(dialog), "filter_combo", filter_combo);
    

    togglebutton2 = gtk_toggle_button_new ();
    gtk_widget_show (togglebutton2);
    gtk_box_pack_start (GTK_BOX (hbox16), togglebutton2, FALSE, FALSE, 0);
    g_object_set_data(G_OBJECT(dialog), "togglebutton2", togglebutton2);
    gtk_widget_set_tooltip_text(togglebutton2, _(filter_text_help));
    g_signal_connect (togglebutton2,
                      "toggled", G_CALLBACK (on_help_filter), 
		      (gpointer) dialog);

    pb = rfm_get_pixbuf ("xffm/stock_dialog-question", SIZE_BUTTON);
    image2 = gtk_image_new_from_pixbuf (pb);
    gtk_widget_show (image2);
    gtk_container_add (GTK_CONTAINER (togglebutton2), image2);

/*    scrolledwindow8 = gtk_scrolled_window_new (NULL, NULL);
    gtk_box_pack_start (GTK_BOX (vbox7a), scrolledwindow8, TRUE, TRUE, 0);
    gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolledwindow8), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
    g_object_set_data(G_OBJECT(togglebutton2), "scrolledwindow8", scrolledwindow8);

    viewport3 = gtk_viewport_new (NULL, NULL);
    gtk_widget_show (viewport3);
    gtk_container_add (GTK_CONTAINER (scrolledwindow8), viewport3);

    filter_help = gtk_label_new ("");

    gtk_container_add (GTK_CONTAINER (viewport3), filter_help);
    g_object_set_data(G_OBJECT(dialog), "filter_help", filter_help);*/

    hbox17 = gtk_hbox_new (FALSE, 0);
    gtk_widget_show (hbox17);
    gtk_box_pack_start (GTK_BOX (vbox7a), hbox17, TRUE, FALSE, 0);

    vbox12 = gtk_vbox_new (FALSE, 0);
    gtk_widget_show (vbox12);
    gtk_box_pack_start (GTK_BOX (hbox17), vbox12, FALSE, FALSE, 0);

    recursive = gtk_check_button_new_with_mnemonic (_("Recursive"));
    gtk_widget_show (recursive);
    gtk_box_pack_start (GTK_BOX (vbox12), recursive, FALSE, FALSE, 0);
    gtk_container_set_border_width (GTK_CONTAINER (recursive), 3);
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (recursive), default_recursive);
    g_object_set_data(G_OBJECT(dialog), "recursive", recursive);

    recursiveH = gtk_check_button_new_with_mnemonic (_( "Find hidden files and directories"));
    gtk_widget_show (recursiveH);
    gtk_box_pack_start (GTK_BOX (vbox12), recursiveH, FALSE, FALSE, 0);
    gtk_container_set_border_width (GTK_CONTAINER (recursiveH), 3);
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (recursiveH), default_recursiveH);
    g_object_set_data(G_OBJECT(dialog), "recursiveH", recursiveH);

    xdev = gtk_check_button_new_with_mnemonic (_("Stay on single filesystem"));
    gtk_widget_show (xdev);
    gtk_box_pack_start (GTK_BOX (vbox12), xdev, FALSE, FALSE, 0);
    gtk_container_set_border_width (GTK_CONTAINER (xdev), 3);
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (xdev), default_xdev);
    g_object_set_data(G_OBJECT(dialog), "xdev", xdev);

    hbox18 = gtk_hbox_new (FALSE, 0);
    gtk_widget_show (hbox18);
    gtk_box_pack_start (GTK_BOX (hbox17), hbox18, TRUE, FALSE, 0);

    spinbutton1_adj = gtk_adjustment_new (result_limit, 0, 4096, 16, 64, 0);
    spinbutton1 = gtk_spin_button_new (GTK_ADJUSTMENT (spinbutton1_adj), 0, 0);
    gtk_widget_show (spinbutton1);
    gtk_box_pack_start (GTK_BOX (hbox18), spinbutton1, TRUE, TRUE, 0);
    gtk_spin_button_set_update_policy (GTK_SPIN_BUTTON (spinbutton1), GTK_UPDATE_IF_VALID);
    g_object_set_data(G_OBJECT(dialog), "spinbutton1", spinbutton1);

    label30 = gtk_label_new ("  ");
    gtk_widget_show (label30);
    gtk_box_pack_start (GTK_BOX (hbox18), label30, FALSE, FALSE, 0);

    hbox27 = gtk_hbox_new (FALSE, 0);
    gtk_widget_show (hbox27);
    gtk_box_pack_start (GTK_BOX (vbox6), hbox27, FALSE, TRUE, 0);
    gtk_container_set_border_width (GTK_CONTAINER (hbox27), 2);

    vbox10 = gtk_vbox_new (FALSE, 0);
    gtk_widget_show (vbox10);
    gtk_box_pack_start (GTK_BOX (hbox27), vbox10, TRUE, TRUE, 0);

    hbox21 = gtk_hbox_new (FALSE, 0);
    gtk_widget_show (hbox21);
    gtk_box_pack_start (GTK_BOX (vbox10), hbox21, TRUE, FALSE, 0);
    gtk_container_set_border_width (GTK_CONTAINER (hbox21), 5);

    label37 = gtk_label_new (_("File type : "));
    gtk_widget_show (label37);
    gtk_box_pack_start (GTK_BOX (hbox21), label37, FALSE, FALSE, 0);

#if GTK_MAJOR_VERSION==2 && GTK_MINOR_VERSION<24
    // this is deprecated...
    file_type_om =  gtk_combo_box_new_text();
#else
    file_type_om =  gtk_combo_box_text_new();
#endif
    gtk_widget_show (file_type_om);
    gtk_box_pack_start (GTK_BOX (hbox21), file_type_om, TRUE, TRUE, 0);
    g_object_set_data(G_OBJECT(dialog), "file_type_om", file_type_om);

    gchar *g=g_strdup_printf("%s. %s", _("Results"), _("Upper limit:"));
    label29 = gtk_label_new (g);
    g_free(g);
    gtk_widget_show (label29);
    gtk_box_pack_start (GTK_BOX (hbox18), label29, TRUE, FALSE, 0);
    
    t=g_strdup_printf("<b>%s</b>", _("Contains"));
    label45 = gtk_label_new (t);
    g_free(t);
    gtk_widget_show (label45);
    gtk_box_pack_start (GTK_BOX (vbox6), label45, FALSE, FALSE, 1);
    gtk_label_set_use_markup (GTK_LABEL (label45), TRUE);
    gtk_misc_set_alignment (GTK_MISC (label45), 0, 0.5);

    hbox26 = gtk_hbox_new (FALSE, 0);
    gtk_widget_show (hbox26);
    gtk_box_pack_start (GTK_BOX (vbox6), hbox26, FALSE, FALSE, 0);
    gtk_container_set_border_width (GTK_CONTAINER (hbox26), 2);

    vbox8 = gtk_vbox_new (FALSE, 5);
    gtk_widget_show (vbox8);
    gtk_box_pack_start (GTK_BOX (hbox26), vbox8, TRUE, TRUE, 0);
    gtk_container_set_border_width (GTK_CONTAINER (vbox8), 5);

    hbox19 = gtk_hbox_new (FALSE, 0);
    gtk_widget_show (hbox19);
    gtk_box_pack_start (GTK_BOX (vbox8), hbox19, FALSE, FALSE, 0);

    t=g_strdup_printf("%s: ",_("Contains the text"));
    label33 = gtk_label_new (t);
    g_free(t);
    gtk_widget_show (label33);
    gtk_box_pack_start (GTK_BOX (hbox19), label33, FALSE, FALSE, 0);

#if GTK_MAJOR_VERSION==2 && GTK_MINOR_VERSION<24
    grep_combo = gtk_combo_box_entry_new ();
#else
    grep_combo = gtk_combo_box_new_with_entry ();
#endif
    g_object_set_data (G_OBJECT (grep_combo), "GladeParentKey", grep_combo);
    gtk_widget_show (grep_combo);
    gtk_box_pack_start (GTK_BOX (hbox19), grep_combo, TRUE, TRUE, 5);
    gtk_widget_set_sensitive (grep_combo, have_grep);
    g_object_set_data(G_OBJECT(dialog), "grep_combo", grep_combo);

    togglebutton3 = gtk_toggle_button_new ();
    gtk_widget_show (togglebutton3);
    gtk_box_pack_start (GTK_BOX (hbox19), togglebutton3, FALSE, FALSE, 0);
    g_object_set_data(G_OBJECT(dialog), "togglebutton3", togglebutton3);
    gtk_widget_set_tooltip_text(togglebutton3, _(grep_text_help));
    g_signal_connect (togglebutton3,
                      "toggled", G_CALLBACK (on_help_grep), 
		      (gpointer) dialog);

    pb = rfm_get_pixbuf ("xffm/stock_dialog-question", SIZE_BUTTON);
    image1 = gtk_image_new_from_pixbuf (pb);
    gtk_widget_show (image1);
    gtk_container_add (GTK_CONTAINER (togglebutton3), image1);
    gtk_widget_set_sensitive (togglebutton3, TRUE);

    hbox20 = gtk_hbox_new (FALSE, 0);
    gtk_widget_show (hbox20);
    gtk_box_pack_start (GTK_BOX (vbox8), hbox20, FALSE, FALSE, 0);

    vbox13 = gtk_vbox_new (FALSE, 0);
    gtk_widget_show (vbox13);
    gtk_box_pack_start (GTK_BOX (hbox20), vbox13, TRUE, TRUE, 0);

    case_sensitive = gtk_check_button_new_with_mnemonic (_("Case Sensitive"));
    gtk_widget_show (case_sensitive);
    gtk_box_pack_start (GTK_BOX (vbox13), case_sensitive, FALSE, FALSE, 0);
    gtk_widget_set_sensitive (case_sensitive, FALSE);
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (case_sensitive), default_case_sensitive);
    g_object_set_data(G_OBJECT(dialog), "case_sensitive", case_sensitive);

    ext_regexp = gtk_check_button_new_with_mnemonic (_("Extended regexp"));
    gtk_widget_show (ext_regexp);
    gtk_box_pack_start (GTK_BOX (vbox13), ext_regexp, FALSE, FALSE, 0);
    gtk_widget_set_sensitive (ext_regexp, FALSE);
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (ext_regexp), default_ext_regexp);
    g_object_set_data(G_OBJECT(dialog), "ext_regexp", ext_regexp);

/*    
    scrolledwindow9 = gtk_scrolled_window_new (NULL, NULL);
    gtk_box_pack_start (GTK_BOX (vbox8), scrolledwindow9, TRUE, TRUE, 0);
    gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolledwindow9), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
    g_object_set_data(G_OBJECT(togglebutton3), "scrolledwindow9", scrolledwindow9);

    viewport4 = gtk_viewport_new (NULL, NULL);
    gtk_widget_show (viewport4);
    gtk_container_add (GTK_CONTAINER (scrolledwindow9), viewport4);

    regexp_help = gtk_label_new ("");
    gtk_widget_show (regexp_help);
    gtk_container_add (GTK_CONTAINER (viewport4), regexp_help);
    g_object_set_data(G_OBJECT(dialog), "regexp_help", regexp_help);*/

    hbox28 = gtk_hbox_new (FALSE, 0);
    gtk_widget_show (hbox28);
    gtk_box_pack_start (GTK_BOX (vbox6), hbox28, TRUE, TRUE, 0);
    gtk_container_set_border_width (GTK_CONTAINER (hbox28), 2);

    vbox11 = gtk_vbox_new (FALSE, 0);
    gtk_widget_show (vbox11);
    gtk_box_pack_start (GTK_BOX (hbox28), vbox11, TRUE, TRUE, 0);

    look_in_binaries = gtk_check_button_new_with_mnemonic (_("Include binary files?"));
    gtk_widget_show (look_in_binaries);
    gtk_box_pack_start (GTK_BOX (vbox11), look_in_binaries, FALSE, FALSE, 0);
    gtk_container_set_border_width (GTK_CONTAINER (look_in_binaries), 5);
    g_object_set_data(G_OBJECT(dialog), "look_in_binaries", look_in_binaries);
    if (have_gnu_grep) {
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (look_in_binaries), default_look_in_binaries);
    }
    gtk_widget_set_sensitive (look_in_binaries, FALSE);

    hbox24 = gtk_hbox_new (FALSE, 0);
    gtk_widget_show (hbox24);
    gtk_box_pack_start (GTK_BOX (vbox11), hbox24, FALSE, FALSE, 0);
    gtk_container_set_border_width (GTK_CONTAINER (hbox24), 5);

    label40 = gtk_label_new (_("Match"));
    gtk_widget_show (label40);
    gtk_box_pack_start (GTK_BOX (hbox24), label40, FALSE, FALSE, 0);
    gtk_misc_set_padding (GTK_MISC (label40), 3, 0);
    gtk_widget_set_sensitive (label40, FALSE);
    g_object_set_data(G_OBJECT(dialog), "label40", label40);

    anywhere = gtk_radio_button_new_with_mnemonic (NULL, _("Anywhere"));
    gtk_widget_show (anywhere);
    gtk_box_pack_start (GTK_BOX (hbox24), anywhere, TRUE, FALSE, 0);
    gtk_radio_button_set_group (GTK_RADIO_BUTTON (anywhere), anywhere_group);
    g_object_set_data(G_OBJECT(dialog), "anywhere", anywhere);
    anywhere_group = gtk_radio_button_get_group (GTK_RADIO_BUTTON (anywhere));
    gtk_widget_set_sensitive (anywhere, FALSE);
    if (default_anywhere) {
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (anywhere), default_anywhere);
    }
     

    match_words = gtk_radio_button_new_with_mnemonic (NULL, _("Whole words only"));
    gtk_widget_show (match_words);
    gtk_box_pack_start (GTK_BOX (hbox24), match_words, TRUE, FALSE, 0);
    gtk_radio_button_set_group (GTK_RADIO_BUTTON (match_words), anywhere_group);
    g_object_set_data(G_OBJECT(dialog), "match_words", match_words);
    anywhere_group = gtk_radio_button_get_group (GTK_RADIO_BUTTON (match_words));
    gtk_widget_set_sensitive (match_words, FALSE);
    if (default_match_words) {
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (match_words), default_match_words);
    }


    match_lines = gtk_radio_button_new_with_mnemonic (NULL, _("lines"));
    gtk_widget_show (match_lines);
    gtk_box_pack_start (GTK_BOX (hbox24), match_lines, TRUE, FALSE, 0);
    gtk_radio_button_set_group (GTK_RADIO_BUTTON (match_lines), anywhere_group);
    g_object_set_data(G_OBJECT(dialog), "match_lines", match_lines);
    anywhere_group = gtk_radio_button_get_group (GTK_RADIO_BUTTON (match_lines));
    gtk_widget_set_sensitive (match_lines, FALSE);
    if (default_match_lines) {
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (match_lines), default_match_lines);
    }

    match_no_match = gtk_radio_button_new_with_mnemonic (NULL, _("No match"));
    gtk_widget_show (match_no_match);
    gtk_box_pack_start (GTK_BOX (hbox24), match_no_match, TRUE, FALSE, 0);
    gtk_radio_button_set_group (GTK_RADIO_BUTTON (match_no_match), anywhere_group);
    g_object_set_data(G_OBJECT(dialog), "match_no_match", match_no_match);
    anywhere_group = gtk_radio_button_get_group (GTK_RADIO_BUTTON (match_no_match));
    gtk_widget_set_sensitive (match_no_match, FALSE);
    if (default_match_no_match) {
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (match_no_match), default_match_no_match);
    }

    label41 = gtk_label_new ("");
    gtk_widget_show (label41);
    gtk_box_pack_start (GTK_BOX (vbox6), label41, FALSE, FALSE, 0);

    hbuttonbox2 = gtk_hbutton_box_new ();
    gtk_widget_show (hbuttonbox2);
    //gtk_box_pack_start (GTK_BOX (vbox7b), hbuttonbox2, FALSE, TRUE, 0);
    gtk_box_pack_start (GTK_BOX (vbox6), hbuttonbox2, FALSE, TRUE, 0);
    gtk_container_set_border_width (GTK_CONTAINER (hbuttonbox2), 3);
    gtk_button_box_set_layout (GTK_BUTTON_BOX (hbuttonbox2), GTK_BUTTONBOX_EDGE);

    GtkWidget *button_image;
    g=g_strdup_printf("%s/%s", _("Find"), _("Close"));
    find_button =  gtk_button_new_with_label(g);
    g_free(g);
    button_image = gtk_image_new_from_pixbuf (rfm_get_pixbuf("xffm/stock_find", SIZE_BUTTON));
    gtk_button_set_image ((GtkButton *)find_button, button_image);
    g_object_set_data(G_OBJECT(dialog), "find_button", find_button);
    g_object_set_data(G_OBJECT(find_button), "dialog", dialog);
    
    apply_button =  gtk_button_new_with_label(_("Find"));
    button_image = gtk_image_new_from_pixbuf (rfm_get_pixbuf("xffm/stock_find", SIZE_BUTTON));
    gtk_button_set_image ((GtkButton *)apply_button, button_image);
    g_object_set_data(G_OBJECT(dialog), "apply_button", apply_button);
    g_object_set_data(G_OBJECT(apply_button), "dialog", dialog);

    close_button =  gtk_button_new_with_label(_("Cancel"));
    button_image = gtk_image_new_from_pixbuf (rfm_get_pixbuf("xffm/stock_cancel", SIZE_BUTTON));
    gtk_button_set_image ((GtkButton *)close_button, button_image);
    g_object_set_data(G_OBJECT(dialog), "close_button", close_button);
    g_object_set_data(G_OBJECT(close_button), "dialog", dialog);

    if (in_widgets_p && in_widgets_p->view_p->flags.type != DESKVIEW_TYPE) {
	gtk_widget_show (find_button);
    }
    gtk_container_add (GTK_CONTAINER (hbuttonbox2), find_button);
    gtk_widget_set_can_default(find_button, TRUE);

    gtk_widget_show (apply_button);
    gtk_container_add (GTK_CONTAINER (hbuttonbox2), apply_button);
    gtk_widget_set_can_default(apply_button, TRUE);

    gtk_widget_show (close_button);
    gtk_container_add (GTK_CONTAINER (hbuttonbox2), close_button);
    gtk_widget_add_accelerator (close_button, "clicked", accel_group, GDK_Escape, 0, GTK_ACCEL_VISIBLE);

    gtk_widget_grab_default (find_button);
    gtk_window_add_accel_group (GTK_WINDOW (dialog), accel_group);

    return dialog;
}
