//
//
/* TODO: chown mountpoint to user after sudo mount.
 * system type test:*/

/* NOTE: linux will fail if HAVE_STATFS is before HAVE_GETMNTENT section */
#if defined(HAVE_GETMNTENT) && defined(HAVE_STRUCT_MNTENT_MNT_DIR)
/* Linux...*/
# define LINUX_CONFIG
#elif defined(HAVE_STATFS) && defined(HAVE_STRUCT_STATFS_F_MNTONNAME)
# define THIS_IS_BSD
/* BSD functions are not thread safe... ergo, disable them for the time being.*/
#endif

/********/
#define MNTTYPE_PROCFS	"proc"
#define MNTTYPE_SMBFS	"smbfs"
#define MNTTYPE_DEV	"devpts"
#define MNTTYPE_SHM	"tmpfs"
#define	MNTTYPE_CDFS	"iso9660"
#define MNTTYPE_SWAP	"swap"
#define MNTTYPE_NFS	"nfs"

#define MNTTYPE_CACHEFS	"cachefs"       // proc
#define MNTTYPE_HSFS	"hsfs"  // proc

#define MNTTYPE_CODAFS	"coda"  //nfs
#define MNTTYPE_CDFS_AIX "cdrfs"        /* AIX */
#define MNTTYPE_KERNFS	"kernfs"        // proc
#define MNTTYPE_MFS	"mfs"   //proc


static int
get_fstab_type (const char *type) {

    if(strstr (type, MNTTYPE_NFS)) {
        TRACE ("    got %s type for fstab item %s\n", MNTTYPE_NFS, type);
        return __NFS_TYPE;
    } else if(strstr (type, MNTTYPE_SMBFS)) {
        TRACE ("    got %s type for fstab item %s\n", MNTTYPE_SMBFS, type);
        return __SMB_TYPE;
    } else if(strstr (type, MNTTYPE_PROCFS)) {
        TRACE ("    got %s type for fstab item %s\n", MNTTYPE_PROCFS, type);
        return __PROC_TYPE;
    } else if(strstr (type, MNTTYPE_DEV)) {
        TRACE ("    got %s type for fstab item %s\n", MNTTYPE_DEV, type);
        return __PROC_TYPE;
    } else if(strstr (type, MNTTYPE_SHM)) {
        TRACE ("    got %s type for fstab item %s\n", MNTTYPE_SHM, type);
        return __PROC_TYPE;
    } else if(strstr (type, MNTTYPE_CDFS)) {
        TRACE ("    got %s type for fstab item%s\n", MNTTYPE_CDFS, type);
        return __CDFS_TYPE;
    } else if(strstr (type, MNTTYPE_CDFS_AIX))
        return __CDFS_TYPE;
    else if(strstr (type, MNTTYPE_CACHEFS))
        return __PROC_TYPE;
    else if(strstr (type, MNTTYPE_HSFS))
        return __PROC_TYPE;
    else if(strstr (type, MNTTYPE_KERNFS))
        return __PROC_TYPE;
    else if(strstr (type, MNTTYPE_MFS))
        return __PROC_TYPE;
    else if(strstr (type, MNTTYPE_MFS))
        return __NFS_TYPE;

    return 1;
}

static void *
private_is_mounted (void *p) {
    if(!p) {
        g_warning ("fstab.i:private_is_mounted() mnt_point != NULL not met!");
        return NULL;
    }
    const gchar *mnt_point = (const gchar *)p;
    const gchar *mnt_fsname = (const gchar *)p;

#if defined LINUX_CONFIG
    // Linux...
    struct mntent *m;
    //const gchar *mnttab;
    FILE *tab_file;

    // try both /etc/mtab and /proc/mounts 
    const gchar *mfile[]={"/proc/mounts", "/etc/mtab", NULL};
    const gchar **pfile;
    for (pfile=mfile; pfile && *pfile; pfile++){
	if((tab_file = fopen (*pfile, "r")) == NULL) {
	    continue;
	}
	fclose(tab_file);
	tab_file = setmntent (*pfile, "r");

	if(!tab_file) {
	    perror ("setmntent:");
	    return NULL;
	}
	while((m = getmntent (tab_file)) != NULL) {
	    TRACE ("FSTAB-TRACE:%s:  %s  or  %s\n", mnt_point, m->mnt_dir, m->mnt_fsname);
	    if(strcmp (m->mnt_dir, mnt_point) == 0 || strcmp (m->mnt_fsname, mnt_fsname) == 0) {
		TRACE ("FSTAB-TRACE: GOTCHA  %s  %s:%s\n", m->mnt_dir, m->mnt_fsname, mnt_point);
		endmntent (tab_file);
		return GINT_TO_POINTER (1);
	    }
	}
	endmntent (tab_file);
    }
    return NULL;
#elif defined THIS_IS_BSD
    // BSD
    struct statfs statfsbuf;
    if(statfs (mnt_point, &statfsbuf) == 0) {
        TRACE ("FSTAB: path=%s mnttoname=%s\n", mnt_point, statfsbuf.f_mntonname);
        if(strcmp (mnt_point, statfsbuf.f_mntonname) == 0) {
            TRACE ("FSTAB: path=%s is MOUNTED\n", mnt_point);
            return GINT_TO_POINTER (1);
        }
    }
    return NULL;
#else
    return GINT_TO_POINTER (-1);
# warning "missing code here. "
#endif
}

#if defined LINUX_CONFIG
# define MTAB "/etc/mtab";

/*
 * HP-UX, Linux, SunOS 4.x
 *
 * Ubuntu linux has both HAVE_GETMNTENT and HAVE_STRUCT_MNTENT_MNT_DIR
 */

static gint
my_strcmp (gconstpointer a, gconstpointer b) {
    TRACE ("FSTAB:     compare %s <=> %s\n", (char *)a, (char *)b);
    return (strcmp ((char *)a, (char *)b));
}

static GList *
destroy_fstab_list (GList ** fstab_list_p) {
    GList *p;
    for(p = *fstab_list_p; p && p->data; p = p->next) {
        g_free (p->data);
    }
    g_list_free (*fstab_list_p);
    return NULL;
}

static gboolean
include_in_xfdir (struct mntent *mnt_struct, GList ** fstab_list_p) {
    /* skip swap */
    if(strstr (mnt_struct->mnt_type, MNTTYPE_SWAP))
        return FALSE;
    /* skip missing mount points */
    if(!rfm_g_file_test (mnt_struct->mnt_dir, G_FILE_TEST_IS_DIR))
        return FALSE;
    if(g_list_find_custom (*fstab_list_p, mnt_struct->mnt_dir, my_strcmp) == NULL) {
        *fstab_list_p = g_list_append (*fstab_list_p, g_strdup (mnt_struct->mnt_dir));
        TRACE ("FSTAB: adding %s\n", mnt_struct->mnt_dir);
        return TRUE;
    } else {
        TRACE ("FSTAB: skipping %s\n", mnt_struct->mnt_dir);
        return FALSE;
    }
    return TRUE;
}

static int
count_elements (void
    ) {

    GList *fstab_list = NULL;
    struct mntent *mnt_struct;
    FILE *fstab_fd;
    int i = 0;
//    char *files[]={"/etc/fstab", NULL};
    char *files[] = { "/etc/fstab", "/etc/mtab", NULL };
    char **p;
    for(p = files; *p; p++) {
        if((fstab_fd = setmntent (*p, "r")) == NULL) {
            g_warning ("Unable to open %s", "/etc/fstab");
            return (0);
        }
        for(mnt_struct = getmntent (fstab_fd); mnt_struct != NULL; mnt_struct = getmntent (fstab_fd)) {
            if(!include_in_xfdir (mnt_struct, &fstab_list))
                continue;
            i++;
        }
        (void)endmntent (fstab_fd);
    }
    fstab_list = destroy_fstab_list (&fstab_list);
    return i;
}

static gboolean
is_user_type (const gchar * mnt_point) {
    FILE *fstab_fd;
    if((fstab_fd = setmntent ("/etc/fstab", "r")) == NULL) {
        g_warning ("Unable to open %s", "/etc/fstab");
        return FALSE;
    }
    struct mntent *mnt_struct;
    gboolean user_type = FALSE;
    for(mnt_struct = getmntent (fstab_fd); mnt_struct != NULL; mnt_struct = getmntent (fstab_fd)) {
        if(strstr (mnt_struct->mnt_type, MNTTYPE_SWAP))
            continue;
        if(!rfm_g_file_test (mnt_struct->mnt_dir, G_FILE_TEST_IS_DIR))
            continue;
        if(strcmp (mnt_point, mnt_struct->mnt_dir) == 0 && strstr (mnt_struct->mnt_opts, "user")) {
            user_type = TRUE;
            break;
        }
    }
    (void)endmntent (fstab_fd);
    return user_type;
}

static void
get_mnt_items (xfdir_t * xfdir_p) {

    GList *fstab_list = NULL;
    char *files[] = { "/etc/fstab", "/etc/mtab", NULL };
    char **p;
    int i = 1;
    struct mntent *mnt_struct;
    for(p = files; *p; p++) {
        TRACE ("FSTAB:  parsing %s\n", *p);
        FILE *fstab_fd = setmntent (*p, "r");

        for(mnt_struct = getmntent (fstab_fd); mnt_struct != NULL; mnt_struct = getmntent (fstab_fd)) {
            if(!include_in_xfdir (mnt_struct, &fstab_list))
                continue;

            xfdir_p->gl[i].en = rfm_stat_entry (mnt_struct->mnt_dir, 0);
            if(strcmp (*p, "/etc/mtab") == 0) {
                TRACE ("MTAB: setting MTAB type for %s\n", xfdir_p->gl[i].en->path);
                SET_MTAB_TYPE (xfdir_p->gl[i].en->type);
            } else {
                xfdir_p->gl[i].en->tag = g_strdup (mnt_struct->mnt_fsname);
            }
            xfdir_p->gl[i].pathv = g_strdup (mnt_struct->mnt_dir);
            TRACE ("FSTAB:  adding item %d %s \n", i, xfdir_p->gl[i].pathv);

            SET_FSTAB_TYPE (xfdir_p->gl[i].en->type);
            if(strstr (mnt_struct->mnt_opts, "user")) {
                SET_USER_TYPE (xfdir_p->gl[i].en->type);
            }
            /* set type */
            TRACE ("FSTAB:   checking %s for fstab type\n", mnt_struct->mnt_type);
            int type = get_fstab_type (mnt_struct->mnt_type);
            switch (type) {
            case __NFS_TYPE:
                SET_NFS_TYPE (xfdir_p->gl[i].en->type);
                break;
            case __CDFS_TYPE:
                SET_CDFS_TYPE (xfdir_p->gl[i].en->type);
                break;
            case __PROC_TYPE:
                SET_PROC_TYPE (xfdir_p->gl[i].en->type);
                break;
            case __SMB_TYPE:
                SET_SMB_TYPE (xfdir_p->gl[i].en->type);
                break;
            default:
                break;
            }
            i++;
        }
        (void)endmntent (fstab_fd);
    }
    fstab_list = destroy_fstab_list (&fstab_list);

}

static gchar *
fstab_get_mnt_dir (gchar * mnt_fsname) {
    FILE *fstab_fd = setmntent ("/etc/mtab", "r");
    if(!fstab_fd)
        return NULL;
    struct mntent *mnt_struct;
    gchar *mnt_dir = NULL;
    for(mnt_struct = getmntent (fstab_fd); mnt_struct != NULL; mnt_struct = getmntent (fstab_fd)) {
        if(strcmp (mnt_fsname, mnt_struct->mnt_fsname) == 0) {
            // hit: multiple entries use first listed 
            // user types have preference and use last listed 
            if(strstr (mnt_struct->mnt_opts, "user")) {
                g_free (mnt_dir);
                mnt_dir = g_strdup (mnt_struct->mnt_dir);
            }
            if(!mnt_dir) {
                mnt_dir = g_strdup (mnt_struct->mnt_dir);
            }
        }
    }
    (void)endmntent (fstab_fd);
    return mnt_dir;
}

static  gchar *
fstab_get_mnt_fsname (gchar * mnt_dir) {
    FILE *fstab_fd = setmntent ("/etc/mtab", "r");
    if(!fstab_fd)
        return NULL;
    struct mntent *mnt_struct;
    gchar *mnt_fsname = NULL;
    for(mnt_struct = getmntent (fstab_fd); mnt_struct != NULL; mnt_struct = getmntent (fstab_fd)) {
        if(strcmp (mnt_dir, mnt_struct->mnt_dir) == 0) {
            // hit: multiple entries use first listed 
            // user types have preference and use last listed 
            if(strstr (mnt_struct->mnt_opts, "user")) {
                g_free (mnt_fsname);
                mnt_fsname = g_strdup (mnt_struct->mnt_fsname);
            }
            if(!mnt_fsname) {
                mnt_fsname = g_strdup (mnt_struct->mnt_fsname);
            }
        }
    }
    (void)endmntent (fstab_fd);
    return mnt_fsname;
}

static void
set_partition_info (record_entry_t * en) {
    FILE *fstab_fd = setmntent ("/etc/fstab", "r");
    if(!fstab_fd)
        return;
    gchar *partition = en->path;
    struct mntent *mnt_struct;
    for(mnt_struct = getmntent (fstab_fd); mnt_struct != NULL; mnt_struct = getmntent (fstab_fd)) {
        if(strcmp (partition, mnt_struct->mnt_fsname) == 0) {
            TRACE ("MOUNT: %s == %s\n", partition, mnt_struct->mnt_fsname);
            // hit: multiple entries use last listed if not user type 
            if(strstr (mnt_struct->mnt_opts, "user")) {
                SET_USER_TYPE (en->type);
                g_free (en->tag);
                en->tag = g_strdup (mnt_struct->mnt_dir);
            } else {
                if(!IS_USER_TYPE (en->type)) {
                    g_free (en->tag);
                    en->tag = g_strdup (mnt_struct->mnt_dir);
                }
            }
        }
    }
    (void)endmntent (fstab_fd);
    return;
}

static int
partition_items (xfdir_t * xfdir_p, int i, gboolean full_read) {
    FILE *partitions = fopen ("/proc/partitions", "r");
    if(partitions) {
        char line[1024];
        memset (line, 0, 1024);
        while(fgets (line, 1023, partitions) && !feof (partitions)) {
            if(strlen (line) < 5)
                continue;
            if(strchr (line, '#'))
                continue;
            TRACE ("FSTAB: %s\n", line);
            gchar *p = strrchr (line, ' ');
            if(p) {
                TRACE ("FSTAB: add --> %s\n", p);
                g_strstrip (p);
                //if (strlen(p) && isdigit(p[strlen(p)-1])) {
                if(strlen (p)) {
                    if(full_read) {
                        xfdir_p->gl[i].pathv = g_strdup_printf ("/dev/%s", p);
                        xfdir_p->gl[i].en = rfm_stat_entry (xfdir_p->gl[i].pathv, 0);
//                xfdir_p->gl[i].en = rfm_mk_entry(0);
                        xfdir_p->gl[i].en->path = g_strdup_printf ("/dev/%s", p);
                        xfdir_p->gl[i].en->mimetype = g_strdup ("xffm/stock_harddisk");
                        TRACE ("FSTAB:  adding item %d %s \n", i, p);
                        SET_PARTITION_TYPE (xfdir_p->gl[i].en->type);
                        set_partition_info (xfdir_p->gl[i].en);
                    }
                    i++;
                }
            }
            memset (line, 0, 1024);
        }
        fclose (partitions);
    }
    return i;
}

static int
count_partitions (void) {
    return partition_items (NULL, 0, FALSE);
}

static int
get_partition_items (xfdir_t * xfdir_p, int i) {
    return partition_items (xfdir_p, i, TRUE);
}

// Linux
static xfdir_t *
private_get_xfdir (xfdir_t * xfdir_p) {

    int elements = count_elements ();
    int partitions = count_partitions ();

    xfdir_p->pathc = elements;
    xfdir_p->pathc += partitions;
    TRACE ("FSTAB:  counted %ld entries for fstab and mtab\n", (long)xfdir_p->pathc);
    /* up item: */
    xfdir_p->pathc++;

    xfdir_p->gl = (dir_t *) malloc (xfdir_p->pathc * sizeof (dir_t));
    memset (xfdir_p->gl, 0, xfdir_p->pathc * sizeof (dir_t));

    // up item:
    xfdir_p->gl[0].en = NULL;
    xfdir_p->gl[0].pathv = g_strdup (g_get_host_name ());
    // fstab and mtab:
    get_mnt_items (xfdir_p);
    // partitions:
    get_partition_items (xfdir_p, elements + 1);

    int i;
    for(i = 1; i < xfdir_p->pathc; i++) {
	// We might have a race here, and pathc is made invalid
	if (!xfdir_p->gl[i].en){
	    xfdir_p->pathc=1;
	} else {
	    xfdir_p->gl[i].en->parent_module = MODULE_NAME;
	}
    }

    TRACE ("FSTAB: gl is 0x%lx with count=%ld\n", (unsigned long)((long)xfdir_p->gl), (long)xfdir_p->pathc);
    // force a icon reload for the monitor thread:
    xfdir_p->invalidate_icons = TRUE;

    return (xfdir_p);
}

/* 
 * This should return the fstab type, -1 if default or NULL is not in fstab...
 * Fstab types *are* (or should be) supported in constants.h.
 */
static void *
fstab_is_in_fstab (void *p) {
    struct mntent *mnt_struct;
    FILE *fstab_fd;
    int result = 0;
    const gchar *path = (const gchar *)p;
    if((fstab_fd = setmntent ("/etc/fstab", "r")) == NULL) {
        fprintf (stderr, "Unable to open %s\n", "/etc/fstab");
        return (NULL);
    }

    for(mnt_struct = getmntent (fstab_fd); mnt_struct != NULL; mnt_struct = getmntent (fstab_fd)) {
        if(strstr (mnt_struct->mnt_type, MNTTYPE_SWAP))
            continue;
        if(!rfm_g_file_test (mnt_struct->mnt_dir, G_FILE_TEST_IS_DIR))
            continue;

        TRACE ("TRACE2:%s --->  %s   or   %s\n", path, mnt_struct->mnt_dir, mnt_struct->mnt_fsname);

        if(strcmp (path, mnt_struct->mnt_dir) == 0 || strcmp (path, mnt_struct->mnt_fsname) == 0) {
            result = get_fstab_type (mnt_struct->mnt_type);
            TRACE ("TRACE2:%s ---> %d %s\n", mnt_struct->mnt_fsname, result, mnt_struct->mnt_type);
            break;
        }
    }

    (void)endmntent (fstab_fd);

    return GINT_TO_POINTER (result);
}

#elif defined THIS_IS_BSD

/*
 * AIX, BSD, OSF1, ULTRIX
 *
 * Ubuntu linux also has HAVE_GETFSENT, but has met preprocessor criteria above.
 */

static gboolean
is_user_type (const gchar * mnt_point) {
    return FALSE;
}

static gboolean
include_in_xfdir (struct fstab *fs) {
    if(strcmp (MNTTYPE_SWAP, fs->fs_vfstype) == 0)
        return FALSE;
    if(!rfm_g_file_test (fs->fs_file, G_FILE_TEST_IS_DIR))
        return FALSE;
    return TRUE;
}

static int
count_elements (void
    ) {
    if(!setfsent ())
        return (0);
    struct fstab *fs;
    int i;
    for(i = 0, fs = getfsent (); fs != NULL; fs = getfsent ()) {
        if(!include_in_xfdir (fs))
            continue;
        i++;
    }
    TRACE ("fstab: count_elements = %d\n", i);
    //endfsent ();
    return i;
}

static int
count_partitions (void) {
    return 0;
}

static gchar *
fstab_get_mnt_dir (gchar * mnt_fsname) {
    struct fstab *fs;

    if(!setfsent ())
        return (NULL);

    gchar *mnt_dir = NULL;
    for(fs = getfsent (); fs != NULL; fs = getfsent ()) {
        if(!include_in_xfdir (fs))
            continue;

        if(strcmp (mnt_fsname, fs->fs_spec) == 0) {
            //if(strcmp (mnt_fsname, mnt_struct->mnt_fsname) == 0) {
            // hit: multiple entries use first listed 
            // user types have preference and use last listed 
            if(strstr (fs->fs_mntops, "user")) {
                g_free (mnt_dir);
                mnt_dir = g_strdup (fs->fs_file);
            }
            if(!mnt_dir) {
                mnt_dir = g_strdup (fs->fs_file);
            }
        }
    }

    //endfsent ();
    return mnt_dir;
}

static  gchar *
fstab_get_mnt_fsname (gchar * mnt_dir) {
    struct fstab *fs;

    if(!setfsent ())
        return (NULL);

    gchar *mnt_fsname = NULL;

    for(fs = getfsent (); fs != NULL; fs = getfsent ()) {
        if(!include_in_xfdir (fs))
            continue;
        if(strcmp (mnt_dir, fs->fs_file) == 0) {
            //if(strcmp (mnt_dir, mnt_struct->mnt_dir) == 0) {
            // hit: multiple entries use first listed 
            // user types have preference and use last listed 
            if(strstr (fs->fs_mntops, "user")) {
                g_free (mnt_fsname);
                mnt_fsname = g_strdup (fs->fs_spec);
            }
            if(!mnt_fsname) {
                mnt_fsname = g_strdup (fs->fs_spec);
            }
        }
    }
    //endfsent ();
    return mnt_fsname;
}

//BSD
static xfdir_t *
private_get_xfdir (xfdir_t * xfdir_p) {
    struct fstab *fs;

    int i = count_elements ();
    if(!setfsent ())
        return (NULL);

    TRACE ("fstab: -count_elements()=%d\n", i);
    /* up item ( + 1): */
    xfdir_p->pathc = i + 1;

    if((xfdir_p->gl = (dir_t *) malloc (xfdir_p->pathc * sizeof (dir_t))) == NULL) {
        endfsent ();
        g_warning ("malloc(%d) failed at fstab-module.i!", xfdir_p->pathc * sizeof (dir_t));
        return (NULL);
    }
    memset (xfdir_p->gl, 0, xfdir_p->pathc * sizeof (dir_t));

    // up item:
    xfdir_p->gl[0].en = NULL;
    xfdir_p->gl[0].pathv = g_strdup (g_get_host_name ());

    for(i = 1, fs = getfsent (); fs != NULL; fs = getfsent ()) {
        if(!include_in_xfdir (fs)) {
            TRACE ("fstab: skipping --> %s\n", fs->fs_file);
            continue;
        }

        xfdir_p->gl[i].en = rfm_stat_entry (fs->fs_file, 0);
        xfdir_p->gl[i].en->parent_module = MODULE_NAME;
        xfdir_p->gl[i].pathv = g_strdup (fs->fs_file);

        SET_FSTAB_TYPE (xfdir_p->gl[i].en->type);
        TRACE ("fstab: %d --> %s\n", i, fs->fs_file);

        if(strcmp (MNTTYPE_CDFS, fs->fs_vfstype) == 0)
            SET_CDFS_TYPE (xfdir_p->gl[i].en->type);
        else if(strcmp (MNTTYPE_CODAFS, fs->fs_vfstype) == 0)
            SET_NFS_TYPE (xfdir_p->gl[i].en->type);
        else if(strcmp (MNTTYPE_KERNFS, fs->fs_vfstype) == 0)
            SET_PROC_TYPE (xfdir_p->gl[i].en->type);
        else if(strcmp (MNTTYPE_MFS, fs->fs_vfstype) == 0)
            SET_PROC_TYPE (xfdir_p->gl[i].en->type);
        else if(strcmp (MNTTYPE_NFS, fs->fs_vfstype) == 0)
            SET_NFS_TYPE (xfdir_p->gl[i].en->type);
        else if(strcmp (MNTTYPE_PROCFS, fs->fs_vfstype) == 0)
            SET_PROC_TYPE (xfdir_p->gl[i].en->type);
        else if(strcmp (MNTTYPE_SMBFS, fs->fs_vfstype) == 0)
            SET_SMB_TYPE (xfdir_p->gl[i].en->type);

        i++;
    }

    //endfsent ();
    // force a icon reload for the monitor thread:
    xfdir_p->invalidate_icons = TRUE;

    return (xfdir_p);
}

static void *
fstab_is_in_fstab (void *p) {
    int result = 0;
    struct fstab *fs;
    const gchar *path = (const gchar *)p;

    if(!setfsent ())
        return (NULL);

    for(fs = getfsent (); fs != NULL; fs = getfsent ()) {
        if(strcmp (MNTTYPE_SWAP, fs->fs_vfstype) == 0)
            continue;
        if(!rfm_g_file_test (fs->fs_file, G_FILE_TEST_IS_DIR))
            continue;

        if(strcmp (path, fs->fs_file) == 0) {
            if(strcmp (MNTTYPE_CDFS, fs->fs_vfstype) == 0)
                result = __CDFS_TYPE;
            else if(strcmp (MNTTYPE_CODAFS, fs->fs_vfstype) == 0)
                result = __NFS_TYPE;
            else if(strcmp (MNTTYPE_KERNFS, fs->fs_vfstype) == 0)
                result = __PROC_TYPE;
            else if(strcmp (MNTTYPE_MFS, fs->fs_vfstype) == 0)
                result = __PROC_TYPE;
            else if(strcmp (MNTTYPE_NFS, fs->fs_vfstype) == 0)
                result = __NFS_TYPE;
            else if(strcmp (MNTTYPE_PROCFS, fs->fs_vfstype) == 0)
                result = __PROC_TYPE;
            else if(strcmp (MNTTYPE_SMBFS, fs->fs_vfstype) == 0)
                result = __SMB_TYPE;
            else
                result = -1;
            break;
        }
    }
    //endfsent ();
    return GINT_TO_POINTER (result);
}

#else
# error "This should not happen: fstab-module.i (Solaris no longer supported)"

#endif

///////////   common code:

static gchar *
fstab_df (void) {
    gchar *df_string = NULL;
    int line_count = 0;
    char line[2048];
#if defined(THIS_IS_BSD)
    FILE *pipe;
    memset ((void *)line, 0, 2048);
    gchar *command = g_find_program_in_path ("df");
    pipe = popen (command, "r");
    g_free (command);
    if(!pipe) {
        g_warning ("unable to pipe df");
        return "";
    }
    while(fgets (line, 2047, pipe) && !feof (pipe)) {
        line_count++;
/*	if (df_string) df_string=g_strconcat(df_string,line,NULL);
	else df_string=g_strdup(line);*/
    }
    pclose (pipe);
#elif defined(LINUX_CONFIG)
    // This is for Linux...*/
    gchar *files[] = { "/proc/partitions", "/proc/mounts", "/etc/mtab", NULL };
    gchar **p;
    for(p = files; p && *p; p++) {
        FILE *partitions = fopen (*p, "r");
        if(partitions) {
            while(fgets (line, 2047, partitions) && !feof (partitions)) {
                line_count++;
            }
            fclose (partitions);
        }
    }
#endif
    df_string = g_strdup_printf ("line_count=%d", line_count);
    TRACE ("DF: %s\n", df_string);
    return df_string;
}

static void *
fstab_entry_tip (void *p) {
    record_entry_t *en = (record_entry_t *) p;
    if(!en)
        return NULL;
    gchar *module_text = NULL;
    gchar *mnt_dir = NULL;
    gchar *mnt_fsname = NULL;
    if(IS_PARTITION_TYPE (en->type)) {
        mnt_fsname = g_strdup (en->path);
        if(en->tag) {
            mnt_dir = g_strdup (en->tag);
        } else {
            mnt_dir = fstab_get_mnt_dir (mnt_fsname);
        }
    } else {
        mnt_dir = g_strdup (en->path);
        if(en->tag) {
            mnt_fsname = g_strdup (en->tag);
        } else {
            mnt_fsname = fstab_get_mnt_fsname (mnt_dir);
        }
    }

    if(!mnt_dir)
        mnt_dir = g_strdup ("");
    if(!mnt_fsname)
        mnt_fsname = g_strdup ("");
    module_text =
        g_strdup_printf ("%s %s\n%s: %s\n%s %s: %s", 
                _("Mount point:"), mnt_dir, 
                _("Partition"), mnt_fsname,
                _("Partition type:"), 
		_("Mount"),
                (IS_USER_TYPE (en->type)) ? _("User") : _("Root"));
    g_free (mnt_dir);
    g_free (mnt_fsname);

    return module_text;
}


static void *
fstab_do_properties (void *p, void *q) {
    
    widgets_t *widgets_p = p;
    if (!q){
	g_warning("!q");
	return NULL;
    }
    GList * select_list=q;
    TRACE("do_prop\n");
    
    record_entry_t *en = select_list->data;

    
    if(IS_PARTITION_TYPE (en->type)) {
        gchar *sudo = g_find_program_in_path ("sudo");
        if(!sudo){
	    g_warning("sudo command not found");
            goto done;
	}
        g_free (sudo);

        gchar *device = g_strdup (en->path);
	TRACE("*device = %s", device);
        while(strlen (device) && isdigit (device[strlen (device) - 1])) {
            device[strlen (device) - 1] = 0;
	    TRACE("device => %s", device);
        }
        gchar *argv[] = { "sudo", "-A", "fdisk", "-l", device, NULL };
	TRACE("%s -A fdisk -l %s", sudo, device);
	rfm_show_text (widgets_p);

        rfm_thread_run_argv (widgets_p, argv, FALSE);
        g_free (device);
        goto done;
    } else {
	TRACE("IS_PARTITION_TYPE (en->type) == FALSE\n");
        return NULL;
    }
done:

    return GINT_TO_POINTER (1);
}

static void *
fstab_hide_local_menu_items (void *p, void *q) {
    record_entry_t *en = q;
    widgets_t *widgets_p = p;
    view_t *view_p=widgets_p->view_p;
    if(!q)
        return NULL;

    // nothing selected:
    gchar *symbols_null[] = {
        "paste_menuitem",
        NULL
    };

    /*gchar *symbols[] = { "copy_menuitem", "cut_menuitem",
       "rename_menuitem", "duplicate_menuitem", "touch_menuitem",
       "symlink_menuitem", "remove_menuitem","autotype_Prun",
       "file_separator",    "ejecutar1",  "ejecutar2", "ejecutar3",
       "ejecutar4", "ejecutar5", "ejecutar6", "ejecutar7",
       "ejecutar8", "ejecutar9", "ejecutar10",
       "file_separator2", "navigation_separator",  "view_separator",
       "file_separator",
       NULL
       };  */
    gchar *symbols[] = {
        "cut_menuitem",
        "rename_menuitem",
        "autotype_Prun",
        "paste_menuitem",
        NULL
    };

    gchar *symbols_partition[] = {
        "open_with_menuitem",
        "copy_menuitem",
        "duplicate_menuitem",
        "symlink_menuitem",
        "touch_menuitem",
        "navigation_separator",
        "view_separator", "file_separator2",
        //"properties2",
        NULL
    };
    gchar *symbols_multiple_select[] = {
        "copy_menuitem",
        "remove_menuitem",
        "properties2", "find2",
        "open_with_menuitem",
        "ejecutar1", "ejecutar2", "ejecutar3",
        "ejecutar4", "ejecutar5", "ejecutar6", "ejecutar7",
        "ejecutar8", "ejecutar9", "ejecutar10",
        "file_separator2", "navigation_separator", "view_separator",
        "file_separator",

        NULL
    };
/*    gchar *symbols_mixed[] = { 
	"copy_menuitem",
	"remove_menuitem",
	NULL
    };*/

    gchar *symbols_dir[] = {
        "module1_menu",
        NULL
    };

    gchar **pp;
    if(!p) {
        for(pp = symbols_null; *pp; pp++) {
            HIDE_IT (widgets_p->paper, *pp);
        }
        return NULL;
    }

    for(pp = symbols; *pp; pp++) {
        HIDE_IT (widgets_p->paper, *pp);
    }
    for(pp = symbols_partition; *pp; pp++) {
        if(IS_PARTITION_TYPE (en->type)) {
            HIDE_IT (widgets_p->paper, *pp);
        }
    }
    for(pp = symbols_dir; *pp; pp++) {
        if(!IS_PARTITION_TYPE (en->type)) {
            HIDE_IT (widgets_p->paper, *pp);
        }
    }
    if(g_slist_length (view_p->selection_list) > 1) {
        for(pp = symbols_multiple_select; *pp; pp++) {
            HIDE_IT (widgets_p->paper, *pp);
        }
    }
    return GINT_TO_POINTER (1);
}

static gchar *
get_mnt_point(widgets_t *widgets_p, record_entry_t *en) {
#define MOUNT_DBH_FILE PLUGIN_DIR,"mountdir-4.6.dbh"
        gchar *f = g_build_filename (g_get_user_data_dir (), MOUNT_DBH_FILE, NULL);
        gchar *device = g_path_get_basename (en->path);
        gchar *default_mount;
        if(rfm_g_file_test (en->tag, G_FILE_TEST_IS_DIR))
            default_mount = g_strdup (en->tag);
        else
            default_mount = g_build_filename (g_get_home_dir (), "mnt", device, NULL);
        g_free (device);
	gchar *extra_text = g_strdup_printf(_("Edit mount point for %s"), en->path);
        const gchar *g = rodent_get_response_history (widgets_p, _("Select mount point"),
						_("Mount Point"), 
						extra_text,
						f,
                                                NULL,
                                                default_mount,
                                                NULL, NULL,
                                                GTK_FILE_CHOOSER_ACTION_SELECT_FOLDER,
                                                "/",
						MATCH_FILE);
	g_free(extra_text);
        g_free (default_mount);
        g_free (f);
        if(g == NULL) {
            return NULL;
        }
        if(g_mkdir_with_parents (g, 0700) < 0) {
	    rfm_show_text(widgets_p);
            rfm_diagnostics (widgets_p, "xffm/stock_dialog-error",  NULL);
            rfm_diagnostics (widgets_p, "xffm_tag/stderr", "mkdir(", g, "): ",strerror (errno), "\n", NULL);
            return NULL;
        }
	// is mount point already mounted?
	if (private_is_mounted((void *)g)){
	    rfm_show_text(widgets_p);
            rfm_diagnostics (widgets_p, "xffm/stock_dialog-error", NULL);
            rfm_diagnostics (widgets_p, "xffm_tag/stderr", g, ": ", _("Unable to mount location"), " (", _("address already in use"), ")", "\n", NULL);
            return NULL;
	}
        return g_strdup (g);

}

static gchar *
get_options (gchar * mnt_src) {
    gchar *sudo = g_find_program_in_path ("sudo");
    if(!sudo)
        return NULL;
    gchar *device = g_strdup (mnt_src);
    while(strlen (device) && isdigit (device[strlen (device) - 1])) {
        device[strlen (device) - 1] = 0;
    }
    gchar *command = g_strdup_printf ("sudo -A fdisk -l %s", device);

    gchar line[2048];
    FILE *pipe;

    gchar *sudo_txt = g_strdup_printf("%s %s\n\n%s\n",
	    _("Verifying:"), _("DOS Compatibility"), command);
    setenv("RFM_SUDO_COMMAND", sudo_txt, 1);
    pipe = popen (command, "r");
    unsetenv("RFM_SUDO_COMMAND");

    g_free (sudo_txt);
    g_free (command);
    gboolean fat = FALSE;
    while(fgets (line, 2047, pipe) && !feof (pipe)) {
        if(strstr (line, mnt_src) && strstr (line, "FAT")) {
            fat = TRUE;
            break;
        }
    }
    pclose (pipe);
    gchar *options = NULL;
    if(fat) {
        options = g_strdup_printf ("uid=%d", geteuid ());
    }
    return options;
}

static void *
private_fstab_mount (widgets_t *widgets_p, record_entry_t *en) {
    gchar *argument[MAX_COMMAND_ARGS];
    gchar **ap;
    gchar *umount = "umount",
        *mount = "mount";
    gchar *mnt_src = NULL;
    gchar *mnt_tgt = NULL;

    SETWD();

    if(!widgets_p)
        return NULL;

    if(!en)
        return NULL;

    mnt_src = g_strdup (en->path);
    mnt_tgt = NULL;

    // user option check for mount point
    // this is necesary for mount/umount command outside of
    // the fstab module directory
    if(is_user_type (mnt_src)) {
        SET_USER_TYPE (en->type);
    }

    gchar *options = NULL;
    if(IS_PARTITION_TYPE (en->type) && private_is_mounted ((void *)mnt_src) == NULL && !IS_USER_TYPE (en->type)) {


	mnt_tgt=get_mnt_point(widgets_p, en);
	if (mnt_tgt == NULL) {
            g_free (mnt_src);
            return NULL;
	}
        options = get_options (mnt_src);
    } // end of partition type condition.

    // isofs:
    if ((en->mimetype && strstr(en->mimetype, "application/x-iso9660-image")) ||
	(en->mimemagic && strstr(en->mimemagic, "application/x-iso9660-image"))) {
	mnt_tgt=get_mnt_point(widgets_p, en);
	if (mnt_tgt == NULL) {
            g_free (mnt_src);
            return NULL;
	}
	options = g_strdup("loop");
    }
    TRACE ("MOUNT: mnt_src=%s mnt_tgt=%s\n", mnt_src, mnt_tgt);

    ap = argument;

    // Sudo check...
    gboolean use_sudo = TRUE;
    // not for root
    if(!getuid ())
        use_sudo = FALSE;
    // not for general user mounts
    else if(IS_USER_TYPE (en->type))
        use_sudo = FALSE;
    // sudo not installed
    else {
        char *p = g_find_program_in_path ("sudo");
        if(p == NULL) {
            use_sudo = FALSE;
            rfm_diagnostics (widgets_p, "xffm/stock_dialog-warning", strerror (ENOENT), ": sudo", "\n", NULL);
        } else {
            g_free (p);
        }
    }

    if(use_sudo) {
        *ap++ = "sudo";
        *ap++ = "-A";
    }

#if 0
    *ap++ = "bash";
    *ap++ = "-c";


    gchar *bash_command=g_strconcat(
	    (private_is_mounted ((void *)mnt_src))?umount:mount,
	    " ",
	    (options)?"-o":"", 
	    " ",
	    (options)?options:"",
	    " ",
	    mnt_src,
	    " ",
	    (mnt_tgt)?mnt_tgt:"",
	    NULL);
#if 10
    // sudo will be dropped after the &&
    if (mnt_tgt && use_sudo) {
	gchar *uid=g_strdup_printf("%d:%d", (gint) geteuid(), (gint)getegid());
	gchar *new_bash_command = g_strconcat(
		bash_command,
		" && ",
		"chown",
		" ",
		uid,
		" ",
		mnt_tgt,
		NULL);
	g_free(uid);
	g_free(bash_command);
	bash_command = new_bash_command;
    }
#endif

    *ap++ = bash_command;
    *ap++ = NULL;
#endif

#if 10
    if(private_is_mounted ((void *)mnt_src)) {
        *ap++ = umount;
    } else {
        *ap++ = mount;
	// Here we set the mounted type. This remains set until
	// monitor unsets it along with the greenball...
	SET_MOUNTED_TYPE(en->type);
        if(options) {
            *ap++ = "-o";
            *ap++ = options;
        }
    }
    memset(en->st, 0, sizeof(struct stat));

    *ap++ = (gchar *) mnt_src;
    if(mnt_tgt) {
        *ap++ = mnt_tgt;
    }
    *ap++ = NULL;
#if 0
    TRACE ("MOUNT:");
    rfm_diagnostics (widgets_p, "xffm/stock_dialog-info", NULL);
    int i;
    for(i = 0; argument[i]; i++) {
        TRACE (" %s", argument[i]);
        if(strstr (argument[i], "username="))
            rfm_diagnostics (widgets_p, NULL, " username=*****", NULL);
        else
            rfm_diagnostics (widgets_p, NULL, " ", argument[i], NULL);
    }
    TRACE ("\n");
    rfm_diagnostics (widgets_p, NULL, "\n", NULL);
#endif

#endif

    rfm_cursor_wait (widgets_p->window);
    rfm_thread_run_argv (widgets_p, argument, FALSE);
    rfm_cursor_reset (widgets_p->window);
    //g_free (bash_command);

#if 0
    if(mnt_tgt && use_sudo){
	ap=argument;
        *ap++ = "sudo";
        *ap++ = "-A";
	*ap++ = "chown";
	gchar *uid=g_strdup_printf("%d:%d", (gint) geteuid(), (gint)getegid());
	*ap++ = uid;
	*ap++ = mnt_tgt;
	*ap++ = NULL;
        rfm_thread_run_argv (widgets_p, argument, FALSE);
	g_free(uid);
    }
#endif

    TRACE ("fstab_mount %s done ", mnt_src);
    g_free (mnt_src);
    g_free (mnt_tgt);
    g_free (options);
    return GINT_TO_POINTER (1);
}

static void *
fstab_double_click (void *p, void *q) {
    record_entry_t *en = (record_entry_t *) q;
    widgets_t *widgets_p = (widgets_t *) p;
    if(IS_PARTITION_TYPE (en->type)) {
        private_fstab_mount (widgets_p, en);
        return GINT_TO_POINTER (1);
    }
    return NULL;
}


static void *
fstab_module_stat (void *p, void *q) {
    TRACE("fstab_module_stat now\n");
    const gchar *path = p;
    // if p is NULL, then the function checks the top level stat file
    // and extra fstab file. (two files, yes)
    if(path == NULL && q != NULL) {
	struct stat st[2];
	if(stat ("/etc/fstab", st) < 0) {
	    g_warning ("stat(%s): %s", "/etc/fstab", strerror (errno));
	    return NULL;
	}
	if(stat ("/etc/mtab", st+1) < 0) {
	    g_warning ("stat(%s): %s", "/etc/mtab", strerror (errno));
	    return NULL;
	}
	st->st_mtime += (st+1)->st_mtime;
	st->st_ctime += (st+1)->st_ctime;
	st->st_size += (st+1)->st_size;
	st->st_mode += (st+1)->st_mode;
	st->st_nlink += (st+1)->st_nlink;
	st->st_uid += (st+1)->st_uid;
	st->st_gid += (st+1)->st_gid;
	memcpy(q, st, sizeof(struct stat));
	return GINT_TO_POINTER (1);
    } else if (path == NULL) {
	path="/etc/mtab";
    }
    struct stat st_v;
    struct stat *st = q;
    // if q is NULL, then the function just checks if the stat is possible.
    if(st == NULL) {
        st = &st_v;
    }
    if(!rfm_g_file_test (path, G_FILE_TEST_EXISTS)){
	// this happens when usb item is unplugged, so
	// it merits not a warning....
	// g_warning("%s does not exist", path);
        return NULL;
    }
    if(stat (path, st) < 0) {
        g_warning ("stat(%s): %s", path, strerror (errno));
        return NULL;
    }
    //rfm_threadwait();
    return GINT_TO_POINTER (1);
}

