/*
 *
 * Edscott Wilson Garcia 2001-2011 edscott@xfce.org
 *
 *
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include "rodent_libs.h"


#if JUST_TRANSLATE 
// translations from applications.xml:
gchar *other_translations[]={
        N_("List files only"),
        N_("Extract files from the archive"),
        N_("Open in New Window"), 
        N_("Open in New Tab"),
        N_("Create a compressed archive with the selected objects"),
        N_("Create a new archive"),
        N_("Install"),
        N_("Uninstall"),
        N_("Information about the program"),
        N_("Simulation of data CD burning"),
        N_("Burn CD/DVD")
};
#endif
#define USER_APPLICATIONS g_get_user_config_dir(),PLUGIN_DIR,"user-applications"
#define CACHE_DIR g_get_user_cache_dir(),PLUGIN_DIR,"modules"

/* this should be first 2 lines after headers: */
G_MODULE_EXPORT LIBRFM_MODULE


#include "mime-module.h"
#include "mime-module.i"
#include "mime-mouse_magic.i"
/****************************************************************************/
 
G_MODULE_EXPORT void *
module_active (void){
    return GINT_TO_POINTER(1);
}
 
G_MODULE_EXPORT 
void *
find_mimetype_in_hash(void *p){
    if (!mimetype_hash) return NULL;
    const gchar *file=p;
    gchar *key = get_hash_key (file);
    const gchar *type = g_hash_table_lookup (mimetype_hash, key);
    g_free (key);
    return (void *)type;
}

static 
void *
put_mimetype_in_hash(const gchar *file, const gchar *mimetype){
    if (!mimetype_hash) return NULL;

    gchar *key = get_hash_key (file);
    g_hash_table_replace (mimetype_hash, (gpointer) key, g_strdup(mimetype));
    return NULL;
}

// This function will return a basic mimetype, never an alias.
G_MODULE_EXPORT 
gchar *
mime_type (void *p, void *q) {
    const gchar *old_mimetype = find_mimetype_in_hash(p);
    if (old_mimetype) {
	// already tabulated. Just return previous value.
	return g_strdup(old_mimetype);
    }
    const gchar *file = p;
    struct stat *st_p = q;
    // If stat information is not provided, then stat the item.
    struct stat st_v;
    if(!st_p) {
        st_p = &st_v;
        stat (file, st_p);
    }
    // Get a mimetype from the stat information, if this applies.
    const gchar *type = mimeable_file (st_p);
    if(type) {
	put_mimetype_in_hash(file, type);
        TRACE ("MIME: stat mime_type(%s) -> %s\n", file, type);
       return (void *)g_strdup(type);
    }

    type = locate_mime_t (file);
    if(type) {
        TRACE ("MIME:locate_mime_t(%s) -> %s\n", file, type);
	put_mimetype_in_hash(file, type);
        return (void *)g_strdup(type);
    }

    /* certain back up types? */
    if(file[strlen (file) - 1] == '~' || file[strlen (file) - 1] == '%') {
        TRACE ("MIME: certain back up types(%s)-> %s\n", file, "application/x-trash");
	put_mimetype_in_hash(file, "application/x-trash");
        return (void *)g_strdup("application/x-trash");
    }

    TRACE ("MIME: locate_mime_t FAILED\n");

    return NULL;
}

#ifndef HAVE_LIBMAGIC
# error "Libmagic support not specified."
#endif

// thread safe 
// This function may obtain a basic or alias mimetype, but will always
// return a basic mimetype.
static const gchar *
lib_magic (const gchar * file, int flags) {
    const char *type;
    g_mutex_lock (magic_mutex);
        magic_setflags (cookie, flags);
        type = magic_file (cookie, file);
    g_mutex_unlock (magic_mutex);
    if(type) {
	gchar *hash_key=get_hash_key(type);
	const gchar *basic_type = g_hash_table_lookup(alias_hash, hash_key);
	if (basic_type) type = basic_type;
    } else {
	type = inode[INODE_UNKNOWN];
    }
    return type;
}

// see "man libmagic" for explantion of flags
G_MODULE_EXPORT 
void *
mime_magic (void *p) {
    const gchar *file = p;
    int flags =
        MAGIC_MIME_TYPE |
        MAGIC_NO_CHECK_APPTYPE |
        MAGIC_NO_CHECK_ENCODING | MAGIC_SYMLINK | MAGIC_NO_CHECK_COMPRESS | MAGIC_NO_CHECK_TAR | MAGIC_PRESERVE_ATIME;
    const gchar *mimemagic = lib_magic (file, flags);
    if (!find_mimetype_in_hash((void *)file)){
	put_mimetype_in_hash(file, mimemagic);
    }
    return (void *)g_strdup(mimemagic);
}

G_MODULE_EXPORT 
void *
mime_encoding (void *p) {
    const gchar *file = p;
    int flags =
        MAGIC_MIME_ENCODING |
        MAGIC_NO_CHECK_APPTYPE | MAGIC_SYMLINK | MAGIC_NO_CHECK_COMPRESS | MAGIC_NO_CHECK_TAR | MAGIC_PRESERVE_ATIME;
    const gchar *encoding = lib_magic (file, flags);
    if (encoding) return (void *)g_strdup(encoding);
    return NULL;
}

G_MODULE_EXPORT 
void *
mime_file (void *p) {
    const gchar *file = p;
    int flags = MAGIC_NO_CHECK_ENCODING | MAGIC_SYMLINK | MAGIC_NO_CHECK_COMPRESS | MAGIC_NO_CHECK_TAR | MAGIC_PRESERVE_ATIME;
    const gchar *f = lib_magic (file, flags);
    if (f) return (void *)g_strdup(f);
    return NULL;
}


G_MODULE_EXPORT 
void *
mime_is_valid_command (void *p) {
    const char *cmd_fmt = p;
    TRACE ("MIME: mime_is_valid_command(%s)\n", cmd_fmt);
    GError *error = NULL;
    int argc;
    gchar *path;
    gchar **argv;
    if(!cmd_fmt)
        return GINT_TO_POINTER (FALSE);
    if(!g_shell_parse_argv (cmd_fmt, &argc, &argv, &error)) {
        gchar *msg = g_strcompress (error->message);
        g_warning ("%s: %s", msg, cmd_fmt);
        g_error_free (error);
        g_free (msg);
        return GINT_TO_POINTER (FALSE);
    }
    path = g_find_program_in_path (argv[0]);
    if(!path) {
        gboolean direct_path = g_file_test (argv[0], G_FILE_TEST_EXISTS) ||
            strncmp (argv[0], "./", strlen ("./")) == 0 || strncmp (argv[0], "../", strlen ("../")) == 0;
        //g_warning("argv[0]=%s",argv[0]);
        if(direct_path) {
            path = g_strdup (argv[0]);
        }
    }
    TRACE ("mime_is_valid_command(): g_find_program_in_path(%s)=%s\n", argv[0], path);

    //if (!path || access(path, X_OK) != 0) {
    if(!path) {
        g_strfreev (argv);
        if(!path)
            errno = ENOENT;
        return GINT_TO_POINTER (FALSE);
    }
    // here we test for execution within sudo
    // XXX we could also check for commands executed in a terminal, but not today...
    void *retval=GINT_TO_POINTER(TRUE);
    if (strcmp(argv[0],"sudo")==0) {
        int i=1;
        if (strcmp(argv[i],"-A")==0) i++;
        retval=mime_is_valid_command((gpointer)argv[i]);
    }

    g_strfreev (argv);
    g_free (path);
    return retval;
}

static
gchar *get_hash_key_strstrip (void *p){
    gchar *pp=g_strdup((char *)p);
    g_strstrip(pp);
    gchar *key=get_hash_key ((gchar *)pp);
    g_free(pp);
    return key;
}

G_MODULE_EXPORT 
void *
mime_command_text (void *p) {
    if (!p) return NULL;
    gchar *key=get_hash_key_strstrip ((gchar *)p);
    void *value=g_hash_table_lookup (application_hash_text, key);
    g_free(key);
    return value;
}

G_MODULE_EXPORT 
void *
mime_command_text2 (void *p) {
    if (!p) return NULL;
    gchar *key=get_hash_key_strstrip ((gchar *)p);
    void *value=g_hash_table_lookup (application_hash_text2, key);
    g_free(key);
    return value;
}

G_MODULE_EXPORT 
void *
mime_command_icon (void *p) {
    if (!p) return NULL;
    gchar *key=get_hash_key_strstrip ((gchar *)p);
    void *value=g_hash_table_lookup (application_hash_icon, key);
    g_free(key);
    return value;
}

G_MODULE_EXPORT
void *
mime_command_output (void *p) {
    if (!p) return NULL;
    gchar *key=get_hash_key_strstrip ((gchar *)p);
    void *value=g_hash_table_lookup (application_hash_output, key);
    g_free(key);
    return value;
}
G_MODULE_EXPORT 
void *
mime_command_output_ext (void *p) {
    if (!p) return NULL;
    gchar *key=get_hash_key_strstrip ((gchar *)p);
    void *value=g_hash_table_lookup (application_hash_output_ext, key);
    g_free(key);
    return value;
}

    
G_MODULE_EXPORT
void *
mime_command (void *p) {
    const char *type = p;
    TRACE ("APPS: mime_command(%s)\n", type);
    gchar **apps;
    int i;
    gchar *cmd_fmt = NULL;
    apps = locate_apps (type);
    if(!apps) {
        TRACE ("APPS: --> NULL\n");
        return NULL;
    }
    if(!apps[0]) {
        TRACE ("APPS: --> NULL\n");
        g_free (apps);
        return NULL;
    }

    for(i = 0; apps[i]; i++) {
        g_free (cmd_fmt);
        cmd_fmt = g_strcompress (apps[i]);
        if(mime_is_valid_command (cmd_fmt)) {
            g_strfreev (apps);
            TRACE ("APPS: --> %s\n", cmd_fmt);
            return (void *)cmd_fmt;
        }
    }
    g_free (cmd_fmt);
    g_strfreev (apps);
    TRACE ("APPS: --> NULL\n");
    return NULL;
}

G_MODULE_EXPORT
void *
mime_apps (void *p) {
    const gchar *type = p;
    TRACE ("MIME: mime_apps(%s)\n", type);
    gchar **apps;
    apps = locate_apps (type);
    if(!apps)
        return NULL;
    if(!apps[0]) {
        g_free (apps);
        return NULL;
    }
    return (void *)apps;
}

#if 0
// FIXME: use of freedesktop file is no longer supported
//        comments are in mime-module.xml (culled from freedesktop)
//        This function may be updated to use such, and then used
//        for tooltip text, maybe
G_MODULE_EXPORT 
void *
mime_typeinfo (void *p) {
    const gchar *type = p;
    xmlNodePtr node;
    xmlNodePtr subnode;
    xmlDocPtr doc;
    gchar *mimefile = NULL;
    gchar *mimetype = NULL;
    gchar *info = NULL;

    mimefile = g_build_filename (FREEDESKTOP_MIME_FILE, NULL);

    TRACE ("MIME: mimefile=%s\n", mimefile);
    if(access (mimefile, R_OK) != 0) {
        g_free (mimefile);
        mimefile = NULL;
        TRACE ("MIME: access(mimefile,R_OK)!=0\n");
        return NULL;
    }
    xmlKeepBlanksDefault (0);

    if((doc = xmlParseFile (mimefile)) == NULL) {
        TRACE ("MIME: rfm-modules: invalid xml file: %s\n", mimefile);
        g_free (mimefile);
        mimefile = NULL;
        return NULL;
    }

    node = xmlDocGetRootElement (doc);
    if(!xmlStrEqual (node->name, (const xmlChar *)"mime-info")) {
        TRACE ("MIME: rfm-modules: invalid xml file %s\n", mimefile);
        g_free (mimefile);
        mimefile = NULL;
        xmlFreeDoc (doc);
        return NULL;
    }

    /* Now parse the xml tree */
    TRACE ("MIME: TRACE: parsing %s\n", mimefile);
    TRACE ("MIME: ------------looking for %s\n", type);
    for(node = node->children; node; node = node->next) {
        if(xmlStrEqual (node->name, (const xmlChar *)"mime-type")) {
            mimetype = (char *)xmlGetProp (node, (const xmlChar *)"type");
            TRACE ("MIME: %s==%s\n", mimetype, type);
            if(xmlStrEqual ((const xmlChar *)mimetype, (const xmlChar *)type)) {
                for(subnode = node->children; subnode; subnode = subnode->next) {
                    if(xmlStrEqual (subnode->name, (const xmlChar *)"comment")) {
                        char *e = (char *)xmlNodeListGetString (doc, subnode->children,
                                                                1);
                        g_free (mimetype);
                        xmlFreeDoc (doc);
                        g_free (mimefile);
                        info = g_strdup (e);
                        if(e)
                            g_free (e);
                        return info;
                    }
                }

            }
            g_free (mimetype);
        }
    }
    xmlFreeDoc (doc);
    g_free (mimefile);
    return NULL;
}
#endif

// Insert a command to a mimetype. This will regenerate the disk
// cache.
G_MODULE_EXPORT 
void *
mime_add (void *p, void *q) {
    gchar *type = p;
    gchar *command = g_strdup((char *)q);
    g_strstrip(command);
    if(!command || !strlen (command)){
	g_free(command);
        return NULL;
    }

    TRACE ("OPEN APPS: adding type %s->%s\n", type, command);
    add_type_to_hashtable(type, command, TRUE);
  
    TRACE ("THREAD: g_thread_create(gencache) from mime-module.c\n");

    GError *error = NULL;
    // thread will dispose of config_command:
    gchar *config_command=g_strdup_printf("%s:%s", type, command);
    g_free(command);
    g_thread_create (gencache, config_command, FALSE, &error); 
    if(error) {
          g_warning ("%s", error->message);
          g_error_free (error);
    }
    return NULL;
}

// Append a command to a mimetype. This will not regenerate the disk
// cache, (see dotdesktop module for the reason why not)
G_MODULE_EXPORT 
void *
mime_append (void *p, void *q) {
    gchar *type = p;
    gchar *command = g_strdup((char *)q);
    g_strstrip(command);
    if(!command || !strlen (command)){
	g_free(command);
        return NULL;
    }
    TRACE ("OPEN APPS: appending type %s->%s\n", type, command);
    add_type_to_hashtable(type, command, FALSE);
    g_free(command);
    return NULL;
}

G_MODULE_EXPORT
void *
mime_mk_command_line (void *p, void *q) {
    const gchar *command_fmt = p;
    const gchar *path = q;

    TRACE ("MIME: mime_mk_command_line(%s)\n", path);
    gchar *command_line = NULL;
    gchar *fmt = NULL;

    if(!command_fmt)
        return NULL;
    if(!path)
        path = "";

    TRACE ("MIME: command_fmt=%s\n", command_fmt);

    /* this is to send path as an argument */

    if(strstr (command_fmt, "%s")) {
        fmt = g_strdup (command_fmt);
    } else {
        fmt = g_strconcat (command_fmt, " %s", NULL);
    }
    TRACE ("MIME: command_fmt fmt=%s\n", fmt);

    TRACE ("MIME: path=%s\n", path);
    gchar *esc_path = rfm_esc_string (path);
    command_line = g_strdup_printf (fmt, esc_path);
    g_free (esc_path);
    TRACE ("MIME2: command_line=%s\n", command_line);

    g_free (fmt);
    return (void *)command_line;
}

G_MODULE_EXPORT 
void *
mime_mk_terminal_line (void *p) {
    const gchar *command = p;
    TRACE ("MIME: mime_mk_command_line(%s)\n", command);
    gchar *command_line = NULL;

    if(!command)
        return NULL;

    const gchar *term = rfm_what_term ();
    const gchar *exec_flag = rfm_term_exec_option(term);
    if(!mime_is_valid_command ((void *)term)) {
        g_warning ("%s == NULL", term);
        return NULL;
    }
    command_line = g_strdup_printf ("%s %s %s", term, exec_flag, command);

    return command_line;
}

/////////////////////////////////////////////////////////////////

G_MODULE_EXPORT
const gchar *
g_module_check_init (GModule * module) {
    TRACE("***************g_module_check_init\n");
    mimetype_hash = g_hash_table_new (g_str_hash, g_str_equal);
    alias_hash = g_hash_table_new (g_str_hash, g_str_equal);
    application_hash_sfx = g_hash_table_new (g_str_hash, g_str_equal);
    application_hash_type = g_hash_table_new (g_str_hash, g_str_equal);
    application_hash_icon = g_hash_table_new (g_str_hash, g_str_equal);
    application_hash_text = g_hash_table_new (g_str_hash, g_str_equal);
    application_hash_text2 = g_hash_table_new (g_str_hash, g_str_equal);
    application_hash_output = g_hash_table_new (g_str_hash, g_str_equal);
    application_hash_output_ext = g_hash_table_new (g_str_hash, g_str_equal);
    if(!load_hashes_from_cache ()) 
    {
            mime_build_hashes ();
            GError *error = NULL;
            TRACE ("THREAD: g_thread_create(gencache) from mime-module.i, load_hashes()\n");
             g_thread_create (gencache, NULL, FALSE, &error); // non-hack
            if(error) {
                g_warning ("%s", error->message);
                g_error_free (error);
            }
    }
    if(!magic_mutex) {
        magic_mutex = g_mutex_new ();
        cookie = magic_open (MAGIC_NONE);
        magic_load (cookie, NULL);
    }
    return NULL;
}

G_MODULE_EXPORT 
void *
module_init (void) {
    return GINT_TO_POINTER (1);
}

G_MODULE_EXPORT
void
g_module_unload (GModule * module) {
    // Too much cleanup to do here
    g_warning("mime-module: This code is not designed for unloading once in use.");
}
