//
//
/*
 * Copyright (C) 2002-2011 Edscott Wilson Garcia
 * EMail: edscott@xfce.org
 *
 * This file include modifications of GPL code provided by
 *  Dov Grobgeld <dov.grobgeld@gmail.com>
 *  Tadej Borovsak tadeboro@gmail.com
 * see below for details.
 *
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

static GMutex *thumbnail_mutex = NULL;

extern GMutex *pixbuf_mutex;

static GdkPixbuf *
fix_pixbuf_scale(GdkPixbuf *in_pixbuf){
    if (!GDK_IS_PIXBUF(in_pixbuf)) return NULL;
    GdkPixbuf *out_pixbuf=in_pixbuf;
    int height = gdk_pixbuf_get_height (in_pixbuf);
    int width = gdk_pixbuf_get_width (in_pixbuf);

    if (!pixbuf_mutex) pixbuf_mutex = g_mutex_new();
    // this is to fix paper size previews for text files and pdfs
    if((width < height && height != PREVIEW_IMAGE_SIZE) || 
	(width >= height && width != PREVIEW_IMAGE_SIZE)) {
        //scale it to size
	if (width < height && height != PREVIEW_IMAGE_SIZE) {
	    width = gdk_pixbuf_get_width (in_pixbuf) * PREVIEW_IMAGE_SIZE / height;
	    g_mutex_lock(pixbuf_mutex);
	    out_pixbuf = gdk_pixbuf_scale_simple (in_pixbuf, width, PREVIEW_IMAGE_SIZE, GDK_INTERP_HYPER);
	    g_mutex_unlock(pixbuf_mutex);
	} else {
	    height = gdk_pixbuf_get_height (in_pixbuf) * PREVIEW_IMAGE_SIZE / width;
	    g_mutex_lock(pixbuf_mutex);
	    out_pixbuf = gdk_pixbuf_scale_simple (in_pixbuf, PREVIEW_IMAGE_SIZE, height, GDK_INTERP_HYPER);
	    g_mutex_unlock(pixbuf_mutex);
	}
        g_object_unref(in_pixbuf);
    }

    return out_pixbuf;
}

static GdkPixbuf *
load_preview_pixbuf_from_disk (const gchar * path, const gchar * thumbnail) {
    GError *error = NULL;
    g_mutex_lock (thumbnail_mutex);
    GdkPixbuf *pixbuf = NULL;
    if (!pixbuf_mutex) pixbuf_mutex = g_mutex_new();
    if (g_file_test(thumbnail, G_FILE_TEST_EXISTS)) {
	g_mutex_lock(pixbuf_mutex);
	pixbuf = gdk_pixbuf_new_from_file (thumbnail, &error);
	g_mutex_unlock(pixbuf_mutex);
	// pixbuf has ref==1
    }
    g_mutex_unlock (thumbnail_mutex);
    if(error) {
        g_warning ("gdk_pixbuf_new_from_file(%s -> %s): %s", path, thumbnail, error->message);
        g_error_free (error);
        return NULL;
    }
    if (!GDK_IS_PIXBUF(pixbuf)) {
	TRACE ("!GDK_IS_PIXBUF(pixbuf)\n");
	return NULL;
    }
    GdkPixbuf *original=pixbuf;
    pixbuf=fix_pixbuf_scale(pixbuf);
    if (original != pixbuf) {
        g_mutex_lock (thumbnail_mutex);
        gdk_pixbuf_save (pixbuf, thumbnail, "png", NULL, "tEXt::Software", "Rodent", NULL);
        g_mutex_unlock (thumbnail_mutex);
    }
    return pixbuf;
}


//static
//void show_tip (const population_t * population_p);

/////////////////////////////////////////////////////////////////
// forked version of libmagicwand... No need to mutex thread unsafe library here...

G_MODULE_EXPORT
const gchar *
want_imagemagick_preview (record_entry_t * en) {
    TRACE ("want_imagemagick_preview\n");
    if(!en) return NULL;
    if (!en->filetype) en->filetype = MIME_file(en->path);
    if (!en->mimemagic) en->mimemagic = MIME_magic (en->path);
    gchar *mimetype = g_strconcat (en->mimemagic, "/", en->mimetype, "/", en->filetype, NULL);
    const gchar *convert_type = NULL;

    if(mimetype && strstr (mimetype, "text") && !(strstr (mimetype, "opendocument"))) {     // decode delegate is ghostscript
        if (!en->encoding) en->encoding=MIME_encoding (en->path);
        TRACE ("MIME_encoding= %s\n", en->encoding);
        if (strcmp(en->encoding,"binary")==0) {
            return NULL;
        } 
        convert_type = "TXT";
    } else if(mimetype && strstr (mimetype, "pdf")) {       // decode delegate is ghostscript
        convert_type = "PDF";
    } else if(mimetype && (strstr (mimetype, "postscript") || strstr (mimetype, "eps")) ){
        // decode delegate is ghostscript
        convert_type = "PS";
    }

    g_free (mimetype);

    if(!convert_type)
        return NULL;
    TRACE ("converttype=%s\n", convert_type);

    static gboolean warned = FALSE;

    gboolean gs_warn = strcmp (convert_type, "PS") == 0 || strcmp (convert_type, "PDF") == 0;
    if(gs_warn) {
        gchar *ghostscript = g_find_program_in_path ("gs");
        if(!ghostscript) {
            if(!warned) {
                g_warning
                    ("\n*** Please install ghostscript for ps and pdf previews\n*** Make sure ghostscript fonts are installed too!\n*** You have been warned.");
                fflush (NULL);
                warned = TRUE;
            }
            return NULL;
        }
        g_free (ghostscript);
    }
    return convert_type;
}

static GdkPixbuf *
convert_wait_f (const population_t *population_p, const gchar *thumbnail) {

    if(thumbnail) {
        if(!g_file_test (thumbnail, G_FILE_TEST_EXISTS)) {
            g_warning ("convert output %s (%s) does not exist!", 
		    thumbnail, population_p->en->path);
            return NULL;
        }
        TRACE ("SHOW_TIP:  convert_wait_f %s (%s)\n", 
		thumbnail, population_p->en->path);
    }

            // pixbuf generated and reffed in routine:
    GdkPixbuf *pixbuf=load_preview_pixbuf_from_disk (population_p->en->path, thumbnail);
    return pixbuf;
}

// this function should take a private population_p
static GdkPixbuf *
image_magic_preview_forked (const population_t * population_p, gchar * thumbnail, const gchar * convert_type) {

    gchar *ghostscript = g_find_program_in_path ("gs");
    gchar *convert = g_find_program_in_path ("convert");
    if(!ghostscript) {
        g_error ("cannot find \"%s\" program in path at rodent_magick.i", ghostscript);
        return NULL;
    }
    int fd = open (population_p->en->path, O_RDONLY);
    if(fd < 0) {
        return 0;
    }
    close (fd);

    population_p = population_p;


// options for pdf/ps and txt previews:
// imagemagick: this depends on ghostscript and gsfonts 
// 
// plain ghostscript is muuch faster than imagemagick: (for ps/pdf)
// gs -dSAFER -dBATCH -dNOPAUSE -sDEVICE=png256 -dFirstPage=1 -dLastPage=1 -dPDFFitPage -r100 -sOutputFile=out.png input.pdf
//
//
// text files...must first convert to ps:
//    groff  -Tps  file  >file.ps
// only problem is the encoding now...
// this works, but how to determine the iso code?
//      iconv -f utf-8 -t iso8859-1 -o file.iso8859 file-utf8
//
// solution: use paps
// paps file >  file.ps
//
// real solution, source paps code to produce preview with cairo/pango

    char *src, *tgt;

    char *arg[13];
    int i = 0;

    //pdf and ps ghostscript conversion
    src = g_strdup (population_p->en->path);
    tgt = g_strdup_printf ("-sOutputFile=%s", thumbnail);
    arg[i++] = ghostscript;
    arg[i++] = "-dSAFER";
    arg[i++] = "-dBATCH";
    arg[i++] = "-dNOPAUSE";
    arg[i++] = "-sPAPERSIZE=letter";    // or a4...
    arg[i++] = "-sDEVICE=png256";
    arg[i++] = "-dFirstPage=1";
    arg[i++] = "-dLastPage=1";
    arg[i++] = "-dPDFFitPage";
    arg[i++] = "-r100";
    arg[i++] = tgt;
    arg[i++] = src;
    arg[i++] = NULL;

#if 0
    fprintf(stderr, "%s", population_p->en->path);for (i=0; arg[i]; i++) fprintf(stderr, " %s",arg[i]); fprintf(stderr, "\n");
#endif
    TRACE ("SHOW_TIPx: %s(%s)\n", arg[0], arg[3]);
    GdkPixbuf * retval;
    // this fork is ok from thread, I guess.
    pid_t pid = fork ();
    if(!pid) {
        execv (arg[0], arg);
        _exit (123);
    } else {
        int status;
        waitpid (pid, &status, WUNTRACED);
	// this function refs retval
        retval = convert_wait_f (population_p, thumbnail);
        TRACE ("SHOW_TIPx: preview created by convert\n");
    }
    g_free (ghostscript);       //arg[0]
    g_free (convert);           //arg[0]
    g_free (src);
    g_free (tgt);
    return retval; 
}

///////  end forked stuff /////

#ifdef HAVE_LIBZIP
#ifdef THIS_IS_LINUX
// XXX this use of shm is not FreeBSD compatible.
// this is for the open document previews as well...
# include <zip.h>
static
gchar *
mouse_shm_open (gchar * in_key, unsigned int size, void **ptr) {

    GString *gs = g_string_new (in_key);
    gchar *key = g_strdup_printf ("/%10u", g_string_hash (gs));
    g_strstrip (key);
    g_string_free (gs, TRUE);

    int shm_fd = shm_open (key, O_CREAT | O_RDWR, 0700);
    if(shm_fd < 0) {
        g_warning ("shm_open: %s: %s", key, strerror (errno));
        close (shm_fd);
        g_free (key);
        return NULL;
    }
    if(ftruncate (shm_fd, size) < 0) {
        g_warning ("shm ftruncate: %s", strerror (errno));
        close (shm_fd);
        shm_unlink (key);
        g_free (key);
        return NULL;
    }
    *ptr = mmap (0, size, PROT_READ | PROT_WRITE, MAP_SHARED, shm_fd, 0);
    if(*ptr == MAP_FAILED) {
        g_warning ("mmap convert_p->shm_p: %s", strerror (errno));
        close (shm_fd);
        shm_unlink (key);
        g_free (key);
        return NULL;
    }
    return key;
}
#endif
#endif
//////////////////////////////////////////////////////////////////////////////////
//
// These routines are heavily modified from the "paps" source
// GPL by Dov Grobgeld <dov.grobgeld@gmail.com>

#define BUFSIZE (4096)
#define DEFAULT_FONT_FAMILY	"Sans"
//#define DEFAULT_FONT_FAMILY   "Arial"
#define DEFAULT_FONT_SIZE	"11"
#define MAKE_FONT_NAME(f,s)	f " " s

#include <locale.h>
#include <pango/pango.h>

typedef struct {
    cairo_t *cr;
    cairo_surface_t *surface;
    PangoContext *context;
    int column_width;
    int column_height;
    int top_margin;
    int bottom_margin;
    int left_margin;
    int right_margin;
    int page_width;
    int page_height;
    PangoDirection pango_dir;
} page_layout_t;

typedef enum {
    PAPER_TYPE_A4 = 0,
    PAPER_TYPE_US_LETTER = 1,
    PAPER_TYPE_US_LEGAL = 2
} paper_type_t;

typedef struct {
    int width;
    int height;
} paper_size_t;

static const paper_size_t paper_sizes[] = {
    {595, 842},                 /* A4 */
    {612, 792},                 /* US letter */
    {612, 1008}                 /* US legal */
};

typedef struct linelink_t {
    PangoLayoutLine *pango_line;
    PangoRectangle logical_rect;
    PangoRectangle ink_rect;
    int formfeed;
} linelink_t;

/* Structure representing a paragraph
 */
typedef struct paragraph_t {
    char *text;
    int length;
    int height;                 /* Height, in pixels */
    int formfeed;
    PangoLayout *layout;
} paragraph_t;


static PangoLanguage *
get_language (void) {
    PangoLanguage *retval;
    gchar *lang = g_strdup (setlocale (LC_CTYPE, NULL));
    gchar *p;

    p = strchr (lang, '.');
    if(p)
        *p = 0;
    p = strchr (lang, '@');
    if(p)
        *p = 0;

    retval = pango_language_from_string (lang);
    g_free (lang);

    return retval;
}

static PangoDirection
get_direction (const gchar * str) {
    PangoDirection direction;
    const gchar *p;
    if(g_utf8_validate (str, -1, NULL)) {
        for(p = str; p != NULL && *p; p = g_utf8_find_next_char (p, NULL)) {
            gunichar ch = g_utf8_get_char (p);
            direction = pango_unichar_direction (ch);
            if(direction == PANGO_DIRECTION_NEUTRAL)
                continue;
            TRACE ("direction found\n");
            return direction;
        }
    }
    // default:
    TRACE ("PANGO_DIRECTION_LTR\n");
    return PANGO_DIRECTION_LTR;
}

static char *
read_file (const gchar * path) {
    char *text;
    char *buffer;
    buffer = (char *)malloc (BUFSIZE);
    memset (buffer, 0, BUFSIZE);

    int fd = open (path, O_RDONLY);
    if(fd < 0) {
        g_warning ("open(%s): %s\n", path, strerror (errno));
        return NULL;
    }
    int bytes = read (fd, buffer, BUFSIZE - 1);
    close (fd);
    TRACE ("bytes=%d (%d)\n", bytes, BUFSIZE - 1);
    if(bytes < 0) {
        g_warning ("%s: Error reading file %s (%s).\n", g_get_prgname (), path, strerror (errno));
        g_free (buffer);
        return NULL;
    }

    if(strrchr (buffer, '\n')) {
        *(strrchr (buffer, '\n') + 1) = 0;
    }

    gchar *encoding = MIME_encoding (path);
    gsize bytes_read,
      bytes_written;
    GError *error = NULL;
    if(encoding && (!strstr (encoding, "utf-8") || !strstr (encoding, "UTF-8"))) {
        gchar *obuffer = g_convert_with_fallback (buffer, -1, "UTF-8", encoding,
                                                 NULL, &bytes_read, &bytes_written, &error);
        if(error) {
            TRACE ("TRACE: g_convert_with_fallback(%s): %s\n", path, error->message);
            g_error_free (error);
	    error=NULL;
	    obuffer = g_convert_with_fallback (buffer, -1, "UTF-8", "iso8859-15",
                                                 NULL, &bytes_read, &bytes_written, &error);
	    if (error) {
		TRACE ("TRACE[2]: g_convert_with_fallback(%s): %s\n", path, error->message);
		g_error_free (error);
	    }
        }
        TRACE ("convert successful from %s to UTF-8\n", encoding);
        g_free (buffer);
        text = obuffer;
    }

    g_free (encoding);
    return text;
}

int
output_page (GList * pango_lines, page_layout_t * page_layout) {
    int column_pos = 0;
    int page_idx = 1;

    int pango_column_height = page_layout->page_height - page_layout->top_margin - page_layout->bottom_margin;

    while(pango_lines) {
        linelink_t *line_link = pango_lines->data;
        PangoLayoutLine *line = line_link->pango_line;
        PangoRectangle logical_rect;
        PangoRectangle ink_rect;
        pango_layout_line_get_extents (line, &ink_rect, &logical_rect);
        /* Assume square aspect ratio for now */
        column_pos += (line_link->logical_rect.height / PANGO_SCALE);
        double y_pos = column_pos + page_layout->top_margin;
        if(y_pos > pango_column_height)
            break;              //just first page
        double x_pos = page_layout->left_margin;
        /* Do RTL column layout for RTL direction */
        if(page_layout->pango_dir == PANGO_DIRECTION_RTL) {
            x_pos = page_layout->page_width - page_layout->right_margin;
        }
        TRACE ("gdk_draw_layout_line: %lf, %lf\n", x_pos, y_pos);
        cairo_move_to (page_layout->cr, x_pos, y_pos);
        pango_cairo_show_layout_line (page_layout->cr, line);
        pango_lines = pango_lines->next;
    }
    return page_idx;
}

/* Split a list of paragraphs into a list of lines.
 */
GList *
split_paragraphs_into_lines (page_layout_t * page_layout, GList * paragraphs) {
    GList *line_list = NULL;
    int max_height = 0;
    /* Read the file */

    /* Now split all the pagraphs into lines */
    GList *par_list;

    par_list = paragraphs;
    while(par_list) {
        int para_num_lines;
        linelink_t *line_link;
        paragraph_t *para = par_list->data;

        para_num_lines = pango_layout_get_line_count (para->layout);
        TRACE ("para_num_lines = %d\n", para_num_lines);

        int i;
        for(i = 0; i < para_num_lines; i++) {
            PangoRectangle logical_rect;
            PangoRectangle ink_rect;

            line_link = g_new (linelink_t, 1);
            line_link->formfeed = 0;
            line_link->pango_line = pango_layout_get_line_readonly (para->layout, i);
            pango_layout_line_get_extents (line_link->pango_line, &ink_rect, &logical_rect);
            line_link->logical_rect = logical_rect;
            if(para->formfeed && i == (para_num_lines - 1))
                line_link->formfeed = 1;
            line_link->ink_rect = ink_rect;
            line_list = g_list_prepend (line_list, line_link);
            if(logical_rect.height > max_height)
                max_height = logical_rect.height;
        }

        par_list = par_list->next;
    }

    return g_list_reverse (line_list);

}

/* Take a UTF8 string and break it into paragraphs on \n characters
 */
static GList *
split_text_into_paragraphs (page_layout_t * page_layout, char *text) {
    char *p = text;
    char *next;
    gunichar wc;
    GList *result = NULL;
    char *last_para = text;
    int width = page_layout->page_width - page_layout->right_margin - page_layout->left_margin;

    while(p != NULL && *p) {
        wc = g_utf8_get_char (p);
        next = g_utf8_next_char (p);
        if(wc == (gunichar) - 1) {
            fprintf (stderr, "%s: Invalid character in input\n", g_get_prgname ());
            wc = 0;
        }
        if(!*p || !wc || wc == '\n' || wc == '\f') {
            paragraph_t *para = g_new (paragraph_t, 1);
            para->text = last_para;
            para->length = p - last_para;
            para->layout = pango_layout_new (page_layout->context);
            pango_layout_set_text (para->layout, para->text, para->length);
            pango_layout_set_justify (para->layout, FALSE);
            pango_layout_set_alignment (para->layout,
                                        page_layout->pango_dir == PANGO_DIRECTION_LTR ? PANGO_ALIGN_LEFT : PANGO_ALIGN_RIGHT);
            pango_layout_set_wrap (para->layout, PANGO_WRAP_WORD_CHAR);

            pango_layout_set_width (para->layout, width * PANGO_SCALE);
            para->height = 0;

            last_para = next;

            if(wc == '\f')
                para->formfeed = 1;
            else
                para->formfeed = 0;

            result = g_list_append (result, para);
        }
        if(!wc)                 /* incomplete character at end */
            break;
        p = next;
    }

    return (result);
}

// this function take a private population_p
static GdkPixbuf *
text_preview (const population_t * population_p, gchar * thumbnail, view_t * view_p) {
    if(!population_p || !population_p->en || !population_p->en->path)
        return NULL;
    if(!g_file_test (population_p->en->path, G_FILE_TEST_IS_REGULAR))
        return NULL;
//1. read a cache page worth of text and convert to utf-8 
    gchar *text = read_file (population_p->en->path);
    if(!text) return NULL;
    page_layout_t page_layout;

    TRACE ("PAP: file read complete:\n%s\n", "");
    TRACE ("PAP: file read complete:\n%s\n", text);

    /* Page layout, either a4 or letter or legal. */
    page_layout.page_width = paper_sizes[PAPER_TYPE_US_LETTER].width;
    page_layout.page_height = paper_sizes[PAPER_TYPE_US_LETTER].height;
    page_layout.top_margin = 36;
    page_layout.bottom_margin = 36;
    page_layout.right_margin = 36;
    page_layout.left_margin = 36;

    // determine pango_dir 
    page_layout.pango_dir = get_direction (text);
    page_layout.column_height = page_layout.page_height - page_layout.top_margin - page_layout.bottom_margin;
    page_layout.column_width = page_layout.page_width - page_layout.left_margin - page_layout.right_margin;

    // create cairo drawable surface and context  
    page_layout.surface = cairo_image_surface_create (CAIRO_FORMAT_ARGB32, page_layout.page_width, page_layout.page_height);
    //CAIRO_FORMAT_A1 CAIRO_FORMAT_A8
    page_layout.cr = cairo_create (page_layout.surface);
    if(cairo_surface_status (page_layout.surface) != CAIRO_STATUS_SUCCESS) {
        g_error ("cairo_surface_status(surface) != CAIRO_STATUS_SUCCESS");
    }
    // clear cairo surface
    //
    cairo_set_source_rgb (page_layout.cr, 1.0, 1.0, 1.0);
    cairo_paint (page_layout.cr);
    // create pango context
    page_layout.context = pango_cairo_create_context (page_layout.cr);
    // Setup pango context
    pango_cairo_context_set_resolution (page_layout.context, -1);
    pango_context_set_language (page_layout.context, get_language ());
    pango_context_set_base_dir (page_layout.context, page_layout.pango_dir);

    // font description 
    gchar *font = MAKE_FONT_NAME (DEFAULT_FONT_FAMILY, DEFAULT_FONT_SIZE);
    PangoFontDescription *font_description = pango_font_description_from_string (font);
    if((pango_font_description_get_set_fields (font_description)
        & PANGO_FONT_MASK_FAMILY) == 0) {
        pango_font_description_set_family (font_description, DEFAULT_FONT_FAMILY);
    }
    if((pango_font_description_get_set_fields (font_description)
        & PANGO_FONT_MASK_SIZE) == 0) {
        pango_font_description_set_size (font_description, atoi (DEFAULT_FONT_SIZE) * PANGO_SCALE);
    }
    pango_context_set_font_description (page_layout.context, font_description);
    pango_font_description_free (font_description);

    TRACE ("pango setup complete\n");
    GList *paragraphs = split_text_into_paragraphs (&page_layout, text);

    TRACE ("pango split_text_into_paragraphs complete--> %d \n", g_list_length (paragraphs));
    GList *pango_lines = split_paragraphs_into_lines (&page_layout, paragraphs);
    TRACE ("pango split_paragraphs_into_lines complete--> %d\n", g_list_length (pango_lines));

    // cairo setup
    cairo_new_path (page_layout.cr);
    cairo_set_line_width (page_layout.cr, 0.5);
    cairo_set_source_rgb (page_layout.cr, 0.0, 0.0, 0.0);

    //output layouts to page ()
    output_page (pango_lines, &page_layout);
    // tmp: write png of pixbuf
    //TRACE("// numpages=%d\n", num_pages);
    cairo_destroy (page_layout.cr);
    TRACE ("// write thumbnail\n");
    g_mutex_lock (thumbnail_mutex);
    if(cairo_surface_write_to_png (page_layout.surface, thumbnail) != CAIRO_STATUS_SUCCESS) {
        g_warning ("cairo_surface_write_to_png(surface,) != CAIRO_STATUS_SUCCESS");
    }
    g_mutex_unlock (thumbnail_mutex);
    cairo_surface_destroy (page_layout.surface);
    g_free (text);
    // destroy GLists
    GList *tmp;
    for(tmp = paragraphs; tmp && tmp->data; tmp = tmp->next) {
        paragraph_t *para = tmp->data;
        if(G_IS_OBJECT (para->layout)) {
            g_object_unref (para->layout);
        }
        g_free (para);
    }
    for(tmp = pango_lines; tmp && tmp->data; tmp = tmp->next) {
        linelink_t *line = tmp->data;
        g_free (line);
    }
    g_list_free (paragraphs);
    g_list_free (pango_lines);
 
    // pixbuf generated and reffed in routine:
    TRACE("load_preview_pixbuf_from_disk %s\n", thumbnail);
    GdkPixbuf *pixbuf = load_preview_pixbuf_from_disk (population_p->en->path, thumbnail);
    return pixbuf;
}


// this function should take a private population_p (thread-safe)
// and returns a newly reffed pixbuf

G_MODULE_EXPORT
GdkPixbuf *
mime_preview (const population_t * population_p) {
    if(!population_p->en || !population_p->en->st) {
        TRACE ("SHOW_TIPx: !population_p->en || !population_p->en->st\n");
        return NULL;
    }
    if(population_p->en->st->st_size == 0) {
        return NULL;
    }

    /*if(population_p->flags  & POPULATION_PREVIEW_BUSY) {
        TRACE ("SHOW_TIPx: POPULATION_PREVIEW_BUSY =%d\n", population_p->flags  & POPULATION_PREVIEW_BUSY);
        return NULL;
    }*/
    if(!thumbnail_mutex) {
        thumbnail_mutex = g_mutex_new ();
    }

    // So it is not already loaded (rodent_preview_if_loaded()
    // should be called beforehand.
    // Is the pixbuf in the thumbnail cache?
    gchar *thumbnail = rfm_get_thumbnail_path (population_p->en->path, PREVIEW_IMAGE_SIZE);
    GdkPixbuf *pixbuf = NULL;

    if(thumbnail && rfm_g_file_test (thumbnail, G_FILE_TEST_EXISTS)) {
        struct stat st;
        if(stat (thumbnail, &st) < 0)
            g_error ("stat(%s): %s", thumbnail, strerror (errno));
        if(st.st_mtime >= population_p->en->st->st_mtime) {
            TRACE ("SHOW_TIPx: %s: thumbnail is up to date\n", population_p->en->path);
            // pixbuf generated and reffed in routine:
            pixbuf = load_preview_pixbuf_from_disk (population_p->en->path, thumbnail);
            if(pixbuf) {
		// pixbuf has ref==1 
                TRACE ("SHOW_TIPx: preview loaded from thumbnail file\n");
                g_free (thumbnail);
                return pixbuf;
            }
        }
    }

    // So it is not in thumbnail cache. So it will have to be regenerated
    // First we shall try internal gtk image manipulation:
    if(rfm_is_image (population_p->en)) {
        pixbuf = rfm_get_pixbuf (population_p->en->path, PREVIEW_IMAGE_SIZE);
		// pixbuf has ref==1 but should not be in the pixbuf hash if
		// PREVIEW_IMAGE_SIZE  > MAX_PIXBUF_HASH_SIZE 
		// (400 > 128 last time I looked)
        TRACE ("SHOW_TIPx: preview created by gtk\n");
        g_free (thumbnail);
        if(pixbuf) {
            // test if pixbuf is in pixbuf hash, if it is, then
	    // we should have an additional ref to that the pixbuf
	    // referenced in the hash remains valid when this
	    // pixbuf is unreffed
            if (PREVIEW_IMAGE_SIZE  <= MAX_PIXBUF_HASH_SIZE){
		if (!G_IS_OBJECT (pixbuf)){
		    g_warning("!G_IS_OBJECT (pixbuf) at rodent_preview()");
		}
		g_object_ref (pixbuf);
	    }
            return pixbuf;
        } else {
            return NULL;
        }
    }

    // So it is not an image type
#ifdef HAVE_LIBZIP
#ifdef THIS_IS_LINUX
    // XXX This use of shm is not FreeBSD compatible.
    //     We should add a FreeBSD method too.
    // Is it an OpenDocument file? get the embeded preview
    // Is it a zip file in disguise?
    if (!population_p->en->filetype) 
        population_p->en->filetype = MIME_file(population_p->en->path);
    gboolean OpenDocument = (population_p->en->filetype && strstr (population_p->en->filetype, "OpenDocument") != NULL);
    if(OpenDocument) {
	if (!pixbuf_mutex) pixbuf_mutex = g_mutex_new();
	pixbuf=NULL;
        TRACE (">>>>> mime file (%s) = %s %s\n", 
		population_p->en->path, population_p->en->filetype, 
		(strstr (population_p->en->filetype, "OpenDocument")) ? "YES" : "NO");
        int errorv;
        void *ptr = NULL;
        struct zip *Z = zip_open (population_p->en->path, 0, &errorv);
        struct zip_stat sb;
        if(zip_stat (Z, "Thumbnails/thumbnail.png", 0, &sb) == 0) {
            gchar *shm_thumbnail = mouse_shm_open (population_p->en->path, sb.size, &ptr);
            TRACE ("shm_thumbnail=%s\n", shm_thumbnail);
            if(shm_thumbnail) {
                struct zip_file *ZF = zip_fopen (Z, "Thumbnails/thumbnail.png", 0);
                if(ZF) {
                    zip_fread (ZF, ptr, sb.size);
                    zip_fclose (ZF);
                    gchar *fs_path = g_strdup_printf ("/dev/shm%s", shm_thumbnail);
		    if (rfm_g_file_test(fs_path, G_FILE_TEST_EXISTS)){
		      g_mutex_lock(pixbuf_mutex);
                      pixbuf = gdk_pixbuf_new_from_file (fs_path, NULL);
		      g_mutex_unlock(pixbuf_mutex);
		      pixbuf=fix_pixbuf_scale(pixbuf);	
		    }    
		    // pixbuf: ref==1
                    if(GDK_IS_PIXBUF(pixbuf)) {
                        g_mutex_lock (thumbnail_mutex);
                        {
                            gdk_pixbuf_save (pixbuf, thumbnail,
				    "png", NULL, "tEXt::Software", "Rodent",
                                             NULL);
                        }
                        g_mutex_unlock (thumbnail_mutex);
                    }

                }
                munmap (ptr, sb.size);
                shm_unlink (shm_thumbnail);
            }
        } else {
            g_warning ("zip_stat(Z, Thumbnails/thumbnail.png, 0, &sb)==0");
        }
        zip_close (Z);
        g_free (thumbnail);
        if(pixbuf) {
            return pixbuf;
        } else {
            return NULL;
        }
    }
#endif
#endif

    // Ok, that didn't work either. Is it ghostscript (ps or pdf) or text?
    // this will construct the thumbnail in disk to load on next mousemove.
    const gchar *convert_type = want_imagemagick_preview (population_p->en);

    if(!convert_type) {
        TRACE ("SHOW_TIPx: convert type=%s\n",convert_type);
        g_free (thumbnail);
        return NULL;
    }

    // pdf forks to ghostscript to create thumbnail file
    pixbuf = NULL;
    if(strcmp (convert_type, "PDF") == 0 || strcmp (convert_type, "PDF") == 0) {
            // pixbuf generated and reffed in routine:
        pixbuf = image_magic_preview_forked (population_p, thumbnail, convert_type);
    }
    // text uses pango cairo to create thumbnail file
    else if(strcmp (convert_type, "TXT") == 0) {
        view_t *view_p = population_p->view_p;
            // pixbuf generated and reffed in routine:
        pixbuf = text_preview (population_p, thumbnail, view_p);
    }
    return pixbuf;
}




#if 0
/////////////////////////////////////////////////////////////////////////////////////////
// This sounds good...
//
//  http://www.gtkforums.com/about5204.html
//  tadej tadeboro from slovenia http://tadeboro.blogspot.com/
//
//  Tadej Borovsak tadeboro@gmail.com

/* Key for automated pixbuf updating and destruction */
static const cairo_user_data_key_t pixbuf_key;

/**
 * pixbuf_cairo_create:
 * @pixbuf: GdkPixbuf that you wish to wrap with cairo context
 *
 * This function will initialize new cairo context with contents of @pixbuf. You
 * can then draw using returned context. When finished drawing, you must call
 * pixbuf_cairo_destroy() or your pixbuf will not be updated with new contents!
 *
 * Return value: New cairo_t context. When you're done with it, call
 * pixbuf_cairo_destroy() to update your pixbuf and free memory.
 */
static cairo_t *
pixbuf_cairo_create (GdkPixbuf * pixbuf) {
    gint width,                 /* Width of both pixbuf and surface */
      height,                   /* Height of both pixbuf and surface */
      p_stride,                 /* Pixbuf stride value */
      p_n_channels,             /* RGB -> 3, RGBA -> 4 */
      s_stride;                 /* Surface stride value */
    guchar *p_pixels,           /* Pixbuf's pixel data */
     *s_pixels;                 /* Surface's pixel data */
    cairo_surface_t *surface;   /* Temporary image surface */
    cairo_t *cr;                /* Final context */

    if (!G_IS_OBJECT (pixbuf)){
	g_warning("!G_IS_OBJECT (pixbuf) at pixbuf_cairo_create()");
    }
    g_object_ref (G_OBJECT (pixbuf));

    /* Inspect input pixbuf and create compatible cairo surface */
    g_object_get (G_OBJECT (pixbuf), "width", &width,
                  "height", &height, "rowstride", &p_stride, "n-channels", &p_n_channels, "pixels", &p_pixels, NULL);
    surface = cairo_image_surface_create (CAIRO_FORMAT_ARGB32, width, height);
    s_stride = cairo_image_surface_get_stride (surface);
    s_pixels = cairo_image_surface_get_data (surface);

    /* Copy pixel data from pixbuf to surface */
    while(height--) {
        gint i;
        guchar *p_iter = p_pixels,
            *s_iter = s_pixels;

        for(i = 0; i < width; i++) {
# if G_BYTE_ORDER == G_LITTLE_ENDIAN
            /* Pixbuf:  RGB(A)
             * Surface: BGRA */
            if(p_n_channels == 3) {
                s_iter[0] = p_iter[2];
                s_iter[1] = p_iter[1];
                s_iter[2] = p_iter[0];
                s_iter[3] = 0xff;
            } else {            /* p_n_channels == 4 */

                gdouble alpha_factor = p_iter[3] / (gdouble) 0xff;

                s_iter[0] = (guchar) (p_iter[2] * alpha_factor + .5);
                s_iter[1] = (guchar) (p_iter[1] * alpha_factor + .5);
                s_iter[2] = (guchar) (p_iter[0] * alpha_factor + .5);
                s_iter[3] = p_iter[3];
            }
# elif G_BYTE_ORDER == G_BIG_ENDIAN
            /* Pixbuf:  RGB(A)
             * Surface: ARGB */
            if(p_n_channels == 3) {
                s_iter[3] = p_iter[2];
                s_iter[2] = p_iter[1];
                s_iter[1] = p_iter[0];
                s_iter[0] = 0xff;
            } else {            /* p_n_channels == 4 */

                gdouble alpha_factor = p_iter[3] / (gdouble) 0xff;

                s_iter[3] = (guchar) (p_iter[2] * alpha_factor + .5);
                s_iter[2] = (guchar) (p_iter[1] * alpha_factor + .5);
                s_iter[1] = (guchar) (p_iter[0] * alpha_factor + .5);
                s_iter[0] = p_iter[3];
            }
# else/* PDP endianness */
            /* Pixbuf:  RGB(A)
             * Surface: RABG */
//    surface = cairo_image_surface_create( CAIRO_FORMAT_ARGB32, width, height );
            if(p_n_channels == 3) {
                s_iter[0] = p_iter[0];
                s_iter[1] = 0xff;
                s_iter[2] = p_iter[2];
                s_iter[3] = p_iter[1];
            } else {            /* p_n_channels == 4 */

                gdouble alpha_factor = p_iter[3] / (gdouble) 0xff;

                s_iter[0] = (guchar) (p_iter[0] * alpha_factor + .5);
                s_iter[1] = p_iter[3];
                s_iter[1] = (guchar) (p_iter[2] * alpha_factor + .5);
                s_iter[2] = (guchar) (p_iter[1] * alpha_factor + .5);
            }
# endif
            s_iter += 4;
            p_iter += p_n_channels;
        }
        s_pixels += s_stride;
        p_pixels += p_stride;
    }

    /* Create context and set user data */
    cr = cairo_create (surface);
    cairo_surface_destroy (surface);
    cairo_set_user_data (cr, &pixbuf_key, pixbuf, g_object_unref);

    /* Return context */
    return (cr);
}

/**
 * pixbuf_cairo_destroy:
 * @cr: Cairo context that you wish to destroy
 * @create_new_pixbuf: If TRUE, new pixbuf will be created and returned. If
 *                     FALSE, input pixbuf will be updated in place.
 *
 * This function will destroy cairo context, created with pixbuf_cairo_create().
 *
 * Return value: New or updated GdkPixbuf. You own a new reference on return
 * value, so you need to call g_object_unref() on returned pixbuf when you don't
 * need it anymore.
 */

#endif

////////////////////////////////////////////////////////////////////////
