/*
 * Copyright (C) 2006-11 Edscott Wilson Garcia
 * EMail: edscott@xfce.org
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "rodent_libs.h"


/* this should be first 2 lines after headers: */
G_MODULE_EXPORT LIBRFM_MODULE



#define MODULE_NAME "ps"
#define CHILD_MODULE "ps"

#define ALL_PROCS 	0x100000
#define PROC_CHILDREN 	0x200000

#define ZOMBIE 0x01
#define STOPPED 0x02
#define RUNNING 0x04
#define SLEEPING 0x08
#define DEEPSLEEP 0x10

#include "ps-module.i"


////////////// General gmodule initialization function 
//

G_MODULE_EXPORT
const gchar * 
g_module_check_init(GModule *module){
#ifdef ENABLE_NLS
    bindtextdomain (GETTEXT_PACKAGE,PACKAGE_LOCALE_DIR);
#ifdef HAVE_BIND_TEXTDOMAIN_CODESET
    bind_textdomain_codeset(GETTEXT_PACKAGE, "UTF-8");
#endif
    TRACE("domain=%s", GETTEXT_PACKAGE);
#endif
    reload_mutex=g_mutex_new();
    return NULL;
}

// void
// This function performs module specific initialization.
// The function is only called when module is loaded.
// Beware:
// If this function is not present, module will not load.
G_MODULE_EXPORT
void *
module_init(void){
    return NULL;
}

//
//////////////  Generalized Root Module functions ///////////////////
//

// Plugin functions. Plugin functions may be specified as one of the
// following types.
//
// void:     These functions all return a void pointer
//   	     and take no argument
// natural:  These functions all return a void pointer
// 	     and take a single void pointer argument.
// 	     The space of natural functions is isomorphic
// 	     to the set of natural numbers.
// rational: These functions all return a void pointer
// 	     and take a two void pointer arguments.
// 	     The space of rational functions is isomorphic
// 	     to the set of rational numbers.
// complex:  These functions all return a void pointer
// 	     and take a three void pointer arguments.
// 	     The space of rational functions is isomorphic
// 	     to the set of complex numbers with integer
// 	     imaginary component.
	

//////////////////////////////////////////////////////////////////////
//                          void functions                          //
//////////////////////////////////////////////////////////////////////
 

// gboolean  
// This function informs caller if module is installed or not.
G_MODULE_EXPORT 
void *
module_active (void){
    return GINT_TO_POINTER(1);
}

// gboolean  
// This function determines whether or not rodent will spawn
// a monitor thread.
G_MODULE_EXPORT 
void *
module_monitor (void) {
    return GINT_TO_POINTER (1);
}

// const gchar *
// This returns a key to set a single preference for all module levels
G_MODULE_EXPORT 
void *
module_preferences_key (void) {
    return "RODENT-PS";
}

// const gchar * 
// This function informs caller of the module parent plugin name.
// This is the name of the library. It should match whatever
// was choosen in the Makefile, without the "lib" prefix. For example,
// a plugin located at "libfoo.so.0.0.0" should return "foo".
G_MODULE_EXPORT
void *
module_name(void){
    return MODULE_NAME;
}

// const gchar * 
// This function informs caller of the module parent plugin name.
// If the elements returned by the module correspond to
// other plugins, you should return the name of the plugin that
// should be used to load each element in turn. The name should 
// follow the rules of the module_name() function.
G_MODULE_EXPORT
void *
submodule_name(void){
    return CHILD_MODULE;
}

// gboolean  
// This function informs caller if module is a root module or not.
G_MODULE_EXPORT
void *
is_root_module(void){
    return GINT_TO_POINTER(1);
}

// const gchar *  
// This function informs caller of the executable name used to summon the module.
G_MODULE_EXPORT
void *
exec_name(void){
    return "rodent-ps";
}

// const gchar * 
// This function informs caller of the module's icon identifier.
// Identifier may be a mimetype or an absolute path.
G_MODULE_EXPORT
void *
module_icon_id(void){
    static gchar *icon=NULL;
    if (icon == NULL){
	icon = g_strdup_printf("%s/icons/Rodent/scalable/actions/system-run.svg", PACKAGE_DATA_DIR);
    }
    return icon;
//    return  "xffm/stock_execute";
}

//  gchar *  
// This function returns a newly allocated string with the general information
// of the module. Rodent uses this to construct the popup tip. Returned value
// should be freed when no longer used.
G_MODULE_EXPORT
void *
plugin_info(void){
    return g_strdup(_("View current processes and monitor system state"));
}


//////////////////////////////////////////////////////////////////////
//                        natural functions                         //
//////////////////////////////////////////////////////////////////////

//  gchar *  
// This function returns a newly allocated string with the general information
// of the entry (parameter p). Rodent uses this to construct the popup tip.
// Returned value should be freed when no longer used.
G_MODULE_EXPORT
void *
entry_tip(void *p){
    return ps_entry_tip(p);
}

// gchar *  
// This function returns a newly allocated string with the entry string
// for the label event tip.
// Rodent uses this to construct the popup tip which appears when the mouse
// hovers over the icon label.
// Parameter p is the item's entry pointer. 
// Returned value should be freed when no longer used.
G_MODULE_EXPORT 
void *entry_label_tip (void *p) {
    record_entry_t *en=p;
    return  g_strdup(en->path);
}


// gboolean  
// This function processes the double click action on the
// icon's label. If it returns FALSE, then the default
// Rodent action for label click is performed. The default
// Rodent action checks for file existance to do a rename
// entry (or symlink or duplicate).
// Parameter p is the item's entry pointer. 
G_MODULE_EXPORT 
void *label_click (void *p) {
    // Do nothing.
    // record_entry_t *en=p;
    return  GINT_TO_POINTER(1);
}


// const gchar * 
// This function returns a const pointer to the entry's icon
// identifier. 
// Parameter p is the item's entry pointer. 
// Identifier may be returned as a mimetype or an absolute path.
G_MODULE_EXPORT
void *
item_icon_id (void *p) {
    return ps_item_icon_id(p);
}

//  gchar *  
// This function returns a newly allocated string with the module label.
// Rodent uses this to construct the name which appears under the icon
// at the rodent root level, and to set window title and icon name when 
// the module is loaded.
// Returned value should be freed when no longer used.
G_MODULE_EXPORT
void *
module_label(void * p){
    return ps_module_label();
}

// gboolean  
// This function informs Rodent if the entry (parameter p) is a 
// selectable icon or not.
G_MODULE_EXPORT
void *
is_selectable(void *p){
	return GINT_TO_POINTER(1);
}

// gboolean
// This function fills in previously allocated xfdir_p
// with glob records and entries of the module population.
// Records which are filled in are:
// xfdir_p->pathc: Number of icons for Rodent to display
// xfdir_p->gl[0 ... pathc-1].pathv: Labels to display with each icon
// xfdir_p->gl[0 ... pathc-1].en: Record_entry_t of each icon 
// 				  (NULL entries will point to Rodent root) 
G_MODULE_EXPORT
void *
module_xfdir_get (void *p) {
    g_mutex_lock(reload_mutex);
    gint retval=ps_xfdir_get(p);
    g_mutex_unlock(reload_mutex);
    return GINT_TO_POINTER(retval);
}

// gboolean  
// This is an unconditional refresh signal
// If the module monitor is enabled (via module_monitor() function
// This will return a refresh signal on each monitor loop.
// Parameter p is used by rodent to tell module that monitor is no 
// longer needed. Here we are using the value for diagnostics window
// which goes to NULL when the view is destroyed.
G_MODULE_EXPORT 
void *
monitor_refresh_action (void *p) {
    widgets_t *widgets_p = p;
    GtkWidget **pp=(widgets_p->diagnostics_window)?
		    widgets_p->diagnostics_window:
		    widgets_p->diagnostics;
    gint i;
    for (i=0; i<5; i++){ 
	rfm_threadwait();
    }
    // if diagnostics is NULL, then the window has disappeared.
    if (pp== NULL || *pp==NULL){
	TRACE("diagnostics gone...\n");
	return NULL;
    }
    /* Is the module private popup widget active? */
    const gchar *id = rfm_get_module_popup_id();
    GtkWidget *popup_widget=g_object_get_data(G_OBJECT(widgets_p->paper), id);
    if(popup_widget && gtk_widget_get_visible (popup_widget)) {
        TRACE ("monitor_refresh_action(): popup_widget is visible, skipping refresh\n");
	return NULL;
    } else {
	TRACE("popup_widget is not visible...\n");
    }
    return GINT_TO_POINTER(1);
}

//////////////////////////////////////////////////////////////////////
//                        rational functions                        //
//////////////////////////////////////////////////////////////////////

// gboolean
// This function informs Rodent by returning TRUE that the double click
// action will be processed by the module. If function returns FALSE
// (or is not defined in the module) then Rodent will attempt the default
// double click action.
// Parameter p is the view's widgets_p pointer.
// Parameter q is the icon's record entry.
G_MODULE_EXPORT
void *
double_click(void * p, void *q){
    return ps_double_click(p,q);
}

// gboolean
// This function is to generate a module specific popup menu, either on
// icon click or on void space click . 
// If this function will generate a popup menu, return value should be TRUE,
// otherwise return FALSE so Rodent will generate the default popup menu.
// Popup menu for modules should be set as view data on the paper object
// identified by "private_popup_menu".
// Parameter p is the view's widgets_p pointer.
// Parameter q is the icon's record entry.
G_MODULE_EXPORT
void *
private_popup(void *p, void *q){
    return ps_popup(p, q);
}


// gint
// This function returns the icon size for the view and module.
// This is so that user can change the icon size and when the
// view comes back to the module, or navigates within the module,
// the size used will be the size set by this function.
// Parameter p is the view's view_p pointer.
// Parameter q is size to set (gint).
G_MODULE_EXPORT 
void *
module_icon_size (void *p, void *q) {
    return ps_module_icon_size(p, q);
}

// gint
// This function sets the layout for the population_p pointer
// when the icon size is equal to zero (list mode).
// In this case, if the stat information is not actual file
// information, and the module which set the alternate stat
// record, should interpret this for Rodent and build
// the appropriate pango layout to display. 
// If this function is not present, Rodent will interpret the
// population stat information as a normal file stat record.
// Parameter p is the view's view_p pointer.
// Parameter q is the population_p pointer.
// Note: This function is summoned with population_pp readlock enabled 
G_MODULE_EXPORT
void *
make_list_layout2(void *p, void *q){
    return ps_list_layout(p, q);
}

// gboolean
// This function is to get the arguments necessary to open a
// new window with the module, when the location to be displayed
// is not the initial module view. 
// Example: to get a nested process from the ps-module.
// Parameter p is a pointer to the element's record entry.
// Parameter q is gchar ** to place the return argv values.
G_MODULE_EXPORT
void *
module_argv(void *p, void *q){
    return ps_module_argv(p, q);
}

